//============================================================================
// File          : D3DUtilAppBasic.cpp
// 
// Content       : See header.
// 
//============================================================================
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "stdafx.h"
#include "D3DUtilAppBasic.hpp"
#include "D3DUtilContextDlg.hpp"

//----------------------------------------------------------------------------
//-- Static Variables --------------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Functions ---------------------------------------------------------------
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// D3DUtilAppBasic()
// ---------------------------------------------------------------------------
// ctor
//----------------------------------------------------------------------------
D3DUtilAppBasic::D3DUtilAppBasic() :
_ShowHelp(false),
_HelpStr(
"BASIC HELP:\n"
"==============\n"
"<Pause> - pause the app\n"
"h       - toggle help display\n"
"\n")
{
    _HelpStr += _Nav.GetHelpStr();

	// Initialize the navigation control
    _Nav.SetViewUp(D3DXVECTOR3(0,1,0));
	_Nav.SetInsideLookingOut(false);
    _Nav.SetNavMode(UtilMouseNav::NAV_MODE_TRACKBALL);
	_Nav.SetCamera(&_Cam);
}


//----------------------------------------------------------------------------
// ~D3DUtilAppBasic
// ---------------------------------------------------------------------------
// dtor
//----------------------------------------------------------------------------
D3DUtilAppBasic::~D3DUtilAppBasic()
{
}

//----------------------------------------------------------------------------
// UpdateScene()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
HRESULT D3DUtilAppBasic::UpdateScene()
{
	if (_Player.IsActive())
	{
		if (!_Player.PlayStep(1))
        {
            _Player.Close();
        }
	}

	if (_Recorder.IsActive())
	{
		if (!_Recorder.RecordStep()) 
		{
			AUXTRACE("Record error! Stopping...");
			StopRecord();
		}
	}
	// do passive movement (use current FPS approximation so we can move even while paused).
	_Nav.DoPassiveMovement(static_cast<DWORD>(1.f/GetFPS()*1000.f));//GetLastFrameTime());

	return D3D_OK;
}

//-----------------------------------------------------------------------------
// WndProc
//-----------------------------------------------------------------------------
// Windows message handling
//-----------------------------------------------------------------------------
LRESULT D3DUtilAppBasic::WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled)
{
	LRESULT res = D3DUtilApp::WndProc(hWnd, uMsg, wParam, lParam, handled);
    if (handled) return res;

	// Update navigation control from user input
    // Disable input if we are playing from a file.
	if (!_Player.IsActive())
	{
		res = _Nav.WndProc(hWnd, uMsg, wParam, lParam, handled);
		if (handled) return res;
	}

	switch(uMsg)
	{
	case WM_CHAR:
		switch (TCHAR(wParam))
		{
		case 27:
			PostQuitMessage(0);
            handled = TRUE;
			return 0;

		case 'h':
		case 'H':
            handled = TRUE;
			return 0;
        case ' ':
            {
                HRESULT hres = D3D_OK;
			    // Change the context. Repeat until the user cancels or we successfully change the context
                // but first make sure we are in windowed mode
                if (D3DUtilContext::Instance()->IsFullscreen())
                {
                    FAIL_ASSERT(D3DUtilContext::Instance()->Clear());
                }
                do
                {
                    if (hres == D3DUTILERR_FAILEDCONFIRM)
                    {
                        MessageBox(NULL, "Selected device does not meet the capabilities required by the application. Please select another one.", "Device Confirmation Failure", MB_OK | MB_ICONEXCLAMATION);
                    }
                } while (D3DUTILERR_FAILEDCONFIRM == (hres=DoContextDlg(GetAppWnd(), GetDeviceWnd())));
                FAIL_ASSERT_MSG(hres, "Device failed initialization. Program aborting.");
                handled = TRUE;
			    return 0;
            }
		}
		break;
	case WM_KEYDOWN:
		switch (wParam)
		{
        case VK_PAUSE:
			if (IsPaused()) UnPause(); else Pause();
            handled = TRUE;
            return 0;
            
		case 'H':
			_ShowHelp = !_ShowHelp;
            handled = TRUE;
            return 0;
		}
	}

    return 0;
}

//----------------------------------------------------------------------------
// Reset()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
HRESULT D3DUtilAppBasic::Reset()
{
    FAIL_RETURN(D3DUtilApp::Reset());
    RECT r;
    GetClientRect(GetAppWnd(), &r);
    _Nav.SetWindowSize(r.right, r.bottom);
    _Cam.SetAspect((float)r.right/r.bottom);
    return D3D_OK;
}

//----------------------------------------------------------------------------
// Record
//----------------------------------------------------------------------------
bool D3DUtilAppBasic::Record(const char *Filename)
{
	return _Recorder.BeginRecord(Filename, &_Cam);
}

//----------------------------------------------------------------------------
// Play
//----------------------------------------------------------------------------
bool D3DUtilAppBasic::Play(const char *Filename)
{
	bool res = _Player.OpenForPlayback(Filename, &_Cam);
	return res;
}

//----------------------------------------------------------------------------
// StopPlay
//----------------------------------------------------------------------------
bool D3DUtilAppBasic::StopPlay(unsigned *numFrames, unsigned* timeMS)
{
	if (!_Player.IsActive()) return false;
	bool res = _Player.Close();
	if (res)
	{
		_Player.GetLastPlayStats(numFrames, timeMS);
	}
	return res;
}

//----------------------------------------------------------------------------
// StopRecord
//----------------------------------------------------------------------------
bool D3DUtilAppBasic::StopRecord()
{
	if (!_Recorder.IsActive()) return false;
	return _Recorder.Close();
}

//----------------------------------------------------------------------------
// SetHelpStr
//----------------------------------------------------------------------------
void D3DUtilAppBasic::SetHelpString(const char *helpStr)  
{ 
	_HelpStr = helpStr; 
}

//----------------------------------------------------------------------------
// AppendHelpStr
//----------------------------------------------------------------------------
void D3DUtilAppBasic::AppendHelpString(const char *helpStr)  
{ 
	_HelpStr += helpStr; 
}

//----------------------------------------------------------------------------
// ShowStats;
//----------------------------------------------------------------------------
// Stat display - call this in RenderScene to show counters, etc.
//----------------------------------------------------------------------------
HRESULT D3DUtilAppBasic::ShowStats()
{
	FAIL_RETURN(D3DUtilApp::ShowStats());
	if (_ShowHelp)
	{
		FAIL_RETURN(GetD3DFont().LockText());
		FAIL_RETURN(GetD3DFont().DrawText(5.f, 5.f+GetD3DFont().GetLineHeight(), 0xFFFFFFFF, _HelpStr.c_str()));
		FAIL_RETURN(GetD3DFont().UnlockText());
	}
	return D3D_OK;
}

