//-----------------------------------------------------------------------------
// D3DUtilContextDlg.cpp
//-----------------------------------------------------------------------------
// Dialog box that allows you to set any options for the context mode.
// Does validation and matching so you can see what modes are actually
// available.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// #includes
//-----------------------------------------------------------------------------
#include "stdafx.h"
#include "D3DUtilContextDlg.hpp"
#include "D3DUtilContext.hpp"
#include "D3DUtilEnumeration.hpp"
#include "D3DUtilResource.h"

#include <string>
#include <sstream>

//-----------------------------------------------------------------------------
// INTERNAL VARIABLES FOR THE DIALOG BOXES
//-----------------------------------------------------------------------------
static D3DDEVICE_CREATION_PARAMETERS d3dcp;
static D3DPRESENT_PARAMETERS d3dpp;
static HWND DlgWnd=NULL;
static HWND DevWnd=NULL;

//-----------------------------------------------------------------------------
// FORWARD DECLARATIONS
//-----------------------------------------------------------------------------
// Used to update the context when a dialog item changes
//-----------------------------------------------------------------------------
void InitSetContextDlg();
HRESULT SetContextFromDlg();
void NotifyAdapterChange(bool update=true);
void NotifyDisplayTypeChange(bool update=true);
void NotifyDeviceTypeChange(bool update=true);
void NotifyVertexProcessingModeChange(bool update=true);
void NotifyDisplayModeChange(bool update=true);
void NotifyDepthFormatChange(bool update=true);
void NotifyMultisampleTypeChange(bool update=true);
//-----------------------------------------------------------------------------
// Used to update the dialog when a context item changes
// These are really only called when the dialog box is opened to sync the
// dialog to the current context.
//-----------------------------------------------------------------------------
void SetDlgFromContext();
void SetDlgAdapter(UINT adapter);
void SetDlgDisplayType(bool windowed);
void SetDlgDeviceType(D3DDEVTYPE type);
void SetDlgDisplayMode(UINT width, UINT height, D3DFORMAT format, UINT refresh);
void SetDlgDepthFormat(bool useDepth, D3DFORMAT format);
void SetDlgMultisampleType(D3DMULTISAMPLE_TYPE type);
void SetDlgCreationParameterFlags(const D3DDEVICE_CREATION_PARAMETERS &d3dcp);
void SetDlgPresentParameterFlags(const D3DPRESENT_PARAMETERS &d3dpp);

//-----------------------------------------------------------------------------
// Utility
//-----------------------------------------------------------------------------
const D3DUtilAdapterInfo *GetAdapterInfo();
const D3DUtilDeviceInfo *GetDeviceInfo();
const D3DUtilDisplayModeInfo *GetDisplayModeInfo();
const D3DUtilBackBufferInfo *GetBackBufferInfo();
const D3DUtilDepthInfo *GetDepthInfo();
const D3DUtilMultisampleInfo *GetMultisampleInfo();

//-----------------------------------------------------------------------------
// FORWARD DECLARATIONS OF DLGFUNCS
//-----------------------------------------------------------------------------
BOOL CALLBACK SetContextDlgFunc(
  HWND hwndDlg,  // handle to dialog box
  UINT uMsg,     // message
  WPARAM wParam, // first message parameter
  LPARAM lParam  // second message parameter
);

//-----------------------------------------------------------------------------
// DoContextDlg
//-----------------------------------------------------------------------------
// Allows a context mode to be set using a dialog box.
//-----------------------------------------------------------------------------
HRESULT DoContextDlg(HWND appWnd, HWND devWnd)
{
    DevWnd = devWnd;
	return DialogBox(NULL, MAKEINTRESOURCE(IDD_SETCONTEXTDLG), appWnd, SetContextDlgFunc);
}

//-----------------------------------------------------------------------------
// SetContextDlgFunc
//-----------------------------------------------------------------------------
// Dialog func for SetContextDlg
//-----------------------------------------------------------------------------
BOOL CALLBACK SetContextDlgFunc(
  HWND hwndDlg,  // handle to dialog box
  UINT uMsg,     // message
  WPARAM wParam, // first message parameter
  LPARAM lParam  // second message parameter
)
{
	switch (uMsg) 
	{
	case WM_INITDIALOG:
		DlgWnd = hwndDlg;
		ZeroMemory(&d3dcp, sizeof(d3dcp));
		ZeroMemory(&d3dpp, sizeof(d3dpp));
		InitSetContextDlg();
		SetDlgFromContext();
        return TRUE;

    case WM_COMMAND: 
        switch (LOWORD(wParam)) 
        { 
        case IDOK: 
			wParam = SetContextFromDlg();
            EndDialog(hwndDlg, wParam);
			return true;

        case IDCANCEL: 
            EndDialog(hwndDlg, IDCANCEL);
            return true; 

		case IDC_ADAPTER:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyAdapterChange();
			}
			return true;

		case IDC_WINDOWED:
		case IDC_FULLSCREEN:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDisplayTypeChange();
			}
			return true;

		case IDC_DISPLAYMODE:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				NotifyDisplayModeChange();
			}
			return true;

		case IDC_USEZBUFFER:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDepthFormatChange();
			}
			return true;

		case IDC_DEPTHBUFFER:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				NotifyDepthFormatChange();
			}
			return true;

		case IDC_MULTISAMPLE:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				NotifyMultisampleTypeChange();
			}
			return true;

		case IDC_HAL:
		case IDC_SW:
		case IDC_REF:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDeviceTypeChange();
			}
			return true;

		case IDC_DISCARD:
		case IDC_FLIP:
		case IDC_COPY:
		case IDC_COPYVSYNC:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDepthFormatChange();
			}
			return true;
		case IDC_HARDWARE:
		case IDC_SOFTWARE:
		case IDC_MIXED:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyVertexProcessingModeChange();
			}
        } 
		return TRUE;
    } 

    return FALSE; 
}

//-----------------------------------------------------------------------------
// InitSetContextDlg
//-----------------------------------------------------------------------------
// Initializes the dialog box with appropriate details.
//-----------------------------------------------------------------------------
void InitSetContextDlg()
{
	unsigned int i;
	for (i=0;i<D3DUtilEnumeration::GetAdapterList().size(); i++)
	{
		SendDlgItemMessage(DlgWnd, IDC_ADAPTER, CB_ADDSTRING, 0, (LPARAM)D3DUtilEnumeration::GetAdapterList()[i].Description);
	}
	SetFocus(GetDlgItem(DlgWnd, IDC_ADAPTER));
}

//-----------------------------------------------------------------------------
// SetDlgFromContext
//-----------------------------------------------------------------------------
// Set the dialog box to reflect the current context settings
//-----------------------------------------------------------------------------
void SetDlgFromContext()
{
	if (!D3DUtilContext::Instance()->IsActive())
	{
		// The context wasn't active so choose some default settings
		d3dcp.AdapterOrdinal = D3DADAPTER_DEFAULT;
		d3dcp.DeviceType = D3DDEVTYPE_HAL;
		d3dcp.BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
		d3dcp.hFocusWindow = GetParent(DlgWnd);
		d3dpp.BackBufferCount = 0;
		d3dpp.EnableAutoDepthStencil = TRUE;
		d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		d3dpp.hDeviceWindow = DevWnd;
		d3dpp.Flags = 0;
		// Set the flags manually since the dialog don't update them.
		SetDlgCreationParameterFlags(d3dcp);
		SetDlgPresentParameterFlags(d3dpp);
		SendDlgItemMessage(DlgWnd, IDC_ADAPTER, CB_SETCURSEL, d3dcp.AdapterOrdinal, 0);
		// Select hardware TnL if possible
		if (IsWindowEnabled(GetDlgItem(DlgWnd, IDC_HARDWARE)))
		{
			d3dcp.BehaviorFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;
		}
		// Force updates of the parameters since we don't have defaults because the context was not active
		NotifyAdapterChange(true);
	} else
	{
		// Update the dialog to reflect the context settings
		d3dcp = D3DUtilContext::Instance()->GetCreationParameters();
		d3dpp = D3DUtilContext::Instance()->GetPresentParameters();
	}

	SetDlgCreationParameterFlags(d3dcp);
	SetDlgPresentParameterFlags(d3dpp);
	SetDlgAdapter(d3dcp.AdapterOrdinal);
	SetDlgDeviceType(d3dcp.DeviceType);
	SetDlgDisplayType(d3dpp.Windowed>0);
	if (!d3dpp.Windowed)
	{
		SetDlgDisplayMode(d3dpp.BackBufferWidth, d3dpp.BackBufferHeight, d3dpp.BackBufferFormat, d3dpp.FullScreen_RefreshRateInHz);
	}
	SetDlgDepthFormat(d3dpp.EnableAutoDepthStencil != 0, d3dpp.AutoDepthStencilFormat);
	SetDlgMultisampleType(d3dpp.MultiSampleType);
}

//-----------------------------------------------------------------------------
// SetDlgAdapter
//-----------------------------------------------------------------------------
// Set the dialog box adapter to reflect the current context adapter
//-----------------------------------------------------------------------------
void SetDlgAdapter(UINT adapter)
{
	if (adapter >= D3DUtilEnumeration::GetAdapterList().size()) 
	{
		AUXTRACE("!!! Adapter out of range: %i", adapter);
		return;
	}
	SendDlgItemMessage(DlgWnd, IDC_ADAPTER, CB_SETCURSEL, adapter, 0);
	NotifyAdapterChange(false);
}

//-----------------------------------------------------------------------------
// SetDlgDeviceType
//-----------------------------------------------------------------------------
// Set the dialog box device type to reflect the current context device type
//-----------------------------------------------------------------------------
void SetDlgDeviceType(D3DDEVTYPE type)
{
	SendDlgItemMessage(DlgWnd, IDC_HAL, BM_SETCHECK, (type==D3DDEVTYPE_HAL?BST_CHECKED:BST_UNCHECKED), 0);
	SendDlgItemMessage(DlgWnd, IDC_SW,  BM_SETCHECK, (type==D3DDEVTYPE_SW ?BST_CHECKED:BST_UNCHECKED), 0);
	SendDlgItemMessage(DlgWnd, IDC_REF, BM_SETCHECK, (type==D3DDEVTYPE_REF?BST_CHECKED:BST_UNCHECKED), 0);

	NotifyDeviceTypeChange(false);
}

//-----------------------------------------------------------------------------
// SetDlgDisplayType
//-----------------------------------------------------------------------------
// Set the dialog box display type to reflect the current context display type
//-----------------------------------------------------------------------------
void SetDlgDisplayType(bool windowed)
{
	if (windowed)
	{
		SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_SETCHECK, BST_CHECKED, 0);
		SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN, BM_SETCHECK, BST_UNCHECKED, 0);
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_SETCHECK, BST_UNCHECKED, 0);
		SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN, BM_SETCHECK, BST_CHECKED, 0);
	}
	NotifyDisplayTypeChange(false);
}

//-----------------------------------------------------------------------------
// SetDlgDisplayMode
//-----------------------------------------------------------------------------
// Set the dialog box display mode to reflect the current context display mode
//-----------------------------------------------------------------------------
void SetDlgDisplayMode(UINT width, UINT height, D3DFORMAT format, UINT refresh)
{
	int i;
	int count = SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETCOUNT, 0, 0);
	for (i=0; i<count; i++)
	{
		UINT dlgWidth, dlgHeight, dlgRefresh;
		D3DFORMAT dlgFormat;
		char buffer[32];
		std::string tempBuffer;
		SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETLBTEXT, i, (LPARAM)&buffer);
		std::istringstream istream(buffer);
		istream >> dlgWidth >> tempBuffer >> dlgHeight >> dlgRefresh >> tempBuffer;
		istream >> tempBuffer;
		dlgFormat = StringToD3DFORMAT(tempBuffer.c_str());
		if (dlgWidth == width && dlgHeight == height && dlgFormat == format && (dlgRefresh == refresh || refresh == 0))
		{
			SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, i, 0);
			break;
		}
	}

	// error check
	if (i == count) 
	{
		AUXTRACE("D3DUtilContextDlg: display mode %d x %d %s not found.", width, height, EnumToString(format));
		SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, -1, 0);
	}

	NotifyDisplayModeChange(false);
}

//-----------------------------------------------------------------------------
// SetDlgDepthFormat
//-----------------------------------------------------------------------------
// Set the dialog box depth format to reflect the current context depth format
//-----------------------------------------------------------------------------
void SetDlgDepthFormat(bool useDepth, D3DFORMAT format)
{
	int i;
	int count = SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_GETCOUNT, 0, 0);
	for (i=0; i<count; i++)
	{
		D3DFORMAT dlgFormat;
		std::string buffer(32,'\0');
		SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_GETLBTEXT, i, (LPARAM)&buffer[0]);
		dlgFormat = StringToD3DFORMAT(buffer.c_str());
		if (dlgFormat == format)
		{
			SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_SETCURSEL, i, 0);
			break;
		}
	}

	// error check
	if (i == count || !useDepth) 
	{
		SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_SETCURSEL, -1, 0);
	}

	NotifyDepthFormatChange(false);
}

//-----------------------------------------------------------------------------
// SetDlgMultisampleType
//-----------------------------------------------------------------------------
// Set the dialog box multisample type to reflect the current context multisample type
//-----------------------------------------------------------------------------
void SetDlgMultisampleType(D3DMULTISAMPLE_TYPE type)
{
	int i;
	int count = SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_GETCOUNT, 0, 0);
	for (i=0; i<count; i++)
	{
		D3DMULTISAMPLE_TYPE dlgType;
		std::string buffer(32,'\0');
		SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_GETLBTEXT, i, (LPARAM)&buffer[0]);
		dlgType = StringToD3DMULTISAMPLE_TYPE(buffer.c_str());
		if (dlgType == type)
		{
			SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_SETCURSEL, i, 0);
			break;
		}
	}

	// error check
	if (i == count) 
	{
		AUXTRACE("D3DUtilContextDlg: Multisample type %s not found.", EnumToString(type));
		SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_SETCURSEL, -1, 0);
	}

	NotifyMultisampleTypeChange(false);
}

//-----------------------------------------------------------------------------
// SetDlgCreationParameterFlags
//-----------------------------------------------------------------------------
// Set additional flags in the dialog box from the context
//-----------------------------------------------------------------------------
void SetDlgCreationParameterFlags(const D3DDEVICE_CREATION_PARAMETERS &d3dcp)
{
	SendDlgItemMessage(DlgWnd, IDC_SOFTWARE, BM_SETCHECK, BST_UNCHECKED, 0);	
	SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_SETCHECK, BST_UNCHECKED, 0);	
	SendDlgItemMessage(DlgWnd, IDC_MIXED,    BM_SETCHECK, BST_UNCHECKED, 0);	
	if (d3dcp.BehaviorFlags & D3DCREATE_SOFTWARE_VERTEXPROCESSING)
	{
		SendDlgItemMessage(DlgWnd, IDC_SOFTWARE, BM_SETCHECK, BST_CHECKED, 0);	
	} else if (d3dcp.BehaviorFlags & D3DCREATE_HARDWARE_VERTEXPROCESSING)
	{
		SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_SETCHECK, BST_CHECKED, 0);	
	} else if (d3dcp.BehaviorFlags & D3DCREATE_MIXED_VERTEXPROCESSING)
	{
		SendDlgItemMessage(DlgWnd, IDC_MIXED,    BM_SETCHECK, BST_CHECKED, 0);	
	}
	
	if (d3dcp.BehaviorFlags & D3DCREATE_FPU_PRESERVE)
	{
		SendDlgItemMessage(DlgWnd, IDC_FPUPRESERVE, BM_SETCHECK, BST_CHECKED, 0);	
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_FPUPRESERVE, BM_SETCHECK, BST_UNCHECKED, 0);	
	}
	
	if (d3dcp.BehaviorFlags & D3DCREATE_PUREDEVICE)
	{
		SendDlgItemMessage(DlgWnd, IDC_PUREDEVICE, BM_SETCHECK, BST_CHECKED, 0);	
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_PUREDEVICE, BM_SETCHECK, BST_UNCHECKED, 0);	
	}
	
	if (d3dcp.BehaviorFlags & D3DCREATE_MULTITHREADED)
	{
		SendDlgItemMessage(DlgWnd, IDC_MULTITHREADED, BM_SETCHECK, BST_CHECKED, 0);	
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_MULTITHREADED, BM_SETCHECK, BST_UNCHECKED, 0);	
	}
}

//-----------------------------------------------------------------------------
// SetDlgPresentParameterFlags
//-----------------------------------------------------------------------------
// Set additional flags in the dialog box from the context
//-----------------------------------------------------------------------------
void SetDlgPresentParameterFlags(const D3DPRESENT_PARAMETERS &d3dpp)
{
	SendDlgItemMessage(DlgWnd, IDC_DISCARD,   BM_SETCHECK, BST_UNCHECKED, 0);	
	SendDlgItemMessage(DlgWnd, IDC_FLIP,      BM_SETCHECK, BST_UNCHECKED, 0);	
	SendDlgItemMessage(DlgWnd, IDC_COPY,      BM_SETCHECK, BST_UNCHECKED, 0);	
	SendDlgItemMessage(DlgWnd, IDC_COPYVSYNC, BM_SETCHECK, BST_UNCHECKED, 0);	
	if (d3dpp.SwapEffect == D3DSWAPEFFECT_DISCARD)
	{
		SendDlgItemMessage(DlgWnd, IDC_DISCARD,   BM_SETCHECK, BST_CHECKED, 0);	
	} else if (d3dpp.SwapEffect == D3DSWAPEFFECT_FLIP)
	{
		SendDlgItemMessage(DlgWnd, IDC_FLIP,      BM_SETCHECK, BST_CHECKED, 0);	
	} else if (d3dpp.SwapEffect == D3DSWAPEFFECT_COPY)
	{
		SendDlgItemMessage(DlgWnd, IDC_COPY,      BM_SETCHECK, BST_CHECKED, 0);	
	} else if (d3dpp.SwapEffect == D3DSWAPEFFECT_COPY_VSYNC)
	{
		SendDlgItemMessage(DlgWnd, IDC_COPYVSYNC, BM_SETCHECK, BST_CHECKED, 0);	
	}

	if (d3dpp.Flags & D3DPRESENTFLAG_LOCKABLE_BACKBUFFER)
	{
		SendDlgItemMessage(DlgWnd, IDC_LOCKABLE, BM_SETCHECK, BST_CHECKED, 0);	
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_LOCKABLE, BM_SETCHECK, BST_UNCHECKED, 0);	
	}

	SetDlgItemInt(DlgWnd, IDC_BACKBUFFERCOUNT, d3dpp.BackBufferCount, FALSE);

	if (d3dpp.EnableAutoDepthStencil)
	{
		SendDlgItemMessage(DlgWnd, IDC_USEZBUFFER, BM_SETCHECK, BST_CHECKED, 0);	
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_USEZBUFFER, BM_SETCHECK, BST_UNCHECKED, 0);	
	}
}

//-----------------------------------------------------------------------------
// SetContextFromDlg
//-----------------------------------------------------------------------------
// Set the context to reflect the current dialog box settings
//-----------------------------------------------------------------------------
HRESULT SetContextFromDlg()
{
	// Get the flags first, since the dialog doesn't update them when they change.
	if (SendDlgItemMessage(DlgWnd, IDC_SOFTWARE, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	} else if (SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.BehaviorFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;
	} else if (SendDlgItemMessage(DlgWnd, IDC_MIXED,    BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.BehaviorFlags = D3DCREATE_MIXED_VERTEXPROCESSING;
	}
	if (SendDlgItemMessage(DlgWnd, IDC_FPUPRESERVE, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.BehaviorFlags |= D3DCREATE_FPU_PRESERVE;
	}
	if (SendDlgItemMessage(DlgWnd, IDC_MULTITHREADED, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.BehaviorFlags |= D3DCREATE_MULTITHREADED;
	}
	if (SendDlgItemMessage(DlgWnd, IDC_PUREDEVICE, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.BehaviorFlags |= D3DCREATE_PUREDEVICE;
	}

	if (SendDlgItemMessage(DlgWnd, IDC_DISCARD, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	} else if (SendDlgItemMessage(DlgWnd, IDC_FLIP, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dpp.SwapEffect = D3DSWAPEFFECT_FLIP;
	} else if (SendDlgItemMessage(DlgWnd, IDC_COPY, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dpp.SwapEffect = D3DSWAPEFFECT_COPY;
	} else if (SendDlgItemMessage(DlgWnd, IDC_COPYVSYNC, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dpp.SwapEffect = D3DSWAPEFFECT_COPY_VSYNC;
	}
	d3dpp.BackBufferCount = GetDlgItemInt(DlgWnd, IDC_BACKBUFFERCOUNT, NULL, FALSE);
	if (SendDlgItemMessage(DlgWnd, IDC_LOCKABLE, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dpp.Flags = D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
	} else
	{
		d3dpp.Flags = 0;
	}

	if (d3dpp.Windowed)
	{
		d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
	} else
	{
		d3dpp.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
	}

	if (D3DUtilContext::Instance()->IsActive())
	{
		D3DDEVICE_CREATION_PARAMETERS contextd3dcp = D3DUtilContext::Instance()->GetCreationParameters();

		if (memcmp(&d3dcp, &contextd3dcp, sizeof(d3dcp)) == 0)
		{
			FAIL_RETURN(D3DUtilContext::Instance()->ResetDevice(d3dpp));
            return D3D_OK;
		}
	}
	FAIL_RETURN(D3DUtilContext::Instance()->InitDevice(d3dcp, d3dpp));
	return D3D_OK;
}

//-----------------------------------------------------------------------------
// NotifyAdapterChange
//-----------------------------------------------------------------------------
// Set the context adapter to reflect the current dialog box adapter
//-----------------------------------------------------------------------------
void NotifyAdapterChange(bool update)
{
	unsigned int i, adapter;
	char buffer[64];
	
	adapter = SendDlgItemMessage(DlgWnd, IDC_ADAPTER, CB_GETCURSEL, 0, 0);

	// update the context
	if (update) d3dcp.AdapterOrdinal = adapter;

	// Set the description text box
	SetDlgItemText(DlgWnd, IDC_DRIVER, D3DUtilEnumeration::GetAdapterList()[adapter].Driver);

	// Clear the old display mode list
	SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_RESETCONTENT, 0, 0);

	// Get a reference to the adapter's list of display modes
	const D3DUtilDisplayModeInfoList &displayModeList = D3DUtilEnumeration::GetAdapterList()[adapter].DisplayModes;

	// Fill in the combo box
	for (i=0; i<displayModeList.size(); i++)
	{
		sprintf(buffer, "%d x %d %d Hz %s", displayModeList[i].Mode.Width, displayModeList[i].Mode.Height, displayModeList[i].Mode.RefreshRate, EnumToString(displayModeList[i].Mode.Format));
		SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_ADDSTRING, 0, (LPARAM)buffer);
	}


	// Set up the device type radio group.
	const D3DUtilDeviceInfoList &deviceList = D3DUtilEnumeration::GetAdapterList()[adapter].Devices;
	EnableWindow(GetDlgItem(DlgWnd, IDC_HAL), FALSE); SendDlgItemMessage(DlgWnd, IDC_HAL, BM_SETCHECK, BST_UNCHECKED, 0);
	EnableWindow(GetDlgItem(DlgWnd, IDC_SW),  FALSE); SendDlgItemMessage(DlgWnd, IDC_SW,  BM_SETCHECK, BST_UNCHECKED, 0);
	EnableWindow(GetDlgItem(DlgWnd, IDC_REF), FALSE); SendDlgItemMessage(DlgWnd, IDC_REF, BM_SETCHECK, BST_UNCHECKED, 0);
	bool found=false;
	for (i=0; i<deviceList.size(); i++)
	{
		int ID;
		switch (deviceList[i].DeviceType)
		{
		case D3DDEVTYPE_HAL: ID = IDC_HAL; break;
		case D3DDEVTYPE_SW:  ID = IDC_SW;  break;
		case D3DDEVTYPE_REF: ID = IDC_REF; break;
		}

		EnableWindow(GetDlgItem(DlgWnd, ID), TRUE);
		if (!found)
		{
			SendDlgItemMessage(DlgWnd, ID, BM_SETCHECK, BST_CHECKED, 0);
			found = true;
		}
	}

	// Notify the dialog of the new options
	NotifyDeviceTypeChange(update);
}

//-----------------------------------------------------------------------------
// NotifyDeviceTypeChange
//-----------------------------------------------------------------------------
// Set the context device type to reflect the current dialog box device type
//-----------------------------------------------------------------------------
void NotifyDeviceTypeChange(bool update)
{
	// update the context to match the dialog settings
	if (update) 
	{
		if (SendDlgItemMessage(DlgWnd, IDC_HAL, BM_GETCHECK, 0, 0) == BST_CHECKED)
		{
			d3dcp.DeviceType = D3DDEVTYPE_HAL;
		} else if (SendDlgItemMessage(DlgWnd, IDC_SW, BM_GETCHECK, 0, 0) == BST_CHECKED)
		{
			d3dcp.DeviceType = D3DDEVTYPE_SW;
		} else if (SendDlgItemMessage(DlgWnd, IDC_REF, BM_GETCHECK, 0, 0) == BST_CHECKED)
		{
			d3dcp.DeviceType = D3DDEVTYPE_REF;
		}
	}

	const D3DUtilDeviceInfo *deviceInfo = GetDeviceInfo();
	// See what vertex processing options are available
	// Software is always available
	if (deviceInfo && deviceInfo->DeviceCaps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT)
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_HARDWARE), TRUE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_MIXED), TRUE);
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_SOFTWARE, BM_SETCHECK, BST_CHECKED, 0);
		EnableWindow(GetDlgItem(DlgWnd, IDC_HARDWARE), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_SETCHECK, BST_UNCHECKED, 0);
		EnableWindow(GetDlgItem(DlgWnd, IDC_MIXED), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_MIXED, BM_SETCHECK, BST_UNCHECKED, 0);
	}

	// See if we can enable windowed mode
	if (deviceInfo && deviceInfo->DeviceCaps.Caps2 & D3DCAPS2_CANRENDERWINDOWED)
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_WINDOWED), TRUE);
		SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_SETCHECK, BST_CHECKED, 0);
		SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN, BM_SETCHECK, BST_UNCHECKED, 0);
	} else
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_WINDOWED), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_SETCHECK, BST_UNCHECKED, 0);
		SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN, BM_SETCHECK, BST_CHECKED, 0);
	}
	// Notify the dialog of the new options
	NotifyVertexProcessingModeChange(update);
	NotifyDisplayTypeChange(update);
}

//-----------------------------------------------------------------------------
// NotifyVertexProcessingModeChange
//-----------------------------------------------------------------------------
// Set the context VP mode to reflect the current dialog box VP mode.
//-----------------------------------------------------------------------------
void NotifyVertexProcessingModeChange(bool update)
{
	const D3DUtilDeviceInfo *deviceInfo = GetDeviceInfo();
	// See if we have a pure device
	if (deviceInfo && (deviceInfo->DeviceCaps.DevCaps & D3DDEVCAPS_PUREDEVICE) && SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_PUREDEVICE), TRUE);
	} else
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_PUREDEVICE), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_PUREDEVICE, BM_SETCHECK, BST_UNCHECKED, 0);
	}
}

//-----------------------------------------------------------------------------
// NotifyDisplayTypeChange
//-----------------------------------------------------------------------------
// Set the context display type to reflect the current dialog box display type
//-----------------------------------------------------------------------------
void NotifyDisplayTypeChange(bool update)
{
	// update the context to match the dialog settings
	if (SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		if (update) d3dpp.Windowed = TRUE;
		EnableWindow(GetDlgItem(DlgWnd, IDC_DISPLAYMODE), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, -1, 0);
	} else
	{
		if (update) d3dpp.Windowed = FALSE;
		EnableWindow(GetDlgItem(DlgWnd, IDC_DISPLAYMODE), TRUE);
//		if (SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETCURSEL, 0, 0) == -1)
//		{
//			SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, 0, 0);
//		}
        // Try and find a 640x480x32 device
        int size = SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETCOUNT, 0, 0);
        int lastNdx = 0;
        for (int i=0; i<size; ++i)
        {
		    UINT dlgWidth, dlgHeight, dlgRefresh;
		    D3DFORMAT dlgFormat;
    		char buffer[32];
	    	std::string tempBuffer;
		    SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETLBTEXT, i, (LPARAM)&buffer);
        	std::istringstream istream(buffer);
		    istream >> dlgWidth >> tempBuffer >> dlgHeight >> dlgRefresh >> tempBuffer;
    		dlgFormat = StringToD3DFORMAT(tempBuffer.c_str());
            if (dlgWidth == 640 && dlgHeight == 480) lastNdx = i;
        }
		SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, lastNdx, 0);

	}

	// Notify the dialog of the new options
	NotifyDisplayModeChange(update);
}

//-----------------------------------------------------------------------------
// NotifyDisplayModeChange
//-----------------------------------------------------------------------------
// Set the context display mode to reflect the current dialog box display mode
//-----------------------------------------------------------------------------
void NotifyDisplayModeChange(bool update)
{
	// update the context to match the dialog settings
	if (update)
	{
		if (!d3dpp.Windowed)
		{
			const D3DUtilDisplayModeInfo *displayModeInfo = GetDisplayModeInfo();
			if (displayModeInfo)
			{
				d3dpp.BackBufferWidth  = displayModeInfo->Mode.Width;
				d3dpp.BackBufferHeight = displayModeInfo->Mode.Height;
				d3dpp.BackBufferFormat = displayModeInfo->Mode.Format;
				d3dpp.FullScreen_RefreshRateInHz = displayModeInfo->Mode.RefreshRate;
			} else
			{
				d3dpp.BackBufferWidth  = 0;
				d3dpp.BackBufferHeight = 0;
				d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
				d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
			}
		} else
		{
			d3dpp.BackBufferWidth  = 0;
			d3dpp.BackBufferHeight = 0;
			d3dpp.BackBufferFormat = D3DUtilContext::Instance()->GetWindowedFormat();
			d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
		}
	}

	// Set up the new depth buffer options
	SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_RESETCONTENT, 0, 0);

	const D3DUtilBackBufferInfo *backBuffers = GetBackBufferInfo();
	const D3DUtilDepthInfoList *depthList = backBuffers ? &backBuffers->DepthBuffers : NULL;
	if (depthList)
	{
		unsigned int depthIndex;
		for (depthIndex=0; depthIndex<depthList->size(); depthIndex++)
		{
			if (depthList->at(depthIndex).Format != D3DFMT_UNKNOWN)
			{
				SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_ADDSTRING, 0, (LPARAM)EnumToString(depthList->at(depthIndex).Format));
			}
		}
	}
	if (SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_GETCOUNT, 0, 0) == 0)
	{
		SendDlgItemMessage(DlgWnd, IDC_USEZBUFFER, BM_SETCHECK, BST_UNCHECKED, 0);
		EnableWindow(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER), FALSE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_USEZBUFFER), FALSE);
	} else
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_USEZBUFFER), TRUE);
		if (SendDlgItemMessage(DlgWnd, IDC_USEZBUFFER, BM_GETCHECK, 0, 0) == BST_CHECKED)
		{
			EnableWindow(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER), TRUE);
			SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_SETCURSEL, 0, 0);
		} else
		{
			EnableWindow(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER), FALSE);
			SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_SETCURSEL, 0, 0);
		}
	}

	// Notify the dialog of the new options
	NotifyDepthFormatChange(update);
}

//-----------------------------------------------------------------------------
// NotifyDepthFormatChange
//-----------------------------------------------------------------------------
// Set the context depth format to reflect the current dialog box depth format
//-----------------------------------------------------------------------------
void NotifyDepthFormatChange(bool update)
{
	// update the context to match the dialog settings
	if (SendDlgItemMessage(DlgWnd, IDC_USEZBUFFER, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER), TRUE);
		if (SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_GETCURSEL, 0, 0) == -1)
		{
			SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_SETCURSEL, 0, 0);
		}
		if (update)
		{
			d3dpp.EnableAutoDepthStencil = TRUE;
			const D3DUtilDepthInfo* depthInfo = GetDepthInfo();
			d3dpp.AutoDepthStencilFormat = depthInfo ? depthInfo->Format : D3DFMT_UNKNOWN;
		}
	} else
	{
		if (update)
		{
			d3dpp.EnableAutoDepthStencil = FALSE;
		}
		EnableWindow(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER), FALSE);
	}
	
	// Set up the new multisample options
	SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_RESETCONTENT, 0, 0);

	const D3DUtilMultisampleInfoList *multisampleList;
	const D3DUtilDepthInfo* depthInfo = GetDepthInfo();
	if (depthInfo)
	{
		if (d3dpp.Windowed)
		{
			multisampleList = &depthInfo->MultisampleWindowedTypes;
		} else
		{
			multisampleList = &depthInfo->MultisampleFullscreenTypes;
		}
	} else
	{
		multisampleList = NULL;
	}

	if (multisampleList)
	{
		unsigned int multisampleIndex;
		for (multisampleIndex=0; multisampleIndex<multisampleList->size(); multisampleIndex++)
		{
			SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_ADDSTRING, 0, (LPARAM)EnumToString((*multisampleList)[multisampleIndex].Type));
		}
	}
	if (SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_GETCOUNT, 0, 0) <= 1 ||
		SendDlgItemMessage(DlgWnd, IDC_DISCARD, BM_GETCHECK, 0, 0) != BST_CHECKED)
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_MULTISAMPLE), FALSE);
	} else
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_MULTISAMPLE), TRUE);
	}
	SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_SETCURSEL, 0, 0);

	// make the dlg update its contents to reflect the change
	NotifyMultisampleTypeChange(update);
}

//-----------------------------------------------------------------------------
// NotifyMultisampleTypeChange
//-----------------------------------------------------------------------------
// Set the context multisample type to reflect the current dialog box multisample type
//-----------------------------------------------------------------------------
void NotifyMultisampleTypeChange(bool update)
{
	if (update)
	{
		const D3DUtilMultisampleInfo* multisampleInfo = GetMultisampleInfo();
		d3dpp.MultiSampleType = multisampleInfo ? multisampleInfo->Type : D3DMULTISAMPLE_NONE;
	}
}

//-----------------------------------------------------------------------------
// GetAdapterInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilAdapterInfo *GetAdapterInfo()
{
	if (D3DUtilEnumeration::GetAdapterList().size() == 0) return NULL;
	return &D3DUtilEnumeration::GetAdapterList()[d3dcp.AdapterOrdinal];
}

//-----------------------------------------------------------------------------
// GetDeviceInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilDeviceInfo *GetDeviceInfo()
{
	const D3DUtilAdapterInfo* adapterInfo = GetAdapterInfo();
	if (adapterInfo == NULL) return NULL;
	const D3DUtilDeviceInfoList &deviceList = adapterInfo->Devices;

	// look for the device type in our enumeration list
	unsigned int deviceIndex;
	for (deviceIndex=0; deviceIndex<deviceList.size(); deviceIndex++)
	{
		if (deviceList[deviceIndex].DeviceType == d3dcp.DeviceType) break;
	}
	if (deviceIndex == deviceList.size())
	{
		AUXTRACE("D3DUtilContextDlg: Couldn't find device type %s in list", EnumToString(d3dcp.DeviceType));
		return NULL;
	}

	return &deviceList[deviceIndex];
}

//-----------------------------------------------------------------------------
// GetDisplayModeInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilDisplayModeInfo *GetDisplayModeInfo()
{
	const D3DUtilAdapterInfo* adapterInfo = GetAdapterInfo();
	if (adapterInfo == NULL) return NULL;
	const D3DUtilDisplayModeInfoList &displayModeList = adapterInfo->DisplayModes;
	
	int displayModeIndex = SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETCURSEL, 0, 0);

	if (displayModeIndex == -1)
	{
		AUXTRACE("D3DUtilContextDlg: Called GetDisplayModeInfo with current selection undefined");
		displayModeIndex = 0;
		return NULL;
	}

	return &displayModeList[displayModeIndex];
}

//-----------------------------------------------------------------------------
// GetBackBufferInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilBackBufferInfo *GetBackBufferInfo()
{
	const D3DUtilDeviceInfo* deviceInfo = GetDeviceInfo();
	if (deviceInfo == NULL) return NULL;
	const D3DUtilBackBufferInfoList &backBufferList = deviceInfo->BackBuffers;

	unsigned int backBufferIndex;
	for (backBufferIndex=0; backBufferIndex<backBufferList.size(); backBufferIndex++)
	{
		if (backBufferList[backBufferIndex].Format == d3dpp.BackBufferFormat) break;
	}
	if (backBufferIndex == backBufferList.size())
	{
		AUXTRACE("D3DUtilContextDlg: Couldn't find back buffer format %s in list", EnumToString(d3dpp.BackBufferFormat));
		return NULL;
	}
	
	return &backBufferList[backBufferIndex];
}

//-----------------------------------------------------------------------------
// GetDepthInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilDepthInfo *GetDepthInfo()
{
	const D3DUtilBackBufferInfo* backBufferInfo = GetBackBufferInfo();
	if (!backBufferInfo) return NULL;
	const D3DUtilDepthInfoList &depthList = backBufferInfo->DepthBuffers;
	if (depthList.size() == 0) return NULL;

	int depthIndex;
	if (SendDlgItemMessage(DlgWnd, IDC_USEZBUFFER, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		depthIndex = SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER, CB_GETCURSEL, 0, 0);
	} else
	{
		depthIndex = depthList.size()-1;
	}

	if (depthIndex == -1)
	{
		AUXTRACE("D3DUtilContextDlg: Called GetDepthInfo with current selection undefined");
		depthIndex = 0;
		return NULL;
	}

	return &depthList[depthIndex];
}

//-----------------------------------------------------------------------------
// GetMultisampleInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilMultisampleInfo *GetMultisampleInfo()
{
	const D3DUtilDepthInfo* depthInfo = GetDepthInfo();
	if (depthInfo == NULL) return NULL;
	const D3DUtilMultisampleInfoList *multisampleList;
	if (d3dpp.Windowed)
	{
		multisampleList = &depthInfo->MultisampleWindowedTypes;
	} else
	{
		multisampleList = &depthInfo->MultisampleFullscreenTypes;
	}

	int multisampleIndex = SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE, CB_GETCURSEL, 0, 0);

	if (multisampleIndex == -1)
	{
		AUXTRACE("D3DUtilContextDlg: Called GetMultisampleInfo with current selection undefined");
		multisampleIndex = 0;
		return NULL;
	}

	return &multisampleList->at(multisampleIndex);
}
