//-----------------------------------------------------------------------------
// D3DUtilEnumeration.hpp
//-----------------------------------------------------------------------------
// Enumerates and creates a list of all the devices and modes on the system.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// #includes
//-----------------------------------------------------------------------------
#include "stdafx.h"
#include "D3DUtilEnumeration.hpp"
#include <algorithm>

//-----------------------------------------------------------------------------
// statics
//-----------------------------------------------------------------------------
D3DUtilAdapterInfoList D3DUtilEnumeration::AdapterList;

//-----------------------------------------------------------------------------
// GetAdapterList
//-----------------------------------------------------------------------------
// Returns the list of adapters and display modes
//-----------------------------------------------------------------------------
const D3DUtilAdapterInfoList &D3DUtilEnumeration::GetAdapterList()
{
	if (AdapterList.size() > 0) return AdapterList;

	IDirect3D8 *pD3D = NULL;
	pD3D = Direct3DCreate8(D3D_SDK_VERSION);
	if (pD3D == NULL) 
    {
        AUXTRACE("D3DUtilEnumeration: Direct3DCreate8 Failed");
    }

	// build the list of adapters
	// for each adapter
		// build the list of display modes
		// sort the list of display modes
		// build the list of devices
		// for each device
			// build the list of back buffer formats (from enumadaptermodes formats)
			// for each back buffer format
				// build the list of depth buffer formats
				// for each depth buffer format
					// build the list of multisample types
				//
			//
		//
	//
	unsigned int adapter;
	BuildAdapterList(pD3D);
	for (adapter=0; adapter<AdapterList.size(); adapter++)
	{
		D3DUtilAdapterInfo &adapterInfo = AdapterList[adapter];
		std::vector<D3DFORMAT> displayFormats;
		BuildDisplayModeList(adapter, displayFormats, pD3D);
		std::sort(adapterInfo.DisplayModes.begin(), adapterInfo.DisplayModes.end());
		BuildDeviceList(adapter, pD3D);
		unsigned int deviceIndex;
		for (deviceIndex=0; deviceIndex<adapterInfo.Devices.size(); deviceIndex++)
		{
			D3DUtilDeviceInfo &deviceInfo = adapterInfo.Devices[deviceIndex];
			BuildBackBufferList(adapter, deviceInfo, displayFormats, pD3D);
			unsigned int formatIndex;
			for (formatIndex=0; formatIndex<deviceInfo.BackBuffers.size(); formatIndex++)
			{
				D3DUtilBackBufferInfo &backBufferInfo = deviceInfo.BackBuffers[formatIndex];
				BuildDepthBufferList(adapter, deviceInfo.DeviceType, backBufferInfo, pD3D);
				unsigned int depthIndex;
				for (depthIndex=0; depthIndex<backBufferInfo.DepthBuffers.size(); depthIndex++)
				{
					D3DUtilDepthInfo &depthInfo = backBufferInfo.DepthBuffers[depthIndex];
					BuildMultisampleList(adapter, deviceInfo.DeviceType, backBufferInfo.Format, depthInfo, pD3D);
				}
			}
		}
	}

	SAFE_RELEASE(pD3D);
	return AdapterList;
}

//-----------------------------------------------------------------------------
// BuildAdapterList
//-----------------------------------------------------------------------------
// Fills the list with the available adapters
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildAdapterList(IDirect3D8 *pD3D)
{
	unsigned int adapter;
	for (adapter=0; adapter<pD3D->GetAdapterCount(); adapter++)
	{
		AdapterList.push_back(D3DUtilAdapterInfo());
		D3DUtilAdapterInfo &adapterInfo = AdapterList.back();
		D3DADAPTER_IDENTIFIER8 adapterIdentifier;
		pD3D->GetAdapterIdentifier(adapter, D3DENUM_NO_WHQL_LEVEL, &adapterIdentifier);
		sprintf(adapterInfo.Driver, "%s", adapterIdentifier.Driver);
		sprintf(adapterInfo.Description, "%s", adapterIdentifier.Description);
	}
}

//-----------------------------------------------------------------------------
// BuildDisplayModeList
//-----------------------------------------------------------------------------
// Fills the list with the available adapters.
// Side effect: fill the given vector with the list of possible formats
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildDisplayModeList(int adapter, std::vector<D3DFORMAT> &displayFormats, IDirect3D8 *pD3D)
{
	D3DUtilAdapterInfo &adapterInfo = AdapterList[adapter];
	unsigned int modeIndex;
	for (modeIndex=0; modeIndex<pD3D->GetAdapterModeCount(adapter); modeIndex++)
	{
        // !!! WRH Don't allow modes over 80 Hz just to keep the list smaller
        D3DUtilDisplayModeInfo modeInfo;
		pD3D->EnumAdapterModes(adapter, modeIndex, &modeInfo.Mode);
        if (modeInfo.Mode.RefreshRate < 80)
        {
		    adapterInfo.DisplayModes.push_back(modeInfo);
		    if (std::find(displayFormats.begin(), displayFormats.end(), modeInfo.Mode.Format) == displayFormats.end())
		    {
			    displayFormats.push_back(modeInfo.Mode.Format);
		    }
        }
	}
}

//-----------------------------------------------------------------------------
// BuildDeviceList
//-----------------------------------------------------------------------------
// Fills the list with the available devices.
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildDeviceList(int adapter, IDirect3D8 *pD3D)
{
	D3DUtilAdapterInfo &adapterInfo = AdapterList[adapter];
	int deviceIndex;
	for (deviceIndex=0; deviceIndex<DeviceTypeCount; deviceIndex++)
	{
		D3DUtilDeviceInfo deviceInfo;
		deviceInfo.DeviceType = DeviceTypes[deviceIndex];
		if (FAILED(pD3D->GetDeviceCaps(adapter, deviceInfo.DeviceType, &deviceInfo.DeviceCaps)))
		{
//			AUXTRACE("Device %s not supported for adapter %d", EnumToString(deviceInfo.DeviceType), adapter);
		} else
		{
			adapterInfo.Devices.push_back(deviceInfo);
		}
	}
}

//-----------------------------------------------------------------------------
// BuildBackBufferList
//-----------------------------------------------------------------------------
// Fills the list with the available back buffer formats.
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildBackBufferList(int adapter, D3DUtilDeviceInfo &deviceInfo, const std::vector<D3DFORMAT> &displayFormats, IDirect3D8 *pD3D)
{
	unsigned int formatIndex;
	for (formatIndex=0; formatIndex<displayFormats.size(); formatIndex++)
	{
		D3DUtilBackBufferInfo backBufferInfo;
		backBufferInfo.Format = displayFormats[formatIndex];
		if (FAILED(pD3D->CheckDeviceFormat(adapter, deviceInfo.DeviceType, backBufferInfo.Format, D3DUSAGE_RENDERTARGET, D3DRTYPE_SURFACE, backBufferInfo.Format)))
		{
//			AUXTRACE("Back buffer format %s is not supported on device %s using adapter %d", EnumToString(backBufferInfo.Format), EnumToString(deviceInfo.DeviceType), adapter);
		} else
		{
			deviceInfo.BackBuffers.push_back(backBufferInfo);
		}
	}
}

//-----------------------------------------------------------------------------
// BuildDepthBufferList
//-----------------------------------------------------------------------------
// Fills the list with the available depth buffer formats.
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildDepthBufferList(int adapter, D3DDEVTYPE deviceType, D3DUtilBackBufferInfo &backBufferInfo, IDirect3D8 *pD3D)
{
	int depthIndex;
	for (depthIndex=0; depthIndex<DepthFormatCount; depthIndex++)
	{
		D3DUtilDepthInfo depthInfo;
		depthInfo.Format = DepthFormats[depthIndex];
		if (depthInfo.Format != D3DFMT_UNKNOWN && (
			FAILED(pD3D->CheckDeviceFormat(adapter, deviceType, backBufferInfo.Format, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, depthInfo.Format)) ||
			FAILED(pD3D->CheckDepthStencilMatch(adapter, deviceType, backBufferInfo.Format, backBufferInfo.Format, depthInfo.Format))))
		{
//			AUXTRACE("Depth buffer format %s with back buffer format %s is not supported on device %s using adapter %d", EnumToString(depthInfo.Format), EnumToString(backBufferInfo.Format), EnumToString(deviceType), adapter);
		} else
		{
			backBufferInfo.DepthBuffers.push_back(depthInfo);
		}
	}
}

//-----------------------------------------------------------------------------
// BuildMultisampleList
//-----------------------------------------------------------------------------
// Fills the list with the available multisample types (windowed and non-windowed).
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildMultisampleList(int adapter, D3DDEVTYPE deviceType, D3DFORMAT backBufferFormat, D3DUtilDepthInfo &depthInfo, IDirect3D8 *pD3D)
{
	int multisampleIndex;
	for (multisampleIndex=0; multisampleIndex<MultisampleCount; multisampleIndex++)
	{
		D3DUtilMultisampleInfo multisampleInfo;
		multisampleInfo.Type = MultisampleTypes[multisampleIndex];
		if (FAILED(pD3D->CheckDeviceMultiSampleType(adapter, deviceType, backBufferFormat, TRUE, multisampleInfo.Type)) || (
			depthInfo.Format != D3DFMT_UNKNOWN && 
			FAILED(pD3D->CheckDeviceMultiSampleType(adapter, deviceType, depthInfo.Format, TRUE, multisampleInfo.Type))))
		{
//			AUXTRACE("%s not supported in windowed mode with depth buffer format %s with back buffer format %s is not supported on device %s using adapter %d", EnumToString(multisampleInfo.Type), EnumToString(depthInfo.Format), EnumToString(backBufferFormat), EnumToString(deviceType), adapter);
		} else
		{
			depthInfo.MultisampleWindowedTypes.push_back(multisampleInfo);
		}
		if (FAILED(pD3D->CheckDeviceMultiSampleType(adapter, deviceType, backBufferFormat, FALSE, multisampleInfo.Type)) || (
			depthInfo.Format != D3DFMT_UNKNOWN && 
			FAILED(pD3D->CheckDeviceMultiSampleType(adapter, deviceType, depthInfo.Format, FALSE, multisampleInfo.Type))))
		{
//			AUXTRACE("%s not supported in fullscreen mode with depth buffer format %s with back buffer format %s is not supported on device %s using adapter %d", EnumToString(multisampleInfo.Type), EnumToString(depthInfo.Format), EnumToString(backBufferFormat), EnumToString(deviceType), adapter);
		} else
		{
			depthInfo.MultisampleFullscreenTypes.push_back(multisampleInfo);
		}
	}
}

/*
	int adapter, modeIndex, deviceIndex, formatIndex, depthIndex, multisampleIndex;
	for (adapter=0; adapter<pD3D->GetAdapterCount(); adapter++)
	{
		// Initialize the AdapterInfo struct for the current adapter
		AdapterList.push_back(D3DUtilAdapterInfo());
		D3DUtilAdapterInfo &adapterInfo = AdapterList.back();
		D3DADAPTER_IDENTIFIER8 adapterIdentifier;
		pD3D->GetAdapterIdentifier(adapter, D3DENUM_NO_WHQL_LEVEL, &adapterIdentifier);
		sprintf(adapterInfo.Driver, "%s", adapterIdentifier.Driver);
		sprintf(adapterInfo.Description, "%s", adapterIdentifier.Description);

		// store the display formats the adapter is capable of using
		std::vector<D3DDISPLAYMODE> DisplayFormats;
		// Check for windowed mode capability
		D3DCAPS8 caps;
		pD3D->GetDeviceCaps(adapter, D3DDEVTYPE_HAL, &caps);
		if ((caps.Caps2 & D3DCAPS2_CANRENDERWINDOWED) != 0)
		{
			// 1st index stores current display mode
			D3DDISPLAYMODE mode;
			pD3D->GetAdapterDisplayMode(adapter, &mode);
			mode.Width = 0;
			mode.Height = 0;
			mode.RefreshRate = 0;
			DisplayFormats.push_back(mode);
		}
		// Check all the fullscreen possibilities
		for (modeIndex=0; modeIndex<pD3D->GetAdapterModeCount(adapter); modeIndex++)
		{
			D3DDISPLAYMODE mode;
			pD3D->EnumAdapterModes(adapter, modeIndex, &mode);
			DisplayFormats.push_back(mode);
		}

		D3DUtilDeviceInfo deviceInfo;
		// Check all the device types
		for (formatIndex=0; formatIndex<DisplayFormats.size(); formatIndex++)
		{
			if (formatIndex==0 && (caps.Caps2 & D3DCAPS2_CANRENDERWINDOWED) != 0)
			{
				deviceInfo.Windowed = true;
			} else
			{
				deviceInfo.Windowed = false;
			}
			deviceInfo.DisplayMode = DisplayFormats[formatIndex];

			for (deviceIndex=0; deviceIndex<DeviceTypeCount; deviceIndex++)
			{
				if (FAILED(pD3D->CheckDeviceType(adapter, DeviceTypes[deviceIndex], deviceInfo.DisplayMode.Format, deviceInfo.DisplayMode.Format, deviceInfo.Windowed)))
				{
					// Can't use this device with this format. Zero the modes out and continue
					ZeroMemory(deviceInfo.TypeMatrix[deviceIndex], sizeof(deviceInfo.TypeMatrix[deviceIndex]));
					continue;
				}
				for (depthIndex=0; depthIndex<DepthFormatCount; depthIndex++)
				{
					if (depthIndex > 0)
					{
						if (FAILED(pD3D->CheckDeviceFormat(adapter, DeviceTypes[deviceIndex], deviceInfo.DisplayMode.Format, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, DepthFormats[depthIndex])) ||
							FAILED(pD3D->CheckDepthStencilMatch(adapter, DeviceTypes[deviceIndex], deviceInfo.DisplayMode.Format, deviceInfo.DisplayMode.Format, DepthFormats[depthIndex])))
						{
							// Can't use this depth buffer with this device. Zero the modes out and continue
							ZeroMemory(deviceInfo.TypeMatrix[deviceIndex][depthIndex], sizeof(deviceInfo.TypeMatrix[deviceIndex][depthIndex]));
							continue;
						}
					}
					for (multisampleIndex=0; multisampleIndex<MultisampleCount; multisampleIndex++)
					{
						if (FAILED(pD3D->CheckDeviceMultiSampleType(adapter, DeviceTypes[deviceIndex], deviceInfo.DisplayMode.Format, deviceInfo.Windowed, MultisampleTypes[multisampleIndex])))
						{
							deviceInfo.TypeMatrix[deviceIndex][depthIndex][multisampleIndex] = false;
							continue;
						}
						// we must also check a depth buffer match if using one
						if (depthIndex > 0)
						{
							if (FAILED(pD3D->CheckDeviceMultiSampleType(adapter, DeviceTypes[deviceIndex], DepthFormats[depthIndex], deviceInfo.Windowed, MultisampleTypes[multisampleIndex])))
							{
								deviceInfo.TypeMatrix[deviceIndex][depthIndex][multisampleIndex] = false;
								continue;
							}
						}
						// Wow. We passed the gauntlet. This is a valid multisampling mode
						deviceInfo.TypeMatrix[deviceIndex][depthIndex][multisampleIndex] = true;
					}
				}
			}
			adapterInfo.Devices.push_back(deviceInfo);
		}
	}

	SAFE_RELEASE(pD3D);
	return AdapterList;
}
*/

