//-----------------------------------------------------------------------------
// D3DUtilFont.cpp
//-----------------------------------------------------------------------------
// Font drawing routines for D3D8.
//
//-----------------------------------------------------------------------------

#include "stdafx.h"
#include "D3DUtilFont.hpp"

//-----------------------------------------------------------------------------
// Custom vertex types for rendering text
//-----------------------------------------------------------------------------
static const int MAX_NUM_CHARACTERS = 256;
static const int MAX_NUM_TRIANGLES = MAX_NUM_CHARACTERS*2;


//-----------------------------------------------------------------------------
// ~D3DUtilFont
//-----------------------------------------------------------------------------
D3DUtilFont::~D3DUtilFont()
{
}

//-----------------------------------------------------------------------------
// OnPrepareReset
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::OnPrepareReset()
{
    SAFE_RELEASE(_VB);

	if(_SavedStateBlock != 0xFFFFFFFF) D3DUtilContext::Instance()->GetD3DDevice()->DeleteStateBlock(_SavedStateBlock);
    if(_DrawTextStateBlock != 0xFFFFFFFF) D3DUtilContext::Instance()->GetD3DDevice()->DeleteStateBlock(_DrawTextStateBlock);

    _SavedStateBlock    = 0xFFFFFFFF;
    _DrawTextStateBlock = 0xFFFFFFFF;

    return S_OK;
}

//-----------------------------------------------------------------------------
// OnReset
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::OnReset()
{
    IDirect3DDevice8* pDevice = D3DUtilContext::Instance()->GetD3DDevice();
	// The derived class MUST allocate the texture before this is called!
	ASSERT(GetTexture() != NULL);

    // Create vertex buffer for the letters
    FAIL_RETURN(pDevice->CreateVertexBuffer(MAX_NUM_TRIANGLES*3*sizeof(FONT2DVERTEX), D3DUSAGE_WRITEONLY | D3DUSAGE_DYNAMIC, 0, D3DPOOL_DEFAULT, &_VB));

    // Create the state blocks for rendering text
    for( unsigned which=0; which<2; which++ )
    {
        pDevice->BeginStateBlock();
        pDevice->SetTexture( 0, GetTexture() );
        pDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
        pDevice->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA );
        pDevice->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA );
        pDevice->SetRenderState( D3DRS_ALPHATESTENABLE,  TRUE );
        pDevice->SetRenderState( D3DRS_ALPHAREF,         0x08 );
        pDevice->SetRenderState( D3DRS_ALPHAFUNC,  D3DCMP_GREATEREQUAL );
        pDevice->SetRenderState( D3DRS_FILLMODE,   D3DFILL_SOLID );
        pDevice->SetRenderState( D3DRS_CULLMODE,   D3DCULL_CCW );
        pDevice->SetRenderState( D3DRS_ZENABLE,          FALSE );
        pDevice->SetRenderState( D3DRS_STENCILENABLE,    FALSE );
        pDevice->SetRenderState( D3DRS_CLIPPING,         TRUE );
        pDevice->SetRenderState( D3DRS_EDGEANTIALIAS,    FALSE );
        pDevice->SetRenderState( D3DRS_CLIPPLANEENABLE,  FALSE );
        pDevice->SetRenderState( D3DRS_VERTEXBLEND,      FALSE );
        pDevice->SetRenderState( D3DRS_INDEXEDVERTEXBLENDENABLE, FALSE );
        pDevice->SetRenderState( D3DRS_FOGENABLE,        FALSE );
        pDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
        pDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
        pDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
        pDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE );
        pDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
        pDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
        pDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
        pDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
        pDevice->SetTextureStageState( 0, D3DTSS_MIPFILTER, D3DTEXF_NONE );
        pDevice->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, 0 );
        pDevice->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
        pDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
        pDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

        if( which==0 )
            pDevice->EndStateBlock( &_SavedStateBlock );
        else
            pDevice->EndStateBlock( &_DrawTextStateBlock );
    }
    return S_OK;
}

//-----------------------------------------------------------------------------
// OnInvalidateDevice
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::OnInvalidateDevice()
{
    SAFE_RELEASE(_pTexture);
    return S_OK;
}


//-----------------------------------------------------------------------------
// OnRebuildDevice
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::OnRebuildDevice()
{
	IDirect3DDevice8 *pDevice = D3DUtilContext::Instance()->GetD3DDevice();

    // Establish the font and texture size

    // Large fonts need larger textures
    if( _FontHeight > 40 )
        _TextureWidth = 1024;
    else if( _FontHeight > 20 )
        _TextureWidth = 512;
    else
        _TextureWidth  = 256;

    // If requested texture is too big, use a smaller texture and smaller font,
    // and scale up when rendering.
    D3DCAPS8 d3dCaps;
    pDevice->GetDeviceCaps( &d3dCaps );

    if( _TextureWidth > d3dCaps.MaxTextureWidth ) FAIL_RETURN(E_FAIL);

    // Create a new texture for the font
    FAIL_RETURN(pDevice->CreateTexture( _TextureWidth, _TextureWidth, 1, 0, D3DFMT_A4R4G4B4, D3DPOOL_MANAGED, &_pTexture ));

    // Prepare to create a bitmap
    unsigned*      pBitmapBits;
    BITMAPINFO bmi;
    ZeroMemory( &bmi.bmiHeader,  sizeof(BITMAPINFOHEADER) );
    bmi.bmiHeader.biSize        = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth       =  (int)_TextureWidth;
    bmi.bmiHeader.biHeight      = -(int)_TextureWidth;
    bmi.bmiHeader.biPlanes      = 1;
    bmi.bmiHeader.biCompression = BI_RGB;
    bmi.bmiHeader.biBitCount    = 32;

    // Create a DC and a bitmap for the font
    HDC     hDC       = CreateCompatibleDC( NULL );
    HBITMAP hbmBitmap = CreateDIBSection( hDC, &bmi, DIB_RGB_COLORS,
                                          (VOID**)&pBitmapBits, NULL, 0 );
    SetMapMode( hDC, MM_TEXT );

    // Create a font.  By specifying ANTIALIASED_QUALITY, we might get an
    // antialiased font, but this is not guaranteed.
    INT nHeight    = -MulDiv( _FontHeight, (INT)GetDeviceCaps(hDC, LOGPIXELSY), 72 );
    unsigned dwBold   = (_FontFlags&D3DFONT_BOLD)   ? FW_BOLD : FW_NORMAL;
    unsigned dwItalic = (_FontFlags&D3DFONT_ITALIC) ? TRUE    : FALSE;
    HFONT hFont    = CreateFont( nHeight, 0, 0, 0, dwBold, dwItalic,
                          FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
                          CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
                          VARIABLE_PITCH, _FontName.c_str() );
    if( NULL==hFont ) return E_FAIL;

    SelectObject( hDC, hbmBitmap );
    SelectObject( hDC, hFont );

	GetTextMetrics(hDC, &_TextMetric);

    // Set text properties
    SetTextColor( hDC, RGB(255,255,255) );
    SetBkColor(   hDC, 0x00000000 );
    SetTextAlign( hDC, TA_TOP );

    // Loop through all printable character and output them to the bitmap..
    // Meanwhile, keep track of the corresponding tex coords for each character.
    unsigned x = 0;
    unsigned y = 0;
    char str[] = "x";
    SIZE size;

    for( char c=32; c<127; c++ )
    {
        str[0] = c;
        GetTextExtentPoint32( hDC, str, 1, &size );

		// !!! WRH Make sure the reported height matched what I think it should be
		// This is just to make sure I understand the font metrics properly
		ASSERT(size.cy == _TextMetric.tmHeight);

        if( (unsigned)(x+size.cx+1) > _TextureWidth )
        {
            x  = 0;
            y += _TextMetric.tmHeight+1;
        }

        ExtTextOut( hDC, x+0, y+0, ETO_OPAQUE, NULL, str, 1, NULL );

        _texCoords[c-32][0] = ((float)(x+0))/_TextureWidth;
        _texCoords[c-32][1] = ((float)(y+0))/_TextureWidth;
        _texCoords[c-32][2] = ((float)(x+0+size.cx))/_TextureWidth;
        _texCoords[c-32][3] = ((float)(y+0+_TextMetric.tmHeight))/_TextureWidth;

        x += size.cx+1;
    }

    // Lock the surface and write the alpha values for the set pixels
    D3DLOCKED_RECT d3dlr;
    _pTexture->LockRect( 0, &d3dlr, 0, 0 );
    WORD* pDst16 = (WORD*)d3dlr.pBits;
    BYTE bAlpha; // 4-bit measure of pixel intensity

    for( y=0; y < _TextureWidth; y++ )
    {
        for( x=0; x < _TextureWidth; x++ )
        {
            bAlpha = (BYTE)((pBitmapBits[_TextureWidth*y + x] & 0xff) >> 4);
            if (bAlpha > 0)
            {
                *pDst16++ = (bAlpha << 12) | 0x0fff;
            }
            else
            {
				// Put a little background alpha on the text so it won't blend into the scene itself.
                *pDst16++ = 0x8000;
            }
        }
    }

    // Done updating texture, so clean up used objects
    _pTexture->UnlockRect(0);
    DeleteObject( hbmBitmap );
    DeleteDC( hDC );
    DeleteObject( hFont );

    return S_OK;
}


//-----------------------------------------------------------------------------
// LockText
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::LockText()
{
	FAIL_RETURN(_LockVB());
	
    // Setup renderstate
    FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->CaptureStateBlock(_SavedStateBlock ));
    FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->ApplyStateBlock(_DrawTextStateBlock));
    FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->SetVertexShader(D3DFVF_FONT2DVERTEX));
    FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->SetStreamSource(0, _VB, sizeof(FONT2DVERTEX)));

	_CursorFound = false;

	return D3D_OK;
}

//-----------------------------------------------------------------------------
// Name: DrawText()
// Desc: Draws 2D text
// NOTE : You must call BeginText() first!
//
// If nextPt != NULL, it returns the screen coordinates of the next character
// if you want to continue writing text.
//
// Draws an inverse color quad over the character at position cursorPos to 
// represent a cursor. Set cursorPos = -1 (the default) to not show a cursor.
// There can only be one cursor between any BeginText/EndText pair.
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::DrawText(float sx, float sy, unsigned dwColor, const char* strText, 
							  float scale, D3DXVECTOR2* nextPt, int cursorPos)
{
	ASSERT(_LockedVerts != NULL);
	
	// Early exit if there is nothing to draw
	if (!strText || *strText == '\0') return D3D_OK;

	// Width/Height of the current character
    float w, h;

	// Current character index (used to see if we have reached the cursor position)
	int ndx = 0;

	// offset the positions for pixel:texel matching
	sx -= 0.5f;
	sy -= 0.5f;

    float fStartX = sx;
	float fStartY = sy;

	// Fill vertex buffer
    while( *strText )
    {

        char c = *strText++;

        if( c == '\n' )
        {
			// track the width of the text.
            sx = fStartX;
            sy += GetLineHeight();
        }
        if( c < ' ' )
            continue;

		float tx1, ty1, tx2, ty2;
		_GetTexCoords(c, tx1, ty1, tx2, ty2);

        w = (tx2-tx1) * GetTextureWidth() * scale;
        h = GetLineHeight() * scale;

		FONT2DVERTEX *p = _LockedVerts;
        _InitVertex( *_LockedVerts++, sx  ,sy+h,0.9f,1.0f, dwColor, tx1, ty2 );
        _InitVertex( *_LockedVerts++, sx  ,sy  ,0.9f,1.0f, dwColor, tx1, ty1 );
        _InitVertex( *_LockedVerts++, sx+w,sy+h,0.9f,1.0f, dwColor, tx2, ty2 );
        _InitVertex( *_LockedVerts++, sx+w,sy  ,0.9f,1.0f, dwColor, tx2, ty1 );
        _InitVertex( *_LockedVerts++, sx+w,sy+h,0.9f,1.0f, dwColor, tx2, ty2 );
        _InitVertex( *_LockedVerts++, sx  ,sy  ,0.9f,1.0f, dwColor, tx1, ty1 );

        _NumTriangles += 2;

		if (!_CursorFound && ndx == cursorPos)
		{
			// This is the cursor character. Save the coordinates of this character so we can go back and
			// highlight it later
			_CursorQuad[0] = p[0].p;
			_CursorQuad[1] = p[1].p;
			_CursorQuad[2] = p[2].p;
			_CursorQuad[3] = p[3].p;
			_CursorFound = true;
		}

        if(_NumTriangles+2 > MAX_NUM_TRIANGLES)
        {
            // Unlock, render, and relock the vertex buffer
			FAIL_RETURN(_UnlockVB());
			FAIL_RETURN(_FlushText());
			FAIL_RETURN(_LockVB());
        }

        sx += w;
		++ndx;
    }

	if (!_CursorFound && cursorPos >= 0)
	{
		// The cursor is at the end of the line, so draw a square one char over.using the last width and height
		_CursorQuad[0] = D3DXVECTOR4(sx  ,sy+h,0.9f,1.0f);
		_CursorQuad[1] = D3DXVECTOR4(sx  ,sy  ,0.9f,1.0f);
		_CursorQuad[2] = D3DXVECTOR4(sx+w,sy+h,0.9f,1.0f);
		_CursorQuad[3] = D3DXVECTOR4(sx+w,sy  ,0.9f,1.0f);
		// !!! WRH Should we update the cursor postion or leave it where it is?
		// sx += w; pSize->cx = sx;
		_CursorFound = true;
	}
		
	if (nextPt != NULL)
	{
		// undo pixe:texel offset
		nextPt->x = sx+0.5f;
		nextPt->y = sy+0.5f;
	}

    return D3D_OK;
}

//-----------------------------------------------------------------------------
// UnlockText
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::UnlockText()
{
    // Unlock and render the vertex buffer
	FAIL_RETURN(_UnlockVB());
    if(_NumTriangles > 0)
	{
        FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->DrawPrimitive(D3DPT_TRIANGLELIST, 0, _NumTriangles));
	}

	// Draw the cursor if we had one
	if (_CursorFound)
	{
		// Highlight the portion of the text that the cursor is over by inverting the color of the square
		// where that character is.
		D3DUtilContext::Instance()->GetD3DDevice()->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_DISABLE );
		D3DUtilContext::Instance()->GetD3DDevice()->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
		D3DUtilContext::Instance()->GetD3DDevice()->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_INVDESTCOLOR );
		D3DUtilContext::Instance()->GetD3DDevice()->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO );
		DWORD cursorColor = 0xFFFFFFFF;
		FAIL_RETURN(_LockVB());
		_InitVertex( *_LockedVerts++, _CursorQuad[0], cursorColor, 0, 0 );
		_InitVertex( *_LockedVerts++, _CursorQuad[1], cursorColor, 0, 0 );
		_InitVertex( *_LockedVerts++, _CursorQuad[2], cursorColor, 0, 0 );
		_InitVertex( *_LockedVerts++, _CursorQuad[3], cursorColor, 0, 0 );
		_InitVertex( *_LockedVerts++, _CursorQuad[2], cursorColor, 0, 0 );
		_InitVertex( *_LockedVerts++, _CursorQuad[1], cursorColor, 0, 0 );
		_NumTriangles += 2;
		FAIL_RETURN(_UnlockVB());
		FAIL_RETURN(_FlushText());
	}
	
    // Restore the modified renderstates
    FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->ApplyStateBlock(_SavedStateBlock));
	
	return D3D_OK;
}


//-----------------------------------------------------------------------------
// Name: GetTextExtent()
// Desc: Get the dimensions of a text string
//-----------------------------------------------------------------------------
HRESULT D3DUtilFont::GetTextExtent(const char* strText, D3DXVECTOR2& size) const
{
    if (NULL==strText) return E_FAIL;

    float fRowWidth  = 0.0f;
    float fWidth     = 0.0f;
    float fHeight    = static_cast<float>(GetLineHeight());

    while( *strText )
    {
        char c = *strText++;

        if( c == '\n' )
        {
            fRowWidth = 0.0f;
            fHeight  += GetLineHeight();
        }
        if( c < ' ' )
            continue;

		float tx1, ty1, tx2, ty2;
		_GetTexCoords(c, tx1, ty1, tx2, ty2);

        fRowWidth += (tx2-tx1)*GetTextureWidth();

        if( fRowWidth > fWidth )
            fWidth = fRowWidth;
    }

    size.x = fWidth;
    size.y = fHeight;

    return S_OK;
}

