//============================================================================
// File          : Util.hpp
// 
// Content       : Helper utilities
// 
//============================================================================
// 
//============================================================================
#ifndef Util_included_
#define Util_included_
#pragma once


//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

// Stop windows.h from defining the min and max macros
#define NOMINMAX
// Fix VC++6 stl min and max problems
#ifndef NOFIXMINMAX
    #if _MSC_VER <= 1200
        #define min _cpp_min
        #define max _cpp_max
    #endif
#endif

// fix crappy MSVC for scoping problem
#ifndef NOFIXFORSCOPING
    #ifdef _MSC_VER
        #define for if (false); else for
    #endif
#endif

#include <windows.h>
#include <dxerr8.h>
#include <d3d8.h>
#include <d3dx8.h>

//----------------------------------------------------------------------------
//-- Forward Declarations ----------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Defines, Constants, Enumerated Types ------------------------------------
//----------------------------------------------------------------------------

#define SAFE_DELETE(p)       { delete   (p); (p)=NULL; }
#define SAFE_DELETE_ARRAY(p) { delete[] (p); (p)=NULL; }
#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p)=NULL; } }

// undefine these to stop debug tracing of FAIL_TEST and FAIL_RETURNs
#define AUX_USE_FAIL_TEST
#define AUX_USE_FAIL_RETURN
#define AUX_USE_FAIL_ASSERT
#define AUX_USE_TRACE

//-----------------------------------------------------------------------------
// AuxAssert
//-----------------------------------------------------------------------------
// For asserts
// Allow pre/post processing funcs so we can, say, enter/exit fullscreen mode
// before presenting the assert message.
//-----------------------------------------------------------------------------
struct AuxAssert
{
    AuxAssert(const char* file, int line) : _file(file), _line(line) {}
    AuxAssert(const char* file, int line, const char* expr) : _file(file), _line(line) {}
    void __cdecl Test(bool expr, const char* format, ...);

    typedef void (*PreProcessFunc)();
    typedef void (*PostProcessFunc)();
    static PreProcessFunc InstallPreProcessFunc(PreProcessFunc func);
    static PostProcessFunc InstallPostProcessFunc(PostProcessFunc func);

private:
    static PreProcessFunc& _CurrentPreProcessFunc();
    static PostProcessFunc& _CurrentPostProcessFunc();
    const char *_file;
    int _line;
};

#if _DEBUG
#define ASSERT(expr)  AuxAssert(__FILE__, __LINE__).Test(expr, #expr)
#define ASSERT_MSG    AuxAssert(__FILE__, __LINE__).Test
#else
#define ASSERT(expr)
#define ASSERT_MSG
#endif

//-----------------------------------------------------------------------------
// AuxTrace
//-----------------------------------------------------------------------------
// Log debug output.
// Handles the ErrorTranlation functions that FAIL_XXX macros rely on.
//-----------------------------------------------------------------------------
struct AuxTrace
{
    AuxTrace(const char* file, int line) : _file(file), _line(line) {}
    void __cdecl Trace(const char* format, ...);

    typedef const char* (*ErrorTranslationFunc)(HRESULT err); 
    static ErrorTranslationFunc GetErrorTranslationFunc();
    static ErrorTranslationFunc SetErrorTranslationFunc(ErrorTranslationFunc errFunc);
    static const char* DefaultErrorTranslationFunc(HRESULT err);
private:
    static ErrorTranslationFunc& _CurrentErrorTranslationFunc();
    const char *_file;
    int _line;
};

#ifdef AUX_USE_TRACE
#define AUXTRACE AuxTrace(__FILE__, __LINE__).Trace
#else
#define AUXTRACE (void)
#endif

#ifdef AUX_USE_FAIL_TEST
    //-----------------------------------------------------------------------------
    // FAIL_TEST
    //-----------------------------------------------------------------------------
    // Similar to FAIL_RETURN except no return call is generated
    //-----------------------------------------------------------------------------
	#define FAIL_TEST(p) \
	{ \
		HRESULT __UTIL__hres__; \
		if ( FAILED( __UTIL__hres__ = (p) ) ) { \
    		AUXTRACE("!!!! FAIL_TEST TRAP !!!! Expr: %s (hr=%s (%#8x))", #p, AuxTrace::GetErrorTranslationFunc()(__UTIL__hres__), __UTIL__hres__); \
		} \
	}
    //-----------------------------------------------------------------------------
    // FAIL_TEST_MSG
    //-----------------------------------------------------------------------------
    // Similar to FAIL_RETURN_MSG except no return call is generated
    //-----------------------------------------------------------------------------
	#define FAIL_TEST_MSG(p,str) \
	{ \
		HRESULT __UTIL__hres__; \
		if ( FAILED( __UTIL__hres__ = (p) ) ) { \
		AUXTRACE("!!!! FAIL_TEST TRAP !!!! Msg: %s (hr=%s (%#8x))", str, AuxTrace::GetErrorTranslationFunc()(__UTIL__hres__), __UTIL__hres__); \
		} \
	}
#else
    #define FAIL_TEST(p) p
    #define FAIL_TEST_MSG(p,str) p
#endif

#ifdef AUX_USE_FAIL_RETURN
    //-----------------------------------------------------------------------------
    // FAIL_RETURN
    //-----------------------------------------------------------------------------
    // Allow debug messages to the WIN32 debug window (see OutputDebugString in the SDK)
    // Translates HRESULTs to DX errors and outputs in a format that can be
    // double-clicked in DevStudio to open the appropriate file and location.
    //-----------------------------------------------------------------------------
    #define FAIL_RETURN(p) \
    { \
	    HRESULT __UTIL__hres__; \
	    if ( FAILED( __UTIL__hres__ = (p) ) ) { \
		    AUXTRACE("!!!! FAIL_RETURN TRAP !!!! Expr: %s (hr=%s (%#8x))", #p, AuxTrace::GetErrorTranslationFunc()(__UTIL__hres__), __UTIL__hres__); \
		    return __UTIL__hres__; \
	    } \
    }

    //-----------------------------------------------------------------------------
    // FAIL_RETURN_MSG
    //-----------------------------------------------------------------------------
    // Similar to FAIL_RETURN, but also appends a user-supplied msg.
    //-----------------------------------------------------------------------------
    #define FAIL_RETURN_MSG(p,str) \
    { \
	    HRESULT __UTIL__hres__; \
	    if ( FAILED( __UTIL__hres__ = (p) ) ) { \
		    AUXTRACE("!!!! FAIL_RETURN TRAP !!!! Msg: %s (hr=%s (%#8x))", str, AuxTrace::GetErrorTranslationFunc()(__UTIL__hres__), __UTIL__hres__); \
		    return __UTIL__hres__; \
	    } \
    }
#else
    #define FAIL_RETURN(p) p
    #define FAIL_RETURN_MSG(p,str) p
#endif

#ifdef AUX_USE_FAIL_ASSERT
    //-----------------------------------------------------------------------------
    // FAIL_ASSERT
    //-----------------------------------------------------------------------------
    // Similar to FAIL_TEST except asserts when fails.
    //-----------------------------------------------------------------------------
	#define FAIL_ASSERT(p) \
	{ \
		HRESULT __UTIL__hres__; \
		if ( FAILED( __UTIL__hres__ = (p) ) ) { \
    		ASSERT_MSG(false, "!!!! FAIL_TEST TRAP !!!! Expr: %s (hr=%s (%#8x))", #p, AuxTrace::GetErrorTranslationFunc()(__UTIL__hres__), __UTIL__hres__); \
		} \
	}
    //-----------------------------------------------------------------------------
    // FAIL_ASSERT_MSG
    //-----------------------------------------------------------------------------
    // Similar to FAIL_TEST_MSG except asserts when fails.
    //-----------------------------------------------------------------------------
	#define FAIL_ASSERT_MSG(p,str) \
	{ \
		HRESULT __UTIL__hres__; \
		if ( FAILED( __UTIL__hres__ = (p) ) ) { \
		ASSERT_MSG(false, "!!!! FAIL_TEST TRAP !!!! Msg: %s (hr=%s (%#8x))", str, AuxTrace::GetErrorTranslationFunc()(__UTIL__hres__), __UTIL__hres__); \
		} \
	}
#else
    #define FAIL_ASSERT(p) p
    #define FAIL_ASSERT_MSG(p,str) p
#endif


//----------------------------------------------------------------------------
//-- Function Declarations ---------------------------------------------------
//----------------------------------------------------------------------------

//---------------------------------------------------------------------------------
// Fast float to long conversion
//---------------------------------------------------------------------------------
// Only works for a certain range of values, else it wraps around...
//---------------------------------------------------------------------------------
static inline long FToL( float f )
{
	const float FLOAT_SHIFT_NUMBER      = 12582912.0f;
	const long  long_SHIFT_NUMBER       = 1262485504;
    float res = f + FLOAT_SHIFT_NUMBER;
	return (*(long*)(&res)) - long_SHIFT_NUMBER ;
}

//-----------------------------------------------------------------------------
// EnumToString
//-----------------------------------------------------------------------------
// Several routines to translate DX8 enums to strings.
//-----------------------------------------------------------------------------
const char * EnumToString(D3DFORMAT format);
const char * EnumToString(D3DMULTISAMPLE_TYPE type);
const char * EnumToString(D3DDEVTYPE type);

//-----------------------------------------------------------------------------
// StringToEnum
//-----------------------------------------------------------------------------
// Several routines to translate strings to DX8 enums.
//-----------------------------------------------------------------------------
D3DFORMAT StringToD3DFORMAT(const char *str);
D3DMULTISAMPLE_TYPE StringToD3DMULTISAMPLE_TYPE(const char *str);
D3DDEVTYPE StringToD3DDEVTYPE(const char *str);

//----------------------------------------------------------------------------
//-- Typedefs, Structs, Classes ----------------------------------------------
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Viewing16v
//----------------------------------------------------------------------------
// Sets up a viewing matrix using basis vectors and position.
//----------------------------------------------------------------------------
inline
void MatrixViewing(D3DXMATRIX& M, const D3DXVECTOR3& X, const D3DXVECTOR3& Y, const D3DXVECTOR3& Z, const D3DXVECTOR3& O)
{
  M[0]=X[0];  M[4]=X[1];   M[8]=X[2];  M[12]=-D3DXVec3Dot(&X,&O);
  M[1]=Y[0];  M[5]=Y[1];   M[9]=Y[2];  M[13]=-D3DXVec3Dot(&Y,&O);
  M[2]=Z[0];  M[6]=Z[1];   M[10]=Z[2]; M[14]=-D3DXVec3Dot(&Z,&O);
  M[3]=0;     M[7]=0;      M[11]=0;    M[15]=1;
}

//----------------------------------------------------------------------------
// InvViewing16v
//----------------------------------------------------------------------------
// Sets up an inverse viewing matrix using basis vectors and position.
//----------------------------------------------------------------------------
inline
void MatrixViewingInv(D3DXMATRIX& M, const D3DXVECTOR3& X, const D3DXVECTOR3& Y, const D3DXVECTOR3& Z, const D3DXVECTOR3& O)
{  
  M[0]=X[0];  M[4]=Y[0];   M[8]=Z[0];  M[12]=O[0];
  M[1]=X[1];  M[5]=Y[1];   M[9]=Z[1];  M[13]=O[1];
  M[2]=X[2];  M[6]=Y[2];   M[10]=Z[2]; M[14]=O[2];
  M[3]=0;     M[7]=0;      M[11]=0;    M[15]=1;
}

//----------------------------------------------------------------------------
// Frustum16vD3D
//----------------------------------------------------------------------------
// Sets up a D3D-style projection matrix ([-1..1] in X & Y, [0..1] in Z).
// D3D viewing volume is looking down the +Z axis, 
// so no right-left handed transfomation is needed along the Z axis.
// Modeled after OGL routine glFrustum.
// Accepts sheared viewing volumes.
//----------------------------------------------------------------------------
inline
void MatrixFrustumD3D(D3DXMATRIX& M, float l, float r, float b, float t, float n, float f)
{
  M[0]=(2*n)/(r-l); M[4]=0;           M[8]=(r+l)/(r-l);    M[12]=0;
  M[1]=0;           M[5]=(2*n)/(t-b); M[9]=(t+b)/(t-b);    M[13]=0;
  M[2]=0;           M[6]=0;           M[10]=f/(f-n);       M[14]=-f*n/(f-n);
  M[3]=0;           M[7]=0;           M[11]=1;             M[15]=0;
}

//----------------------------------------------------------------------------
// InvFrustum16vD3D
//----------------------------------------------------------------------------
// Sets up an inverse D3D-style projection matrix.
//----------------------------------------------------------------------------
inline
void MatrixFrustumInvD3D(D3DXMATRIX& M, float l, float r, float b, float t, float n, float f)
{
  M[0]=(r-l)/(2*n); M[4]=0;           M[8]=0;              M[12]=(r+l)/(-2*n);
  M[1]=0;           M[5]=(t-b)/(2*n); M[9]=0;              M[13]=(t+b)/(-2*n);
  M[2]=0;           M[6]=0;           M[10]=0;             M[14]=1;
  M[3]=0;           M[7]=0;           M[11]=(-f+n)/(f*n);  M[15]=1/n;
}

//----------------------------------------------------------------------------
// Frustum16vD3D
//----------------------------------------------------------------------------
// Sets up a D3D-style projection matrix ([-1..1] in X & Y, [0..1] in Z).
// Creates only non-sheared viewing volumes.
//----------------------------------------------------------------------------
inline
void MatrixFrustumD3D(D3DXMATRIX& M, float w, float h, float n, float f)
{
  M[0]=(2*n)/w;     M[4]=0;           M[8]=0;              M[12]=0;
  M[1]=0;           M[5]=(2*n)/h;     M[9]=0;              M[13]=0;
  M[2]=0;           M[6]=0;           M[10]=f/(f-n);       M[14]=-f*n/(f-n);
  M[3]=0;           M[7]=0;           M[11]=1;             M[15]=0;
}

//----------------------------------------------------------------------------
// InvFrustum16vD3D
//----------------------------------------------------------------------------
// Sets up an inverse D3D-style projection matrix.
//----------------------------------------------------------------------------
inline
void MatrixFrustumInvD3D(D3DXMATRIX& M, float w, float h, float n, float f)
{
  M[0]=w/(2*n);     M[4]=0;           M[8]=0;              M[12]=0;
  M[1]=0;           M[5]=h/(2*n);     M[9]=0;              M[13]=0;
  M[2]=0;           M[6]=0;           M[10]=0;             M[14]=1;
  M[3]=0;           M[7]=0;           M[11]=(-f+n)/(f*n);  M[15]=1/n;
}

//----------------------------------------------------------------------------
// Common setup code for constructing projection matrices
//----------------------------------------------------------------------------
inline
void SetupPerspectiveForFrustum(float YfovDeg, float Aspect, float Ndist, float& ww, float& wh)
{
    float YfovRad = D3DXToRadian(YfovDeg);
	wh=2.f*tanf(YfovRad*0.5f)*Ndist;
	ww=wh*Aspect;
}

//----------------------------------------------------------------------------
// Perspective16vD3D
//----------------------------------------------------------------------------
// Sets up a D3D-style projection matrix ([-1..1] in X & Y, [0..1] in Z).
// Creates only non-sheared viewing volumes.
//----------------------------------------------------------------------------
inline
void MatrixPerspectiveD3D(D3DXMATRIX& M, float Yfov, float Aspect, float Ndist, float Fdist)
{
  float ww, wh;
  SetupPerspectiveForFrustum(Yfov, Aspect, Ndist, ww, wh);
  MatrixFrustumD3D(M,ww,wh,Ndist,Fdist);
}

//----------------------------------------------------------------------------
// InvPerspective16vD3D
//----------------------------------------------------------------------------
// Sets up an inverse D3D-style projection matix.
//----------------------------------------------------------------------------
inline
void MatrixPerspectiveInvD3D(D3DXMATRIX& M, float Yfov, float Aspect, float Ndist, float Fdist)
{
  float ww, wh;
  SetupPerspectiveForFrustum(Yfov, Aspect, Ndist, ww, wh);
  MatrixFrustumInvD3D(M,ww,wh,Ndist,Fdist);
}

void MatrixRotate(D3DXMATRIX& M, float RadAng, const D3DXVECTOR3& Axis);

// ---------------------------------------------------------------------------
// Rotation followed by translation
//----------------------------------------------------------------------------
// [ R R R 0 ] * [ 1 0 0 0 ] = [ R R R 0 ]
// [ R R R 0 ]   [ 0 1 0 0 ]   [ R R R 0 ]
// [ R R R 0 ]   [ 0 0 1 0 ]   [ R R R 0 ]
// [ 0 0 0 1 ]   [ T T T 1 ]   [ T T T 1 ]
//----------------------------------------------------------------------------
inline
void MatrixRotTrans(D3DXMATRIX& M, const D3DXMATRIX& rot, const D3DXVECTOR3& v)
{
	// First 9 entries are simply the rot
	M[0] = rot[0]; M[1] = rot[1]; M[2] = rot[2];
	M[4] = rot[4]; M[5] = rot[5]; M[6] = rot[6];
	M[8] = rot[8]; M[9] = rot[9]; M[10] = rot[10];

    // just the translation
    M[12] = v.x; M[13] = v.y; M[14] = v.z;

    // 0 0 0 1
	M[3] = M[7] = M[11] = 0.f; M[15] = 1.f;
}

// ---------------------------------------------------------------------------
// Translation followed by rotation
//----------------------------------------------------------------------------
// [ 1 0 0 0 ] * [ R R R 0 ] = [  R    R    R   0 ]
// [ 0 1 0 0 ]   [ R R R 0 ]   [  R    R    R   0 ]
// [ 0 0 1 0 ]   [ R R R 0 ]   [  R    R    R   0 ]
// [ T T T 1 ]   [ 0 0 0 1 ]   [ T*C1 T*C2 T*C3 1 ]
//
// Where Cx = column x of the rotation matrix.
//----------------------------------------------------------------------------
inline
void MatrixTransRot(D3DXMATRIX& M, const D3DXVECTOR3& v, const D3DXMATRIX& rot)
{
	// First 9 entries are simply the rot
	M[0] = rot[0]; M[1] = rot[1]; M[2] = rot[2];
	M[4] = rot[4]; M[5] = rot[5]; M[6] = rot[6];
	M[8] = rot[8]; M[9] = rot[9]; M[10] = rot[10];

	// the translation section gets the product
    M[12] = rot[0]*v.x + rot[4]*v.y + rot[8]*v.z;
    M[13] = rot[1]*v.x + rot[5]*v.y + rot[9]*v.z;
    M[14] = rot[2]*v.x + rot[6]*v.y + rot[10]*v.z;

    // 0 0 0 1
	M[3] = M[7] = M[11] = 0.f; M[15] = 1.f;
}

// ---------------------------------------------------------------------------
// Translation followed by rotation followed by translation
//----------------------------------------------------------------------------
// [  1  0  0 0 ] * [ R R R 0 ] * [  1  0  0 0 ] = [     R          R          R      0 ]
// [  0  1  0 0 ]   [ R R R 0 ]   [  0  1  0 0 ]   [     R          R          R      0 ]
// [  0  0  1 0 ]   [ R R R 0 ]   [  0  0  1 0 ]   [     R          R          R      0 ]
// [  T  T  T 1 ]   [ 0 0 0 1 ]   [ T2 T2 T2 1 ]   [ T*C1+T2.x  T*C2+T2.y  T*C3+T2.z  1 ]
//
// Where Cx = column x of the rotation matrix.
//----------------------------------------------------------------------------
inline
void MatrixTransRotTrans(D3DXMATRIX& M, const D3DXVECTOR3& v1, const D3DXMATRIX& rot, const D3DXVECTOR3& v2)
{
	// First 9 entries are simply the rot
	M[0] = rot[0]; M[1] = rot[1]; M[2] = rot[2];
	M[4] = rot[4]; M[5] = rot[5]; M[6] = rot[6];
	M[8] = rot[8]; M[9] = rot[9]; M[10] = rot[10];

	// the translation section gets the product
    M[12] = rot[0]*v1.x + rot[4]*v1.y + rot[8]*v1.z + v2.x;
    M[13] = rot[1]*v1.x + rot[5]*v1.y + rot[9]*v1.z + v2.y;
    M[14] = rot[2]*v1.x + rot[6]*v1.y + rot[10]*v1.z + v2.z;

    // 0 0 0 1
	M[3] = M[7] = M[11] = 0.f; M[15] = 1.f;
}

inline 
void Vec2Set(D3DXVECTOR2& v, float x, float y)
{
	v.x = x;
	v.y = y;
}

inline 
void Vec3Set(D3DXVECTOR3& v, float x, float y, float z)
{
	v.x = x;
	v.y = y;
	v.z = z;
}

inline 
void Vec4Set(D3DXVECTOR4& v, float x, float y, float z, float w=1.f)
{
	v.x = x;
	v.y = y;
	v.z = z;
	v.w = w;
}

//----------------------------------------------------------------------------
//-- External Variables ------------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Function Definitions ----------------------------------------------------
//----------------------------------------------------------------------------


#endif
