//============================================================================
// File          : UtilCamera.hpp
// 
// Content       : UtilCamera class. Uses D3D-style matrix math.
// 
//                  camera space (view space) is defined as:
//                      RightDir = +x
//                      UpDir    = +y
//                      ViewDir  = +z
//                  The camera viewing windows is defined as:
//                      UpperLeft  = (t,l,n)
//                      UpperRight = (t,r,n)
//                      LowerLeft  = (b,l,n)
//                      LowerRight = (b,r,n)
//                  The camera viewing window is the rectangle in worldspace
//                  that defines the plane that the scene in projected onto.
//
// *** I'VE MADE NOTES LIKE THIS IN ROUTINES THAT ARE EXPLICITLY D3D SPECIFIC
//     WITH COMMENTS ON WHAT D3D-SPECIFIC ASSUMPTIONS ARE BEING MADE ***
//============================================================================
#ifndef UtilCamera_included_
#define UtilCamera_included_
#pragma once


//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "Util.hpp"

//----------------------------------------------------------------------------
//-- Forward Declarations ----------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Defines, Constants, Enumerated Types ------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Function Declarations ---------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Typedefs, Structs, Classes ----------------------------------------------
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Class UtilCamera
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
class UtilCamera
{
public:
	//----------------------------------------------------------------------------
	// ctor/dtor
	//----------------------------------------------------------------------------
	UtilCamera();
	~UtilCamera();

	//----------------------------------------------------------------------------
	// The matrices that the camera actually controls
	//----------------------------------------------------------------------------
    const D3DXMATRIX    GetWorldToViewMatrix() const;
    const D3DXMATRIX    GetProjectionMatrix() const;
    const D3DXMATRIX    GetInvWorldToViewMatrix() const;
    const D3DXMATRIX    GetInvProjectionMatrix() const;

    //----------------------------------------------------------------------------
	// Viewing parameters
	//----------------------------------------------------------------------------
	void            LookAt(const D3DXVECTOR3& eyePt, const D3DXVECTOR3& viewRefPt, const D3DXVECTOR3& viewUpDir);
	void            SetViewSideUp(const D3DXVECTOR3& vectView, const D3DXVECTOR3& vectSide, const D3DXVECTOR3& vectUp);
    void            SetPosition(const D3DXVECTOR3& p);
    const D3DXVECTOR3&    GetView() const;
    const D3DXVECTOR3&    GetSide() const;
    const D3DXVECTOR3&    GetUp() const;
    const D3DXVECTOR3&    GetPosition() const;

	//----------------------------------------------------------------------------
	// Projection parameters
	//----------------------------------------------------------------------------
	void		    SetProjection(float degFovY, float aspect, float nearDist, float farDist);
	void		    SetProjection(float l, float r, float b, float t, float n, float f);
    void		    SetNear(float nearDist);
    void		    SetFar(float farDist);
	void		    SetAspect(float aspect);
    void		    SetDegFovY(float fovY);
    float           GetNear() const;
    float           GetFar() const;
	float			GetAspect() const;
    float           GetDegFovY() const;
	//----------------------------------------------------------------------------
	// Viewing window coordinates
	//----------------------------------------------------------------------------
	float           GetWindowWidth() const;
	float           GetWindowHeight() const;
	float           GetWindowLeft() const;
	float           GetWindowRight() const;
	float           GetWindowBottom() const;
	float           GetWindowTop() const;

	//----------------------------------------------------------------------------
	// camera geometry
	//----------------------------------------------------------------------------
	void            CalcWorldspaceVerts(D3DXVECTOR3 verts[8]) const;
	void            CalcWorldspacePlanes(D3DXVECTOR4 planes[6]) const;
	void            TransformWorldToCam(const D3DXVECTOR3 &worldPt, D3DXVECTOR3 &camPt) const;
	void            TransformCamToWorld(const D3DXVECTOR3 &camPt, D3DXVECTOR3 &worldPt) const;
	void            Transform(const D3DXMATRIX &mat);

	//----------------------------------------------------------------------------
    // Create a worldspace ray going through a pixel location starting at 
	// the pixel location (integers represent pixel centers).
    // Have to supply the viewport parameters
	//----------------------------------------------------------------------------
	void            CalcPixelRay(const D3DXVECTOR2& screenPos, D3DXVECTOR3& worldPos, D3DXVECTOR3& worldDir, const D3DXVECTOR2& viewportOrigin, const D3DXVECTOR2& viewportSize);

protected:
    void           _SyncViewingRectWithProjection();
    void           _SyncProjectionWithViewingRect();
private:
    D3DXVECTOR3          _x, _y, _z, _o;
    float          _fovY, _aspect, _l, _r, _b, _t, _n, _f;

};


//----------------------------------------------------------------------------
//-- External Variables ------------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Function Definitions ----------------------------------------------------
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// UtilCamera()
// ---------------------------------------------------------------------------
// ctor
//----------------------------------------------------------------------------
inline
UtilCamera::UtilCamera() :
_x(1,0,0), 
_y(0,1,0), 
_z(0,0,1),
_o(0,0,0),
_fovY(90), 
_aspect(1),
_l(-1), _r(1),
_b(-1), _t(1), 
_n(1), _f(10)
{
}


//----------------------------------------------------------------------------
// ~UtilCamera
// ---------------------------------------------------------------------------
// dtor
//----------------------------------------------------------------------------
inline
UtilCamera::~UtilCamera()
{
}

//----------------------------------------------------------------------------
// GetWorldToViewMatrix()
// ---------------------------------------------------------------------------
// *** THIS ROUTINE IS DEPENDEENT ON VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
inline
const D3DXMATRIX UtilCamera::GetWorldToViewMatrix() const
{
    D3DXMATRIX t;
	MatrixViewing(t, _x, _y, _z, _o);
    return t;
}

//----------------------------------------------------------------------------
// GetProjectionMatrix()
// ---------------------------------------------------------------------------
// *** THIS ROUTINE IS DEPENDEENT ON VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
inline
const D3DXMATRIX UtilCamera::GetProjectionMatrix() const
{
    D3DXMATRIX t;
    MatrixFrustumD3D(t, _l, _r, _b, _t, _n, _f);
    return t;
}

//----------------------------------------------------------------------------
// GetInvWorldToViewMatrix()
// ---------------------------------------------------------------------------
// *** THIS ROUTINE IS DEPENDEENT ON VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
inline
const D3DXMATRIX UtilCamera::GetInvWorldToViewMatrix() const
{
    D3DXMATRIX t;
	MatrixViewingInv(t, _x, _y, _z, _o);
    return t;
}

//----------------------------------------------------------------------------
// GetProjectionMatrix()
// ---------------------------------------------------------------------------
// *** THIS ROUTINE IS DEPENDEENT ON VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
inline
const D3DXMATRIX UtilCamera::GetInvProjectionMatrix() const
{
    D3DXMATRIX t;
    MatrixFrustumInvD3D(t, _l, _r, _b, _t, _n, _f);
    return t;
}

//----------------------------------------------------------------------------
// LookAt()
// ---------------------------------------------------------------------------
// *** THIS ROUTINE IS DEPENDEENT ON VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
inline
void UtilCamera::LookAt(const D3DXVECTOR3& eyePt, const D3DXVECTOR3& viewRefPt, const D3DXVECTOR3& viewUpDir)
{
    _o = eyePt;
    _z = viewRefPt-eyePt; D3DXVec3Normalize(&_z, &_z);
    D3DXVec3Cross(&_x, &viewUpDir, &_z);  D3DXVec3Normalize(&_x, &_x);
    D3DXVec3Cross(&_y, &_z, &_x);         D3DXVec3Normalize(&_y, &_y);
}

//----------------------------------------------------------------------------
// SetViewSideUp()
// ---------------------------------------------------------------------------
// Assumes the vectors are already normalized.
// *** THIS ROUTINE IS DEPENDEENT ON VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
inline
void UtilCamera::SetViewSideUp(const D3DXVECTOR3& vectView, const D3DXVECTOR3& vectSide, const D3DXVECTOR3& vectUp)
{
    _x = vectSide;
    _y = vectUp;
    _z = vectView;
}

//----------------------------------------------------------------------------
// SetPosition()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilCamera::SetPosition(const D3DXVECTOR3& p)
{
    _o = p;
}

//----------------------------------------------------------------------------
// GetView()
// ---------------------------------------------------------------------------
// *** THIS ROUTINE IS DEPENDEENT ON VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
inline
const D3DXVECTOR3& UtilCamera::GetView() const
{
    return _z;
}

//----------------------------------------------------------------------------
// GetSide()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
const D3DXVECTOR3& UtilCamera::GetSide() const
{
    return _x;
}

//----------------------------------------------------------------------------
// GetUp()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
const D3DXVECTOR3& UtilCamera::GetUp() const
{
    return _y;
}

//----------------------------------------------------------------------------
// GetPosition()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
const D3DXVECTOR3& UtilCamera::GetPosition() const
{
    return _o;
}

//----------------------------------------------------------------------------
// SetProjection()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilCamera::SetProjection(float degFovY, float aspect, float nearDist, float farDist)
{
    _fovY = degFovY;
    _aspect = aspect;
    _n = nearDist;
    _f = farDist;
    _SyncViewingRectWithProjection();
}

//----------------------------------------------------------------------------
// SetProjection()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilCamera::SetProjection(float l, float r, float b, float t, float n, float f)
{
    _l = l; _r = r;
    _b = b; _t = t;
    _n = n; _f = f;
    _SyncProjectionWithViewingRect();
}

//----------------------------------------------------------------------------
// SetNear()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilCamera::SetNear(float nearDist)
{
    _n = nearDist;
    _SyncProjectionWithViewingRect();
}

//----------------------------------------------------------------------------
// SetFar()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilCamera::SetFar(float farDist)
{
    _f = farDist;
}

//----------------------------------------------------------------------------
// SetAspect()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilCamera::SetAspect(float aspect)
{
    _aspect = aspect;
    _SyncViewingRectWithProjection();
}

//----------------------------------------------------------------------------
// SetFovY()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilCamera::SetDegFovY(float fovY)
{
    _fovY = fovY;
    _SyncViewingRectWithProjection();
}

//----------------------------------------------------------------------------
// GetNear()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetNear() const
{
    return _n;
}

//----------------------------------------------------------------------------
// GetFar()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetFar() const
{
    return _f;
}

//----------------------------------------------------------------------------
// GetAspect()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetAspect() const
{
    return _aspect;
}

//----------------------------------------------------------------------------
// GetDegFovY()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetDegFovY() const
{
    return _fovY;
}

//----------------------------------------------------------------------------
// GetWindowWidth()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetWindowWidth() const
{
    return _r-_l;
}

//----------------------------------------------------------------------------
// GetWindowHeight()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetWindowHeight() const
{
    return _t-_b;
}

//----------------------------------------------------------------------------
// GetWindowLeft()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetWindowLeft() const
{
    return _l;
}

//----------------------------------------------------------------------------
// GetWindowRight()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetWindowRight() const
{
    return _r;
}

//----------------------------------------------------------------------------
// GetWindowBottom()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetWindowBottom() const
{
    return _b;
}

//----------------------------------------------------------------------------
// GetWindowTop()
// ---------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilCamera::GetWindowTop() const
{
    return _t;
}

#endif
