//============================================================================
// UtilMouseNav.hpp:
//
// Encapsulates "mouse-style" movement using a UtilCamera.
//
// This class is VERY D3D specific. It also relies on the D3D-specific UtilCamera class.
//
// This class implements 7 common navigation modes, grouped into two categories:
// 1. Active - Must click-n-drag the mouse to navigate.
// 2. Passive - Mouse is captured and can only be used to navigate.
//
//	The active modes are:
//
//	1. Trackball - trackball-style movement. The object is a virtual sphere and
//	   dragging the mouse simulates rotating that sphere around. Similar to arcball,
//	   but path-dependent. I tend to like it better than arcball.
//	2. Drive - Dragging the mouse moves forward/back and turns left/right like a car.
//	3. Translate - Essentially a pan mode like found on common CAD programs.
//	4. Look - Dragging is equivalent to looking left/right/up/down.
//
//	The passive modes are:
//
//	1. Descent - 6 DOF movement. See the help string for key bindings.
//	2. Doom - Really more like Quake with mouselook. Movement always constrained 
//	   to a level plane. See the help string for key bindings.
//	3. Observe - Like Quake III Observe mode, which is in turn like Doom mode but
//	   movement is not constrained to a level plane. See the help string for key
//	   bindings. 
//
//	Notes:
//
//	* before using the UtilCamera, you must set a worldcenter and world radius, that is
//	  used for movement calculations. Also, a global up direction is needed for modes
//	  that require it (look, drive, doom, observe, etc.).
//
//	* Custom key bindings are not currently supported, but there's no reason they
//	  couldn't be. Since I am the only one using this stuff, I don't need it right now.
//
//	* The help string explains the key bindings. 
//
//	* Holding down the control and shift keys while using the active modes will alter
//	  the axes that are affected by the movement. Useful for cool camera movement for
//	  path recordings.
//
//	* Switching the InsideLookingOut mode (active modes only) will conceptually 
//	  switch the "sense" of the movement. Think of a 1st-person camera 
//	  (inside-looking-out) and a god's-eye-view of an object (outside-looking-in).
//
//	* The camera needs to know the window size to properly interpret the mouse messages.
//	  If you change the viewport in any way, be sure to call SetWindowSize().
//
//============================================================================
#ifndef UtilMouseNav__Included__
#define UtilMouseNav__Included__
#pragma once

#include "Util.hpp"
#include <windows.h>
#include <windowsx.h>
#include "UtilCamera.hpp"

class UtilMouseNav
{
public:
	inline UtilMouseNav();
	inline ~UtilMouseNav();

	//----------------------------------------------------------------------------
	// WndProc. The class must receive all messages from the message loop.
	//----------------------------------------------------------------------------
	LRESULT WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled);

    enum {
      NAV_MODE_TRACKBALL = 0, 
	  NAV_MODE_DRIVE     = 1, 
	  NAV_MODE_TRANSLATE = 2, 
	  NAV_MODE_LOOK    = 3,
	  NAV_MODE_PASSIVE = 4,
	  NAV_MODE_DESCENT = 4,
	  NAV_MODE_DOOM    = 5,
	  NAV_MODE_OBSERVE = 6
	};

	void SetNavMode(int mode);
	int GetNavMode()			{ return NavMode; }

	void SetActive(int active);
	int IsActive()				{ return Active; }

	void DoPassiveMovement(DWORD lastFrameTime);

    //----------------------------------------------------------------------------
	// "Mouse-style" UtilMouseNav routines.
    //----------------------------------------------------------------------------
    // Screenspace Panning
	void TranslateX(float OldX, float NewX);
	void TranslateY(float OldY, float NewY);
    // Walking Along Camera View Vector
	void DriveY(float OldY, float NewY);
    // Mouselook Around ViewUp direction
	void LookX(float OldX, float NewX);
	void LookY(float OldY, float NewY);
    // Orbiting Around The World Center
	void TrackBallX(float OldX, float NewX);
	void TrackBallY(float OldY, float NewY);
    // Twist Around Viewing Direction
    void TwistX(float OldX, float NewX);
    // Look Left/Right Around 
    void LookUserRefX(float OldX, float NewX);
    // Walking Constrained To Level Plane Defined By ViewUp Vector
    void DriveUserRefY(float OldY, float NewY);
    // Mouselook Around ViewUp Direction, Constrained to < 87 Degrees
    void LookConstrainedY(float OldY, float NewY);

	const char* GetHelpStr() const { return HelpStr; }

    // Navigation parameters
    void SetWindowSize(int width, int height);
    void SetViewUp(const D3DXVECTOR3& up);
    void SetWorldCenter(const D3DXVECTOR3& center);
    void SetWorldRadius(float radius);
    void SetInsideLookingOut(bool IO);
    void SetCamera(UtilCamera *pNewCam);

    int GetWindowWidth() const;
    int GetWidthHeight() const;
    const D3DXVECTOR3& GetViewUp() const;
    const D3DXVECTOR3& GetWorldCenter() const;
    float GetWorldRadius() const;
    bool GetInsideLookingOut() const;
    UtilCamera* GetCamera() const;

protected:
	LRESULT WndProcActive(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled);
	LRESULT WndProcDescent(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled);
	LRESULT WndProcDoom(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled);
	LRESULT WndProcObserve(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled);

	// public properties
	int	   NavMode;
	int	   Active;

	// Active movement state
    bool LeftButtonDown;
    float OldX, OldY, NewX, NewY;  // CURRENT OLD AND NEW MOUSE POSITIONS
    bool CtrlPressed;
	float Mult;

	// Passive movement state
	float PassiveMult;
	int PassiveBorderWidth;
    int SlidingLR;
    int SlidingUD;
    int SlidingFB;
    int Twisting;

	const char* HelpStr;

	int WW, WH;
	D3DXVECTOR3  ViewUp;
	D3DXVECTOR3  WorldCenter;
	float  WorldRadius;
	bool   InsideLookingOut;
	UtilCamera *pCam;

    // WARP CURSOR TO LEGAL PORTIONS OF THE SCREEN.
    void ConstrainCursor(HWND hWnd, int x, int y);

};

//----------------------------------------------------------------------------
// UtilMouseNav()
//----------------------------------------------------------------------------
// Ctor
//----------------------------------------------------------------------------
inline UtilMouseNav::UtilMouseNav() : 
Active(true),
ViewUp(0,1,0), 
InsideLookingOut(false),
WorldCenter(0,0,0),
WorldRadius(1.f),
LeftButtonDown(false),
Mult(1.f),
OldX(0), OldY(0), NewX(0), NewY(0),
CtrlPressed(false),
pCam(NULL),
PassiveMult(1.f),
NavMode(NAV_MODE_TRACKBALL),
PassiveBorderWidth(0),
SlidingLR(0),
SlidingUD(0),
SlidingFB(0),
Twisting(0),
HelpStr(
"MOVEMENT HELP:\n"
"==============\n"
"Active Movement Modes (Click and drag):\n"
"    F1 - Orbit\n"
"    F2 - Drive\n"
"    F3 - Pan\n"
"    F4 - Look\n"
"\n"
"Passive Movement Modes (No need to click):\n"
"    F5 - Descent (6-DOF)\n"
"         AWSD Slides, QE twists, mouse buttons thrust\n"
"    F6 - Doom (Constrained to level plane)\n"
"         AWSD Moves, C crouch, <SPACE> jump\n"
"    F7 - Quake Observe (Constrained up vector)\n"
"         AWSD Moves, C crouch, <SPACE> jump\n"
"    * Use mouse wheel (SLOWLY!) to change scale.\n"
"\n")
{}

//----------------------------------------------------------------------------
// ~UtilMouseNav()
//----------------------------------------------------------------------------
// Dtor
//----------------------------------------------------------------------------
inline UtilMouseNav::~UtilMouseNav()
{}

//----------------------------------------------------------------------------
// GetWindowWidth()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
int UtilMouseNav::GetWindowWidth() const
{
    return WW;
}

//----------------------------------------------------------------------------
// GetWindowHeight()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
int UtilMouseNav::GetWidthHeight() const
{
    return WH;
}


//----------------------------------------------------------------------------
// SetViewUp()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilMouseNav::SetViewUp(const D3DXVECTOR3& up)
{
    ViewUp = up;
}

//----------------------------------------------------------------------------
// GetViewUp()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
const D3DXVECTOR3& UtilMouseNav::GetViewUp() const
{
    return ViewUp;
}

//----------------------------------------------------------------------------
// SetWorldCenter()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilMouseNav::SetWorldCenter(const D3DXVECTOR3& center)
{
    WorldCenter = center;
}

//----------------------------------------------------------------------------
// GetWorldCenter()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
const D3DXVECTOR3& UtilMouseNav::GetWorldCenter() const
{
    return WorldCenter;
}

//----------------------------------------------------------------------------
// SetWorldRadius()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilMouseNav::SetWorldRadius(float radius)
{
    WorldRadius = radius;
}

//----------------------------------------------------------------------------
// GetWorldRadius()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
float UtilMouseNav::GetWorldRadius() const
{
    return WorldRadius;
}

//----------------------------------------------------------------------------
// SetInsideLookingOut()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilMouseNav::SetInsideLookingOut(bool IO)
{
    InsideLookingOut = IO;
}

//----------------------------------------------------------------------------
// GetInsideLookingOut()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
bool UtilMouseNav::GetInsideLookingOut() const
{
    return InsideLookingOut;
}

//----------------------------------------------------------------------------
// SetCamera()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
void UtilMouseNav::SetCamera(UtilCamera *pNewCam)
{
    pCam = pNewCam;
}

//----------------------------------------------------------------------------
// GetCamera()
//----------------------------------------------------------------------------
// 
//----------------------------------------------------------------------------
inline
UtilCamera* UtilMouseNav::GetCamera() const
{
    return pCam;
}

#endif
