//-----------------------------------------------------------------------------
// D3DUtilApp.cpp
//-----------------------------------------------------------------------------
// Simple engine utilizing D3DUtilContext.
// Handles device mode change notification and simple frame-rate
// independent animation control using timeGetTime.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// #includes
//-----------------------------------------------------------------------------
#include "stdafx.h"
#include "D3DUtilApp.hpp"
#include "D3DUtilContextDlg.hpp"
#include "D3DUtilEnumeration.hpp"
#include <mmsystem.h>

//-----------------------------------------------------------------------------
// statics
//-----------------------------------------------------------------------------
D3DUtilApp* D3DUtilApp::GlobalD3DUtilApp=NULL;

namespace Internal
{
    bool s_wasFullScreen = false;
    void AssertPreProcess()
    {
        if (D3DUtilContext::IsInitialized())
        {
            if (D3DUtilContext::Instance()->IsFullscreen())
            {
                s_wasFullScreen = true;
                FAIL_TEST(D3DUtilContext::Instance()->Clear());
            }
        }
    }

    void AssertPostProcess()
    {
        if (s_wasFullScreen)
        {
            s_wasFullScreen = false;
            D3DPRESENT_PARAMETERS d3dpp = D3DUtilContext::Instance()->GetPresentParameters();
            FAIL_ASSERT(D3DUtilContext::Instance()->InitDevice(D3DUtilContext::Instance()->GetCreationParameters(), d3dpp));
        }
    }

    bool RegisterAssertProcessing()
    {
        AuxAssert::InstallPreProcessFunc(&AssertPreProcess);
        AuxAssert::InstallPostProcessFunc(AssertPostProcess);
        return true;
    }

    bool s_RegisterAssertProcessing = RegisterAssertProcessing();
};

//-----------------------------------------------------------------------------
// StaticWndProc
//-----------------------------------------------------------------------------
// WndProc that links the WIN32 callback to the D3DUtilApp.
//-----------------------------------------------------------------------------
LRESULT CALLBACK D3DUtilApp::StaticWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    if (GlobalD3DUtilApp)
    {
        BOOL handled = FALSE;
	    LRESULT res = GlobalD3DUtilApp->WndProc(hWnd, uMsg, wParam, lParam, handled);
        if (handled) return res;
    }
    return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

//----------------------------------------------------------------------------
// D3DUtilApp
//----------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------
D3DUtilApp::D3DUtilApp() :
Font("Courier New", 8, 0), 
FPSWindowDuration(1000)
{
    ASSERT_MSG(GlobalD3DUtilApp == NULL, "D3DUtilApp: Instance already created");

    GlobalD3DUtilApp = this;
	AppWnd = NULL;
    DeviceWnd = NULL;
    AppWndStyle = 0;
    AppAccel = NULL;
	Ready = false;
    OneTimeInit = false;
	MainMenu = NULL;
    AppWndHooked = false;
    DeviceLost = false;
    
	Minimized   = false;
	InSizeMove  = false;
	SizeChanged = false;

	FPSWindowStart=0;
	FPSWindowEnd=0;
	FPSWindowCurrent=0;
	FPSWindowFrameCount=0;
	FPSLastFrameTime=0;
	FPS=0.f;
	TotalFrameTime=0;

	PauseCount=0;

    D3DUtilContext::Create(this);
}

//----------------------------------------------------------------------------
// ~D3DUtilApp
//----------------------------------------------------------------------------
// Destructor
//----------------------------------------------------------------------------
D3DUtilApp::~D3DUtilApp()
{
	if (GlobalD3DUtilApp == this) 
	{
        if (AppWnd) 
        {
            DestroyWindow(AppWnd); AppWnd = NULL;
        }
		GlobalD3DUtilApp = NULL;
	}
    D3DUtilContext::Destroy();
}

//-----------------------------------------------------------------------------
// WindowCreate
//-----------------------------------------------------------------------------
// Creates the window
//-----------------------------------------------------------------------------
BOOL D3DUtilApp::WindowCreate(HINSTANCE hInst, 
							   const char *windowTitle, 
							   INT width, 
							   INT height, 
							   DWORD MainMenuID /*=0*/, 
							   DWORD IconID /*=0*/,
                               DWORD MainAccelID /*=0*/)
{
	if (AppWnd != NULL) return FALSE;

    // Register the window class
    WNDCLASSEX wndClass = {sizeof(wndClass)};
	wndClass.style			= CS_CLASSDC | CS_HREDRAW | CS_VREDRAW;
	wndClass.lpfnWndProc	= StaticWndProc;
	wndClass.cbClsExtra		= 0;
	wndClass.cbWndExtra		= 0;
	wndClass.hInstance		= hInst;
	wndClass.hIcon			= LoadIcon(hInst, MAKEINTRESOURCE(IconID));
	wndClass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wndClass.hbrBackground	= (HBRUSH)GetStockObject(BLACK_BRUSH);
	wndClass.lpszMenuName	= NULL;
	wndClass.lpszClassName	= "D3DUtilApp App";
	wndClass.hIconSm		= NULL;

    RegisterClassEx(&wndClass);

	// Load the menu
    // Create the render window

	MainMenu = LoadMenu( hInst, MAKEINTRESOURCE(MainMenuID) );
    AppWndStyle = WS_OVERLAPPEDWINDOW|WS_VISIBLE;
    RECT rc;
    SetRect(&rc, 0, 0, width, height);
    AdjustWindowRect(&rc, AppWndStyle, MainMenu != NULL);
    AppWnd = CreateWindow( "D3DUtilApp App", windowTitle,
                           AppWndStyle,
                           0, 0, rc.right-rc.left, rc.bottom-rc.top,
                           NULL, 
						   MainMenu,
                           hInst, 0L );

	if (AppWnd == NULL) return FALSE;

    UpdateWindow(AppWnd);

    AppAccel = LoadAccelerators( hInst, MAKEINTRESOURCE(MainAccelID) );

    Ready = true;

    return TRUE;
}

//----------------------------------------------------------------------------
// HookWindow()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
BOOL
D3DUtilApp::HookWindow(HWND wnd, HACCEL MainAccelID)
{
    AppWnd = wnd;
	AppAccel = MainAccelID;
    Ready = true;
    AppWndHooked = true;
    AppWndStyle = GetWindowLong(wnd, GWL_STYLE);
    
    return TRUE;
}

//-----------------------------------------------------------------------------
// Run
//-----------------------------------------------------------------------------
// The WIN32 message loop. Handles timing and rendering as well.
//-----------------------------------------------------------------------------
HRESULT D3DUtilApp::Run(D3DUtilAppRunFlags deviceCreation, HWND deviceWnd)
{
    // Load keyboard accelerators
	HRESULT hres = D3D_OK;

    ASSERT_MSG(AppWnd != NULL, "D3DUtilApp: Window not created before calling Run()");
    DeviceWnd = deviceWnd;
	if (deviceCreation == DIALOG_DEVICE_CREATION)
	{
        do 
        {
            if (hres == D3DUTILERR_FAILEDCONFIRM)
            {
                MessageBox(NULL, "Selected device does not meet the capabilities required by the application. Please select another one.", "Device Confirmation Failure", MB_OK | MB_ICONEXCLAMATION);
            }
            hres = DoContextDlg(GetAppWnd(), DeviceWnd);
            if (hres == IDCANCEL) return D3D_OK;
        } while (hres == D3DUTILERR_FAILEDCONFIRM);
    } else if (deviceCreation == AUTOMATIC_DEVICE_CREATION)
    {
        hres = InternalAutomaticSearchForDevice();
    }

    if (SUCCEEDED(hres))
    {
	    hres=InternalRun();
    }
    FAIL_TEST(D3DUtilContext::Instance()->Clear());
    FAIL_RETURN(hres);

    if (!AppWndHooked)
    {
    	DestroyWindow(AppWnd);
    }
	AppWnd = NULL;
    if (AppAccel && !AppWndHooked) DestroyAcceleratorTable(AppAccel);
    
    OneTimeInit = false;
    return D3D_OK;
}

//-----------------------------------------------------------------------------
// InternalRun
//-----------------------------------------------------------------------------
// Won't call OneTimeSceneInit until the context becomes active.
//-----------------------------------------------------------------------------
HRESULT D3DUtilApp::InternalRun()
{
    BOOL bGotMsg;
    MSG  msg;
	msg.message = !WM_QUIT;
	BOOL goodToGo = FALSE;
	BOOL prevGoodToGo = FALSE;
	DeviceLost = false;

    while( WM_QUIT != msg.message  )
    {
		prevGoodToGo = goodToGo;
        goodToGo = D3DUtilContext::Instance()->IsActive() && Ready && !Minimized;
		if (goodToGo && !prevGoodToGo)
		{
			InitFPSWindow(timeGetTime());
		}

        // Use PeekMessage() if the app is active, so we can use idle time to render the scene.
        if (goodToGo)
        {
            bGotMsg = PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE );
        } else
        {
            bGotMsg = GetMessage( &msg, NULL, 0U, 0U );
        }

        if( bGotMsg )
        {
            // Translate and dispatch the message
            if( 0 == TranslateAccelerator( AppWnd, AppAccel, &msg ) )
            {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
            }
        }
        else
        {
            // Render a frame during idle time (no messages are waiting)
            if( goodToGo )
            {
				// Call the one-only scene init
				if (!OneTimeInit)
				{
					FAIL_RETURN_MSG(OneTimeSceneInit(), "D3DUtilApp: OneTimeSceneInit Failed");
					OneTimeInit = true;
				}
				// update the FPS counter window
				UpdateFPSWindow();
                FAIL_RETURN_MSG(UpdateScene(), "D3DUtilApp: UpdateScene Failed");
                FAIL_RETURN_MSG(InternalRenderScene(),"D3DUtilApp: RenderScene Failed");
                // Give up some time if we ware waiting on a lost device
                if (DeviceLost)
                {
                    Sleep(100);
                }
            }
        }
    }

    return D3D_OK;
}

//----------------------------------------------------------------------------
// InternalRenderScene()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
D3DUtilApp::InternalRenderScene()
{
    HRESULT hres;
	// Render the scene
    if (DeviceLost)
    {
        // If we lost the device, see if we can reset it
        FAIL_RETURN_MSG(HandleDeviceLost(), "D3DUtilApp: Failed while handling lost device");
    }
    // device might still be lost
	if (!DeviceLost)
    {
        if (FAILED(hres = RenderScene()))
	    {
		    // Handle lost devices possibly
		    if (hres == D3DERR_DEVICELOST)
		    {
			    DeviceLost = TRUE;
                Pause();
		    } else
		    {
                FAIL_RETURN(hres);
		    }
        }
    }
    return D3D_OK;
}

//-----------------------------------------------------------------------------
// HandleDeviceLost
//-----------------------------------------------------------------------------
// Handles lost devices. deviceLost is assumed intialized to TRUE. Returns new value.
//-----------------------------------------------------------------------------
HRESULT D3DUtilApp::HandleDeviceLost()
{
	HRESULT hr = D3DUtilContext::Instance()->GetD3DDevice()->TestCooperativeLevel();
	if (hr == D3DERR_DEVICENOTRESET || hr == D3D_OK)
	{
		// if TestCooperativeLevel shouldn't return success because the device was lost
		if (hr == D3D_OK)
		{
			FAIL_RETURN_MSG(E_FAIL, "D3DUtilApp: Device was lost, but TestCooperativeLevel returns no error");
		}
		// It's ok to reset the device now
		D3DPRESENT_PARAMETERS d3dpp = D3DUtilContext::Instance()->GetPresentParameters();
		FAIL_RETURN_MSG(D3DUtilContext::Instance()->ResetDevice(d3dpp), "D3DUtilApp: Failed to reset device after recovering from lost.");
        UnPause();
		DeviceLost = false;
        return D3D_OK;
	} else
	{
		//AUXTRACE("D3DUtilApp: Device is lost. Waiting on TestCooperativeLevel...");
	}
	return D3D_OK;
}

//----------------------------------------------------------------------------
// FullscreenToggle()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
bool
D3DUtilApp::FullscreenToggle()
{
    if (D3DUtilContext::Instance()->IsFullscreen())
    {
        SetWindowLong(GetAppWnd(), GWL_STYLE, WS_POPUP|WS_VISIBLE);
    } else
    {
        SetWindowLong(GetAppWnd(), GWL_STYLE, AppWndStyle);
    }
    SetWindowPos(GetAppWnd(), HWND_TOP, 0, 0, 0, 0, SWP_SHOWWINDOW | SWP_NOMOVE | SWP_NOSIZE | SWP_FRAMECHANGED);
    return true;
}

//-----------------------------------------------------------------------------
// Pause
//-----------------------------------------------------------------------------
// Pause the application. This works like a general semaphore (nestable).
//-----------------------------------------------------------------------------
void D3DUtilApp::Pause()
{
	PauseCount++;
}

//-----------------------------------------------------------------------------
// UnPause
//-----------------------------------------------------------------------------
// UnPause the application. This works like a general semaphore (nestable).
//-----------------------------------------------------------------------------
void D3DUtilApp::UnPause()
{
	if (PauseCount <= 0) return;
	PauseCount--;
}

//----------------------------------------------------------------------------
// ShowStats;
//----------------------------------------------------------------------------
// Stat display - call this in RenderScene to show counters, etc.
//----------------------------------------------------------------------------
HRESULT D3DUtilApp::ShowStats()
{
	static char buffer[128];
	sprintf(buffer, "FPS: %.2f", GetFPS());
    FAIL_RETURN(Font.LockText());
	FAIL_RETURN(Font.DrawText(5, 5, 0xFFFFFF00, buffer));
    FAIL_RETURN(Font.UnlockText());
	return D3D_OK;
}

//-----------------------------------------------------------------------------
// WndProc
//-----------------------------------------------------------------------------
// WIN32 message handling
//-----------------------------------------------------------------------------
LRESULT D3DUtilApp::WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled )
{
    if (!D3DUtilContext::IsInitialized()) return 0;
    D3DUtilContext* pCon = D3DUtilContext::Instance();

    switch( uMsg )
    {
        case WM_CLOSE:
			PostQuitMessage(0);
            handled = TRUE;
            return 0;
        
        case WM_DESTROY:
            AppWnd = NULL;
            handled = TRUE;
            return 0;

        case WM_PAINT:
            // If we are paused and windowed, force a scene rendering. 
            // Otherwise it will happen when it can.
            if (pCon->IsActive() && !pCon->IsFullscreen() && IsPaused() && OneTimeInit)
            {
                FAIL_RETURN(InternalRenderScene());
            } 
            {
                PAINTSTRUCT paint;
                BeginPaint(hWnd, &paint);
                EndPaint(hWnd, &paint);
            }
            handled = TRUE;
            return 0;

		case WM_ENTERSIZEMOVE:
            if (!pCon->IsActive() || pCon->IsFullscreen()) break;
			InSizeMove = true;
			SizeChanged = false;
			Pause();
            handled = TRUE;
			return 0;

		case WM_EXITSIZEMOVE:
		case WM_SIZE:

            if (!pCon->IsActive() || pCon->IsFullscreen()) break;
            if (uMsg == WM_EXITSIZEMOVE)
            {
			    InSizeMove = false;
    			UnPause();
            } else
            {
			    if (InSizeMove)
			    {
				    SizeChanged = true;
			    } else
			    {
                    if (wParam == SIZE_MINIMIZED)
                    {
					    Minimized = true;
                        Pause();
                    } else
                    {
					    if (Minimized)
					    {
						    Minimized = false;
                            UnPause();
					    }
                        SizeChanged = true;
                    }
                }
            }

			if (!InSizeMove && SizeChanged)
			{
                SizeChanged = false;
                // Only reset the device if we are using the default window size.
                D3DPRESENT_PARAMETERS d3dpp = pCon->GetPresentParameters();
                if (d3dpp.BackBufferWidth == 0 || d3dpp.BackBufferHeight == 0)
                {
    				FAIL_ASSERT_MSG(pCon->ResetDevice(d3dpp), "D3DUtilApp: Failed to reset device after window resize");
                }
			}
            handled = TRUE;
			return 0;

        case WM_NCHITTEST:
            // Prevent the user from selecting the menu in fullscreen mode
            if(pCon->IsFullscreen())
            {
                handled = TRUE;
                return HTCLIENT;
            }
            break;

        case WM_SYSCOMMAND:
            // Prevent moving/sizing and power loss in fullscreen mode
            switch( wParam )
            {
                case SC_MOVE:
                case SC_SIZE:
                case SC_MAXIMIZE:
                case SC_KEYMENU:
                case SC_MONITORPOWER:
                    if(D3DUtilContext::Instance()->IsFullscreen())
                    {
                        handled = TRUE;
                        return 1;
                    }
            }
            break;
	}

	return 0;
}

//-----------------------------------------------------------------------------
// InitFPSWindow
//-----------------------------------------------------------------------------
// Resets the FPS timing window. Called every time the timing window ends 
// or is reset.
//-----------------------------------------------------------------------------
void D3DUtilApp::InitFPSWindow(DWORD startTime)
{
	FPSWindowFrameCount=0;
	FPSWindowStart = startTime;
	FPSWindowCurrent = FPSWindowStart;
	FPSWindowEnd = FPSWindowStart + FPSWindowDuration;
}

//-----------------------------------------------------------------------------
// UpdateFPSWindow
//-----------------------------------------------------------------------------
// Called every frmae to update the FPS timing window for the last frame time.
//-----------------------------------------------------------------------------
void D3DUtilApp::UpdateFPSWindow()
{
	DWORD FPSWindowLast = FPSWindowCurrent;
	FPSWindowCurrent = timeGetTime();
    if (PauseCount == 0)
    {
	    FPSLastFrameTime = FPSWindowCurrent - FPSWindowLast;
    	TotalFrameTime += FPSLastFrameTime;
    } else
    {
        FPSLastFrameTime = 0;
    }
	// !!! WRH This will fail every 49.71 days when the timer wraps around !!!
	if (FPSWindowCurrent > FPSWindowEnd)
	{
		FPS = static_cast<float>(FPSWindowFrameCount)/static_cast<float>(FPSWindowCurrent-FPSWindowStart)*1000.f;
		// Update the display of the FPS
		InitFPSWindow(FPSWindowCurrent);
	}

	FPSWindowFrameCount++;
}

//----------------------------------------------------------------------------
// InvalidateDevice()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
D3DUtilApp::InvalidateDevice()
{
	FAIL_RETURN(Font.OnInvalidateDevice());
    return D3D_OK;
}

//----------------------------------------------------------------------------
// RebuildDevice()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
D3DUtilApp::RebuildDevice()
{
	FAIL_RETURN(Font.OnRebuildDevice());
    return D3D_OK;
}

//----------------------------------------------------------------------------
// PrepareReset()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
D3DUtilApp::PrepareReset()
{
	FAIL_RETURN(Font.OnPrepareReset());
    return D3D_OK;
}

//----------------------------------------------------------------------------
// Reset()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
D3DUtilApp::Reset()
{
	FAIL_RETURN(Font.OnReset());
    return D3D_OK;
}

//----------------------------------------------------------------------------
// InternalAutomaticSearchForDevice()
// ---------------------------------------------------------------------------
//  This routine could be a lot better. Work in progress...
//----------------------------------------------------------------------------
HRESULT
D3DUtilApp::InternalAutomaticSearchForDevice()
{
    // Notes for future enhancements:
    //  * only use the default adapter for now since I won't change the window location to match the monitor for that adapter
    //  * only check windowed modes
    //  * don't check multisampling flags (which do we check first).
    // Might ultimately need a GetMinSpecs() call for listeners to report their needed specs.
    D3DDEVICE_CREATION_PARAMETERS d3dcp;
    D3DPRESENT_PARAMETERS d3dpp;
    d3dcp.AdapterOrdinal = 0;
    d3dcp.hFocusWindow = GetAppWnd();
    d3dpp.hDeviceWindow = GetAppWnd();
    d3dpp.Windowed = TRUE;
    d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
    d3dpp.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
    d3dpp.BackBufferFormat = D3DUtilContext::Instance()->GetWindowedFormat();
    d3dpp.BackBufferWidth = 0;
    d3dpp.BackBufferHeight = 0;
    d3dpp.BackBufferCount = 1;
    d3dpp.MultiSampleType = D3DMULTISAMPLE_NONE;
    d3dpp.MultiSampleQuality = 0;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.Flags = 0;

//    d3dpp.EnableAutoDepthStencil;
//    d3dpp.AutoDepthStencilFormat;
//    d3dcp.BehaviorFlags;
//    d3dcp.DeviceType;
//    D3DFORMAT preferredBackBuffers[] = 
//    {
//	    D3DFMT_A8R8G8B8,
//        D3DFMT_A4R4G4B4,
//        D3DFMT_A1R5G5B5,
//        D3DFMT_X8R8G8B8,
//        D3DFMT_R8G8B8,
//        D3DFMT_X4R4G4B4,
//        D3DFMT_R5G6B5,
//        D3DFMT_X1R5G5B5,
//        D3DFMT_A8R3G3B2,
//        D3DFMT_R3G3B2,
//        D3DFMT_A8,
//    };
//    DWORD numPreferredBackBuffers = sizeof(preferredBackBuffers)/sizeof(preferredBackBuffers[0]);

    D3DFORMAT preferredDepthBuffers[] = 
    {
        D3DFMT_D24S8,
        D3DFMT_D24X4S4,
        D3DFMT_D15S1,
        D3DFMT_D32,
        D3DFMT_D24X8,
        D3DFMT_D16,
        D3DFMT_D16_LOCKABLE,
        D3DFMT_UNKNOWN,
    };
    DWORD numPreferredDepthBuffers = sizeof(preferredDepthBuffers)/sizeof(preferredDepthBuffers[0]);
    DWORD preferredBehaviors[] =
    {
        D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_PUREDEVICE,
        D3DCREATE_HARDWARE_VERTEXPROCESSING,
        D3DCREATE_MIXED_VERTEXPROCESSING,
        D3DCREATE_SOFTWARE_VERTEXPROCESSING,
    };
    DWORD numPreferredBehaviors = sizeof(preferredBehaviors)/sizeof(preferredBehaviors[0]);
    D3DDEVTYPE preferredDevices[] =
    {
        D3DDEVTYPE_HAL,
        D3DDEVTYPE_REF,

    };
    DWORD numPreferredDevices = sizeof(preferredDevices)/sizeof(preferredDevices[0]);

    // use the enumeration to pick a valid device (only use the default adapter for now...)
    for (int iDev=0;iDev<numPreferredDevices;++iDev)
    {
        const D3DUtilDeviceInfo* device = D3DUtilEnumeration::GetAdapterList()[0].FindDevice(preferredDevices[iDev]);
        if (device == NULL) continue;

        d3dcp.DeviceType = device->DeviceType;
        for (int iBehav=0;iBehav<numPreferredBehaviors;++iBehav)
        {
            if (((preferredBehaviors[iBehav] & D3DCREATE_SOFTWARE_VERTEXPROCESSING) != 0) ||
                ((device->DeviceCaps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) != 0))
            {
                const D3DUtilAdapterModeInfo* adapterMode = device->FindAdapterMode(d3dpp.BackBufferFormat, TRUE);
                if (adapterMode == NULL) continue;

                const D3DUtilBackBufferInfo* backBuffer = adapterMode->FindBackBuffer(d3dpp.BackBufferFormat);
                if (backBuffer == NULL) continue;

                d3dcp.BehaviorFlags = preferredBehaviors[iBehav];
                for (int iDepth=0;iDepth<numPreferredDepthBuffers;++iDepth)
                {
                    const D3DUtilDepthInfo* depth = backBuffer->FindDepth(preferredDepthBuffers[iDepth]);
                    if (depth == NULL) continue;

                    d3dpp.EnableAutoDepthStencil = depth->Format != D3DFMT_UNKNOWN;
                    d3dpp.AutoDepthStencilFormat = depth->Format;
                    const D3DUtilMultisampleInfo* multi = depth->FindMultisample(D3DMULTISAMPLE_NONE);
                    if (multi == NULL) continue;

                    if (device->DeviceType == D3DDEVTYPE_REF)
                    {
                        MessageBox(GetAppWnd(), "Could not find a suitable HAL device to support the needs of the current application. Switching to the (VERY SLOW) reference rasterizer.", "Automatic Device Selection Warning", MB_OK | MB_ICONWARNING);
                    }
                    HRESULT hres = D3DUtilContext::Instance()->InitDevice(d3dcp, d3dpp);
                    if (hres == D3DUTILERR_FAILEDCONFIRM) continue;
                    FAIL_RETURN(hres);
                    return D3D_OK;
                }
            }
        }
    }
    MessageBox(GetAppWnd(), "Could not instantiate a device (even the REF device). Contact the application creator for support.", "FATAL ERROR", MB_OK | MB_ICONERROR);
    FAIL_RETURN_MSG(E_FAIL, "!!! Failed to find a suitable device (even REF device) !!!");
    return D3D_OK;
}
