//-----------------------------------------------------------------------------
// D3DUtilApp.hpp
//-----------------------------------------------------------------------------
// Simple engine utilizing D3DUtilContext.
// Handles device mode change notification and simple frame-rate
// independent animation control using timeGetTime.
//
//-----------------------------------------------------------------------------
#ifndef D3DUtilApp_hpp_
#define D3DUtilApp_hpp_
#pragma once

//-----------------------------------------------------------------------------
// #includes
//-----------------------------------------------------------------------------
#include "Util.hpp"
#include "D3DUtilContext.hpp"
#include "D3DUtilFont.hpp"

enum D3DUtilAppRunFlags
{
    MANUAL_DEVICE_CREATION,   // the app will initialize the device itself somehow
    DIALOG_DEVICE_CREATION,   // a dialog should be displayed to let the user choose the device
    AUTOMATIC_DEVICE_CREATION, // the device should be determined automatically using ConfirmDevice
};

//-----------------------------------------------------------------------------
// D3DUtilApp
//-----------------------------------------------------------------------------
class D3DUtilApp : public D3DUtilContextListener
{
public:
	//----------------------------------------------------------------------------
	// Constructors/Destructors - Implemented to allow only one of these to be created
	//----------------------------------------------------------------------------
	D3DUtilApp();
	virtual ~D3DUtilApp();

	//----------------------------------------------------------------------------
	// This is a singleton, so supply an accessor
	//----------------------------------------------------------------------------
	static		D3DUtilApp* Instance()	{ return GlobalD3DUtilApp; }

	//----------------------------------------------------------------------------
	// Win32 application control - call then in your WinMan 
	//----------------------------------------------------------------------------
	BOOL WindowCreate(HINSTANCE hInst, const char *windowTitle, INT width=512, INT height=512, DWORD MainMenuID=0, DWORD IconID=0, DWORD MainAccelID=0);
    // or supply your own window
    BOOL HookWindow(HWND wnd, HACCEL=0);
	HRESULT Run(D3DUtilAppRunFlags deviceCreation=AUTOMATIC_DEVICE_CREATION, HWND deviceWnd=NULL);

	//----------------------------------------------------------------------------
	// Scene control - Override to manage the 3D scene
	//----------------------------------------------------------------------------
	virtual HRESULT OneTimeSceneInit()	{ return D3D_OK; }
	virtual HRESULT UpdateScene()		{ return D3D_OK; }
	virtual HRESULT RenderScene()		= 0;

	//----------------------------------------------------------------------------
	// WndProc. The class must receive all messages from the message loop.
    // You should always call the inherited WndProc first, then check the handled
    // flag to see if further processing is necessary.
    // If you then handle the message, set the handled flag.
	//----------------------------------------------------------------------------
	virtual LRESULT WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled);

	//----------------------------------------------------------------------------
	// Pause control (allows nesting)
	//----------------------------------------------------------------------------
	void Pause();
	void UnPause();
    bool IsPaused() const { return PauseCount>0; }

	//----------------------------------------------------------------------------
	// Stat display - call this in RenderScene to show counters, etc.
	//----------------------------------------------------------------------------
	virtual HRESULT ShowStats();

	//----------------------------------------------------------------------------
	// Property accessors
	//----------------------------------------------------------------------------
	bool		IsReady()				{ return Ready; }
	HWND		GetAppWnd()				{ return AppWnd; }
	HWND		GetDeviceWnd()			{ return DeviceWnd; }
	HMENU		GetMainMenu()			{ return MainMenu; }
	DWORD		GetLastFrameTime()		{ return FPSLastFrameTime; }
	DWORD       GetTotalFrameTime()     { return TotalFrameTime; }
	float		GetFPS()				{ return FPS; }
	D3DUtilFont&GetD3DFont()		    { return Font; }

	//----------------------------------------------------------------------------
	// Listener response routines.
	//----------------------------------------------------------------------------
	virtual HRESULT InvalidateDevice();
	virtual HRESULT RebuildDevice();
	virtual HRESULT PrepareReset();
	virtual HRESULT Reset();

    virtual bool FullscreenToggle();
    
protected:
private:
	//----------------------------------------------------------------------------
	// WIN32 state
	//----------------------------------------------------------------------------
	HWND AppWnd;
    HWND DeviceWnd;
    DWORD AppWndStyle;
    HACCEL AppAccel;
	bool Ready;
	bool OneTimeInit;
    bool AppWndHooked;
    bool DeviceLost;
	bool Minimized;
	bool InSizeMove;
	bool SizeChanged;
	HMENU MainMenu;
	D3DUtilFont Font;

	//----------------------------------------------------------------------------
	// Run Helpers
	//----------------------------------------------------------------------------
	HRESULT InternalRun();
    HRESULT InternalRenderScene();
	HRESULT HandleDeviceLost();
    HRESULT InternalAutomaticSearchForDevice();

	//----------------------------------------------------------------------------
	// Frame rate control
	//----------------------------------------------------------------------------
	void InitFPSWindow(DWORD startTime);
	void UpdateFPSWindow();

	//----------------------------------------------------------------------------
	// Frame rate state
	//----------------------------------------------------------------------------
	const unsigned FPSWindowDuration;		// ms per FPS timing window
	DWORD FPSWindowStart;					// FPS timing window start count
	DWORD FPSWindowEnd;						// FPS timing window end count
	DWORD FPSWindowCurrent;					// FPS timing window current count
	DWORD FPSWindowFrameCount;				// Number of frames rendered since FPSWindowStart
	DWORD FPSLastFrameTime;					// Duration of last frame in ms.
	float FPS;								// current FPS estimate (from last FPS timing window)
	DWORD TotalFrameTime;                   // Accumulation of all FPSLastFrameTimes since start of program.

	//----------------------------------------------------------------------------
	// pause state
	//----------------------------------------------------------------------------
	DWORD PauseCount;

	static D3DUtilApp* GlobalD3DUtilApp;
	static LRESULT CALLBACK StaticWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
};

#endif
