//-----------------------------------------------------------------------------
// D3DUtilContext.hpp
//-----------------------------------------------------------------------------
// Encapsulates the devices that encompass a D3D9 scene.
//
//-----------------------------------------------------------------------------
#ifndef D3DUtilContext_hpp_
#define D3DUtilContext_hpp_
#pragma once

//-----------------------------------------------------------------------------
// #includes
//-----------------------------------------------------------------------------
#include "Util.hpp"

//#pragma warning( disable : 4786)
//#include <stack>

#pragma warning( push )
#pragma warning( disable : 4786)

class D3DUtilContextListener;

#define D3DUTILERR_FAILEDCONFIRM          0x82000001

//-----------------------------------------------------------------------------
// D3DUtilContext
//-----------------------------------------------------------------------------
class D3DUtilContext
{
public:
	//----------------------------------------------------------------------------
	// Singleton methods
	//----------------------------------------------------------------------------
    // If you want, pass a listener class to be notified when the device status changes.
	//----------------------------------------------------------------------------
    static void Create(D3DUtilContextListener* listener = NULL);
    static void Destroy();
    static bool IsInitialized();
    static D3DUtilContext* Instance();

	//----------------------------------------------------------------------------
	// D3DDevice context control
	//----------------------------------------------------------------------------
	HRESULT InitDevice(const D3DDEVICE_CREATION_PARAMETERS& d3dcp, D3DPRESENT_PARAMETERS &d3dpp);
	HRESULT ResetDevice(D3DPRESENT_PARAMETERS &newParams);
	HRESULT Clear();

	//DWORD PushRenderState(D3DRENDERSTATETYPE stateType);
	//void  PopRenderState();
	//DWORD PushTextureStageState(DWORD index, D3DTEXTURESTAGESTATETYPE stateType);
	//void  PopTextureStageState();

	//----------------------------------------------------------------------------
	// Property accessors
	//----------------------------------------------------------------------------
	IDirect3D9*			GetD3D()					{ return pD3D; }
	IDirect3DDevice9*	GetD3DDevice()				{ return pD3DDevice; }
    const D3DCAPS9&     GetD3DCaps()                { return D3DCaps; }
	bool				IsActive() const			{ return Active; }
	const D3DDEVICE_CREATION_PARAMETERS& 
						GetCreationParameters()const{ return D3DCP; }
	const D3DPRESENT_PARAMETERS& 
						GetPresentParameters() const{ return D3DPP; }

	D3DFORMAT			GetWindowedFormat()	const	{ return WindowedFormat; }

	bool				IsStencil()	const			{ return Active && Stencil; }
	bool				IsDepth() const				{ return Active && D3DPP.EnableAutoDepthStencil == TRUE; }
	bool				IsFullscreen() const		{ return Active && !D3DPP.Windowed; }
    DWORD				GetRenderWidth() const		{ return Active?RenderWidth:0; }
    DWORD				GetRenderHeight() const		{ return Active?RenderHeight:0; }


protected:	
    //----------------------------------------------------------------------------
	// Contructors/Destructors
	//----------------------------------------------------------------------------
	D3DUtilContext(D3DUtilContextListener* listener);
	~D3DUtilContext();

private:
	//------------------------------------------------------------------------------
	// No assignment
	//------------------------------------------------------------------------------
	D3DUtilContext(const D3DUtilContext& context);
	D3DUtilContext& operator=(const D3DUtilContext& context);

	//------------------------------------------------------------------------------
	// Intialization
	//------------------------------------------------------------------------------
	void Init();
	void SetCreationParameters(const D3DDEVICE_CREATION_PARAMETERS& d3dcp);
	void SetPresentParameters(const D3DPRESENT_PARAMETERS& d3dpp);
    enum ActivationType
    {
        AT_Reset, AT_Rebuild
    };
    HRESULT Activate(ActivationType type, bool togglingFullscreen);
    HRESULT TryConfirmDevice(const D3DDEVICE_CREATION_PARAMETERS& d3dcp, const D3DPRESENT_PARAMETERS& d3dpp);

	HRESULT InitRenderStates();

	//------------------------------------------------------------------------------
	// Stack to hold renderstates
	//------------------------------------------------------------------------------
	//std::stack< std::pair<D3DRENDERSTATETYPE, DWORD> > RenderStateStack;
	//std::stack< std::pair<std::pair<DWORD, D3DTEXTURESTAGESTATETYPE>, DWORD> > TextureStageStateStack;

	//------------------------------------------------------------------------------
	// D3D objects
	//------------------------------------------------------------------------------
	IDirect3D9 *pD3D;
	IDirect3DDevice9 *pD3DDevice;
	D3DDEVICE_CREATION_PARAMETERS D3DCP;
	D3DPRESENT_PARAMETERS D3DPP;
    D3DCAPS9 D3DCaps;
	D3DFORMAT WindowedFormat;
	bool Stencil;

	//------------------------------------------------------------------------------
	// WIN32 state
	//------------------------------------------------------------------------------
	void SaveWindowPosition(const D3DDEVICE_CREATION_PARAMETERS &d3dcp, const D3DPRESENT_PARAMETERS &d3dpp);
	void RestoreWindowPosition(const D3DPRESENT_PARAMETERS& newParams);

    D3DUtilContextListener* Listener;
	bool Active;
	RECT OldWinRect;
	RECT OldClientRect;
	DWORD RenderWidth;
	DWORD RenderHeight;

    static D3DUtilContext* s_Context;
};

//----------------------------------------------------------------------------
// D3DUtilContextListener
//----------------------------------------------------------------------------
// These class can register with the context for notification when the context changes
//----------------------------------------------------------------------------
class D3DUtilContextListener
{
public:
	//----------------------------------------------------------------------------
	// Listener response routines.
	//----------------------------------------------------------------------------
	virtual HRESULT InvalidateDevice() { return D3D_OK; }
	virtual HRESULT RebuildDevice()    { return D3D_OK; }
	virtual HRESULT PrepareReset()     { return D3D_OK; }
	virtual HRESULT Reset()			   { return D3D_OK; }

    // Before a device is fully activated, check with the listener if it can actually use it.
    virtual bool ConfirmDevice(const D3DCAPS9& caps, const D3DDEVICE_CREATION_PARAMETERS& d3dcp, const D3DPRESENT_PARAMETERS& d3dpp) { return true; }

    /*
    // !!! WRH UNIMPLEMENTED This might be used to allow a slightly more flexible way of confirming the device.
    // First, let the listener report the min specs to pare down the list...
    struct MinSpecs
    {
        MinSpecs();

        bool Windowed;
        UINT FullScreen_PresentationInterval;
        UINT FullScreen_RefreshRateInHz;
        UINT BackBufferFormat;
        UINT BackBufferWidth;
        UINT BackBufferHeight;
        UINT BackBufferCount;
        BOOL EnableAutoDepthStencil;
        D3DMULTISAMPLE_TYPE MultiSampleType;
        D3DSWAPEFFECT SwapEffect;
        DWORD PresentFlags;
    };

    // Before trying to confirm the device, we can ask the listener for a set of minimum specs
    virtual void GetMinimumSpecs(MinSpecs& specs) {}
    */

    // After a window/fullscreen transition, this is called. Use Context::IsFullscreen() to test new value.
    // Allows the window styles to be changed as appropriate (like converting to a popup window in fullscreen)
    virtual bool FullscreenToggle() { return true; }
private:
};

inline
void D3DUtilContext::Create(D3DUtilContextListener* listener)
{
    ASSERT_MSG(s_Context == NULL, "!!! Tried to create more than one D3DUtilContext !!!");
    s_Context = new D3DUtilContext(listener);
}

inline
void D3DUtilContext::Destroy()
{
    SAFE_DELETE(s_Context);
}

inline
bool D3DUtilContext::IsInitialized()
{
    return s_Context != NULL;
}

inline
D3DUtilContext* D3DUtilContext::Instance()
{
    ASSERT_MSG(s_Context != NULL, "!!! Tried to reference NULL D3DUtilContext::Instance !!!");
    return s_Context;
}

/*
//------------------------------------------------------------------------------
// PushRenderState
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
inline DWORD D3DUtilContext::PushRenderState(D3DRENDERSTATETYPE stateType)
{
	std::pair<D3DRENDERSTATETYPE, DWORD> item(stateType, 0);
	FAIL_TEST(pD3DDevice->GetRenderState(stateType, &item.second));
	RenderStateStack.push(item);
	return item.second;
}

//------------------------------------------------------------------------------
// PopRenderState
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
inline void D3DUtilContext::PopRenderState()
{
	std::pair<D3DRENDERSTATETYPE, DWORD> item = RenderStateStack.top();
	FAIL_TEST(pD3DDevice->SetRenderState(item.first, item.second));
	RenderStateStack.pop();
}

//------------------------------------------------------------------------------
// PushTextureStageState
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
inline DWORD D3DUtilContext::PushTextureStageState(DWORD index, D3DTEXTURESTAGESTATETYPE stateType)
{
	std::pair<std::pair<DWORD, D3DTEXTURESTAGESTATETYPE>, DWORD> item(std::pair<DWORD, D3DTEXTURESTAGESTATETYPE>(index, stateType), 0);
	FAIL_TEST(pD3DDevice->GetTextureStageState(index, stateType, &item.second));
	TextureStageStateStack.push(item);
	return item.second;
}

//------------------------------------------------------------------------------
// PopTextureStageState
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
inline void D3DUtilContext::PopTextureStageState()
{
	std::pair<std::pair<DWORD, D3DTEXTURESTAGESTATETYPE>, DWORD> item = TextureStageStateStack.top();
	FAIL_TEST(pD3DDevice->SetTextureStageState(item.first.first, item.first.second, item.second));
	TextureStageStateStack.pop();
}
*/

#pragma warning( pop )

#endif
