//-----------------------------------------------------------------------------
// D3DUtilContextDlg.cpp
//-----------------------------------------------------------------------------
// Dialog box that allows you to set any options for the context mode.
// Does validation and matching so you can see what modes are actually
// available.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// #includes
//-----------------------------------------------------------------------------
#include "stdafx.h"
#include "D3DUtilContextDlg.hpp"
#include "D3DUtilContext.hpp"
#include "D3DUtilEnumeration.hpp"
#include "D3DUtilResource.h"

#include <string>
#include <sstream>

//-----------------------------------------------------------------------------
// INTERNAL VARIABLES FOR THE DIALOG BOXES
//-----------------------------------------------------------------------------
static D3DDEVICE_CREATION_PARAMETERS d3dcp;
static D3DPRESENT_PARAMETERS d3dpp;
static HWND DlgWnd=NULL;
static HWND AppWnd=NULL;
static HWND DevWnd=NULL;
static const char* STR_NONE = "None";
static const char* STR_DEFAULT = "Default";
static const char* STR_IMMEDIATE = "Immediate";
static const char* STR_ONE = "One";
static const char* STR_TWO = "Two";
static const char* STR_THREE = "Three";
static const char* STR_FOUR = "Four";

//-----------------------------------------------------------------------------
// FORWARD DECLARATIONS OF DLGFUNCS
//-----------------------------------------------------------------------------
BOOL CALLBACK SetContextDlgFunc(
  HWND hwndDlg,  // handle to dialog box
  UINT uMsg,     // message
  WPARAM wParam, // first message parameter
  LPARAM lParam  // second message parameter
);

//-----------------------------------------------------------------------------
// FORWARD DECLARATIONS
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Used to update the context when a dialog item changes
//-----------------------------------------------------------------------------
void InitSetContextDlg();
HRESULT SetContextFromDlg();

//-----------------------------------------------------------------------------
// certain notifications are recursively passed down because their data is dependent.
// such dependencies are noted by indentions below.
//-----------------------------------------------------------------------------
void NotifyAdapterChange();
  void NotifyDeviceTypeChange();
    void NotifyVertexProcessingModeChange();
      void NotifyPureDeviceChange();
      void NotifyDriverMgmtChange();
    void NotifyDisplayTypeChange();
      void NotifyBackBufferSizeChange();
      void NotifyDisplayFormatChange();
        void NotifyDisplayResolutionChange();
          void NotifyDisplayRefreshChange();
        void NotifyBackBufferFormatChange();
          void NotifyDepthFormatChange();
          void NotifyDepthDiscardChange();
            void NotifyMultisampleTypeChange();
              void NotifyMultisampleQualityChange();
      void NotifyPresentationIntervalChange();
    void NotifyAdapterGroupChange();
void NotifyBackBufferLockableChange();
void NotifyBackBufferCountChange();
void NotifyMiscFlagsChange();
void NotifySwapEffectChange();

//-----------------------------------------------------------------------------
// Used to update the dialog when a context item changes
// These are really only called when the dialog box is opened to sync the
// dialog to the current context. 
// Same dependencies as above.
//-----------------------------------------------------------------------------
void SetDlgFromContext();
void SetDlgAdapter(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
  void SetDlgDeviceType(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
    void SetDlgVertexProcessingMode(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
      void SetDlgPureDevice(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
      void SetDlgDriverMgmt(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
    void SetDlgDisplayType(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
      void SetDlgBackBufferSize(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
      void SetDlgDisplayFormat(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
        void SetDlgDisplayResolution(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
          void SetDlgDisplayRefresh(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
        void SetDlgBackBufferFormat(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
          void SetDlgDepthFormat(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
          void SetDlgDepthDiscard(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
            void SetDlgMultisampleType(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
              void SetDlgMultisampleQuality(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
      void SetDlgPresentationInterval(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
    void SetDlgAdapterGroup(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
void SetDlgBackBufferLockable(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
void SetDlgBackBufferCount(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
void SetDlgMiscFlags(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);
void SetDlgSwapEffect(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp);

//-----------------------------------------------------------------------------
// Utility
//-----------------------------------------------------------------------------
const D3DUtilAdapterInfo *GetAdapterInfo();
const D3DUtilDeviceInfo *GetDeviceInfo();
const D3DUtilAdapterModeInfo *GetAdapterModeInfo();
const D3DUtilBackBufferInfo *GetBackBufferInfo();
const D3DUtilDepthInfo *GetDepthInfo();
const D3DUtilMultisampleInfo *GetMultisampleInfo();

//-----------------------------------------------------------------------------
// DoContextDlg
//-----------------------------------------------------------------------------
// Allows a context mode to be set using a dialog box.
//-----------------------------------------------------------------------------
HRESULT DoContextDlg(HWND appWnd, HWND deviceWnd)
{
    AppWnd = appWnd;
    DevWnd = deviceWnd;
	return DialogBox(NULL, MAKEINTRESOURCE(IDD_SETCONTEXTDLG), appWnd, SetContextDlgFunc);
}

//-----------------------------------------------------------------------------
// SetContextDlgFunc
//-----------------------------------------------------------------------------
// Dialog func for SetContextDlg
//-----------------------------------------------------------------------------
BOOL CALLBACK SetContextDlgFunc(
  HWND hwndDlg,  // handle to dialog box
  UINT uMsg,     // message
  WPARAM wParam, // first message parameter
  LPARAM lParam  // second message parameter
)
{
	switch (uMsg) 
	{
	case WM_INITDIALOG:
		DlgWnd = hwndDlg;
		ZeroMemory(&d3dcp, sizeof(d3dcp));
		ZeroMemory(&d3dpp, sizeof(d3dpp));
		InitSetContextDlg();
        if (D3DUtilContext::Instance()->IsActive())
        {
		    SetDlgFromContext();
        }
		return TRUE;

    case WM_COMMAND: 
        switch (LOWORD(wParam)) 
        { 
        case IDOK: 
			wParam = SetContextFromDlg();
            EndDialog(hwndDlg, wParam);
			return true;

        case IDCANCEL: 
            EndDialog(hwndDlg, IDCANCEL);
            return true; 
		case IDC_ADAPTER_NAME:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyAdapterChange();
			}
			return true;
		case IDC_HAL:
		case IDC_REF:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDeviceTypeChange();
			}
			return true;
		case IDC_HARDWARE:
		case IDC_SOFTWARE:
		case IDC_MIXED:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyVertexProcessingModeChange();
			}
            return true;
        case IDC_PUREDEVICE:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyPureDeviceChange();
			}
            return true;
        case IDC_DISABLEDRIVERMGMT:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDriverMgmtChange();
			}
            return true;
        case IDC_ADAPTER_GROUP_DEVICE:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyAdapterGroupChange();
			}
            return true;
        case IDC_FPUPRESERVE:
        case IDC_MULTITHREADED:
        case IDC_DEVICECLIP:
        case IDC_VIDEO_HINT:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyMiscFlagsChange();
			}
            return true;
        case IDC_FULLSCREEN:
        case IDC_WINDOWED:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDisplayTypeChange();
			}
            return true;
		case IDC_FULLSCREEN_FORMAT:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyDisplayFormatChange();
			}
			return true;
		case IDC_DISPLAYMODE:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyDisplayResolutionChange();
			}
			return true;
		case IDC_REFRESH:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyDisplayRefreshChange();
			}
			return true;
        case IDC_BACKBUFFER_LOCKABLE:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyBackBufferLockableChange();
			}
            return true;
		case IDC_BACKBUFFER_FORMAT:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyBackBufferFormatChange();
			}
			return true;
		case IDC_BACKBUFFER_WIDTH:
		case IDC_BACKBUFFER_HEIGHT:
			if (HIWORD(wParam) == EN_CHANGE) 
			{
				NotifyBackBufferSizeChange();
			}
			return true;
		case IDC_BACKBUFFER_COUNT:
			if (HIWORD(wParam) == EN_CHANGE) 
			{
				NotifyBackBufferCountChange();
			}
			return true;
        case IDC_DISCARD:
        case IDC_FLIP:
        case IDC_COPY:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifySwapEffectChange();
			}
            return true;
		case IDC_DEPTHBUFFER_FORMAT:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyDepthFormatChange();
			}
			return true;
        case IDC_DEPTHBUFFER_DISCARD:
			if (HIWORD(wParam) == BN_CLICKED)
			{
				NotifyDepthDiscardChange();
			}
            return true;
		case IDC_MULTISAMPLE_TYPE:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyMultisampleTypeChange();
			}
			return true;
		case IDC_MULTISAMPLE_QUALITY:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyMultisampleQualityChange();
			}
			return true;
		case IDC_PRESENTATION_INTERVAL:
			if (HIWORD(wParam) == CBN_SELCHANGE) 
			{
				NotifyPresentationIntervalChange();
			}
			return true;
        } 
		return TRUE;
    } 

    return FALSE; 
}

//-----------------------------------------------------------------------------
// InitSetContextDlg
//-----------------------------------------------------------------------------
// Initializes the dialog box with appropriate details.
//-----------------------------------------------------------------------------
void InitSetContextDlg()
{
	unsigned int i;

    SetDlgItemText(DlgWnd, IDC_BACKBUFFER_WIDTH, "0");
    SetDlgItemText(DlgWnd, IDC_BACKBUFFER_HEIGHT, "0");
    SetDlgItemText(DlgWnd, IDC_BACKBUFFER_COUNT, "0");

    SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_SETCHECK, BST_CHECKED, 0);
    SendDlgItemMessage(DlgWnd, IDC_DISCARD, BM_SETCHECK, BST_CHECKED, 0);
    SendDlgItemMessage(DlgWnd, IDC_PUREDEVICE, BM_SETCHECK, BST_CHECKED, 0);
    
    // Clear the old adapter list
	SendDlgItemMessage(DlgWnd, IDC_ADAPTER_NAME, CB_RESETCONTENT, 0, 0);
	for (i=0;i<D3DUtilEnumeration::GetAdapterList().size(); i++)
	{
		SendDlgItemMessage(DlgWnd, IDC_ADAPTER_NAME, CB_ADDSTRING, 0, (LPARAM)D3DUtilEnumeration::GetAdapterList()[i].Description);
	}
	SetFocus(GetDlgItem(DlgWnd, IDC_ADAPTER_NAME));
	SendDlgItemMessage(DlgWnd, IDC_ADAPTER_NAME, CB_SETCURSEL, 0, 0);

    // Perform all notifications to set up a detault context setting.
    // Some notifications are recursively called, some arent.
    NotifyAdapterChange();
    NotifyBackBufferLockableChange();
    NotifyBackBufferCountChange();
    NotifyMiscFlagsChange();
    NotifySwapEffectChange();

    d3dcp.hFocusWindow = AppWnd;
    d3dpp.hDeviceWindow = DevWnd;
}

//-----------------------------------------------------------------------------
// SetDlgFromContext
//-----------------------------------------------------------------------------
// Set the dialog box to reflect the current context settings.
// Done when the dialog is first displayed.
//-----------------------------------------------------------------------------
void SetDlgFromContext()
{
	// Update the dialog to reflect the context settings
	D3DDEVICE_CREATION_PARAMETERS localcp = D3DUtilContext::Instance()->GetCreationParameters();
	D3DPRESENT_PARAMETERS localpp = D3DUtilContext::Instance()->GetPresentParameters();

    SetDlgAdapter(localcp, localpp);
    SetDlgBackBufferLockable(localcp, localpp);
    SetDlgBackBufferCount(localcp, localpp);
    SetDlgMiscFlags(localcp, localpp);
    SetDlgSwapEffect(localcp, localpp);
    d3dpp.hDeviceWindow = localpp.hDeviceWindow;
    d3dcp.hFocusWindow = localcp.hFocusWindow;
}

//-----------------------------------------------------------------------------
// SetDlgAdapter
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgAdapter(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    if (static_cast<LRESULT>(localcp.AdapterOrdinal) >= SendDlgItemMessage(DlgWnd, IDC_ADAPTER_NAME, CB_GETCOUNT, 0, 0))
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. Adapter index out of range");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_ADAPTER_NAME, CB_SETCURSEL, localcp.AdapterOrdinal, 0);
        NotifyAdapterChange();
        SetDlgDeviceType(localcp, localpp);
    }
}

//-----------------------------------------------------------------------------
// SetDlgDeviceType
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDeviceType(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    INT type[] = {D3DDEVTYPE_HAL, D3DDEVTYPE_REF};
    INT ctrl[] = {IDC_HAL, IDC_REF};
    bool valid = false;
    for (int i=0; i<2; ++i)
    {
        if (localcp.DeviceType == type[i])
        {
            if (!IsWindowEnabled(GetDlgItem(DlgWnd, ctrl[i])))
            {
                ASSERT_MSG(false, "Current settings not valiated by device enumeration. device type invalid.");
            } else
            {
                SendDlgItemMessage(DlgWnd, ctrl[i], BM_SETCHECK, BST_CHECKED, 0);
                NotifyDeviceTypeChange();
                SetDlgVertexProcessingMode(localcp, localpp);
                SetDlgAdapterGroup(localcp, localpp);
                SetDlgDisplayType(localcp, localpp);
                valid = true;
            }
        } else
        {
            SendDlgItemMessage(DlgWnd, ctrl[i], BM_SETCHECK, BST_UNCHECKED, 0);
        }
    }
    if (!valid)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. device type invalid.");
    }
}

//-----------------------------------------------------------------------------
// SetDlgVertexProcessingMode
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgVertexProcessingMode(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    INT type[] = {D3DCREATE_HARDWARE_VERTEXPROCESSING, D3DCREATE_MIXED_VERTEXPROCESSING, D3DCREATE_SOFTWARE_VERTEXPROCESSING};
    INT ctrl[] = {IDC_HARDWARE, IDC_MIXED, IDC_SOFTWARE};
    bool valid = false;
    for (int i=0;i<3;++i)
    {
        if (localcp.BehaviorFlags & type[i])
        {
            if (!IsWindowEnabled(GetDlgItem(DlgWnd, ctrl[i])))
            {
                ASSERT_MSG(false, "Current settings not valiated by device enumeration. vertex processing type invalid.");
            } else
            {
                SendDlgItemMessage(DlgWnd, ctrl[i], BM_SETCHECK, BST_CHECKED, 0);
                NotifyVertexProcessingModeChange();
                SetDlgPureDevice(localcp, localpp);
                SetDlgDriverMgmt(localcp, localpp);
                valid = true;
            }
        } else
        {
            SendDlgItemMessage(DlgWnd, ctrl[i], BM_SETCHECK, BST_UNCHECKED, 0);
        }
    }
    if (!valid)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. vertex processing type invalid.");
    }
}

//-----------------------------------------------------------------------------
// SetDlgPureDevice
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgPureDevice(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    if (localcp.BehaviorFlags & D3DCREATE_PUREDEVICE && !IsWindowEnabled(GetDlgItem(DlgWnd, IDC_PUREDEVICE)))
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. pure device flag is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_PUREDEVICE, BM_SETCHECK, localcp.BehaviorFlags & D3DCREATE_PUREDEVICE ? BST_CHECKED : BST_UNCHECKED, 0);
        NotifyPureDeviceChange();
    }
}

//-----------------------------------------------------------------------------
// SetDlgDriverMgmt
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDriverMgmt(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    if (localcp.BehaviorFlags & D3DCREATE_DISABLE_DRIVER_MANAGEMENT && !IsWindowEnabled(GetDlgItem(DlgWnd, IDC_DISABLEDRIVERMGMT)))
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. driver management flag is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_DISABLEDRIVERMGMT, BM_SETCHECK, localcp.BehaviorFlags & D3DCREATE_DISABLE_DRIVER_MANAGEMENT ? BST_CHECKED : BST_UNCHECKED, 0);
        NotifyDriverMgmtChange();
    }
}

//-----------------------------------------------------------------------------
// SetDlgAdapterGroup
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgAdapterGroup(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    if (localcp.BehaviorFlags & D3DCREATE_ADAPTERGROUP_DEVICE && !IsWindowEnabled(GetDlgItem(DlgWnd, IDC_ADAPTER_GROUP_DEVICE)))
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. adapter group device flag is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_ADAPTER_GROUP_DEVICE, BM_SETCHECK, localcp.BehaviorFlags & D3DCREATE_ADAPTERGROUP_DEVICE ? BST_CHECKED : BST_UNCHECKED, 0);
        NotifyAdapterGroupChange();
    }
}

//-----------------------------------------------------------------------------
// SetDlgDisplayType
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDisplayType(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    if (localpp.Windowed == TRUE)
    {
        if (!IsWindowEnabled(GetDlgItem(DlgWnd, IDC_WINDOWED)))
        {
            ASSERT_MSG(false, "Current settings not valiated by device enumeration. display type is invalid.");
        } else
        {
            SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_SETCHECK, BST_CHECKED, 0);
            NotifyDisplayTypeChange();
            SetDlgBackBufferSize(localcp, localpp);
            SetDlgPresentationInterval(localcp, localpp);
            SetDlgBackBufferFormat(localcp, localpp);
        }
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN, BM_SETCHECK, BST_CHECKED, 0);
        NotifyDisplayTypeChange();
        SetDlgPresentationInterval(localcp, localpp);
        SetDlgDisplayFormat(localcp, localpp);
    }
}

//-----------------------------------------------------------------------------
// SetDlgBackBufferSize
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgBackBufferSize(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    char buffer[32];
    if (localpp.Windowed)
    {
        sprintf(buffer, "%d", localpp.BackBufferWidth);
        SetDlgItemText(DlgWnd, IDC_BACKBUFFER_WIDTH, buffer);
        sprintf(buffer, "%d", localpp.BackBufferHeight);
        SetDlgItemText(DlgWnd, IDC_BACKBUFFER_HEIGHT, buffer);
        NotifyBackBufferSizeChange();
    }
}

//-----------------------------------------------------------------------------
// SetDlgPresentationInterval
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgPresentationInterval(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    const char* str = "";
    if (localpp.PresentationInterval == D3DPRESENT_INTERVAL_DEFAULT) 
    {
        str = STR_DEFAULT;
    } else if (localpp.PresentationInterval == D3DPRESENT_INTERVAL_IMMEDIATE)
    {
        str = STR_IMMEDIATE;
    } else if (localpp.PresentationInterval == D3DPRESENT_INTERVAL_ONE)
    {
        str = STR_ONE;
    } else if (localpp.PresentationInterval == D3DPRESENT_INTERVAL_TWO)
    {
        str = STR_TWO;
    } else if (localpp.PresentationInterval == D3DPRESENT_INTERVAL_THREE)
    {
        str = STR_THREE;
    } else if (localpp.PresentationInterval == D3DPRESENT_INTERVAL_THREE)
    {
        str = STR_FOUR;
    } else if (localpp.PresentationInterval == D3DPRESENT_INTERVAL_FOUR)
    {
        str = STR_FOUR;
    }
    INT ret = SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_FINDSTRINGEXACT, -1, (LPARAM)str);
    if (ret == CB_ERR)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. presentation interval is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_SETCURSEL, ret, 0);
        NotifyPresentationIntervalChange();
    }
}

//-----------------------------------------------------------------------------
// SetDlgDisplayFormat
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDisplayFormat(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    // strip off alpha from format
    D3DFORMAT format = localpp.BackBufferFormat;
    if (format == D3DFMT_A8R8G8B8) 
    {
        format = D3DFMT_X8R8G8B8;
    } else if (format == D3DFMT_A1R5G5B5)
    {
        format = D3DFMT_X1R5G5B5;
    }
    const char* str = EnumToString(format);

    INT ret = SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_FINDSTRINGEXACT, -1, (LPARAM)str);
    if (ret == CB_ERR)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. fullscreen display format is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_SETCURSEL, ret, 0);
        NotifyDisplayFormatChange();
        SetDlgDisplayResolution(localcp, localpp);
    }
}

//-----------------------------------------------------------------------------
// SetDlgDisplayResolution
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDisplayResolution(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    char buffer[32];
    sprintf(buffer, "%d x %d", localpp.BackBufferWidth, localpp.BackBufferHeight);
    INT ret = SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_FINDSTRINGEXACT, -1, (LPARAM)buffer);
    if (ret == CB_ERR)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. fullscreen resolution is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, ret, 0);
        NotifyDisplayResolutionChange();
        SetDlgDisplayRefresh(localcp, localpp);
    }
}

//-----------------------------------------------------------------------------
// SetDlgDisplayRefresh
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDisplayRefresh(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    char buffer[32];
    sprintf(buffer, "%d", localpp.FullScreen_RefreshRateInHz);
    INT ret = SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_FINDSTRINGEXACT, -1, (LPARAM)buffer);
    if (ret == CB_ERR)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. fullscreen refresh is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_SETCURSEL, ret, 0);
        NotifyDisplayRefreshChange();
        SetDlgBackBufferFormat(localcp, localpp);
    }
}

//-----------------------------------------------------------------------------
// SetDlgBackBufferFormat
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgBackBufferFormat(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    const char* str = EnumToString(localpp.BackBufferFormat);

    INT ret = SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_FINDSTRINGEXACT, -1, (LPARAM)str);
    if (ret == CB_ERR)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. Backbuffer format is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_SETCURSEL, ret, 0);
        NotifyBackBufferFormatChange();
        SetDlgDepthFormat(localcp, localpp);
    }
}

//-----------------------------------------------------------------------------
// SetDlgDepthFormat
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDepthFormat(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    const char* str = EnumToString(localpp.AutoDepthStencilFormat);
    if (!localpp.EnableAutoDepthStencil)
    {
        str = STR_NONE;
    }
    INT ret = SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_FORMAT, CB_FINDSTRINGEXACT, -1, (LPARAM)str);
    if (ret == CB_ERR)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. depth buffer format is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_FORMAT, CB_SETCURSEL, ret, 0);
        NotifyDepthFormatChange();
        SetDlgDepthDiscard(localcp, localpp);
        SetDlgMultisampleType(localcp, localpp);
    }
}

//-----------------------------------------------------------------------------
// SetDlgDepthDiscard
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgDepthDiscard(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    if (localpp.Flags & D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL && !IsWindowEnabled(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER_DISCARD)))
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. disard depth/stencil flag is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_DISCARD, BM_SETCHECK, localpp.Flags & D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL ? BST_CHECKED : BST_UNCHECKED, 0);
        NotifyDepthDiscardChange();
    }
}

//-----------------------------------------------------------------------------
// SetDlgMultisampleType
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgMultisampleType(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    const char* str = EnumToString(localpp.MultiSampleType);
    INT ret = SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_TYPE, CB_FINDSTRINGEXACT, -1, (LPARAM)str);
    if (ret == CB_ERR)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. multisample type is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_TYPE, CB_SETCURSEL, ret, 0);
        NotifyMultisampleTypeChange();
        if (localpp.MultiSampleType != D3DMULTISAMPLE_NONE)
        {
            SetDlgMultisampleQuality(localcp, localpp);
        }
    }

}

//-----------------------------------------------------------------------------
// SetDlgMultisampleQuality
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgMultisampleQuality(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    if (static_cast<LRESULT>(localpp.MultiSampleQuality) >= SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_QUALITY, CB_GETCOUNT, 0, 0))
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. multisample quality is invalid.");
    } else
    {
        SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_QUALITY, CB_SETCURSEL, localpp.MultiSampleQuality, 0);
        NotifyMultisampleQualityChange();
    }
}

//-----------------------------------------------------------------------------
// SetDlgBackBufferLockable
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgBackBufferLockable(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_LOCKABLE, BM_SETCHECK, localpp.Flags & D3DPRESENTFLAG_LOCKABLE_BACKBUFFER ? BST_CHECKED : BST_UNCHECKED, 0);
    NotifyBackBufferLockableChange();
}

//-----------------------------------------------------------------------------
// SetDlgBackBufferCount
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgBackBufferCount(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    char buffer[32];
    sprintf(buffer, "%d", localpp.BackBufferCount);
    SetDlgItemText(DlgWnd, IDC_BACKBUFFER_COUNT, buffer);
    NotifyBackBufferCountChange();
}

//-----------------------------------------------------------------------------
// SetDlgMiscFlags
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgMiscFlags(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    SendDlgItemMessage(DlgWnd, IDC_DEVICECLIP, BM_SETCHECK, localpp.Flags & D3DPRESENTFLAG_DEVICECLIP ? BST_CHECKED : BST_UNCHECKED, 0);
    SendDlgItemMessage(DlgWnd, IDC_VIDEO_HINT, BM_SETCHECK, localpp.Flags & D3DPRESENTFLAG_VIDEO? BST_CHECKED : BST_UNCHECKED, 0);
    SendDlgItemMessage(DlgWnd, IDC_FPUPRESERVE, BM_SETCHECK, localcp.BehaviorFlags & D3DCREATE_FPU_PRESERVE ? BST_CHECKED : BST_UNCHECKED, 0);
    SendDlgItemMessage(DlgWnd, IDC_MULTITHREADED, BM_SETCHECK, localcp.BehaviorFlags & D3DCREATE_MULTITHREADED ? BST_CHECKED : BST_UNCHECKED, 0);
    NotifyMiscFlagsChange();
}

//-----------------------------------------------------------------------------
// SetDlgSwapEffect
//-----------------------------------------------------------------------------
// Used to validate and sync the dialog to the local parameters.
//-----------------------------------------------------------------------------
void SetDlgSwapEffect(const D3DDEVICE_CREATION_PARAMETERS& localcp, const D3DPRESENT_PARAMETERS& localpp)
{
    D3DSWAPEFFECT type[] = {D3DSWAPEFFECT_DISCARD, D3DSWAPEFFECT_FLIP, D3DSWAPEFFECT_COPY};
    INT ctrl[] = {IDC_DISCARD, IDC_FLIP, IDC_COPY};
    bool valid = false;
    for (int i=0;i<3;++i)
    {
        if (localpp.SwapEffect == type[i])
        {
            SendDlgItemMessage(DlgWnd, ctrl[i], BM_SETCHECK, BST_CHECKED, 0);
            NotifySwapEffectChange();
            valid = true;
        } else
        {
            SendDlgItemMessage(DlgWnd, ctrl[i], BM_SETCHECK, BST_UNCHECKED, 0);
        }
    }
    if (!valid)
    {
        ASSERT_MSG(false, "Current settings not valiated by device enumeration. swap effect is invalid.");
    }
}

//-----------------------------------------------------------------------------
// SetContextFromDlg
//-----------------------------------------------------------------------------
// Set the context to reflect the current dialog box settings.
// Done before the dialog box closes if the user hits OK.
//-----------------------------------------------------------------------------
HRESULT SetContextFromDlg()
{
	if (D3DUtilContext::Instance()->IsActive())
	{
		D3DDEVICE_CREATION_PARAMETERS contextd3dcp = D3DUtilContext::Instance()->GetCreationParameters();

		if (memcmp(&d3dcp, &contextd3dcp, sizeof(d3dcp)) == 0)
		{
			FAIL_RETURN(D3DUtilContext::Instance()->ResetDevice(d3dpp));
            return D3D_OK;
		}
	}
	FAIL_RETURN(D3DUtilContext::Instance()->InitDevice(d3dcp, d3dpp));
	return D3D_OK;
}

//-----------------------------------------------------------------------------
// NotifyAdapterChange
//-----------------------------------------------------------------------------
// Set the context adapter to reflect the current dialog box adapter
//-----------------------------------------------------------------------------
void NotifyAdapterChange()
{
	unsigned int adapter = SendDlgItemMessage(DlgWnd, IDC_ADAPTER_NAME, CB_GETCURSEL, 0, 0);

    if (adapter < D3DUtilEnumeration::GetAdapterList().size())
    {
	    // update the context
	    d3dcp.AdapterOrdinal = adapter;

	    // Set the description text box
	    SetDlgItemText(DlgWnd, IDC_ADAPTER_DRIVER, D3DUtilEnumeration::GetAdapterList()[adapter].Driver);
	    SetDlgItemText(DlgWnd, IDC_ADAPTER_DESC, D3DUtilEnumeration::GetAdapterList()[adapter].Description);

        //////////////////////////////////
	    // Set up the device type radio group.
        //////////////////////////////////

        // Clear the radio buttons
	    const D3DUtilDeviceInfoList &deviceList = D3DUtilEnumeration::GetAdapterList()[adapter].Devices;
	    EnableWindow(GetDlgItem(DlgWnd, IDC_HAL), FALSE); SendDlgItemMessage(DlgWnd, IDC_HAL, BM_SETCHECK, BST_UNCHECKED, 0);
	    EnableWindow(GetDlgItem(DlgWnd, IDC_REF), FALSE); SendDlgItemMessage(DlgWnd, IDC_REF, BM_SETCHECK, BST_UNCHECKED, 0);
	    bool found=false;
	    for (unsigned int i=0; i<deviceList.size(); i++)
	    {
		    int ID;
		    switch (deviceList[i].DeviceType)
		    {
		    case D3DDEVTYPE_HAL: ID = IDC_HAL; break;
		    case D3DDEVTYPE_REF: ID = IDC_REF; break;
		    }

		    EnableWindow(GetDlgItem(DlgWnd, ID), TRUE);
            // Select the first one we find.
		    if (!found)
		    {
			    SendDlgItemMessage(DlgWnd, ID, BM_SETCHECK, BST_CHECKED, 0);
			    found = true;
		    }
	    }
	    NotifyDeviceTypeChange();
    }
}

//-----------------------------------------------------------------------------
// NotifyDeviceTypeChange
//-----------------------------------------------------------------------------
// Set the context device type to reflect the current dialog box device type.
// Also enable device creation flags that make sense for that device.
//-----------------------------------------------------------------------------
void NotifyDeviceTypeChange()
{
    //////////////////////////////////
	// update the context to match the dialog settings
    //////////////////////////////////

    if (SendDlgItemMessage(DlgWnd, IDC_HAL, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.DeviceType = D3DDEVTYPE_HAL;
	} else if (SendDlgItemMessage(DlgWnd, IDC_REF, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dcp.DeviceType = D3DDEVTYPE_REF;
	}

    //////////////////////////////////
	// See what vertex processing options are available
    //////////////////////////////////

    const D3DUtilDeviceInfo *deviceInfo = GetDeviceInfo();
	// Software is always available
	if (deviceInfo && deviceInfo->DeviceCaps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT)
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_HARDWARE), TRUE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_MIXED), TRUE);
	} else
	{
		SendDlgItemMessage(DlgWnd, IDC_SOFTWARE, BM_SETCHECK, BST_CHECKED, 0);
		EnableWindow(GetDlgItem(DlgWnd, IDC_HARDWARE), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_SETCHECK, BST_UNCHECKED, 0);
		EnableWindow(GetDlgItem(DlgWnd, IDC_MIXED), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_MIXED, BM_SETCHECK, BST_UNCHECKED, 0);
	}
	NotifyVertexProcessingModeChange();

    //////////////////////////////////
	// Set up the display type
    //////////////////////////////////

	// See if we can enable windowed mode ??? can ALL devices in DX9 ???
//	if (deviceInfo && deviceInfo->DeviceCaps.Caps2 & D3DCAPS2_CANRENDERWINDOWED)
    if (true)
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_WINDOWED), TRUE);
		SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_SETCHECK, BST_CHECKED, 0);
		SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN, BM_SETCHECK, BST_UNCHECKED, 0);
	} else
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_WINDOWED), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_SETCHECK, BST_UNCHECKED, 0);
		SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN, BM_SETCHECK, BST_CHECKED, 0);
	}
	NotifyDisplayTypeChange();

    //////////////////////////////////
	// Set up master adapter flag
    //////////////////////////////////
    if (deviceInfo && deviceInfo->DeviceCaps.AdapterOrdinalInGroup == 0)
    {
        EnableWindow(GetDlgItem(DlgWnd, IDC_ADAPTER_GROUP_DEVICE), TRUE);
    } else
    {
        EnableWindow(GetDlgItem(DlgWnd, IDC_ADAPTER_GROUP_DEVICE), FALSE);
    }
    NotifyAdapterGroupChange();
}

//-----------------------------------------------------------------------------
// NotifyAdapterGroupChange
//-----------------------------------------------------------------------------
// Set the context adapter group flag to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyAdapterGroupChange()
{
    const D3DUtilDeviceInfo* deviceInfo = GetDeviceInfo();
    if (deviceInfo)
    {
        if (SendDlgItemMessage(DlgWnd, IDC_ADAPTER_GROUP_DEVICE, BM_GETCHECK, 0, 0) == BST_CHECKED)
        {
            d3dcp.BehaviorFlags |= D3DCREATE_ADAPTERGROUP_DEVICE;
        } else
        {
            d3dcp.BehaviorFlags &= ~D3DCREATE_ADAPTERGROUP_DEVICE;
        }
    }
}

//-----------------------------------------------------------------------------
// NotifyVertexProcessingModeChange
//-----------------------------------------------------------------------------
// Set the context VP mode to reflect the current dialog box VP mode.
//-----------------------------------------------------------------------------
void NotifyVertexProcessingModeChange()
{
    //////////////////////////////////
	// update the context to match the dialog settings
    //////////////////////////////////

    d3dcp.BehaviorFlags &= ~(D3DCREATE_HARDWARE_VERTEXPROCESSING | D3DCREATE_SOFTWARE_VERTEXPROCESSING | D3DCREATE_MIXED_VERTEXPROCESSING);
    if (SendDlgItemMessage(DlgWnd, IDC_HARDWARE, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dcp.BehaviorFlags |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
    } else
    if (SendDlgItemMessage(DlgWnd, IDC_MIXED, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dcp.BehaviorFlags |= D3DCREATE_MIXED_VERTEXPROCESSING;
    } else
    {
        d3dcp.BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    }

    //////////////////////////////////
	// update the dialog based on the device type
    //////////////////////////////////

	const D3DUtilDeviceInfo *deviceInfo = GetDeviceInfo();
	// See if we have a pure device
	if (deviceInfo && (deviceInfo->DeviceCaps.DevCaps & D3DDEVCAPS_PUREDEVICE) && 
        (d3dcp.BehaviorFlags & D3DCREATE_HARDWARE_VERTEXPROCESSING))
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_PUREDEVICE), TRUE);
	} else
	{
		EnableWindow(GetDlgItem(DlgWnd, IDC_PUREDEVICE), FALSE);
	}

    if (deviceInfo && (deviceInfo->DeviceCaps.Caps2 & D3DCAPS2_CANMANAGERESOURCE))
    {
        EnableWindow(GetDlgItem(DlgWnd, IDC_DISABLEDRIVERMGMT), TRUE);
    } else
    {
        EnableWindow(GetDlgItem(DlgWnd, IDC_DISABLEDRIVERMGMT), FALSE);
    }

    NotifyPureDeviceChange();
    NotifyDriverMgmtChange();
}

//-----------------------------------------------------------------------------
// NotifyPureDeviceChange
//-----------------------------------------------------------------------------
// Set the context pure device mode to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyPureDeviceChange()
{
    if (IsWindowEnabled(GetDlgItem(DlgWnd, IDC_PUREDEVICE)) && SendDlgItemMessage(DlgWnd, IDC_PUREDEVICE, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dcp.BehaviorFlags |= D3DCREATE_PUREDEVICE;
    } else
    {
        d3dcp.BehaviorFlags &= ~D3DCREATE_PUREDEVICE;
    }
}

//-----------------------------------------------------------------------------
// NotifyDriverMgmtChange
//-----------------------------------------------------------------------------
// Set the context driver mgmt mode to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyDriverMgmtChange()
{
    if (IsWindowEnabled(GetDlgItem(DlgWnd, IDC_DISABLEDRIVERMGMT)) && SendDlgItemMessage(DlgWnd, IDC_DISABLEDRIVERMGMT, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dcp.BehaviorFlags |= D3DCREATE_DISABLE_DRIVER_MANAGEMENT;
    } else
    {
        d3dcp.BehaviorFlags &= ~D3DCREATE_DISABLE_DRIVER_MANAGEMENT;
    }
}

//-----------------------------------------------------------------------------
// NotifyDisplayTypeChange
//-----------------------------------------------------------------------------
// Set the context display type to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyDisplayTypeChange()
{
    //////////////////////////////////
	// update the context to match the dialog settings
    //////////////////////////////////
	if (SendDlgItemMessage(DlgWnd, IDC_WINDOWED, BM_GETCHECK, 0, 0) == BST_CHECKED)
	{
		d3dpp.Windowed = TRUE;
		EnableWindow(GetDlgItem(DlgWnd, IDC_FULLSCREEN_FORMAT), FALSE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_FULLSCREEN_FORMAT_STR), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_SETCURSEL, -1, 0);
		EnableWindow(GetDlgItem(DlgWnd, IDC_DISPLAYMODE), FALSE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_DISPLAYMODE_STR), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, -1, 0);
		EnableWindow(GetDlgItem(DlgWnd, IDC_REFRESH), FALSE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_REFRESH_STR), FALSE);
		SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_SETCURSEL, -1, 0);

        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_WIDTH), TRUE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_WIDTH_STR), TRUE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_HEIGHT), TRUE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_HEIGHT_STR), TRUE);

        EnableWindow(GetDlgItem(DlgWnd, IDC_PRESENTATION_INTERVAL), FALSE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_PRESENTATION_INTERVAL_STR), FALSE);

        D3DFORMAT windowedFormat = D3DUtilContext::Instance()->GetWindowedFormat();

        //////////////////////////////////
	    // fill in the back buffer formats with compatible formats.
        //////////////////////////////////

        // Clear the back buffer formats list
	    SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_RESETCONTENT, 0, 0);
        
        const D3DUtilAdapterModeInfo* adapterModeInfo = GetAdapterModeInfo();
        if (adapterModeInfo)
        {
            // prefer back buffers that match (alpha buffers even better)
            UINT preferredNdx = 0;
            for (unsigned int i=0;i<adapterModeInfo->BackBuffers.size();++i)
            {
                D3DFORMAT format = adapterModeInfo->BackBuffers[i].Format;
                if ((format == windowedFormat) ||
                    (windowedFormat == D3DFMT_X8R8G8B8 && format == D3DFMT_A8R8G8B8) ||
                    (windowedFormat == D3DFMT_X1R5G5B5 && format == D3DFMT_A1R5G5B5))
                {
                    preferredNdx = i;
                }
                SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_ADDSTRING, 0, (LPARAM)EnumToString(format));
            }
            SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_SETCURSEL, preferredNdx, 0);

            NotifyBackBufferFormatChange();
        }
        // Reset the current back buffer size parameters also.
        NotifyBackBufferSizeChange();

        //////////////////////////////////
	    // fill in the presentation interval list.
        //////////////////////////////////

        // clear the old presentation interval list.
	    SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_RESETCONTENT, 0, 0);

        // The new list can only contain default
        SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_ADDSTRING, 0, (LPARAM)STR_IMMEDIATE);
        SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_SETCURSEL, 0, 0);

        NotifyPresentationIntervalChange();
	} else
	{
		d3dpp.Windowed = FALSE;
		EnableWindow(GetDlgItem(DlgWnd, IDC_FULLSCREEN_FORMAT), TRUE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_FULLSCREEN_FORMAT_STR), TRUE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_DISPLAYMODE), TRUE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_DISPLAYMODE_STR), TRUE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_REFRESH), TRUE);
		EnableWindow(GetDlgItem(DlgWnd, IDC_REFRESH_STR), TRUE);

        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_WIDTH), FALSE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_WIDTH_STR), FALSE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_HEIGHT), FALSE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_BACKBUFFER_HEIGHT_STR), FALSE);

        EnableWindow(GetDlgItem(DlgWnd, IDC_PRESENTATION_INTERVAL), TRUE);
        EnableWindow(GetDlgItem(DlgWnd, IDC_PRESENTATION_INTERVAL_STR), TRUE);

        //////////////////////////////////
	    // fill in the fullscreen modes combo box
        //////////////////////////////////

        // Clear the old fullscreen format list
	    SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_RESETCONTENT, 0, 0);
        
        const D3DUtilAdapterInfo* adapterInfo = GetAdapterInfo();
        if (adapterInfo)
        {
            for (unsigned int i=0;i<adapterInfo->DisplayModes.size();++i)
            {
                SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_ADDSTRING, 0, (LPARAM)EnumToString(adapterInfo->DisplayModes[i].DisplayModeFormat));
            }
            SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_SETCURSEL, 0, 0);
        
            // Notify the dialog of the new options
        	NotifyDisplayFormatChange();
        }

        //////////////////////////////////
	    // fill in the presentation interval list.
        //////////////////////////////////

        // clear the old presentation interval list.
	    SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_RESETCONTENT, 0, 0);

        // The new list can only contain default
        const D3DUtilDeviceInfo* deviceInfo = GetDeviceInfo();

        SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_ADDSTRING, 0, (LPARAM)STR_DEFAULT);
        SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_SETCURSEL, 0, 0);
        if (deviceInfo->DeviceCaps.PresentationIntervals & D3DPRESENT_INTERVAL_IMMEDIATE)
        {
            SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_ADDSTRING, 0, (LPARAM)STR_IMMEDIATE);
            // prefer immediate presentation.
            SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_SETCURSEL, 1, 0);
        }
        if (deviceInfo->DeviceCaps.PresentationIntervals & D3DPRESENT_INTERVAL_ONE)
        {
            SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_ADDSTRING, 0, (LPARAM)STR_ONE);
        }
        if (deviceInfo->DeviceCaps.PresentationIntervals & D3DPRESENT_INTERVAL_TWO)
        {
            SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_ADDSTRING, 0, (LPARAM)STR_TWO);
        }
        if (deviceInfo->DeviceCaps.PresentationIntervals & D3DPRESENT_INTERVAL_THREE)
        {
            SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_ADDSTRING, 0, (LPARAM)STR_THREE);
        }
        if (deviceInfo->DeviceCaps.PresentationIntervals & D3DPRESENT_INTERVAL_FOUR)
        {
            SendDlgItemMessage(DlgWnd, IDC_PRESENTATION_INTERVAL, CB_ADDSTRING, 0, (LPARAM)STR_FOUR);
        }

        NotifyPresentationIntervalChange();
    }
}

//-----------------------------------------------------------------------------
// NotifyPresentationIntervalChange
//-----------------------------------------------------------------------------
// Set the context presentation interval to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyPresentationIntervalChange()
{
    char buffer[32];
    GetDlgItemText(DlgWnd, IDC_PRESENTATION_INTERVAL, buffer, 31);
    if (strcmp(buffer, STR_DEFAULT) == 0)
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
    } else
    if (strcmp(buffer, STR_IMMEDIATE) == 0)
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
    } else
    if (strcmp(buffer, STR_ONE) == 0)
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
    } else
    if (strcmp(buffer, STR_TWO) == 0)
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_TWO;
    } else
    if (strcmp(buffer, STR_THREE) == 0)
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_THREE;
    } else
    if (strcmp(buffer, STR_FOUR) == 0)
    {
        d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_FOUR;
    }
}

//-----------------------------------------------------------------------------
// NotifyBackBufferSizeChange
//-----------------------------------------------------------------------------
// Set the context back buffer sizes to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyBackBufferSizeChange()
{
    BOOL success;
    UINT width = GetDlgItemInt(DlgWnd, IDC_BACKBUFFER_WIDTH, &success, FALSE);
    if (!success) width = 0;
    UINT height = GetDlgItemInt(DlgWnd, IDC_BACKBUFFER_HEIGHT, &success, FALSE);
    if (!success) height = 0;

    d3dpp.BackBufferWidth = width;
    d3dpp.BackBufferHeight = height;
}

//-----------------------------------------------------------------------------
// NotifyDisplayFormatChange
//-----------------------------------------------------------------------------
// Set the context display format to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyDisplayFormatChange()
{
    int modeNdx = SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_GETCURSEL, 0, 0);
    
    const D3DUtilAdapterInfo* adapterInfo = GetAdapterInfo();
    if (adapterInfo)
    {
        // Clear the old display mode list
	    SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_RESETCONTENT, 0, 0);
    
        const D3DUtilDisplayModeInfo* displayModeInfo = &adapterInfo->DisplayModes[modeNdx];
        if (displayModeInfo)
        {
            //////////////////////////////////
            // Set up the Display Mode list
            //////////////////////////////////
            D3DFORMAT fullscreenMode = displayModeInfo->DisplayModeFormat;

	        // Fill in the combo box
            UINT width=0, height=0, preferredNdx=0;
            bool matchedDesktopRes = false;
            D3DDISPLAYMODE dm;
            D3DUtilContext::Instance()->GetD3D()->GetAdapterDisplayMode(d3dcp.AdapterOrdinal, &dm);
            char buffer[32];
	        for (unsigned int i=0; i<displayModeInfo->Modes.size(); i++)
	        {
                // We know the list is sorted by display size, so just find changes in display size and add those.
                const D3DDISPLAYMODE& mode = displayModeInfo->Modes[i];
                if (mode.Width != width || mode.Height != height)
                {
                    width = mode.Width;
                    height = mode.Height;
                    if (width == dm.Width && height == dm.Height)
                    {
                        preferredNdx = SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETCOUNT, 0, 0);
                        matchedDesktopRes = true;
                    }
                    if (!matchedDesktopRes && (
                        (width == 640 && height == 480) ||
                        (width == 800 && height == 600) ||
                        (width == 1024&& height == 768)))
                    {
                        preferredNdx = SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_GETCOUNT, 0, 0);
                    }
    		        sprintf(buffer, "%d x %d", width, height);
	    	        SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_ADDSTRING, 0, (LPARAM)buffer);
                }
	        }
            SendDlgItemMessage(DlgWnd, IDC_DISPLAYMODE, CB_SETCURSEL, preferredNdx, 0);

            NotifyDisplayResolutionChange();

            //////////////////////////////////
            // Set up the back buffer mode list (must match fullscreen display format)
            //////////////////////////////////

            // Clear the old back buffer mode list
            SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_RESETCONTENT, 0, 0);

            // Fill in the combo box.
            const D3DUtilAdapterModeInfo* adapterModeInfo = GetAdapterModeInfo();
            if (adapterModeInfo)
            {
                // find back buffers that match (alpha buffers even better)
                UINT preferredNdx = 0;
                for (unsigned int i=0;i<adapterModeInfo->BackBuffers.size();++i)
                {
                    D3DFORMAT format = adapterModeInfo->BackBuffers[i].Format;
                    if ((format == fullscreenMode) ||
                        (fullscreenMode == D3DFMT_X8R8G8B8 && format == D3DFMT_A8R8G8B8) ||
                        (fullscreenMode == D3DFMT_X1R5G5B5 && format == D3DFMT_A1R5G5B5))
                    {
                        preferredNdx = SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_GETCOUNT, 0, 0);
                        SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_ADDSTRING, 0, (LPARAM)EnumToString(format));
                    }
                }
                SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_FORMAT, CB_SETCURSEL, preferredNdx, 0);

                NotifyBackBufferFormatChange();
            }
        }
    }
}

//-----------------------------------------------------------------------------
// NotifyDisplayResolutionChange
//-----------------------------------------------------------------------------
// Set the context display mode to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyDisplayResolutionChange()
{
    int modeNdx = SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_GETCURSEL, 0, 0);
    const D3DUtilAdapterInfo* adapterInfo = GetAdapterInfo();
    if (adapterInfo)
    {
        //////////////////////////////////
        // Set up the Refresh list
        //////////////////////////////////

        // Clear the old refresh list
	    SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_RESETCONTENT, 0, 0);
    
        const D3DUtilDisplayModeInfo* displayModeInfo = &adapterInfo->DisplayModes[modeNdx];
	    // Find the current width and height in the dialog
		char buffer[32];
        std::string tempBuffer;
        UINT width, height;
		GetDlgItemText(DlgWnd, IDC_DISPLAYMODE, buffer, 31);
		std::istringstream istream(buffer);
		istream >> width >> tempBuffer >> height;

        // find the start of the display modes that match the current width and height
        unsigned int modeNdx = 0;
        for (modeNdx=0;modeNdx<displayModeInfo->Modes.size();++modeNdx)
        {
            if (displayModeInfo->Modes[modeNdx].Width == width && displayModeInfo->Modes[modeNdx].Height == height)
            {
                break;
            }
        }
        // using that starting point, fill in all the refresh rates supported.
        if (modeNdx<displayModeInfo->Modes.size())
        {
            sprintf(buffer, STR_DEFAULT);
       	    SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_ADDSTRING, 0, (LPARAM)buffer);
            for (; modeNdx<displayModeInfo->Modes.size(); ++modeNdx)
            {
                if (displayModeInfo->Modes[modeNdx].Width == width && displayModeInfo->Modes[modeNdx].Height == height)
                {
                    UINT refresh = displayModeInfo->Modes[modeNdx].RefreshRate;
        		    sprintf(buffer, "%d Hz", refresh);
	        	    SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_ADDSTRING, 0, (LPARAM)buffer);
                } else
                {
                    // since we are sorted on display mode, stop when we change to a different one.
                    break;
                }

            }
            SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_SETCURSEL, 0, 0);
        }
    }
  
    NotifyDisplayRefreshChange();
}

//-----------------------------------------------------------------------------
// NotifyDisplayRefreshChange
//-----------------------------------------------------------------------------
// Set the context refresh to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyDisplayRefreshChange()
{
    unsigned int modeNdx = SendDlgItemMessage(DlgWnd, IDC_FULLSCREEN_FORMAT, CB_GETCURSEL, 0, 0);
    const D3DUtilAdapterInfo* adapterInfo = GetAdapterInfo();
    if (adapterInfo)
    {
	    // Find the current width and height and refresh in the dialog
        if (modeNdx < adapterInfo->DisplayModes.size())
        {
            const D3DUtilDisplayModeInfo* displayModeInfo = &adapterInfo->DisplayModes[modeNdx];
		    char buffer[32];
            std::string tempBuffer;
            UINT width, height, refresh;
		    GetDlgItemText(DlgWnd, IDC_DISPLAYMODE, buffer, 31);
		    std::istringstream istream(buffer);
		    istream >> width >> tempBuffer >> height;

            if (SendDlgItemMessage(DlgWnd, IDC_REFRESH, CB_GETCURSEL, 0, 0) == 0)
            {
                refresh = 0;
            } else
            {
                GetDlgItemText(DlgWnd, IDC_REFRESH, buffer, 31);
                std::istringstream istream(buffer);
                istream >> refresh;
            }
            d3dpp.BackBufferWidth = width;
            d3dpp.BackBufferHeight = height;
            d3dpp.FullScreen_RefreshRateInHz = refresh;
        }
    }
}

//-----------------------------------------------------------------------------
// NotifyBackBufferLockableChange
//-----------------------------------------------------------------------------
// Set the context lockable flag to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyBackBufferLockableChange()
{
    if (SendDlgItemMessage(DlgWnd, IDC_BACKBUFFER_LOCKABLE, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dpp.Flags |= D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
    } else
    {
        d3dpp.Flags &= ~D3DPRESENTFLAG_LOCKABLE_BACKBUFFER;
    }
}

//-----------------------------------------------------------------------------
// NotifyBackBufferCountChange
//-----------------------------------------------------------------------------
// Set the context back buffer count to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyBackBufferCountChange()
{
    BOOL success;
    UINT count = GetDlgItemInt(DlgWnd, IDC_BACKBUFFER_COUNT, &success, FALSE);
    if (!success) count = 0;

    d3dpp.BackBufferCount = count;
}

//-----------------------------------------------------------------------------
// NotifyBackBufferFormatChange
//-----------------------------------------------------------------------------
// Set the context back buffer format to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyBackBufferFormatChange()
{
    const D3DUtilAdapterModeInfo* adapterModeInfo = GetAdapterModeInfo();
    if (adapterModeInfo)
    {
        //////////////////////////////////
	    // update the context to match the dialog settings
        //////////////////////////////////
		char buffer[32];
        std::string tempBuffer;
		GetDlgItemText(DlgWnd, IDC_BACKBUFFER_FORMAT, buffer, 31);
        d3dpp.BackBufferFormat = StringToD3DFORMAT(buffer);
        if (d3dpp.BackBufferFormat != D3DFMT_UNKNOWN)
        {
            //////////////////////////////////
	        // update the back buffer conversion type
            //////////////////////////////////
            const D3DUtilBackBufferInfo* backBufferInfo = GetBackBufferInfo();
            BOOL hwConv = backBufferInfo->HardwareConversion;
            const char* str = "Hardware";
            if (!hwConv)
            {
                if (!d3dpp.Windowed)
                {
                    // hardware can't do conversions. This should not have passed the caps.
                    str = "Error! Software";
                } else
                {
                    str = "Software";
                }
            }
            SetDlgItemText(DlgWnd, IDC_CONVERSION_TYPE, str);

            //////////////////////////////////
	        // Set up the depth format list.
            //////////////////////////////////

            // clear the old depth format list.
            SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_FORMAT, CB_RESETCONTENT, 0, 0);

            // build the new depth format list.
            INT preferredNdx = -1;
            for (unsigned int i=0;i<backBufferInfo->DepthBuffers.size(); ++i)
            {
                D3DFORMAT depthFormat = backBufferInfo->DepthBuffers[i].Format;
                // prefer depth buffers of the same bit depth as the back buffer.
                if ((backBufferInfo->Format == D3DFMT_A8R8G8B8 ||
                    backBufferInfo->Format == D3DFMT_X8R8G8B8) &&
                    (depthFormat == D3DFMT_D24S8 ||
                    depthFormat == D3DFMT_D24X4S4 ||
                    depthFormat == D3DFMT_D24X8 ||
                    depthFormat == D3DFMT_D32) &&
                    preferredNdx == -1)
                {
                    preferredNdx = i;
                } else
                if ((backBufferInfo->Format == D3DFMT_A1R5G5B5 ||
                    backBufferInfo->Format == D3DFMT_X1R5G5B5 ||
                    backBufferInfo->Format == D3DFMT_R5G6B5) &&
                    (depthFormat == D3DFMT_D16 ||
                    depthFormat == D3DFMT_D15S1 ||
                    depthFormat == D3DFMT_D16_LOCKABLE) &&
                    preferredNdx == -1)
                {
                    preferredNdx = i;
                }

                if (depthFormat == D3DFMT_UNKNOWN)
                {
                    SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_FORMAT, CB_ADDSTRING, 0, (LPARAM)STR_NONE);
                } else
                {
                    SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_FORMAT, CB_ADDSTRING, 0, (LPARAM)EnumToString(backBufferInfo->DepthBuffers[i].Format));
                }
            }
            if (preferredNdx == -1)
            {
                preferredNdx = 0;
            }
            SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_FORMAT, CB_SETCURSEL, preferredNdx, 0);

            NotifyDepthFormatChange();
        }
    }
}

//-----------------------------------------------------------------------------
// NotifyDepthFormatChange
//-----------------------------------------------------------------------------
// Set the context depth format to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyDepthFormatChange()
{
    //////////////////////////////////
	// update the context to match the dialog settings
    //////////////////////////////////

    const D3DUtilBackBufferInfo* backBufferInfo = GetBackBufferInfo();
    if (backBufferInfo)
    {
        UINT depthNdx = SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_FORMAT, CB_GETCURSEL, 0, 0);
        if (depthNdx < backBufferInfo->DepthBuffers.size())
        {
            const D3DUtilDepthInfo* depthInfo = &backBufferInfo->DepthBuffers[depthNdx];

            //////////////////////////////////
	        // update the context to match the depth format
            //////////////////////////////////
            if (depthInfo->Format == D3DFMT_UNKNOWN)
            {
                d3dpp.EnableAutoDepthStencil = FALSE;
            } else
            {
                d3dpp.EnableAutoDepthStencil = TRUE;
                d3dpp.AutoDepthStencilFormat = depthInfo->Format;
            }

            //////////////////////////////////
	        // update the discard depth buffer flag
            //////////////////////////////////

            if (depthInfo->Format == D3DFMT_UNKNOWN)
            {
                EnableWindow(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER_DISCARD), FALSE);
            } else
            {
                EnableWindow(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER_DISCARD), TRUE);
            }

            NotifyDepthDiscardChange();

            //////////////////////////////////
	        // create the multisample type list
            //////////////////////////////////

            // clear the old multisample list
            SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_TYPE, CB_RESETCONTENT, 0, 0);
            
            // build the new multisample list.
            for (unsigned int i=0;i<depthInfo->MultisampleTypes.size();++i)
            {
                const D3DUtilMultisampleInfo* multisampleInfo = &depthInfo->MultisampleTypes[i];
                SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_TYPE, CB_ADDSTRING, 0, (LPARAM)EnumToString(multisampleInfo->Type));
            }
            SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_TYPE, CB_SETCURSEL, 0, 0);

            NotifyMultisampleTypeChange();
        }
    }
}

//-----------------------------------------------------------------------------
// NotifyDepthDiscardChange
//-----------------------------------------------------------------------------
// Set the context discard depth flag to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyDepthDiscardChange()
{
    if (IsWindowEnabled(GetDlgItem(DlgWnd, IDC_DEPTHBUFFER_DISCARD)) && SendDlgItemMessage(DlgWnd, IDC_DEPTHBUFFER_DISCARD, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dpp.Flags |= D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
    } else
    {
        d3dpp.Flags &= ~D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
    }
}

//-----------------------------------------------------------------------------
// NotifyMultisampleTypeChange
//-----------------------------------------------------------------------------
// Set the context multisample type to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyMultisampleTypeChange()
{
    //////////////////////////////////
	// update the context to match the dialog settings
    //////////////////////////////////

    const D3DUtilDepthInfo* depthInfo = GetDepthInfo();
    if (depthInfo)
    {
        UINT multisampleNdx = SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_TYPE, CB_GETCURSEL, 0, 0);
        if (multisampleNdx < depthInfo->MultisampleTypes.size())
        {
            d3dpp.MultiSampleType = depthInfo->MultisampleTypes[multisampleNdx].Type;

            if (d3dpp.MultiSampleType != D3DMULTISAMPLE_NONE)
            {
                EnableWindow(GetDlgItem(DlgWnd, IDC_MULTISAMPLE_QUALITY), TRUE);
                EnableWindow(GetDlgItem(DlgWnd, IDC_MULTISAMPLE_QUALITY_STR), TRUE);

                //////////////////////////////////
	            // Create the quality levels list
                //////////////////////////////////

                // clear the old quality levels list
                SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_QUALITY, CB_RESETCONTENT, 0, 0);

                // build the new multisample quality list.
                const D3DUtilMultisampleInfo* multisampleInfo = depthInfo->FindMultisample(d3dpp.MultiSampleType);
                if (multisampleInfo)
                {
                    for (unsigned int i=0;i<multisampleInfo->QualityLevels;++i)
                    {
                        char buffer[32];
                        sprintf(buffer, "%d", i);
                        SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_QUALITY, CB_ADDSTRING, 0, (LPARAM)buffer);
                    }
                    SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_QUALITY, CB_SETCURSEL, 0, 0);
                }
            } else
            {
                EnableWindow(GetDlgItem(DlgWnd, IDC_MULTISAMPLE_QUALITY), FALSE);
                EnableWindow(GetDlgItem(DlgWnd, IDC_MULTISAMPLE_QUALITY_STR), FALSE);
        	    SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_QUALITY, CB_RESETCONTENT, 0, 0);
            }
            NotifyMultisampleQualityChange();
        }
    }
}

//-----------------------------------------------------------------------------
// NotifyMultisampleQualityChange
//-----------------------------------------------------------------------------
// Set the context multisample quality  to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyMultisampleQualityChange()
{
    //////////////////////////////////
	// update the context to match the dialog settings
    //////////////////////////////////

    const D3DUtilMultisampleInfo* multisampleInfo = GetMultisampleInfo();
    if (multisampleInfo)
    {
        int quality = SendDlgItemMessage(DlgWnd, IDC_MULTISAMPLE_QUALITY, CB_GETCURSEL, 0, 0);
        d3dpp.MultiSampleQuality = quality < 0 ? 0 : quality;
    }
}

//-----------------------------------------------------------------------------
// NotifyMiscFlagsChange
//-----------------------------------------------------------------------------
// Set the context misc flags to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifyMiscFlagsChange()
{
    if (SendDlgItemMessage(DlgWnd, IDC_DEVICECLIP, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dpp.Flags |= D3DPRESENTFLAG_DEVICECLIP;
    } else
    {
        d3dpp.Flags &= ~D3DPRESENTFLAG_DEVICECLIP;
    }

    if (SendDlgItemMessage(DlgWnd, IDC_VIDEO_HINT, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dpp.Flags |= D3DPRESENTFLAG_VIDEO;
    } else
    {
        d3dpp.Flags &= ~D3DPRESENTFLAG_VIDEO;
    }

    if (SendDlgItemMessage(DlgWnd, IDC_FPUPRESERVE, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dcp.BehaviorFlags |= D3DCREATE_FPU_PRESERVE;
    } else
    {
        d3dcp.BehaviorFlags &= ~D3DCREATE_FPU_PRESERVE;
    }

    if (SendDlgItemMessage(DlgWnd, IDC_MULTITHREADED, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dcp.BehaviorFlags |= D3DCREATE_MULTITHREADED;
    } else
    {
        d3dcp.BehaviorFlags &= ~D3DCREATE_MULTITHREADED;
    }
}

//-----------------------------------------------------------------------------
// NotifySwapEffectChange
//-----------------------------------------------------------------------------
// Set the context swap effect to reflect the current dialog.
//-----------------------------------------------------------------------------
void NotifySwapEffectChange()
{
    if (SendDlgItemMessage(DlgWnd, IDC_DISCARD, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    } else
    if (SendDlgItemMessage(DlgWnd, IDC_FLIP, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dpp.SwapEffect = D3DSWAPEFFECT_FLIP;
    } else
    if (SendDlgItemMessage(DlgWnd, IDC_COPY, BM_GETCHECK, 0, 0) == BST_CHECKED)
    {
        d3dpp.SwapEffect = D3DSWAPEFFECT_COPY;
    }
}
















//-----------------------------------------------------------------------------
// GetAdapterInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilAdapterInfo *GetAdapterInfo()
{
	if (D3DUtilEnumeration::GetAdapterList().size() == 0) return NULL;
	return &D3DUtilEnumeration::GetAdapterList()[d3dcp.AdapterOrdinal];
}

//-----------------------------------------------------------------------------
// GetDeviceInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilDeviceInfo *GetDeviceInfo()
{
	const D3DUtilAdapterInfo* adapterInfo = GetAdapterInfo();
	if (adapterInfo == NULL) return NULL;
	return adapterInfo->FindDevice(d3dcp.DeviceType);
}

//-----------------------------------------------------------------------------
// GetAdapterModeInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilAdapterModeInfo *GetAdapterModeInfo()
{
	const D3DUtilDeviceInfo* deviceInfo = GetDeviceInfo();
	if (deviceInfo == NULL) return NULL;
    D3DFORMAT displayFormat;
    if (d3dpp.Windowed)
    {
        displayFormat = D3DUtilContext::Instance()->GetWindowedFormat();
    } else
    {
        char buffer[32];
        GetDlgItemText(DlgWnd, IDC_FULLSCREEN_FORMAT, buffer, 31);
        displayFormat = StringToD3DFORMAT(buffer);
    }
    return deviceInfo->FindAdapterMode(displayFormat, d3dpp.Windowed);
}

//-----------------------------------------------------------------------------
// GetBackBufferInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilBackBufferInfo *GetBackBufferInfo()
{
	const D3DUtilAdapterModeInfo* adapterModeInfo = GetAdapterModeInfo();
	if (adapterModeInfo == NULL) return NULL;
    return adapterModeInfo->FindBackBuffer(d3dpp.BackBufferFormat);
}

//-----------------------------------------------------------------------------
// GetDepthInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilDepthInfo *GetDepthInfo()
{
	const D3DUtilBackBufferInfo* backBufferInfo = GetBackBufferInfo();
	if (backBufferInfo == NULL) return NULL;
    return backBufferInfo->FindDepth(d3dpp.EnableAutoDepthStencil ? d3dpp.AutoDepthStencilFormat : D3DFMT_UNKNOWN);
}

//-----------------------------------------------------------------------------
// GetMultisampleInfo
//-----------------------------------------------------------------------------
// Deduce Enumeration info from the dialog settings
//-----------------------------------------------------------------------------
const D3DUtilMultisampleInfo *GetMultisampleInfo()
{
	const D3DUtilDepthInfo* depthInfo = GetDepthInfo();
	if (depthInfo == NULL) return NULL;
    return depthInfo->FindMultisample(d3dpp.MultiSampleType);
}
