//-----------------------------------------------------------------------------
// D3DUtilEnumeration.hpp
//-----------------------------------------------------------------------------
// Enumerates and creates a list of all the devices and modes on the system.
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// #includes
//-----------------------------------------------------------------------------
#include "stdafx.h"
#include "D3DUtilEnumeration.hpp"
#include <algorithm>

//-----------------------------------------------------------------------------
// statics
//-----------------------------------------------------------------------------
D3DUtilAdapterInfoList D3DUtilEnumeration::AdapterList;

namespace
{
    bool EmptyMultisamples(const D3DUtilDepthInfo& depthInfo)
    {
        return depthInfo.MultisampleTypes.empty();
    }
    bool EmptyDepthFormats(const D3DUtilBackBufferInfo& backBufferInfo)
    {
        return backBufferInfo.DepthBuffers.empty();
    }
    bool EmptyBackBufferFormats(const D3DUtilAdapterModeInfo& adapterModeInfo)
    {
        return adapterModeInfo.BackBuffers.empty();
    }
    bool EmptyAdapterModes(const D3DUtilDeviceInfo& deviceInfo)
    {
        return deviceInfo.AdapterModes.empty();
    }
    bool EmptyDevice(const D3DUtilAdapterInfo& adapterInfo)
    {
        return adapterInfo.Devices.empty();
    }
};

//-----------------------------------------------------------------------------
// GetAdapterList
//-----------------------------------------------------------------------------
// Returns the list of adapters and display modes
//-----------------------------------------------------------------------------
const D3DUtilAdapterInfoList &D3DUtilEnumeration::GetAdapterList()
{
	if (AdapterList.size() > 0) return AdapterList;

	IDirect3D9 *pD3D = NULL;
	pD3D = Direct3DCreate9(D3D_SDK_VERSION);
    ASSERT_MSG(pD3D != NULL, "D3DUtilEnumeration: Direct3DCreate9 Failed");

	unsigned int adapter;
	BuildAdapterList(pD3D);
	for (adapter=0; adapter<AdapterList.size(); adapter++)
	{
		D3DUtilAdapterInfo &adapterInfo = AdapterList[adapter];
		BuildDisplayModeList(adapter, pD3D);
		BuildDeviceList(adapter, pD3D);
		for (unsigned int deviceIndex=0; deviceIndex<adapterInfo.Devices.size(); deviceIndex++)
		{
			D3DUtilDeviceInfo &deviceInfo = adapterInfo.Devices[deviceIndex];
            BuildAdapterModeList(adapter, deviceInfo, pD3D);
            for (unsigned int adapterModeIndex=0; adapterModeIndex<deviceInfo.AdapterModes.size(); ++adapterModeIndex)
            {
                D3DUtilAdapterModeInfo &adapterModeInfo = deviceInfo.AdapterModes[adapterModeIndex];
			    BuildBackBufferList(adapter, deviceInfo.DeviceType, adapterModeInfo, pD3D);
			    for (unsigned int formatIndex=0; formatIndex<adapterModeInfo.BackBuffers.size(); formatIndex++)
			    {
				    D3DUtilBackBufferInfo &backBufferInfo = adapterModeInfo.BackBuffers[formatIndex];
				    BuildDepthBufferList(adapter, deviceInfo.DeviceType, adapterModeInfo.Mode, adapterModeInfo.Windowed, backBufferInfo, pD3D);
				    for (unsigned int depthIndex=0; depthIndex<backBufferInfo.DepthBuffers.size(); depthIndex++)
				    {
					    D3DUtilDepthInfo &depthInfo = backBufferInfo.DepthBuffers[depthIndex];
					    BuildMultisampleList(adapter, deviceInfo.DeviceType, adapterModeInfo.Mode, adapterModeInfo.Windowed, backBufferInfo.Format, depthInfo, pD3D);
                    }
                    backBufferInfo.DepthBuffers.erase(std::remove_if(backBufferInfo.DepthBuffers.begin(), backBufferInfo.DepthBuffers.end(), &EmptyMultisamples), backBufferInfo.DepthBuffers.end());
				}
                adapterModeInfo.BackBuffers.erase(std::remove_if(adapterModeInfo.BackBuffers.begin(), adapterModeInfo.BackBuffers.end(), &EmptyDepthFormats), adapterModeInfo.BackBuffers.end());
			}
            deviceInfo.AdapterModes.erase(std::remove_if(deviceInfo.AdapterModes.begin(), deviceInfo.AdapterModes.end(), &EmptyBackBufferFormats), deviceInfo.AdapterModes.end());
		}
        adapterInfo.Devices.erase(std::remove_if(adapterInfo.Devices.begin(), adapterInfo.Devices.end(), &EmptyAdapterModes), adapterInfo.Devices.end());
	}
    AdapterList.erase(std::remove_if(AdapterList.begin(), AdapterList.end(), &EmptyDevice), AdapterList.end());

	SAFE_RELEASE(pD3D);
	return AdapterList;
}

//-----------------------------------------------------------------------------
// BuildAdapterList
//-----------------------------------------------------------------------------
// Fills the list with the available adapters
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildAdapterList(IDirect3D9 *pD3D)
{
	unsigned int adapter;
	for (adapter=0; adapter<pD3D->GetAdapterCount(); adapter++)
	{
		AdapterList.push_back(D3DUtilAdapterInfo());
		D3DUtilAdapterInfo &adapterInfo = AdapterList.back();
		D3DADAPTER_IDENTIFIER9 adapterIdentifier;
		pD3D->GetAdapterIdentifier(adapter, 0, &adapterIdentifier);
		sprintf(adapterInfo.Driver, "%s", adapterIdentifier.Driver);
		sprintf(adapterInfo.Description, "%s", adapterIdentifier.Description);
		sprintf(adapterInfo.DeviceName, "%s", adapterIdentifier.DeviceName);
	}
}

//-----------------------------------------------------------------------------
// BuildDisplayModeList
//-----------------------------------------------------------------------------
// Fills the list with the available adapters.
// Side effect: fill the given vector with the list of possible formats
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildDisplayModeList(int adapter, IDirect3D9 *pD3D)
{
	D3DUtilAdapterInfo &adapterInfo = AdapterList[adapter];
    for (int bbMode=0; bbMode<BackBufferFormatCount; ++bbMode)
    {
        // skip alpha display modes since they are equivalent to non-alpha ones.
        if (BackBufferFormats[bbMode] != D3DFMT_A8R8G8B8 && BackBufferFormats[bbMode] != D3DFMT_A1R5G5B5)
        {
            D3DUtilDisplayModeInfo modeInfo;
            modeInfo.DisplayModeFormat = BackBufferFormats[bbMode];
            if (pD3D->GetAdapterModeCount(adapter, modeInfo.DisplayModeFormat) > 0)
            {
	            for (unsigned int modeIndex=0; modeIndex<pD3D->GetAdapterModeCount(adapter, modeInfo.DisplayModeFormat); modeIndex++)
	            {
                    D3DDISPLAYMODE mode;
		            if (SUCCEEDED(pD3D->EnumAdapterModes(adapter, modeInfo.DisplayModeFormat, modeIndex, &mode)))
                    {
	                    modeInfo.Modes.push_back(mode);
                    }
                    std::sort(modeInfo.Modes.begin(), modeInfo.Modes.end(), &LessThan);
                }
                adapterInfo.DisplayModes.push_back(modeInfo);
            }
        }
	}
}

//-----------------------------------------------------------------------------
// BuildDeviceList
//-----------------------------------------------------------------------------
// Fills the list with the available devices.
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildDeviceList(int adapter, IDirect3D9 *pD3D)
{
	D3DUtilAdapterInfo &adapterInfo = AdapterList[adapter];
	int deviceIndex;
	for (deviceIndex=0; deviceIndex<DeviceTypeCount; deviceIndex++)
	{
		D3DUtilDeviceInfo deviceInfo;
		deviceInfo.DeviceType = DeviceTypes[deviceIndex];
		if (SUCCEEDED(pD3D->GetDeviceCaps(adapter, deviceInfo.DeviceType, &deviceInfo.DeviceCaps)))
		{
			adapterInfo.Devices.push_back(deviceInfo);
		}
	}
}

//-----------------------------------------------------------------------------
// BuildBackBufferList
//-----------------------------------------------------------------------------
// Fills the list with the available adapter modes.
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildAdapterModeList(int adapter, D3DUtilDeviceInfo &deviceInfo, IDirect3D9 *pD3D)
{
    for (unsigned int formatIndex=0; formatIndex<AdapterList[adapter].DisplayModes.size(); ++formatIndex)
    {
        D3DFORMAT mode = AdapterList[adapter].DisplayModes[formatIndex].DisplayModeFormat;
        for (BOOL windowed=FALSE; windowed<=TRUE; ++windowed)
        {
            D3DUtilAdapterModeInfo adapterModeInfo;
            adapterModeInfo.Mode = mode;
            adapterModeInfo.Windowed = windowed;
            deviceInfo.AdapterModes.push_back(adapterModeInfo);
        }
    }
}

//-----------------------------------------------------------------------------
// BuildBackBufferList
//-----------------------------------------------------------------------------
// Fills the list with the available back buffer formats.
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildBackBufferList(int adapter, D3DDEVTYPE deviceType, D3DUtilAdapterModeInfo &adapterModeInfo, IDirect3D9 *pD3D)
{
	for (int formatIndex=0; formatIndex<BackBufferFormatCount; ++formatIndex)
	{
		D3DUtilBackBufferInfo backBufferInfo;
		backBufferInfo.Format = BackBufferFormats[formatIndex];
        backBufferInfo.HardwareConversion = FALSE;
		if (SUCCEEDED(pD3D->CheckDeviceFormat(adapter, deviceType, adapterModeInfo.Mode, D3DUSAGE_RENDERTARGET, D3DRTYPE_SURFACE, backBufferInfo.Format)))
		{
            // in windowed modes, format conversion could occur. Check for hardware support.
            if (adapterModeInfo.Mode == backBufferInfo.Format || !adapterModeInfo.Windowed ||
                SUCCEEDED(pD3D->CheckDeviceFormatConversion(adapter, deviceType, backBufferInfo.Format, adapterModeInfo.Mode)))
            {
                backBufferInfo.HardwareConversion = TRUE;
            }
			adapterModeInfo.BackBuffers.push_back(backBufferInfo);
		}
	}
}

//-----------------------------------------------------------------------------
// BuildDepthBufferList
//-----------------------------------------------------------------------------
// Fills the list with the available depth buffer formats.
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildDepthBufferList(int adapter, D3DDEVTYPE deviceType, D3DFORMAT adapterMode, BOOL windowed, D3DUtilBackBufferInfo &backBufferInfo, IDirect3D9 *pD3D)
{
	for (int depthIndex=0; depthIndex<DepthFormatCount; depthIndex++)
	{
		D3DUtilDepthInfo depthInfo;
		depthInfo.Format = DepthFormats[depthIndex];
        // D3DFMT_UNKNOWN means that we are not using a depth buffer
		if (depthInfo.Format == D3DFMT_UNKNOWN || (
			SUCCEEDED(pD3D->CheckDeviceFormat(adapter, deviceType, adapterMode, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, depthInfo.Format)) &&
			SUCCEEDED(pD3D->CheckDepthStencilMatch(adapter, deviceType, adapterMode, backBufferInfo.Format, depthInfo.Format))))
		{
			backBufferInfo.DepthBuffers.push_back(depthInfo);
		}
	}
}

//-----------------------------------------------------------------------------
// BuildMultisampleList
//-----------------------------------------------------------------------------
// Fills the list with the available multisample types (windowed and non-windowed).
//-----------------------------------------------------------------------------
void D3DUtilEnumeration::BuildMultisampleList(int adapter, D3DDEVTYPE deviceType, D3DFORMAT adapterMode, BOOL windowed, D3DFORMAT backBufferFormat, D3DUtilDepthInfo &depthInfo, IDirect3D9 *pD3D)
{
	for (int multisampleIndex=0; multisampleIndex<MultisampleCount; multisampleIndex++)
	{
		D3DUtilMultisampleInfo multisampleInfo;
		multisampleInfo.Type = MultisampleTypes[multisampleIndex];
		if (SUCCEEDED(pD3D->CheckDeviceMultiSampleType(adapter, deviceType, backBufferFormat, windowed, multisampleInfo.Type, &multisampleInfo.QualityLevels)))
        {
            if (depthInfo.Format == D3DFMT_UNKNOWN || 
			SUCCEEDED(pD3D->CheckDeviceMultiSampleType(adapter, deviceType, depthInfo.Format, windowed, multisampleInfo.Type, &multisampleInfo.QualityLevels)))
            {
                if (multisampleInfo.QualityLevels > 0)
                {
                    depthInfo.MultisampleTypes.push_back(multisampleInfo);
                }
            }
		}
	}
}
