//-----------------------------------------------------------------------------
// D3DUtilFont.hpp
//-----------------------------------------------------------------------------
// Font drawing routines.
//
//-----------------------------------------------------------------------------
#ifndef D3DUtilFont_hpp_
#define D3DUtilFont_hpp_
#pragma once

#include "Util.hpp"
#include "D3DUtilContext.hpp"
#include <utility>
#include <atlbase.h>
#include <string>

// Font creation flags
#define D3DFONT_BOLD        0x0001
#define D3DFONT_ITALIC      0x0002

// Font rendering flags
#define D3DFONT_CENTERED    0x0001
#define D3DFONT_TWOSIDED    0x0002
#define D3DFONT_NOFILTER    0x0004

//-----------------------------------------------------------------------------
// D3DUtilFont
//-----------------------------------------------------------------------------
class D3DUtilFont
{
public:
	//-----------------------------------------------------------------------------
    // Constructor / destructor
	//-----------------------------------------------------------------------------
    D3DUtilFont(const char* strFontName, unsigned dwHeight, unsigned dwFlags = 0);
    ~D3DUtilFont();

	//-----------------------------------------------------------------------------
    // 2D and 3D text drawing functions
	//-----------------------------------------------------------------------------
	HRESULT LockText();
    HRESULT DrawText(float x, float y, unsigned dwColor, const char* strText, float scale=1.f, D3DXVECTOR2* nextPt=NULL, int cursorPos=-1);
	HRESULT UnlockText();
    
	//-----------------------------------------------------------------------------
    // Function to get extent of text
	//-----------------------------------------------------------------------------
    HRESULT          GetTextExtent(const char* strText, D3DXVECTOR2& size) const;
	int              GetLineHeight() const         { return _TextMetric.tmHeight + _TextMetric.tmExternalLeading; }
    IDirect3DTexture9* GetTexture() const          { return _pTexture; }
    DWORD            GetTextureWidth() const       { return _TextureWidth; }

    //-----------------------------------------------------------------------------
	// routines to call when the device is being reset/destroyed/rebuilt
	//-----------------------------------------------------------------------------
	HRESULT OnPrepareReset();
	HRESULT OnReset();
	HRESULT OnInvalidateDevice();
	HRESULT OnRebuildDevice();

private:
	//-----------------------------------------------------------------------------
	// helper routines
	//-----------------------------------------------------------------------------
	HRESULT _LockVB();
	HRESULT _AddChar(char c, DWORD color, float& sx, float sy);
	HRESULT _FlushText();
	HRESULT _UnlockVB();

	// Stores the quad that represents the cursor location.
	bool                    _CursorFound;
	D3DXVECTOR4             _CursorQuad[4];

    IDirect3DVertexBuffer9 *_VB;         // VertexBuffer for rendering text

	// Stores the lock position of the VB
	struct FONT2DVERTEX;
	FONT2DVERTEX           *_LockedVerts;
	unsigned                _NumTriangles;

	//-----------------------------------------------------------------------------
    // Stateblocks for setting and restoring render states
	//-----------------------------------------------------------------------------
    IDirect3DStateBlock9*   _SavedStateBlock;
    IDirect3DStateBlock9*   _DrawTextStateBlock;

	//-----------------------------------------------------------------------------
    // The vertex type for character quads
	//-----------------------------------------------------------------------------
	struct FONT2DVERTEX 
	{ 
		D3DXVECTOR4 p;   
		unsigned color;     
		float tu, tv; 
	};
	
	enum 
	{
		D3DFVF_FONT2DVERTEX = D3DFVF_XYZRHW | 
			                  D3DFVF_DIFFUSE | 
							  D3DFVF_TEX1
	};

	static void _InitVertex( FONT2DVERTEX& v, float x, float y, float z, float w, D3DCOLOR color, float tu, float tv );
	static void _InitVertex( FONT2DVERTEX& v, const D3DXVECTOR4& p, D3DCOLOR color, float tu, float tv );
	void _GetTexCoords(unsigned uChar, float &u0, float &v0, float &u1, float &v1) const;

	//-----------------------------------------------------------------------------
    // Font data itself
	//-----------------------------------------------------------------------------
	std::string             _FontName;            // Font properties
    unsigned                _FontHeight;
    unsigned                _FontFlags;
	TEXTMETRIC				_TextMetric;

	IDirect3DTexture9*      _pTexture;
	unsigned                _TextureWidth;

    float					_texCoords[128-32][4];	// UL and BR corners of the character
													// Each character is always the same height
};

//-----------------------------------------------------------------------------
// D3DUtilFont
//-----------------------------------------------------------------------------
inline
D3DUtilFont::D3DUtilFont(const char* strFontName, unsigned dwHeight, unsigned dwFlags) :
_VB(NULL),
_LockedVerts(NULL),
_CursorFound(false),
_NumTriangles(0),
_SavedStateBlock(NULL),
_DrawTextStateBlock(NULL),
_FontName(strFontName), 
_FontHeight(dwHeight),
_FontFlags(dwFlags),
_pTexture(NULL),
_TextureWidth(0)
{
}

//-----------------------------------------------------------------------------
// _LockVB
//-----------------------------------------------------------------------------
inline HRESULT D3DUtilFont::_LockVB()
{
	ASSERT(_LockedVerts == NULL);
#ifdef _DEBUG
	FAIL_RETURN(_VB->Lock(0, 0, reinterpret_cast<void**>(&_LockedVerts), D3DLOCK_DISCARD | D3DLOCK_NOSYSLOCK));
#else
	FAIL_RETURN(_VB->Lock(0, 0, reinterpret_cast<void**>(&_LockedVerts), D3DLOCK_DISCARD));
#endif
    FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->SetFVF(D3DFVF_FONT2DVERTEX));
    FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->SetStreamSource(0, _VB, 0, sizeof(FONT2DVERTEX)));
	_NumTriangles = 0;
	return D3D_OK;
}

//-----------------------------------------------------------------------------
// _FlushText
//-----------------------------------------------------------------------------
inline HRESULT D3DUtilFont::_FlushText()
{
	FAIL_RETURN(D3DUtilContext::Instance()->GetD3DDevice()->DrawPrimitive(D3DPT_TRIANGLELIST, 0, _NumTriangles));
	return D3D_OK;
}

//-----------------------------------------------------------------------------
// _UnlockVB
//-----------------------------------------------------------------------------
inline HRESULT D3DUtilFont::_UnlockVB()
{
	ASSERT(_LockedVerts != NULL);
    FAIL_RETURN(_VB->Unlock());
	_LockedVerts = NULL;
	return D3D_OK;
}

//-----------------------------------------------------------------------------
// _InitVertex
//-----------------------------------------------------------------------------
inline void D3DUtilFont::_InitVertex( FONT2DVERTEX& v, float x, float y, float z, float w, D3DCOLOR color, float tu, float tv )
{
	v.p.x = x; 
	v.p.y = y; 
	v.p.z = z; 
	v.p.w = w; 
	
	v.color = color; 
	
	v.tu = tu; 
	v.tv = tv;
}

//-----------------------------------------------------------------------------
// _InitVertex
//-----------------------------------------------------------------------------
inline void D3DUtilFont::_InitVertex( FONT2DVERTEX& v, const D3DXVECTOR4& p, D3DCOLOR color, float tu, float tv )
{
	v.p = p;
	v.color = color;
	v.tu = tu;
	v.tv = tv;
}

//----------------------------------------------------------------------------
// GetTexCoords
// ---------------------------------------------------------------------------
// Get the texture coordinates for a single character
//----------------------------------------------------------------------------
inline void
D3DUtilFont::_GetTexCoords(unsigned uChar, float &u0, float &v0, float &u1, float &v1) const
{
	u0 = _texCoords[uChar-' '][0];
	v0 = _texCoords[uChar-' '][1];
	u1 = _texCoords[uChar-' '][2];
	v1 = _texCoords[uChar-' '][3];
}
#endif
