//============================================================================
// File          : UtilCamera.cpp
// 
// Content       : UtilCamera Class
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "stdafx.h"
#include "UtilCamera.hpp"

//----------------------------------------------------------------------------
//-- Static Variables --------------------------------------------------------
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
//-- Functions ---------------------------------------------------------------
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// _SyncViewingRectWithProjection()
// ---------------------------------------------------------------------------
// When a projection parameter changes, we need to sync the viewing rectangle
// parameters.
// Syncs _l, _r, _b, and _t.
//----------------------------------------------------------------------------
void UtilCamera::_SyncViewingRectWithProjection()
{
    float ww2, wh2;
    // This creates ww2 and wh2 as the full window width and height.
    SetupPerspectiveForFrustum(_fovY, _aspect, _n, ww2, wh2);
    // but we actually want the half width and height
    ww2 *= 0.5f;
    wh2 *= 0.5f;
    _l = -ww2; _r = ww2;
    _b = -wh2; _t = wh2;
}

//----------------------------------------------------------------------------
// _SyncProjectionWithViewingRect()
// ---------------------------------------------------------------------------
// When a viewing rectangle parameter changes, we need to sync the projection 
// parameters.
// Syncs _fovY and _aspect.
//----------------------------------------------------------------------------
void UtilCamera::_SyncProjectionWithViewingRect()
{
    // compute the fovY and aspect ratio
    float ww2 = (_r-_l)*0.5f;
    float wh2 = (_t-_b)*0.5f;
    _fovY = D3DXToDegree(atan2f(_n,wh2));
    _aspect = ww2 / wh2;
}

//----------------------------------------------------------------------------
// CalcWorldspaceVerts()
// ---------------------------------------------------------------------------
// Returns the 8 verts of the camera, starting with the 4 near plane verts
// from UpperRight to BottomRight.
// ---------------------------------------------------------------------------
void UtilCamera::CalcWorldspaceVerts(D3DXVECTOR3 verts[8]) const
{
	D3DXVECTOR3 camSpaceVerts[8];
    float ww2 = (_r-_l)*0.5f;
    float wh2 = (_t-_b)*0.5f;

	Vec3Set(camSpaceVerts[0], ww2,wh2,GetNear());		// TR Near
	Vec3Set(camSpaceVerts[1], -ww2,wh2,GetNear());		// TL Near
	Vec3Set(camSpaceVerts[2], -ww2,-wh2,GetNear());		// BL Near
	Vec3Set(camSpaceVerts[3], ww2,-wh2,GetNear());		// BR Near

	// CALC FAR PTS (IN CAM COORDS)
	float FN=GetFar()/GetNear();
	float fwL=-ww2*FN, fwR=ww2*FN, fwB=-wh2*FN, fwT=wh2*FN;
	Vec3Set(camSpaceVerts[4], fwR,fwT,GetFar());	// TR Far
	Vec3Set(camSpaceVerts[5], fwL,fwT,GetFar());	// TL Far
	Vec3Set(camSpaceVerts[6], fwL,fwB,GetFar());	// BL Far
	Vec3Set(camSpaceVerts[7], fwR,fwB,GetFar());	// BR Far

	// transform verts to worldspace
	for (int i=0; i<8; i++)
	{
		TransformCamToWorld(camSpaceVerts[i], verts[i]);
	}
}

//----------------------------------------------------------------------------
// CalcWorldspacePlanes()
// ---------------------------------------------------------------------------
//
// ---------------------------------------------------------------------------
void UtilCamera::CalcWorldspacePlanes(D3DXVECTOR4 planes[6]) const
{
    D3DXMATRIX mtx = GetWorldToViewMatrix(), temp = GetProjectionMatrix();
    
    // gives us our world->clip transform
    mtx *= temp;
    const D3DXVECTOR4& row1 = reinterpret_cast<const D3DXVECTOR4&>(mtx.m[0][0]);
    const D3DXVECTOR4& row2 = reinterpret_cast<const D3DXVECTOR4&>(mtx.m[1][0]);
    const D3DXVECTOR4& row3 = reinterpret_cast<const D3DXVECTOR4&>(mtx.m[2][0]);
    const D3DXVECTOR4& row4 = reinterpret_cast<const D3DXVECTOR4&>(mtx.m[3][0]);
    // left   clipping plane:  (x<-w)  =>  x+w<0  =>  -v*(row1 + row4)>0
    planes[0] = -(row1 + row4);
    // right  clipping plane:  (x>w)   =>  x-w>0  =>   v*(row1 - row4)>0
    planes[1] = row1 - row4;
    // top    clipping plane:  (y>w)   =>  y-w>0  =>   v*(row2 - row4)>0
    planes[2] = row2 - row4;
    // bottom clipping plane:  (y<-w)  =>  y+w<0  =>  -v*(row2 + row4)>0
    planes[3] = -(row2 + row4);
    // near   clipping plane:  (z<0)   =>   -z>0  =>  -v*(row3       )>0
    planes[4] = -row3;
    // far    clipping plane:  (z>w)   =>  z-w>0  =>   v*(row3 - row4)>0
    planes[5] = row3 - row4;
}

//----------------------------------------------------------------------------
// TransformWorldToCam()
// ---------------------------------------------------------------------------
// See Viewing16v for an explanation of how this works
// ---------------------------------------------------------------------------
void UtilCamera::TransformWorldToCam(const D3DXVECTOR3 &worldPt, D3DXVECTOR3 &camPt) const
{
	// camPt = (worldPt - Orig) * Rot
    camPt = worldPt;
    camPt -= _o;
    Vec3Set(camPt, D3DXVec3Dot(&camPt,&_x), D3DXVec3Dot(&camPt,&_y), D3DXVec3Dot(&camPt,&_z));
}

//----------------------------------------------------------------------------
// TransformCamToWorld()
// ---------------------------------------------------------------------------
// See InvViewing16v for an explanation of how this works
// ---------------------------------------------------------------------------
void UtilCamera::TransformCamToWorld(const D3DXVECTOR3 &camPt, D3DXVECTOR3 &worldPt) const
{
	// worldPt = (camPt * InvRot) + Orig
	worldPt = _x*camPt.x + _y*camPt.y + _z*camPt.z + GetPosition();
}

//----------------------------------------------------------------------------
// Transform()
// ---------------------------------------------------------------------------
// Transforms the camera by the specified matrix. Only rotates and translates
// are allowed. The transform looks like the following:
// [ X X X 0 ]   [ R R R 1 ]
// [ Y Y Y 0 ] * [ R R R 1 ]
// [ Z Z Z 0 ]   [ R R R 1 ]
// [ O O O 1 ]   [ T T T 1 ]
//
// This is essentially three vector rotations (the basis vectors) and a
// vector transformation (the origin).
// *** THIS ROUTINE IS DEPENDEENT ON mat BEING A ROW-MAJOR MATRIX ***
// ---------------------------------------------------------------------------
void UtilCamera::Transform(const D3DXMATRIX &mat)
{
    Vec3Set(_x,
        _x.x*mat[0] + _x.y*mat[4] + _x.z*mat[ 8],
        _x.x*mat[1] + _x.y*mat[5] + _x.z*mat[ 9],
        _x.x*mat[2] + _x.y*mat[6] + _x.z*mat[10]);
    Vec3Set(_y,
        _y.x*mat[0] + _y.y*mat[4] + _y.z*mat[ 8],
        _y.x*mat[1] + _y.y*mat[5] + _y.z*mat[ 9],
        _y.x*mat[2] + _y.y*mat[6] + _y.z*mat[10]);
    Vec3Set(_z,
        _z.x*mat[0] + _z.y*mat[4] + _z.z*mat[ 8],
        _z.x*mat[1] + _z.y*mat[5] + _z.z*mat[ 9],
        _z.x*mat[2] + _z.y*mat[6] + _z.z*mat[10]);
    Vec3Set(_o,
        _o.x*mat[0] + _o.y*mat[4] + _o.z*mat[ 8] + mat[12],
        _o.x*mat[1] + _o.y*mat[5] + _o.z*mat[ 9] + mat[13],
        _o.x*mat[2] + _o.y*mat[6] + _o.z*mat[10] + mat[14]);
}

//----------------------------------------------------------------------------
// CalcPixelRay()
// ---------------------------------------------------------------------------
// Create a worldspace ray going through a pixel location starting at 
// the pixel location (integers represent pixel centers).
// Have to supply the viewport parameters
// *** THIS ROUTINE IS DEPENDEENT ON D3D VIEW DIR BEING +Z ***
//----------------------------------------------------------------------------
void UtilCamera::CalcPixelRay(const D3DXVECTOR2& screenPos, D3DXVECTOR3& worldPos, D3DXVECTOR3& worldDir, const D3DXVECTOR2& viewportOrigin, const D3DXVECTOR2& viewportSize)
{
	// Calculate the normalized pixel position
	float fx = (screenPos.x + 0.5f - viewportOrigin.x) / viewportSize.x;
	float fy = (screenPos.y + 0.5f - viewportOrigin.y) / viewportSize.y;

	// Locate that point in camera space, starting with the Top-Left-Near point on camera
	D3DXVECTOR3 camPos(-GetWindowWidth()*0.5f, GetWindowHeight()*0.5f, GetNear());  
	// Offset the TLN point by the normalized pixel postion re-scaled to the cameraspace viewport size.
	camPos.x += fx * GetWindowWidth();
	camPos.y -= fy * GetWindowHeight();
	// Transform the point to world space. This is our starting point
	TransformCamToWorld(camPos, worldPos);
	// Deterimine the worldspace direction by forming a vector from the camera center to the worldPos
	worldDir = worldPos; 
	worldDir -= GetPosition();
}
