//============================================================================
// UtilMouseNav.hpp:
//
// Encapsulates "mouse-style" movement using a UtilCamera.
//============================================================================
#include "stdafx.h"
#include "UtilMouseNav.hpp"
#include <zmouse.h>
#include <algorithm>

//----------------------------------------------------------------------------
// SetNavMode
//----------------------------------------------------------------------------
void UtilMouseNav::SetNavMode(int mode)
{
	if (mode == NavMode) return;
	static bool oldInsideLookingOut;
	static HCURSOR oldCursor = NULL;

	// set up new mode
	switch(mode)
	{
	case NAV_MODE_TRACKBALL:
	case NAV_MODE_DRIVE:
	case NAV_MODE_TRANSLATE:
	case NAV_MODE_LOOK:
		// if we are transitioning to an active mode from a passive mode:
		//   1. release the mouse capture and show it again.
		//   2. restore the old iLookO mode
		if (NavMode >= NAV_MODE_PASSIVE)
		{
			ReleaseCapture();
			SetCursor(oldCursor);
			InsideLookingOut = oldInsideLookingOut;
			LeftButtonDown = false;
		}
		break;

	case NAV_MODE_DOOM:
	case NAV_MODE_OBSERVE:
        {
			D3DXVECTOR3 view = pCam->GetView();
			D3DXVECTOR3 side; D3DXVec3Cross(&side, &ViewUp, &pCam->GetView()); D3DXVec3Normalize(&side, &side);
			D3DXVECTOR3 up; D3DXVec3Cross(&up, &view, &side); D3DXVec3Normalize(&up, &up);
			pCam->SetViewSideUp(view, side, up);
        }
		// fall through

	case NAV_MODE_DESCENT:
		// if we are transitioning to a passive mode from an active mode:
		//   1. capture the mouse and hide it.
		//   2. save the old iLookO mode
		if (NavMode < NAV_MODE_PASSIVE)
		{
			SetCapture(GetActiveWindow());
			oldCursor = SetCursor(NULL);
			POINT p;
			GetCursorPos(&p); 
            ScreenToClient(GetActiveWindow(), &p); 
            NewX = static_cast<float>(p.x); 
            NewY = static_cast<float>(p.y);

			oldInsideLookingOut = InsideLookingOut;
			InsideLookingOut = true;
		}

		// always reset our movement indicators
		SlidingLR =
		SlidingUD =
		SlidingFB =
		Twisting  = 0;

		break;
	}

	NavMode = mode;
}

void UtilMouseNav::SetActive(int active)
{
	Active = active;
}

void UtilMouseNav::SetWindowSize(int width, int height)
{
    WW = width;
    WH = height;
	PassiveBorderWidth = std::min(WW, WH) / 10;
}

//----------------------------------------------------------------------------
// HANDLE WINDOWS MESSAGES AND TRANSLATE THEM INTO Camera MOVEMENT
//----------------------------------------------------------------------------
LRESULT UtilMouseNav::WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled)
{
	if (!Active) return S_OK;

	switch (uMsg)
	{
	case WM_CHAR:
		switch (TCHAR(wParam))
		{
		case VK_ESCAPE: 
			if (NavMode >= NAV_MODE_PASSIVE) 
			{
				SetNavMode(NAV_MODE_TRACKBALL); handled = true; return S_OK;
			}
		}
		break;

	case WM_KEYDOWN:
		switch (wParam)
		{
		case VK_F1: SetNavMode(NAV_MODE_TRACKBALL); handled = true; return S_OK;
		case VK_F2: SetNavMode(NAV_MODE_DRIVE); handled = true; return S_OK;
		case VK_F3: SetNavMode(NAV_MODE_TRANSLATE); handled = true; return S_OK;
		case VK_F4: SetNavMode(NAV_MODE_LOOK); handled = true; return S_OK;
		case VK_F5: SetNavMode(NAV_MODE_DESCENT); handled = true; return S_OK;
		case VK_F6: SetNavMode(NAV_MODE_DOOM); handled = true; return S_OK;
		case VK_F7: SetNavMode(NAV_MODE_OBSERVE); handled = true; return S_OK;
		}
		break;

	case WM_MOUSEWHEEL:
		// increase/decrease the movement scale when the scroll wheel is used
		int zDelta; zDelta = (short)HIWORD(wParam);
		float m; m = 2.f;
		if (zDelta < 0) m = 0.5f;
		if (NavMode >= NAV_MODE_PASSIVE) PassiveMult *= m;
		else Mult *= m;
		handled = true; return S_OK;

	case WM_SETCURSOR:
		// Turn off window cursor 
		if (NavMode >= NAV_MODE_PASSIVE) 
        {
            SetCursor(NULL);
		    handled = true;
        }
        return S_OK;

	}

	// return the hres no matter what because it encodes whether the function completely handled the message.
	LRESULT res;
	if (NavMode < NAV_MODE_PASSIVE)
	{
		res = WndProcActive(hWnd, uMsg, wParam, lParam, handled);
	} else if (NavMode == NAV_MODE_DESCENT)
	{
		res = WndProcDescent(hWnd, uMsg, wParam, lParam, handled);
	} else if (NavMode == NAV_MODE_DOOM)
	{
		res = WndProcDoom(hWnd, uMsg, wParam, lParam, handled);
	} else if (NavMode == NAV_MODE_OBSERVE)
	{
		res = WndProcObserve(hWnd, uMsg, wParam, lParam, handled);
	}
	return res;
}

LRESULT UtilMouseNav::WndProcActive(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled)
{
	switch (uMsg)
	{
    case WM_LBUTTONDOWN:
		LeftButtonDown	= true; // SET LEFT-BUTTON-DOWN FLAG
		NewX			= static_cast<float>(GET_X_LPARAM(lParam));				// STORE THE NEW MOUSE POS
		NewY			= static_cast<float>(GET_Y_LPARAM(lParam));				
		CtrlPressed		= (wParam & MK_CONTROL) != 0;	// CTRL BIT SET
		OldX			= NewX;	// RESET THE OLD TO THE CURRENT (starting over)
		OldY			= NewY;      

		SetCapture(hWnd);
		handled = true; return S_OK;

    case WM_LBUTTONUP:
		NewX			= static_cast<float>(GET_X_LPARAM(lParam));				// STORE THE NEW MOUSE POS
		NewY			= static_cast<float>(GET_Y_LPARAM(lParam));				
		LeftButtonDown	= false;						// UNSET LEFT-BUTTON-DOWN FLAG

		ReleaseCapture();
		handled = true; return S_OK;

    case WM_MOUSEMOVE:
		if (LeftButtonDown)
		{
			OldX = NewX;
			OldY = NewY;
			NewX = static_cast<float>(GET_X_LPARAM(lParam));				// STORE THE NEW MOUSE POS
			NewY = static_cast<float>(GET_Y_LPARAM(lParam));

			// -------  WORLD NAVIGATION -------
			switch(NavMode)
			{
			case NAV_MODE_TRACKBALL:
				if (CtrlPressed) { TrackBallX(OldX,NewX); DriveY(OldY,NewY); }
				else { TrackBallX(OldX,NewX); TrackBallY(OldY,NewY); }
				break;
			
			case NAV_MODE_DRIVE:
				if (CtrlPressed) { TranslateX(Mult*OldX,Mult*NewX); TranslateY(Mult*OldY,Mult*NewY); }
				else { LookX(OldX,NewX); DriveY(Mult*OldY,Mult*NewY); }
				break;
			
			case NAV_MODE_TRANSLATE:
				TranslateX(Mult*OldX,Mult*NewX); 
				if (CtrlPressed) DriveY(Mult*OldY,Mult*NewY); else TranslateY(Mult*OldY,Mult*NewY);
				break;

			case NAV_MODE_LOOK:
				if (CtrlPressed) { TranslateX(Mult*OldX,Mult*NewX); TranslateY(Mult*OldY,Mult*NewY); }
				else { LookX(OldX,NewX); LookY(OldY,NewY); }
				break;
			};
			handled = true; return S_OK;
		}
		break;
	}

	return S_OK;
}

LRESULT UtilMouseNav::WndProcDescent(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled)
{
	switch (uMsg)
	{
	case WM_CHAR:
		switch (TCHAR(wParam))
		{
		case 'a': case 'A':
		case 'd': case 'D':
		case 'w': case 'W':
		case 's': case 'S':
		case 'q': case 'Q':
		case 'e': case 'E':
		handled = true; return S_OK;
		}
		break;

	case WM_KEYDOWN:
		{
			bool firstPress = (lParam & (1 << 30)) == 0;
			switch (wParam)
			{
			case 'A': if (firstPress) ++SlidingLR; handled = true; return S_OK;
			case 'D': if (firstPress) --SlidingLR; handled = true; return S_OK;
			case 'W': if (firstPress) ++SlidingUD; handled = true; return S_OK;
			case 'S': if (firstPress) --SlidingUD; handled = true; return S_OK;
			case 'Q': if (firstPress) ++Twisting; handled = true; return S_OK;
			case 'E': if (firstPress) --Twisting; handled = true; return S_OK;
			}
		}
		break;

	case WM_KEYUP:
		switch (wParam)
		{
			case 'A': --SlidingLR; handled = true; return S_OK;
			case 'D': ++SlidingLR; handled = true; return S_OK;
			case 'W': --SlidingUD; handled = true; return S_OK;
			case 'S': ++SlidingUD; handled = true; return S_OK;
			case 'Q': --Twisting; handled = true; return S_OK;
			case 'E': ++Twisting; handled = true; return S_OK;
		}
		break;

    case WM_LBUTTONDOWN:
		++SlidingFB; handled = true; return S_OK;

    case WM_LBUTTONUP:
		--SlidingFB; handled = true; return S_OK;

    case WM_RBUTTONDOWN:
		--SlidingFB; handled = true; return S_OK;

    case WM_RBUTTONUP:
		++SlidingFB; handled = true; return S_OK;

    case WM_MOUSEMOVE:
		ConstrainCursor(hWnd, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		handled = true; return S_OK;

	}

	return S_OK;
}

LRESULT UtilMouseNav::WndProcDoom(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled)
{
	switch (uMsg)
	{
	case WM_CHAR:
		switch (TCHAR(wParam))
		{
		case 'a': case 'A':
		case 'd': case 'D':
		case 'w': case 'W':
		case 's': case 'S':
		case 'c': case 'C':
		case ' ': 
		handled = true; return S_OK;
		}
		break;

	case WM_KEYDOWN:
		{
			bool firstPress = (lParam & (1 << 30)) == 0;
			switch (wParam)
			{
			case 'A': if (firstPress) ++SlidingLR; handled = true; return S_OK;
			case 'D': if (firstPress) --SlidingLR; handled = true; return S_OK;
			case ' ': if (firstPress) ++SlidingUD; handled = true; return S_OK;
			case 'C': if (firstPress) --SlidingUD; handled = true; return S_OK;
			case 'W': if (firstPress) ++SlidingFB;  handled = true; return S_OK;
			case 'S': if (firstPress) --SlidingFB;  handled = true; return S_OK;
			}
		}
		break;

	case WM_KEYUP:
		switch (wParam)
		{
			case 'A': --SlidingLR; handled = true; return S_OK;
			case 'D': ++SlidingLR; handled = true; return S_OK;
			case ' ': --SlidingUD; handled = true; return S_OK;
			case 'C': ++SlidingUD; handled = true; return S_OK;
			case 'W': --SlidingFB;  handled = true; return S_OK;
			case 'S': ++SlidingFB;  handled = true; return S_OK;
		}
		break;

    case WM_MOUSEMOVE:
		ConstrainCursor(hWnd, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		handled = true; return S_OK;
		break;

	}

	return S_OK;
}

LRESULT UtilMouseNav::WndProcObserve(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& handled)
{
	switch (uMsg)
	{
	case WM_CHAR:
		switch (TCHAR(wParam))
		{
		case 'a': case 'A':
		case 'd': case 'D':
		case 'w': case 'W':
		case 's': case 'S':
		case 'c': case 'C':
		case ' ': 
		handled = true; return S_OK;
		}
		break;

	case WM_KEYDOWN:
		{
			bool firstPress; firstPress = (lParam & (1 << 30)) == 0;
			switch (wParam)
			{
			case 'A': 
				if (firstPress) ++SlidingLR; handled = true; return S_OK;
			case 'D': 
				if (firstPress) --SlidingLR; handled = true; return S_OK;
			case ' ': 
				if (firstPress) ++SlidingUD; handled = true; return S_OK;
			case 'C': 
				if (firstPress) --SlidingUD; handled = true; return S_OK;
			case 'W': 
				if (firstPress) ++SlidingFB;  handled = true; return S_OK;
			case 'S': 
				if (firstPress) --SlidingFB;  handled = true; return S_OK;
			}
		}
		break;

	case WM_KEYUP:
		switch (wParam)
		{
			case 'A': --SlidingLR; handled = true; return S_OK;
			case 'D': ++SlidingLR; handled = true; return S_OK;
			case ' ': --SlidingUD; handled = true; return S_OK;
			case 'C': ++SlidingUD; handled = true; return S_OK;
			case 'W': --SlidingFB;  handled = true; return S_OK;
			case 'S': ++SlidingFB;  handled = true; return S_OK;
		}
		break;

    case WM_MOUSEMOVE:
		ConstrainCursor(hWnd, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
		handled = true; return S_OK;
		break;

	}

	return S_OK;
}

static bool s_waiting=false;
static POINT s_waitingPt;
void UtilMouseNav::ConstrainCursor(HWND hWnd, int x, int y)
{
	if (s_waiting)
	{
		if (s_waitingPt.x == x && s_waitingPt.y == y)
		{
			NewX = static_cast<float>(x);
			NewY = static_cast<float>(y);
			s_waiting = false;
			return;
		}
	}

	int move=0;

	int waitingX=x, waitingY=y;
	OldX=NewX; OldY=NewY;
	NewX=static_cast<float>(x); NewY=static_cast<float>(y);
	float dx=NewX-OldX;
	float dy=NewY-OldY;
	if (dx == 0 && dy == 0) return;

	if (x > WW-PassiveBorderWidth) {
		waitingX = PassiveBorderWidth+(x-WW+PassiveBorderWidth);
		move=1;
	} else if (x < PassiveBorderWidth) {
		waitingX = WW-PassiveBorderWidth + x-PassiveBorderWidth;
		move=1;
	}
	if (y > WH-PassiveBorderWidth) {
		waitingY = PassiveBorderWidth+(y-WH+PassiveBorderWidth);
		move=1;
	} else if (y < PassiveBorderWidth) {
		waitingY = WH-PassiveBorderWidth + y-PassiveBorderWidth;
		move=1;
	}

	if (move) {
		POINT p; p.x = waitingX; p.y = waitingY;
		s_waiting = true;
		s_waitingPt = p;
		ClientToScreen(hWnd, &p);
		SetCursorPos(p.x, p.y);
	}

	if (NavMode==NAV_MODE_DESCENT) {
		LookUserRefX(OldX,NewX);
		// invert the mouselook orientation
		LookY(OldY, OldY-(NewY-OldY));
	} else if (NavMode==NAV_MODE_DOOM) {
		LookX(OldX,NewX);
		LookConstrainedY(OldY,NewY);
	} else if (NavMode==NAV_MODE_OBSERVE) {
		LookX(OldX,NewX);
		LookConstrainedY(OldY,NewY);
	}
}

//----------------------------------------------------------------------------
// DoPassiveMovement
//----------------------------------------------------------------------------
void UtilMouseNav::DoPassiveMovement(DWORD lastFrameTime)
{
	// skip this if we aren't in a passive mode
	if (NavMode < NAV_MODE_PASSIVE) return;

	float f = lastFrameTime * 0.001f * PassiveMult;
	float fW = f * WW;
	float fH = f * WH;
	if (SlidingLR) TranslateX(0, SlidingLR * fW);
	if (SlidingUD) TranslateY(0, SlidingUD * fH);
	if (NavMode==NAV_MODE_DOOM)
	{
		if (SlidingFB) DriveUserRefY(0, -SlidingFB * fH);
	} else if (NavMode==NAV_MODE_DESCENT)
	{
		if (SlidingFB) DriveY(0, -SlidingFB * fH);
		if (Twisting)  TwistX(0, Twisting * (float) lastFrameTime * WW * (360 / 180) * 0.5f * 0.001f);
	} else if (NavMode==NAV_MODE_OBSERVE)
	{
		if (SlidingFB) DriveY(0, -SlidingFB * fH);
	}
}

//----------------------------------------------------------------------------
// MOUSE-TO-CAMERA MOVEMENT ROUTINE: TRANSLATE
//----------------------------------------------------------------------------
void UtilMouseNav::TranslateX(float OldX, float NewX)
{
  float RelX = (NewX-OldX)/(float)WW;
  if (RelX == 0) return;
  float dX;

  if (InsideLookingOut)
  {
    dX = RelX * (-WorldRadius*0.25f);
  }
  else // FOR OUTSIDE-LOOKING-IN
  {
    D3DXVECTOR3 EyeToWorldCntr = WorldCenter - pCam->GetPosition();
    float DistToCntr = D3DXVec3Length(&EyeToWorldCntr);
    if (DistToCntr<WorldRadius) DistToCntr=WorldRadius;

    float vpX = RelX * pCam->GetWindowWidth();
    dX = DistToCntr * vpX / -pCam->GetNear();
  }

  D3DXMATRIX Mat;
  D3DXVECTOR3 Trans = pCam->GetSide() * dX;
  D3DXMatrixTranslation(&Mat, Trans.x, Trans.y, Trans.z);
  pCam->Transform(Mat);
}

void UtilMouseNav::TranslateY(float OldY, float NewY)
{
  float RelY = (NewY-OldY)/(float)WH;
  if (RelY == 0) return;
  float dY;

  if (InsideLookingOut)
  {
    dY = RelY * (WorldRadius*0.25f);
  }
  else // FOR OUTSIDE-LOOKING-IN
  {
    D3DXVECTOR3 EyeToWorldCntr = WorldCenter - pCam->GetPosition();
    float DistToCntr = D3DXVec3Length(&EyeToWorldCntr);
    if (DistToCntr<WorldRadius) DistToCntr=WorldRadius;

    float vpY = RelY * pCam->GetWindowHeight();
    dY = DistToCntr * vpY / pCam->GetNear();
  }

  D3DXMATRIX Mat;
  D3DXVECTOR3 Trans = pCam->GetUp() * dY;
  D3DXMatrixTranslation(&Mat, Trans.x, Trans.y, Trans.z);
  pCam->Transform(Mat);
}

//----------------------------------------------------------------------------
// MOUSE-TO-CAMERA MOVEMENT ROUTINE: DRIVE MODE
//----------------------------------------------------------------------------
void UtilMouseNav::DriveY(float OldY, float NewY)
{
  float RelY = (OldY-NewY)/(float)WH;
  if (RelY == 0) return;
  float dY;
  
  if (InsideLookingOut)
  {
    dY = RelY * (WorldRadius*0.25f);
  }
  else // FOR OUTSIDE-LOOKING-IN
  {
    D3DXVECTOR3 EyeToWorldCntr = WorldCenter - pCam->GetPosition();
    float DistToCntr = D3DXVec3Length(&EyeToWorldCntr);
    if (DistToCntr<WorldRadius) DistToCntr=WorldRadius;
	dY = DistToCntr*RelY*2.0f;
  }

  D3DXMATRIX Mat;
  D3DXVECTOR3 Trans = pCam->GetView() * dY;
  D3DXMatrixTranslation(&Mat, Trans.x, Trans.y, Trans.z);
  pCam->Transform(Mat);
}


void UtilMouseNav::DriveUserRefY(float OldY, float NewY)
{
  float RelY = (NewY-OldY)/(float)WH;
  if (RelY == 0) return;
  float dY;
  
  if (InsideLookingOut)
  {
    dY = RelY * (WorldRadius*0.25f);
  }
  else // FOR OUTSIDE-LOOKING-IN
  {
    D3DXVECTOR3 EyeToWorldCntr = WorldCenter - pCam->GetPosition();
    float DistToCntr = D3DXVec3Length(&EyeToWorldCntr);
    if (DistToCntr<WorldRadius) DistToCntr=WorldRadius;
	dY = DistToCntr*RelY*2.0f;
  }

  D3DXMATRIX Mat;
  D3DXVECTOR3 Trans; D3DXVec3Cross(&Trans, &ViewUp, &pCam->GetSide());
  // Trans.Normalize();
  Trans *= dY;

  D3DXMatrixTranslation(&Mat, Trans.x, Trans.y, Trans.z);
  pCam->Transform(Mat);
}

//----------------------------------------------------------------------------
// MOUSE-TO-CAMERA MOVEMENT ROUTINE: LOOK MODE
//----------------------------------------------------------------------------
void UtilMouseNav::LookX(float OldX, float NewX)
{
  float RelX = (OldX-NewX)/(float)WW;
  if (RelX == 0) return;
 
  D3DXMATRIX Mat;
  D3DXMATRIX Rot;
  D3DXMatrixRotationAxis(&Rot, &ViewUp, D3DXToRadian(-RelX*90.f));

  MatrixTransRotTrans(Mat, -pCam->GetPosition(), Rot, pCam->GetPosition());
  pCam->Transform(Mat);
}

void UtilMouseNav::LookY(float OldY, float NewY)
{
  float RelY = (OldY-NewY)/(float)WH;
  if (RelY == 0) return;

  D3DXMATRIX Mat;
  D3DXMATRIX Rot;
  D3DXMatrixRotationAxis(&Rot, &pCam->GetSide(), D3DXToRadian(-RelY*90.f));
  MatrixTransRotTrans(Mat, -pCam->GetPosition(), Rot, pCam->GetPosition());
  pCam->Transform(Mat);
}


//----------------------------------------------------------------------------
// MOUSE-TO-CAMERA MOVEMENT ROUTINE: TRACKBALL MODE
//----------------------------------------------------------------------------
void UtilMouseNav::TrackBallX(float OldX, float NewX)
{
  float RelX = (OldX-NewX)/(float)WW;
  if (RelX == 0) return;

  D3DXVECTOR3 Up = InsideLookingOut ? ViewUp : pCam->GetUp();
  D3DXMATRIX Mat;
  D3DXMATRIX Rot;
  D3DXMatrixRotationAxis(&Rot, &Up, D3DXToRadian(-RelX*180.f));
  MatrixTransRotTrans(Mat, -WorldCenter, Rot, WorldCenter);
  pCam->Transform(Mat);
}

void UtilMouseNav::TrackBallY(float OldY, float NewY)
{
  float RelY = (OldY-NewY)/(float)WH;
  if (RelY == 0) return;

  D3DXMATRIX Mat;
  D3DXMATRIX Rot;
  D3DXMatrixRotationAxis(&Rot, &pCam->GetSide(), D3DXToRadian(-RelY*180.f));
  MatrixTransRotTrans(Mat, -WorldCenter, Rot, WorldCenter);
  pCam->Transform(Mat);
}

void UtilMouseNav::TwistX(float OldX, float NewX)
{
  float RelX = (OldX-NewX)/(float)WW;
  if (RelX == 0) return;

  D3DXMATRIX Mat;
  D3DXMATRIX Rot;
  D3DXMatrixRotationAxis(&Rot, &pCam->GetView(), D3DXToRadian(-RelX*180.f));
  MatrixTransRotTrans(Mat, -pCam->GetPosition(), Rot, pCam->GetPosition());
  pCam->Transform(Mat);
}

void UtilMouseNav::LookUserRefX(float OldX, float NewX)
{
  float RelX = (NewX-OldX)/(float)WW;
  if (RelX == 0) return;
 
  D3DXMATRIX Mat;
  D3DXMATRIX Rot;
  D3DXMatrixRotationAxis(&Rot, &pCam->GetUp(), D3DXToRadian(RelX*90.f));
  MatrixTransRotTrans(Mat, -pCam->GetPosition(), Rot, pCam->GetPosition());
  pCam->Transform(Mat);
}

void UtilMouseNav::LookConstrainedY(float OldY, float NewY)
{
  float RelY = (NewY-OldY)/(float)WH;
  if (RelY == 0) return;
 
  // DON'T ROTATE IF WE'VE ALREADY ROTATED MORE THAN 87 DEGREES UP OR DOWN
  float dot = D3DXVec3Dot(&pCam->GetView(),&ViewUp);
  if (dot > ((float)0.9986295) && (RelY < 0)) return;
  if (dot < ((float)-0.9986295) && (RelY > 0)) return;

  D3DXMATRIX Mat;
  D3DXMATRIX Rot;
  D3DXMatrixRotationAxis(&Rot, &pCam->GetSide(), D3DXToRadian(RelY*90.f));
  MatrixTransRotTrans(Mat, -pCam->GetPosition(), Rot, pCam->GetPosition());
  pCam->Transform(Mat);
}
