//============================================================================
// File          : VertexLighting.cpp
// 
// Content       : All the setup used for the vertex lighting Cg programs.
// 
//============================================================================
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "d3d8Expanded.hpp"
#include <vector>
#include <iterator>

extern D3DXCOLOR MatDiffuse;
extern D3DXCOLOR MatAmbientEmissive;
extern D3DXCOLOR MatSpecular;
extern D3DXVECTOR4 MatPower;

static DWORD FToDW(float f)
{
    return *((DWORD*)&f);
}

const char* CG_VERT_LIGHTING = "..\\common\\shaders\\VertexLighting.cg";
const char* CG_SIMPLE_BLENDING = "..\\common\\shaders\\SimpleBlending.cg";

//----------------------------------------------------------------------------
// _LoadVertexLightingPrograms()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d8Expanded::_LoadVertexLightingPrograms()
{
    IDirect3DDevice8* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    {
        const char* profileOpts[] =
        {
            cgD3D8GetOptimalOptions(VertexProfile),
            NULL,
        };

        vertLightParms_.vertprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_VERT_LIGHTING, VertexProfile, "main", profileOpts);
	    if (vertLightParms_.vertprog_ == NULL)
	    {
		    const char* listing = cgGetLastListing(context_);
		    if (listing == NULL)
		    {
			    listing = "No listing available!";
		    }
            OutputErrors(listing);
		    FAIL_RETURN(E_FAIL);
	    }
    }

    OutputErrors(cgGetProgramString(vertLightParms_.vertprog_, CG_COMPILED_PROGRAM));

    // validate the declaration
    std::vector<DWORD> decl;
    decl.push_back(D3DVSD_STREAM(0));
    decl.push_back(D3DVSD_REG(cgD3D8ResourceToInputRegister(cgGetParameterResource(cgGetNamedParameter(
        vertLightParms_.vertprog_, "In.Position"))), D3DVSDT_FLOAT3));
    decl.push_back(D3DVSD_REG(cgD3D8ResourceToInputRegister(cgGetParameterResource(cgGetNamedParameter(
        vertLightParms_.vertprog_, "In.Normal"  ))), D3DVSDT_FLOAT3));
    decl.push_back(D3DVSD_REG(cgD3D8ResourceToInputRegister(cgGetParameterResource(cgGetNamedParameter(
        vertLightParms_.vertprog_, "In.Tex0"    ))), D3DVSDT_FLOAT2));
    decl.push_back(D3DVSD_SKIP(3)); // skip the tangent vector
    decl.push_back(D3DVSD_END());
    FAIL_RETURN(cgD3D8ValidateVertexDeclaration(vertLightParms_.vertprog_, &decl.front()) ? D3D_OK : E_FAIL);

	// Assemble the shader
    FAIL_RETURN(cgD3D8LoadProgram(vertLightParms_.vertprog_, true, 0, 
        (D3DUtilContext::Instance()->GetCreationParameters().BehaviorFlags & D3DCREATE_SOFTWARE_VERTEXPROCESSING) ? D3DUSAGE_SOFTWAREPROCESSING : 0, 
        &decl.front()));

    // Verify the parameter sizes are what we expect (optional)
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj")))              == 16 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale")))          ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower")))              ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition")))        ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange")))           ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation")))     ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"))) ==  4 ? D3D_OK : E_FAIL);

    // Get the parameters that changes every frame.
	vertLightParms_.mvp_                  = cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj");              FAIL_RETURN(vertLightParms_.mvp_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.scale_                = cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale");          FAIL_RETURN(vertLightParms_.scale_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.lightPosition_        = cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition");        FAIL_RETURN(vertLightParms_.lightPosition_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.lightRange_           = cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange");           FAIL_RETURN(vertLightParms_.lightRange_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.lightAttenuation_     = cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation");     FAIL_RETURN(vertLightParms_.lightAttenuation_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.camPositonModelspace_ = cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"); FAIL_RETURN(vertLightParms_.camPositonModelspace_ == NULL ? E_FAIL:D3D_OK);

    // Bind the parameters that are constant.
    CGparameter param;
    D3DXCOLOR dummy(0.5f,0.5f,0.5f,1.f);
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D8SetUniform(param, &dummy);
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D8SetUniform(param, &dummy);
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D8SetUniform(param, &dummy);
	param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D8SetUniform(param, &MatPower);

    // Try binding an array
    D3DXCOLOR colorArray[3] = {MatDiffuse, MatSpecular, MatAmbientEmissive};
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    FAIL_RETURN(cgD3D8SetUniformArray(param, 0, 3, colorArray));
    
    // Try binding a struct
    struct MatStruct
    {
        D3DXCOLOR colors[3];
        float power;
    };
    MatStruct myMat;
    myMat.colors[0] = MatDiffuse;
    myMat.colors[1] = MatSpecular;
    myMat.colors[2] = MatAmbientEmissive;
    myMat.power = MatPower.x;
    FAIL_RETURN(cgD3D8SetUniform(param, &myMat));


    // Now load the pixel shader
    {
        const char* profileOpts[] =
        {
            cgD3D8GetOptimalOptions(PixelProfile),
            NULL,
        };

        vertLightParms_.pixprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_SIMPLE_BLENDING, PixelProfile, "main", profileOpts);
	    if (vertLightParms_.pixprog_ == NULL)
	    {
		    const char* listing = cgGetLastListing(context_);
		    if (listing == NULL)
		    {
			    listing = "No listing available!";
		    }
            OutputErrors(listing);
		    FAIL_RETURN(E_FAIL);
	    }
    }

	// Assemble the shader
    FAIL_RETURN(cgD3D8LoadProgram(vertLightParms_.pixprog_, true, 0, 0, NULL));

    // Bind the parameters that are constant.
    CGparameter texParam = cgGetNamedParameter(vertLightParms_.pixprog_, "baseMap");
	FAIL_RETURN(cgD3D8SetTexture(texParam, meshTexture_));
    FAIL_RETURN(cgD3D8SetTextureStageState(texParam, D3DTSS_MAGFILTER, D3DTEXF_LINEAR));
    FAIL_RETURN(cgD3D8SetTextureStageState(texParam, D3DTSS_MINFILTER, D3DTEXF_LINEAR));
    FAIL_RETURN(cgD3D8SetTextureStageState(texParam, D3DTSS_MIPFILTER, D3DTEXF_LINEAR));

    return D3D_OK;
}

//----------------------------------------------------------------------------
// _SetupForVertexLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d8Expanded::_SetupForVertexLighting(const D3DXMATRIX& modelviewproj, D3DXVECTOR4& lightPos, D3DXVECTOR4& lightRange, D3DXVECTOR4& lightAttenuation, D3DXVECTOR4& cameraPosition, D3DXVECTOR4& scale)
{
    IDirect3DDevice8* pDev = D3DUtilContext::Instance()->GetD3DDevice();

	// bind the constants to the program (this must be done before activating the shaders)
    FAIL_RETURN(cgD3D8SetUniformMatrix(vertLightParms_.mvp_, &modelviewproj));

	FAIL_RETURN(cgD3D8SetUniform(vertLightParms_.scale_, &scale));
	FAIL_RETURN(cgD3D8SetUniform(vertLightParms_.lightPosition_, &lightPos));
	FAIL_RETURN(cgD3D8SetUniform(vertLightParms_.lightRange_, &lightRange));
	FAIL_RETURN(cgD3D8SetUniform(vertLightParms_.lightAttenuation_, &lightAttenuation));
	FAIL_RETURN(cgD3D8SetUniform(vertLightParms_.camPositonModelspace_, &cameraPosition));
	
    // activate the shaders
    FAIL_RETURN(cgD3D8BindProgram(vertLightParms_.vertprog_));
    FAIL_RETURN(cgD3D8BindProgram(vertLightParms_.pixprog_));

    return D3D_OK;
}
