//============================================================================
// File          : VertexLighting.cpp
// 
// Content       : All the setup used for the vertex lighting Cg programs.
// 
//============================================================================
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "d3d8Minimal.hpp"
#include <vector>
#include <iterator>

extern D3DXCOLOR MatDiffuse;
extern D3DXCOLOR MatAmbientEmissive;
extern D3DXCOLOR MatSpecular;
extern D3DXVECTOR4 MatPower;

static DWORD FToDW(float f)
{
    return *((DWORD*)&f);
}

const char* CG_VERT_LIGHTING = "..\\common\\shaders\\VertexLighting.cg";
const char* CG_SIMPLE_BLENDING = "..\\common\\shaders\\SimpleBlending.cg";

//----------------------------------------------------------------------------
// _LoadVertexLightingPrograms()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d8Minimal::_LoadVertexLightingPrograms()
{
    IDirect3DDevice8* pDev = D3DUtilContext::Instance()->GetD3DDevice();

	char entry[512] = "main";
    vertLightParms_.vertprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_VERT_LIGHTING, VertexProfile, entry, NULL);
	if (vertLightParms_.vertprog_ == NULL)
	{
		const char* listing = cgGetLastListing(context_);
		if (listing == NULL)
		{
			listing = "No listing available!";
		}
        OutputErrors(listing);
		FAIL_RETURN(E_FAIL);
	}

	// Assemble the shader
    {
    	const char* code = cgGetProgramString(vertLightParms_.vertprog_, CG_COMPILED_PROGRAM);
        OutputErrors(code);
        CComPtr<ID3DXBuffer> constants, compiled, errors;
        HRESULT hres = D3D_OK;
        if (FAILED(hres=D3DXAssembleShader(code, strlen(code), 0, &constants, &compiled, &errors)))
        {
            if (errors != NULL)
            {
                OutputDebugString((LPCSTR)errors->GetBufferPointer());
            	OutputDebugString("\n");
                FAIL_RETURN(hres);
            }
        }

	    // We load position as a Float3, but it is used as a float4 in the program. Let D3D implicitly expand to a float4.
        std::vector<DWORD> decl;
        if (constants != NULL)
        {
            const DWORD* constBegin = static_cast<const DWORD*>(constants->GetBufferPointer());
            const DWORD* constEnd = constBegin + (constants->GetBufferSize() / sizeof(DWORD));
            std::copy(constBegin, constEnd, std::back_inserter(decl));
        }
        decl.push_back(D3DVSD_STREAM(0));
        decl.push_back(D3DVSD_REG(
            cgD3D8ResourceToInputRegister(cgGetParameterResource(cgGetNamedParameter(
            vertLightParms_.vertprog_, "In.Position"))), D3DVSDT_FLOAT3));
        decl.push_back(D3DVSD_REG(
            cgD3D8ResourceToInputRegister(cgGetParameterResource(cgGetNamedParameter(
            vertLightParms_.vertprog_, "In.Normal"  ))), D3DVSDT_FLOAT3));
        decl.push_back(D3DVSD_REG(
            cgD3D8ResourceToInputRegister(cgGetParameterResource(cgGetNamedParameter(
            vertLightParms_.vertprog_, "In.Tex0"    ))), D3DVSDT_FLOAT2));
        decl.push_back(D3DVSD_SKIP(3)); // skip the tangent vector
        decl.push_back(D3DVSD_END());
        
        FAIL_RETURN(cgD3D8ValidateVertexDeclaration(vertLightParms_.vertprog_, &decl.front()) ? D3D_OK : E_FAIL);

        FAIL_RETURN(pDev->CreateVertexShader(&decl.front(), static_cast<const DWORD*>(compiled->GetBufferPointer()), &vertLightParms_.vertprogHandle_, (D3DUtilContext::Instance()->GetCreationParameters().BehaviorFlags & D3DCREATE_SOFTWARE_VERTEXPROCESSING) ? D3DUSAGE_SOFTWAREPROCESSING : 0));
    }

    // bind the vertex shader parameters to iterators
	vertLightParms_.mvp_                  = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj"));
	vertLightParms_.scale_                = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale"));
	vertLightParms_.matDiffuse_           = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]"));
	vertLightParms_.matSpecular_          = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]"));
	vertLightParms_.matAmbientEmissive_   = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]"));
	vertLightParms_.matSpecPower_         = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower"));
	vertLightParms_.lightPosition_        = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition"));
	vertLightParms_.lightRange_           = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange"));
	vertLightParms_.lightAttenuation_     = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation"));
	vertLightParms_.camPositonModelspace_ = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"));
    // Verify the parameter sizes are what we expect (optional)
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj")))              == 16 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale")))          ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower")))              ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition")))        ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange")))           ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation")))     ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D8TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"))) ==  4 ? D3D_OK : E_FAIL);

    // Now load the pixel shader
	sprintf(entry, "main");
    vertLightParms_.pixprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_SIMPLE_BLENDING, PixelProfile, entry, NULL);
	if (vertLightParms_.pixprog_ == NULL)
	{
		const char* listing = cgGetLastListing(context_);
		if (listing == NULL)
		{
			listing = "No listing available!";
		}
        OutputErrors(listing);
		FAIL_RETURN(E_FAIL);
	}

	// Assemble the shader
    {
    	const char* code = cgGetProgramString(vertLightParms_.pixprog_, CG_COMPILED_PROGRAM);
        OutputErrors(code);
        CComPtr<ID3DXBuffer> constants, compiled, errors;
        HRESULT hres = D3D_OK;
        if (FAILED(hres=D3DXAssembleShader(code, strlen(code), 0, &constants, &compiled, &errors)))
        {
            if (errors != NULL)
            {
                OutputDebugString((LPCSTR)errors->GetBufferPointer());
            	OutputDebugString("\n");
                FAIL_RETURN(hres);
            }
        }

        FAIL_RETURN(pDev->CreatePixelShader(static_cast<const DWORD*>(compiled->GetBufferPointer()), &vertLightParms_.pixprogHandle_));
    }

    // bind the pixel shader parameters to iterators
	vertLightParms_.baseTex_ = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.pixprog_, "baseMap"));

    return D3D_OK;
}

//----------------------------------------------------------------------------
// _SetupForVertexLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d8Minimal::_SetupForVertexLighting(const D3DXMATRIX& modelviewproj, D3DXVECTOR4& lightPos, D3DXVECTOR4& lightRange, D3DXVECTOR4& lightAttenuation, D3DXVECTOR4& cameraPosition, D3DXVECTOR4& scale)
{
    IDirect3DDevice8* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    // activate the shaders
    FAIL_RETURN(pDev->SetVertexShader(vertLightParms_.vertprogHandle_));
    FAIL_RETURN(pDev->SetPixelShader(vertLightParms_.pixprogHandle_));

	// bind the constants to the program
    FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.mvp_, &modelviewproj, 4));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.scale_, &scale, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.matDiffuse_, &MatDiffuse, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.matSpecular_, &MatSpecular, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.matAmbientEmissive_, &MatAmbientEmissive, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.matSpecPower_, &MatPower, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.lightPosition_, &lightPos, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.lightRange_, &lightRange, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.lightAttenuation_, &lightAttenuation, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstant(vertLightParms_.camPositonModelspace_, &cameraPosition, 1));

    FAIL_RETURN(pDev->SetTexture(vertLightParms_.baseTex_, meshTexture_));
	
    return D3D_OK;
}
