//============================================================================
// File          : PixelLighting.cpp
// 
// Content       : All the setup used for the pixel lighting Cg programs.
// 
//============================================================================
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "d3d9Expanded.hpp"
#include <vector>
#include <iterator>

extern D3DXCOLOR MatDiffuse;
extern D3DXCOLOR MatAmbientEmissive;
extern D3DXCOLOR MatSpecular;
extern D3DXVECTOR4 MatPower;

static DWORD FToDW(float f)
{
    return *((DWORD*)&f);
}

const char* CG_PIX_LIGHTING_SETUP = "..\\common\\shaders\\PixelLightingSetup.cg";
const char* CG_PIX_LIGHTING = "..\\common\\shaders\\PixelLighting.cg";

//----------------------------------------------------------------------------
// _LoadPixelLightingPrograms()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Expanded::_LoadPixelLightingPrograms()
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    {
        const char **profileOpts = cgD3D9GetOptimalOptions(VertexProfile);

        pixLightParms_.vertprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_PIX_LIGHTING_SETUP, VertexProfile, "main", profileOpts);
	    if (pixLightParms_.vertprog_ == NULL)
	    {
		    const char* listing = cgGetLastListing(context_);
		    if (listing == NULL)
		    {
			    listing = "No listing available!";
		    }
            OutputErrors(listing);
		    FAIL_RETURN(E_FAIL);
	    }
    }


    // Assemble the shader
    FAIL_RETURN(cgD3D9LoadProgram(pixLightParms_.vertprog_, true, 0));

    // Verify the parameter sizes are what we expect (optional)
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "ModelViewProj")))              == 16 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.Scale")))          ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightPosition")))        ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightRange")))           ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightAttenuation")))     ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.CameraPosition"))) ==  4 ? D3D_OK : E_FAIL);

    // Get the parameters that changes every frame.
	pixLightParms_.mvp_                  = cgGetNamedParameter(pixLightParms_.vertprog_, "ModelViewProj");              FAIL_RETURN(pixLightParms_.mvp_ == NULL ? E_FAIL:D3D_OK);
	pixLightParms_.scale_                = cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.Scale");          FAIL_RETURN(pixLightParms_.scale_ == NULL ? E_FAIL:D3D_OK);
	pixLightParms_.lightPosition_        = cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightPosition");        FAIL_RETURN(pixLightParms_.lightPosition_ == NULL ? E_FAIL:D3D_OK);
	pixLightParms_.lightRange_           = cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightRange");           FAIL_RETURN(pixLightParms_.lightRange_ == NULL ? E_FAIL:D3D_OK);
	pixLightParms_.lightAttenuation_     = cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightAttenuation");     FAIL_RETURN(pixLightParms_.lightAttenuation_ == NULL ? E_FAIL:D3D_OK);
	pixLightParms_.camPositonModelspace_ = cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.CameraPosition"); FAIL_RETURN(pixLightParms_.camPositonModelspace_ == NULL ? E_FAIL:D3D_OK);

    {
        const char **profileOpts = cgD3D9GetOptimalOptions(PixelProfile);

        // Now load the pixel shader
        pixLightParms_.pixprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_PIX_LIGHTING, PixelProfile, "main", profileOpts);
	    if (pixLightParms_.pixprog_ == NULL)
	    {
		    const char* listing = cgGetLastListing(context_);
		    if (listing == NULL)
		    {
			    listing = "No listing available!";
		    }
            OutputErrors(listing);
		    FAIL_RETURN(E_FAIL);
	    }
    }

	// Assemble the shader
    FAIL_RETURN(cgD3D9LoadProgram(pixLightParms_.pixprog_, true, 0));

    // Bind the parameters that are constant.
    CGparameter texParams[3];
    texParams[0] = cgGetNamedParameter(pixLightParms_.pixprog_, "baseMap");
    texParams[1] = cgGetNamedParameter(pixLightParms_.pixprog_, "bumpMap");
    texParams[2] = cgGetNamedParameter(pixLightParms_.pixprog_, "specPowMap");
	FAIL_RETURN(cgD3D9SetTexture(texParams[0], meshTexture_));
	FAIL_RETURN(cgD3D9SetTexture(texParams[1], meshBump_));
	FAIL_RETURN(cgD3D9SetTexture(texParams[2], meshSpecPow_));
    FAIL_RETURN(cgD3D9SetUniform(cgGetNamedParameter(pixLightParms_.pixprog_, "Diffuse"), &MatDiffuse));
	FAIL_RETURN(cgD3D9SetUniform(cgGetNamedParameter(pixLightParms_.pixprog_, "Specular"), &MatSpecular));
	FAIL_RETURN(cgD3D9SetUniform(cgGetNamedParameter(pixLightParms_.pixprog_, "AmbientEmissive"), &MatAmbientEmissive));

    for (int i=0;i<3;++i)
    {
        FAIL_RETURN(cgD3D9SetSamplerState(texParams[i], D3DSAMP_MAGFILTER, D3DTEXF_LINEAR));
        FAIL_RETURN(cgD3D9SetSamplerState(texParams[i], D3DSAMP_MINFILTER, D3DTEXF_LINEAR));
        FAIL_RETURN(cgD3D9SetSamplerState(texParams[i], D3DSAMP_MIPFILTER, D3DTEXF_LINEAR));
    }
    FAIL_RETURN(cgD3D9SetSamplerState(texParams[2], D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP));
    FAIL_RETURN(cgD3D9SetSamplerState(texParams[2], D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP));
    FAIL_RETURN(cgD3D9SetSamplerState(texParams[2], D3DSAMP_ADDRESSW, D3DTADDRESS_CLAMP));

    // Bind this constant which is actually unused (just to show that this works)
	FAIL_RETURN(cgD3D9SetUniform(cgGetNamedParameter(pixLightParms_.pixprog_, "SpecPower"), &MatPower));

    // This can be separate in D3D9 since VDecls aren't linked to shaders anymore.
    FAIL_RETURN(_CreateVDeclForPixelLighting());
    
    return D3D_OK;
}

//----------------------------------------------------------------------------
// _SetupForPixelLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Expanded::_SetupForPixelLighting(const D3DXMATRIX& modelviewproj, D3DXVECTOR4& lightPos, D3DXVECTOR4& lightRange, D3DXVECTOR4& lightAttenuation, D3DXVECTOR4& cameraPosition, D3DXVECTOR4& scale)
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    // bind the constants to the program (this must be done before activating the shaders)
    FAIL_RETURN(cgD3D9SetUniformMatrix(pixLightParms_.mvp_, &modelviewproj));

	FAIL_RETURN(cgD3D9SetUniform(pixLightParms_.scale_, &scale));
	FAIL_RETURN(cgD3D9SetUniform(pixLightParms_.lightPosition_, &lightPos));
	FAIL_RETURN(cgD3D9SetUniform(pixLightParms_.lightRange_, &lightRange));
	FAIL_RETURN(cgD3D9SetUniform(pixLightParms_.lightAttenuation_, &lightAttenuation));
	FAIL_RETURN(cgD3D9SetUniform(pixLightParms_.camPositonModelspace_, &cameraPosition));
	
    // activate the shaders
    FAIL_RETURN(cgD3D9BindProgram(pixLightParms_.vertprog_));
    FAIL_RETURN(cgD3D9BindProgram(pixLightParms_.pixprog_));
    FAIL_RETURN(pDev->SetVertexDeclaration(pixLightParms_.vertprogDecl_));

    return D3D_OK;
    
}

//----------------------------------------------------------------------------
// _CreateVDeclForPixelLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Expanded::_CreateVDeclForPixelLighting()
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    // validate the declaration
    CGparameter position = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Position");
    CGparameter normal   = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Normal");
    CGparameter tex0     = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Tex0");
    CGparameter tangent  = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Tangent");
    const D3DVERTEXELEMENT9 decl[] =
    {
        { 0, 0*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
            cgD3D9ResourceToDeclUsage(cgGetParameterResource(position)),
            static_cast<BYTE>(cgGetParameterResourceIndex(position)) },
        { 0, 3*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
            cgD3D9ResourceToDeclUsage(cgGetParameterResource(normal)),
            static_cast<BYTE>(cgGetParameterResourceIndex(normal)) },
        { 0, 6*sizeof(float), D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, 
            cgD3D9ResourceToDeclUsage(cgGetParameterResource(tex0)),
            static_cast<BYTE>(cgGetParameterResourceIndex(tex0)) },
        { 0, 8*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
            cgD3D9ResourceToDeclUsage(cgGetParameterResource(tangent)),
            static_cast<BYTE>(cgGetParameterResourceIndex(tangent)) },
        D3DDECL_END(),
    };
    
    FAIL_RETURN(cgD3D9ValidateVertexDeclaration(pixLightParms_.vertprog_, decl) ? D3D_OK : E_FAIL);
    FAIL_RETURN(pDev->CreateVertexDeclaration(decl, &pixLightParms_.vertprogDecl_));

    return D3D_OK;
}

