//============================================================================
// File          : VertexLighting.cpp
// 
// Content       : All the setup used for the vertex lighting Cg programs.
// 
//============================================================================
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "d3d9Expanded.hpp"
#include <vector>
#include <iterator>

extern D3DXCOLOR MatDiffuse;
extern D3DXCOLOR MatAmbientEmissive;
extern D3DXCOLOR MatSpecular;
extern D3DXVECTOR4 MatPower;

static DWORD FToDW(float f)
{
    return *((DWORD*)&f);
}

const char* CG_VERT_LIGHTING = "..\\common\\shaders\\VertexLighting.cg";
const char* CG_SIMPLE_BLENDING = "..\\common\\shaders\\SimpleBlending.cg";

//----------------------------------------------------------------------------
// _LoadVertexLightingPrograms()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Expanded::_LoadVertexLightingPrograms()
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    {
        const char **profileOpts = cgD3D9GetOptimalOptions(VertexProfile);

        vertLightParms_.vertprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_VERT_LIGHTING, VertexProfile, "main", profileOpts);
	    if (vertLightParms_.vertprog_ == NULL)
	    {
		    const char* listing = cgGetLastListing(context_);
		    if (listing == NULL)
		    {
			    listing = "No listing available!";
		    }
            OutputErrors(listing);
		    FAIL_RETURN(E_FAIL);
	    }
    }

    OutputErrors(cgGetProgramString(vertLightParms_.vertprog_, CG_COMPILED_PROGRAM));

	// Assemble the shader
    FAIL_RETURN(cgD3D9LoadProgram(vertLightParms_.vertprog_, true, 0));

    // Verify the parameter sizes are what we expect (optional)
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj")))              == 16 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale")))          ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower")))              ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition")))        ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange")))           ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation")))     ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"))) ==  4 ? D3D_OK : E_FAIL);

    // Get the parameters that changes every frame.
	vertLightParms_.mvp_                  = cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj");              FAIL_RETURN(vertLightParms_.mvp_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.scale_                = cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale");          FAIL_RETURN(vertLightParms_.scale_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.lightPosition_        = cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition");        FAIL_RETURN(vertLightParms_.lightPosition_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.lightRange_           = cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange");           FAIL_RETURN(vertLightParms_.lightRange_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.lightAttenuation_     = cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation");     FAIL_RETURN(vertLightParms_.lightAttenuation_ == NULL ? E_FAIL:D3D_OK);
	vertLightParms_.camPositonModelspace_ = cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"); FAIL_RETURN(vertLightParms_.camPositonModelspace_ == NULL ? E_FAIL:D3D_OK);

    // Bind the parameters that are constant.
    CGparameter param;
    D3DXCOLOR dummy(0.5f,0.5f,0.5f,1.f);
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D9SetUniform(param, &dummy);
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D9SetUniform(param, &dummy);
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D9SetUniform(param, &dummy);
	param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    cgD3D9SetUniform(param, &MatPower);

    // Try binding an array
    D3DXCOLOR colorArray[3] = {MatDiffuse, MatSpecular, MatAmbientEmissive};
    param = cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors"); FAIL_RETURN(param == NULL ? E_FAIL:D3D_OK);
    FAIL_RETURN(cgD3D9SetUniformArray(param, 0, 3, colorArray));
    
    // Try binding a struct
    struct MatStruct
    {
        D3DXCOLOR colors[3];
        float power;
    };
    MatStruct myMat;
    myMat.colors[0] = MatDiffuse;
    myMat.colors[1] = MatSpecular;
    myMat.colors[2] = MatAmbientEmissive;
    myMat.power = MatPower.x;
    FAIL_RETURN(cgD3D9SetUniform(param, &myMat));


    // Now load the pixel shader
    {
        const char **profileOpts = cgD3D9GetOptimalOptions(PixelProfile);

        vertLightParms_.pixprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_SIMPLE_BLENDING, PixelProfile, "main", profileOpts);
	    if (vertLightParms_.pixprog_ == NULL)
	    {
		    const char* listing = cgGetLastListing(context_);
		    if (listing == NULL)
		    {
			    listing = "No listing available!";
		    }
            OutputErrors(listing);
		    FAIL_RETURN(E_FAIL);
	    }
    }

	// Assemble the shader
    FAIL_RETURN(cgD3D9LoadProgram(vertLightParms_.pixprog_, true, 0));

    // Bind the parameters that are constant.

    CGparameter texParam = cgGetNamedParameter(vertLightParms_.pixprog_, "baseMap");
	FAIL_RETURN(cgD3D9SetTexture(texParam, meshTexture_));

    FAIL_RETURN(cgD3D9SetSamplerState(texParam, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR));
    FAIL_RETURN(cgD3D9SetSamplerState(texParam, D3DSAMP_MINFILTER, D3DTEXF_LINEAR));
    FAIL_RETURN(cgD3D9SetSamplerState(texParam, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR));

    // This can be separate in D3D9 since VDecls aren't linked to shaders anymore.
    FAIL_RETURN(_CreateVDeclForVertexLighting());

    return D3D_OK;
}

//----------------------------------------------------------------------------
// _SetupForVertexLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Expanded::_SetupForVertexLighting(const D3DXMATRIX& modelviewproj, D3DXVECTOR4& lightPos, D3DXVECTOR4& lightRange, D3DXVECTOR4& lightAttenuation, D3DXVECTOR4& cameraPosition, D3DXVECTOR4& scale)
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

	// bind the constants to the program (this must be done before activating the shaders)
    FAIL_RETURN(cgD3D9SetUniformMatrix(vertLightParms_.mvp_, &modelviewproj));

	FAIL_RETURN(cgD3D9SetUniform(vertLightParms_.scale_, &scale));
	FAIL_RETURN(cgD3D9SetUniform(vertLightParms_.lightPosition_, &lightPos));
	FAIL_RETURN(cgD3D9SetUniform(vertLightParms_.lightRange_, &lightRange));
	FAIL_RETURN(cgD3D9SetUniform(vertLightParms_.lightAttenuation_, &lightAttenuation));
	FAIL_RETURN(cgD3D9SetUniform(vertLightParms_.camPositonModelspace_, &cameraPosition));
	
    // activate the shaders
    FAIL_RETURN(cgD3D9BindProgram(vertLightParms_.vertprog_));
    FAIL_RETURN(cgD3D9BindProgram(vertLightParms_.pixprog_));
    FAIL_RETURN(pDev->SetVertexDeclaration(vertLightParms_.vertprogDecl_));

    return D3D_OK;
}

//----------------------------------------------------------------------------
// _CreateVDeclForVertexLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Expanded::_CreateVDeclForVertexLighting()
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    // validate the declaration
    CGparameter position = cgGetNamedParameter(vertLightParms_.vertprog_, "In.Position");
    CGparameter normal   = cgGetNamedParameter(vertLightParms_.vertprog_, "In.Normal");
    CGparameter tex0     = cgGetNamedParameter(vertLightParms_.vertprog_, "In.Tex0");
    const D3DVERTEXELEMENT9 decl[] =
    {
        { 0, 0*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
            cgD3D9ResourceToDeclUsage(cgGetParameterResource(position)),
            static_cast<BYTE>(cgGetParameterResourceIndex(position)) },
        { 0, 3*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
            cgD3D9ResourceToDeclUsage(cgGetParameterResource(normal)),
            static_cast<BYTE>(cgGetParameterResourceIndex(normal)) },
        { 0, 6*sizeof(float), D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, 
            cgD3D9ResourceToDeclUsage(cgGetParameterResource(tex0)),
            static_cast<BYTE>(cgGetParameterResourceIndex(tex0)) },
        D3DDECL_END(),
    };
    FAIL_RETURN(cgD3D9ValidateVertexDeclaration(vertLightParms_.vertprog_, decl) ? D3D_OK : E_FAIL);
    FAIL_RETURN(pDev->CreateVertexDeclaration(decl, &vertLightParms_.vertprogDecl_));
    return D3D_OK;
}

