//============================================================================
// File          : PixelLighting.cpp
// 
// Content       : All the setup used for the pixel lighting Cg programs.
// 
//============================================================================
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "d3d9Minimal.hpp"
#include <vector>
#include <iterator>

extern D3DXCOLOR MatDiffuse;
extern D3DXCOLOR MatAmbientEmissive;
extern D3DXCOLOR MatSpecular;
extern D3DXVECTOR4 MatPower;

static DWORD FToDW(float f)
{
    return *((DWORD*)&f);
}

const char* CG_PIX_LIGHTING_SETUP = "..\\common\\shaders\\PixelLightingSetup.cg";
const char* CG_PIX_LIGHTING = "..\\common\\shaders\\PixelLighting.cg";

//----------------------------------------------------------------------------
// _LoadPixelLightingPrograms()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Minimal::_LoadPixelLightingPrograms()
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

	char entry[512] = "main";
    const char* profileOpts[] = 
    {
        "-profileopts", "dcls", NULL,
    };

    pixLightParms_.vertprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_PIX_LIGHTING_SETUP, VertexProfile, entry, profileOpts);
	if (pixLightParms_.vertprog_ == NULL)
	{
		const char* listing = cgGetLastListing(context_);
		if (listing == NULL)
		{
			listing = "No listing available!";
		}
        OutputErrors(listing);
		FAIL_RETURN(E_FAIL);
	}

	// Assemble the shader
    {
    	const char* code = cgGetProgramString(pixLightParms_.vertprog_, CG_COMPILED_PROGRAM);
        OutputErrors(code);
        CComPtr<ID3DXBuffer> compiled, errors;
        HRESULT hres = D3D_OK;
        if (FAILED(hres=D3DXAssembleShader(code, strlen(code), NULL, NULL, 0, &compiled, &errors)))
        {
            if (errors != NULL)
            {
                OutputDebugString((LPCSTR)errors->GetBufferPointer());
            	OutputDebugString("\n");
                FAIL_RETURN(hres);
            }
        }

	    // We load position as a Float3, but it is used as a float4 in the program. Let D3D implicitly expand to a float4.
        CGparameter position = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Position");
        CGparameter normal   = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Normal");
        CGparameter tex0     = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Tex0");
        CGparameter tangent  = cgGetNamedParameter(pixLightParms_.vertprog_, "In.Tangent");
        const D3DVERTEXELEMENT9 decl[] =
        {
            { 0, 0*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
                cgD3D9ResourceToDeclUsage(cgGetParameterResource(position)),
                static_cast<BYTE>(cgGetParameterResourceIndex(position)) },
            { 0, 3*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
                cgD3D9ResourceToDeclUsage(cgGetParameterResource(normal)),
                static_cast<BYTE>(cgGetParameterResourceIndex(normal)) },
            { 0, 6*sizeof(float), D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, 
                cgD3D9ResourceToDeclUsage(cgGetParameterResource(tex0)),
                static_cast<BYTE>(cgGetParameterResourceIndex(tex0)) },
            { 0, 8*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
                cgD3D9ResourceToDeclUsage(cgGetParameterResource(tangent)),
                static_cast<BYTE>(cgGetParameterResourceIndex(tangent)) },
            D3DDECL_END(),
        };
        
        FAIL_RETURN(cgD3D9ValidateVertexDeclaration(pixLightParms_.vertprog_, decl) ? D3D_OK : E_FAIL);
        FAIL_RETURN(pDev->CreateVertexDeclaration(decl, &pixLightParms_.vertprogDecl_));
        FAIL_RETURN(pDev->CreateVertexShader(static_cast<const DWORD*>(compiled->GetBufferPointer()), 
            &pixLightParms_.vertprogHandle_));
    }

    // bind the vertex shader parameters to iterators
	pixLightParms_.mvp_                  = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.vertprog_, "ModelViewProj"));
	pixLightParms_.scale_                = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.Scale"));
	pixLightParms_.lightPosition_        = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightPosition"));
	pixLightParms_.lightRange_           = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightRange"));
	pixLightParms_.lightAttenuation_     = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightAttenuation"));
	pixLightParms_.camPositonModelspace_ = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.CameraPosition"));
    
    // Verify the parameter sizes are what we expect (optional)
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "ModelViewProj")))               == 16 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.Scale")))           ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightPosition")))         ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightRange")))            ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "light.LightAttenuation")))      ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(pixLightParms_.vertprog_, "modelParams.CameraPosition")))  ==  4 ? D3D_OK : E_FAIL);

    // Now load the pixel shader
	sprintf(entry, "main");
    pixLightParms_.pixprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_PIX_LIGHTING, PixelProfile, entry, NULL);
	if (pixLightParms_.pixprog_ == NULL)
	{
		const char* listing = cgGetLastListing(context_);
		if (listing == NULL)
		{
			listing = "No listing available!";
		}
        OutputErrors(listing);
		FAIL_RETURN(E_FAIL);
	}

	// Assemble the shader
    {
    	const char* code = cgGetProgramString(pixLightParms_.pixprog_, CG_COMPILED_PROGRAM);
        OutputErrors(code);
        CComPtr<ID3DXBuffer> compiled, errors;
        HRESULT hres = D3D_OK;
        if (FAILED(hres=D3DXAssembleShader(code, strlen(code), NULL, NULL, 0, &compiled, &errors)))
        {
            if (errors != NULL)
            {
                OutputDebugString((LPCSTR)errors->GetBufferPointer());
            	OutputDebugString("\n");
                FAIL_RETURN(hres);
            }
        }

        FAIL_RETURN(pDev->CreatePixelShader(static_cast<const DWORD*>(compiled->GetBufferPointer()), &pixLightParms_.pixprogHandle_));
    }

    // bind the pixel shader parameters to iterators
	pixLightParms_.baseTex_              = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.pixprog_, "baseMap"));
	pixLightParms_.bumpTex_              = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.pixprog_, "bumpMap"));
	pixLightParms_.specPowTex_           = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.pixprog_, "specPowMap"));
	pixLightParms_.matDiffuse_           = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.pixprog_, "Diffuse"));
	pixLightParms_.matSpecular_          = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.pixprog_, "Specular"));
	pixLightParms_.matAmbientEmissive_   = cgGetParameterResourceIndex(cgGetNamedParameter(pixLightParms_.pixprog_, "AmbientEmissive"));
    
    return D3D_OK;
}

//----------------------------------------------------------------------------
// _SetupForPixelLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Minimal::_SetupForPixelLighting(const D3DXMATRIX& modelviewproj, D3DXVECTOR4& lightPos, D3DXVECTOR4& lightRange, D3DXVECTOR4& lightAttenuation, D3DXVECTOR4& cameraPosition, D3DXVECTOR4& scale)
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    // activate the shaders
    FAIL_RETURN(pDev->SetVertexShader(pixLightParms_.vertprogHandle_));
    FAIL_RETURN(pDev->SetPixelShader(pixLightParms_.pixprogHandle_));
    FAIL_RETURN(pDev->SetVertexDeclaration(pixLightParms_.vertprogDecl_));

    // bind the constants to the program
    FAIL_RETURN(pDev->SetVertexShaderConstantF(pixLightParms_.mvp_, &modelviewproj.m[0][0], 4));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(pixLightParms_.scale_, &scale.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(pixLightParms_.lightPosition_, &lightPos.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(pixLightParms_.lightRange_, &lightRange.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(pixLightParms_.lightAttenuation_, &lightAttenuation.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(pixLightParms_.camPositonModelspace_, &cameraPosition.x, 1));

	FAIL_RETURN(pDev->SetPixelShaderConstantF(pixLightParms_.matDiffuse_, &MatDiffuse.r, 1));
	FAIL_RETURN(pDev->SetPixelShaderConstantF(pixLightParms_.matSpecular_, &MatSpecular.r, 1));
	FAIL_RETURN(pDev->SetPixelShaderConstantF(pixLightParms_.matAmbientEmissive_, &MatAmbientEmissive.r, 1));

    FAIL_RETURN(pDev->SetTexture(pixLightParms_.baseTex_, meshTexture_));
    FAIL_RETURN(pDev->SetTexture(pixLightParms_.bumpTex_, meshBump_));
    FAIL_RETURN(pDev->SetTexture(pixLightParms_.specPowTex_, meshSpecPow_));
    FAIL_RETURN(pDev->SetSamplerState(pixLightParms_.specPowTex_, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP));
	FAIL_RETURN(pDev->SetSamplerState(pixLightParms_.specPowTex_, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP));
	FAIL_RETURN(pDev->SetSamplerState(pixLightParms_.specPowTex_, D3DSAMP_ADDRESSW, D3DTADDRESS_CLAMP));
	
    return D3D_OK;
    
}
