//============================================================================
// File          : VertexLighting.cpp
// 
// Content       : All the setup used for the vertex lighting Cg programs.
// 
//============================================================================
// 
//============================================================================

//----------------------------------------------------------------------------
//-- Includes ----------------------------------------------------------------
//----------------------------------------------------------------------------

#include "d3d9Minimal.hpp"
#include <vector>
#include <iterator>

extern D3DXCOLOR MatDiffuse;
extern D3DXCOLOR MatAmbientEmissive;
extern D3DXCOLOR MatSpecular;
extern D3DXVECTOR4 MatPower;

static DWORD FToDW(float f)
{
    return *((DWORD*)&f);
}

const char* CG_VERT_LIGHTING = "..\\common\\shaders\\VertexLighting.cg";
const char* CG_SIMPLE_BLENDING = "..\\common\\shaders\\SimpleBlending.cg";

//----------------------------------------------------------------------------
// _LoadVertexLightingPrograms()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Minimal::_LoadVertexLightingPrograms()
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

	char entry[512] = "main";
    const char* profileOpts[] = 
    {
        "-profileopts", "dcls", NULL,
    };

    vertLightParms_.vertprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_VERT_LIGHTING, VertexProfile, entry, profileOpts);
	if (vertLightParms_.vertprog_ == NULL)
	{
		const char* listing = cgGetLastListing(context_);
		if (listing == NULL)
		{
			listing = "No listing available!";
		}
        OutputErrors(listing);
		FAIL_RETURN(E_FAIL);
	}

	// Assemble the shader
    {
    	const char* code = cgGetProgramString(vertLightParms_.vertprog_, CG_COMPILED_PROGRAM);
        OutputErrors(code);
        CComPtr<ID3DXBuffer> compiled, errors;
        HRESULT hres = D3D_OK;
        if (FAILED(hres=D3DXAssembleShader(code, strlen(code), NULL, NULL, 0, &compiled, &errors)))
        {
            if (errors != NULL)
            {
                OutputDebugString((LPCSTR)errors->GetBufferPointer());
            	OutputDebugString("\n");
                FAIL_RETURN(hres);
            }
        }

        CGparameter position = cgGetNamedParameter(vertLightParms_.vertprog_, "In.Position");
        CGparameter normal   = cgGetNamedParameter(vertLightParms_.vertprog_, "In.Normal");
        CGparameter tex0     = cgGetNamedParameter(vertLightParms_.vertprog_, "In.Tex0");
        const D3DVERTEXELEMENT9 decl[] =
        {
            { 0, 0*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
                cgD3D9ResourceToDeclUsage(cgGetParameterResource(position)),
                static_cast<BYTE>(cgGetParameterResourceIndex(position)) },
            { 0, 3*sizeof(float), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, 
                cgD3D9ResourceToDeclUsage(cgGetParameterResource(normal)),
                static_cast<BYTE>(cgGetParameterResourceIndex(normal)) },
            { 0, 6*sizeof(float), D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, 
                cgD3D9ResourceToDeclUsage(cgGetParameterResource(tex0)),
                static_cast<BYTE>(cgGetParameterResourceIndex(tex0)) },
            D3DDECL_END(),
        };
        FAIL_RETURN(cgD3D9ValidateVertexDeclaration(vertLightParms_.vertprog_, decl) ? D3D_OK : E_FAIL);
        FAIL_RETURN(pDev->CreateVertexDeclaration(decl, &vertLightParms_.vertprogDecl_));
        FAIL_RETURN(pDev->CreateVertexShader(static_cast<const DWORD*>(compiled->GetBufferPointer()), &vertLightParms_.vertprogHandle_));
    }

    // bind the vertex shader parameters to iterators
	vertLightParms_.mvp_                  = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj"));
	vertLightParms_.scale_                = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale"));
	vertLightParms_.matDiffuse_           = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]"));
	vertLightParms_.matSpecular_          = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]"));
	vertLightParms_.matAmbientEmissive_   = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]"));
	vertLightParms_.matSpecPower_         = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower"));
	vertLightParms_.lightPosition_        = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition"));
	vertLightParms_.lightRange_           = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange"));
	vertLightParms_.lightAttenuation_     = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation"));
	vertLightParms_.camPositonModelspace_ = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"));
    // Verify the parameter sizes are what we expect (optional)
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "ModelViewProj")))              == 16 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.Scale")))          ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[0]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[1]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.Colors[2]")))              ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "mat.SpecPower")))              ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightPosition")))        ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightRange")))           ==  1 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "light.LightAttenuation")))     ==  4 ? D3D_OK : E_FAIL);
    FAIL_RETURN(cgD3D9TypeToSize(cgGetParameterType(cgGetNamedParameter(vertLightParms_.vertprog_, "modelParams.CameraPosition"))) ==  4 ? D3D_OK : E_FAIL);

    // Now load the pixel shader
	sprintf(entry, "main");
    vertLightParms_.pixprog_ = cgCreateProgramFromFile(context_, CG_SOURCE, CG_SIMPLE_BLENDING, PixelProfile, entry, NULL);
	if (vertLightParms_.pixprog_ == NULL)
	{
		const char* listing = cgGetLastListing(context_);
		if (listing == NULL)
		{
			listing = "No listing available!";
		}
        OutputErrors(listing);
		FAIL_RETURN(E_FAIL);
	}

	// Assemble the shader
    {
    	const char* code = cgGetProgramString(vertLightParms_.pixprog_, CG_COMPILED_PROGRAM);
        OutputErrors(code);
        CComPtr<ID3DXBuffer> compiled, errors;
        HRESULT hres = D3D_OK;
        if (FAILED(hres=D3DXAssembleShader(code, strlen(code), NULL, NULL, 0, &compiled, &errors)))
        {
            if (errors != NULL)
            {
                OutputDebugString((LPCSTR)errors->GetBufferPointer());
            	OutputDebugString("\n");
                FAIL_RETURN(hres);
            }
        }

        FAIL_RETURN(pDev->CreatePixelShader(static_cast<const DWORD*>(compiled->GetBufferPointer()), &vertLightParms_.pixprogHandle_));
    }

    // bind the pixel shader parameters to iterators
	vertLightParms_.baseTex_ = cgGetParameterResourceIndex(cgGetNamedParameter(vertLightParms_.pixprog_, "baseMap"));

    return D3D_OK;
}

//----------------------------------------------------------------------------
// _SetupForVertexLighting()
// ---------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------
HRESULT
d3d9Minimal::_SetupForVertexLighting(const D3DXMATRIX& modelviewproj, D3DXVECTOR4& lightPos, D3DXVECTOR4& lightRange, D3DXVECTOR4& lightAttenuation, D3DXVECTOR4& cameraPosition, D3DXVECTOR4& scale)
{
    IDirect3DDevice9* pDev = D3DUtilContext::Instance()->GetD3DDevice();

    // activate the shaders
    FAIL_RETURN(pDev->SetVertexShader(vertLightParms_.vertprogHandle_));
    FAIL_RETURN(pDev->SetPixelShader(vertLightParms_.pixprogHandle_));
    FAIL_RETURN(pDev->SetVertexDeclaration(vertLightParms_.vertprogDecl_));

	// bind the constants to the program
    FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.mvp_, &modelviewproj.m[0][0], 4));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.scale_, &scale.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.matDiffuse_, &MatDiffuse.r, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.matSpecular_, &MatSpecular.r, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.matAmbientEmissive_, &MatAmbientEmissive.r, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.matSpecPower_, &MatPower.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.lightPosition_, &lightPos.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.lightRange_, &lightRange.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.lightAttenuation_, &lightAttenuation.x, 1));
	FAIL_RETURN(pDev->SetVertexShaderConstantF(vertLightParms_.camPositonModelspace_, &cameraPosition.x, 1));

    FAIL_RETURN(pDev->SetTexture(vertLightParms_.baseTex_, meshTexture_));
	
    return D3D_OK;
}
