#include "BlackbodyRadiationTable.h"

#include <cmath>

#include "math/Definitions.h"

namespace GpuThesis {

const double BlackbodyRadiationTable::H = 6.626e-34;
const double BlackbodyRadiationTable::C = 3.0e8;
const double BlackbodyRadiationTable::K = 1.38e-23;

const double BlackbodyRadiationTable::RED_WAVE = 0.7e-6;
const double BlackbodyRadiationTable::GREEN_WAVE = 0.56e-6;
const double BlackbodyRadiationTable::BLUE_WAVE = 0.47e-6;

BlackbodyRadiationTable::BlackbodyRadiationTable(unsigned int dimension, 
                                                 double minTemp, double maxTemp, double exposureLevel) :

	dimension(dimension), minTemp(minTemp), maxTemp(maxTemp), 
	exposureLevel(exposureLevel), colorPixels(4*dimension)
{
	initializeColors();
}

const std::vector<float> &BlackbodyRadiationTable::getColorPixels() const
{
	return colorPixels;
}


double BlackbodyRadiationTable::lightPower(double waveLength, double temperature)
{
	double l = waveLength, t = temperature;

	return 2 * Math::pi * H *std::pow(C, 2) / (std::pow(l, 5) * (std::exp((H * C) / (K * l * t)) - 1.0));
}


double BlackbodyRadiationTable::expose(double light, double exposure)
{
	return (1.0 - std::exp(-light * exposure)) * 255.0f;
}


void BlackbodyRadiationTable::initializeColors()
{
	double temp = minTemp;
	double delta = (maxTemp - minTemp) / static_cast<float>(dimension);

	float minAlpha = 0.1f;

	float c1 = 2.0f;
	float c2 = static_cast<float>(std::log(minAlpha / c1) / (minTemp - maxTemp));

	for (unsigned int i = 0; i < colorPixels.size(); i+=4, temp += delta) {

		unsigned int red = static_cast<unsigned int>(expose(lightPower(RED_WAVE, temp), exposureLevel));
		unsigned int green = static_cast<unsigned int>(expose(lightPower(GREEN_WAVE, temp), exposureLevel));
		unsigned int blue = static_cast<unsigned int>(expose(lightPower(BLUE_WAVE, temp), exposureLevel));

		colorPixels[i] = static_cast<float>(red) / 255.0f;
		colorPixels[i + 1] = static_cast<float>(green) / 255.0f;
		colorPixels[i + 2] = static_cast<float>(blue) / 255.0f;
		colorPixels[i + 3] = c1 *  static_cast<float>(std::exp((minTemp - temp) * c2));

		if (colorPixels[i + 3] > 1.0f)
			colorPixels[i + 3] = 1.0f;
	}
}

}