#ifndef GPUTHESIS_FIRE_PALETTE_H
#define GPUTHESIS_FIRE_PALETTE_H

#include <vector>

/**
 * GpuThesis namespace, contains classes directly tied to the thesis.
 */
namespace GpuThesis {

/**
 * Generates a black-body radiation table using Planck's formula.
 */
class BlackbodyRadiationTable {

public:
	/**
	 * Construct a black-body radiation table.
	 * @param dimension the discrete dimension of the black-body radiation table.
	 * @param minTemp the minimum temperature of the black-body.
	 * @param maxTemp the maximum temperature of the black-body.
	 * @param exposureLevel the exposure used to map the intensities onto a limited dynamic range.
	 */
	BlackbodyRadiationTable(unsigned int dimension, double minTemp, double maxTemp, double exposureLevel);

	/**
	 * Returns the color values of the resulting black-body radiation table. 
	 * The pixels are stored in the following way: (red, green, blue) * dimension, 
	 * from low to high temperatures. The color values are in the range 0-1.
	 */
	const std::vector<float>& getColorPixels() const;

private:	
	double lightPower(double waveLength, double temperature);
	double expose(double light, double exposure);
	void initializeColors();
	
	static const double H;// = 6.626e-34;
	static const double C;// = 3.0e8;
	static const double K;// = 1.38e-23;

	static const double RED_WAVE;// = 0.7e-6;
	static const double GREEN_WAVE;// = 0.56e-6;
	static const double BLUE_WAVE;// = 0.47e-6;

	unsigned int dimension;
	double exposureLevel;// = 16;
	double minTemp;
	double maxTemp;
	
	std::vector<float> colorPixels;
};

};

#endif
