#include "Engine.h"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iostream>
#include <iterator>
#include <memory>
#include <numeric>
#include <sstream>
#include <vector>

// gpu-thesis
#include "BlackbodyRadiationTable.h"
#include "EngineParameters.h"
#include "EventHandler.h"
#include "Fire3dSimulator.h"
#include "VolumetricExtrusion.h"
#include "Fluid3dSimulator.h"
#include "FluidFactory.h"
#include "Light.h"
#include "ModelObject.h"
#include "ParticleSystem.h"
#include "RayMarcher.h"
#include "Runner.h"
#include "Scene.h"
#include "SimpleVolumeRenderer.h"
#include "Terrain.h"
#include "VolumeRenderer.h"

#include "test-suite/Fluid3dTests.h"
#include "test-suite/FluidSliceTests.h"
#include "test-suite/TestSuite.h"


// library
//#include "console/Action.h"
#include "console/Console.h"
#include "console/Variable.h"

#include "geometry/ModelLoader.h"

#include "image/Image.h"

#include "gpgpu/framebufferObject.h"

#include "math/Vector3d.h"
#include "math/Matrix4x4.h"

#include "sdl/Events.h"
#include "sdl/OpenGLContext.h"
#include "sdl/Timer.h"

#include "simple-xml/XmlDocument.h"
#include "simple-xml/XmlNode.h"
#include "simple-xml/XmlNodeWrapper.h"

#include "opengl/Camera.h"
#include "opengl/CgContext.h"
#include "opengl/CgFragmentProgram.h"
#include "opengl/CgOperation.h"
#include "opengl/CgVertexProgram.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/FpsCounter.h"
#include "opengl/OpenGL.h"
#include "opengl/OrthonormalProjection.h"
#include "opengl/PrimaryTextureContainer.h"
#include "opengl/SecondaryTextureContainer.h"
#include "opengl/Texture2dOperation.h"
#include "opengl/TextureContainer.h"
#include "opengl/TextureContainerActions.h"

#include "util/Util.h"
#include "util/VoxelGrid.h"


using namespace GpuThesis;
using namespace Math;
using namespace OpenGL;
using namespace SimpleXml;
using namespace Util;

namespace GpuThesis {

Engine::Engine(XmlDocument &xmlDoc) :
	xmlDoc(xmlDoc),	
	deltaDist(1.0f * 24.0f),
	deltaAngle(24.0f * Math::pi / 60.0f),
	paused(false),
	showFps(false),
	debug(false),
	textColor(1, 1, 1),
	separatorColor(1, 0, 0),
	gridColor(1, 1, 1),
	smoothDisplay(false),
	fireLocation(Math::Vector3d(0, -10, 0)),
	fireScale(100),
	hasAccumulatedDistance(false),
	lightPosition(Math::Vector3d(0, 40, 0)),
	fixedTimestep(false),
	timestep(0.02f)
{
	initialize();

	updateFireLocation();
}

Engine::~Engine()
{
	std::for_each(ownedModels.begin(), ownedModels.end(), Util::destroy<Geometry::Model>);
	std::for_each(ownedModelObjects.begin(), ownedModelObjects.end(), Util::destroy<ModelObject>);
	std::for_each(ownedTerrains.begin(), ownedTerrains.end(), Util::destroy<Terrain>);
	std::for_each(ownedLights.begin(), ownedLights.end(), Util::destroy<Light>);
}

void Engine::handleInput(float timestep)
{
	if (eventHandler->consoleActive())
		return;

	float fireDelta = 10.0f;

	if (eventHandler->isPressed('w'))
		camera->move(deltaDist * timestep);
	if(eventHandler->isPressed('s'))
		camera->move(-deltaDist * timestep);
	if (eventHandler->isPressed('d'))
		camera->strafe(deltaDist * timestep);
	if (eventHandler->isPressed('a'))
		camera->strafe(-deltaDist * timestep);
	if (eventHandler->isPressed(' '))
		camera->up(deltaDist * timestep);
	if (eventHandler->isPressed('r'))
		camera->up(-deltaDist * timestep);
	if (eventHandler->isPressed('i'))
		camera->pitch(deltaAngle * timestep);
	if (eventHandler->isPressed('k'))
		camera->pitch(-deltaAngle * timestep);
	if (eventHandler->isPressed('j'))
		camera->yaw(-deltaAngle * timestep);
	if (eventHandler->isPressed('l'))
		camera->yaw(deltaAngle * timestep);
	if (eventHandler->isPressed(SDL::KEY_LEFT))
		moveFireSimulation(Vector3d(-timestep * fireDelta, 0, 0));
	if (eventHandler->isPressed(SDL::KEY_RIGHT))
		moveFireSimulation(Vector3d(timestep * fireDelta, 0, 0));
	if (eventHandler->isPressed(SDL::KEY_UP))
		moveFireSimulation(Vector3d(0, 0, -timestep * fireDelta));
	if (eventHandler->isPressed(SDL::KEY_DOWN))
		moveFireSimulation(Vector3d(0, 0, timestep * fireDelta));
}

void Engine::moveFireSimulation(const Math::Vector3d &distance)
{
	fireLocation += distance;

	Math::Vector3d particleDistance = distance * (1.0f / fireScale);

	if (particleSystem.get())
		particleSystem->advect(-particleDistance);

	if (lightSystem.get())
		lightSystem->advect(-particleDistance);

	if (smokeSystem.get())
		smokeSystem->advect(-particleDistance);

	float computationWidth = fireSimulator->getFlat3dTextureSet()->getComputationWidth();

	accumulatedDistance -= distance * (computationWidth / fireScale);

	hasAccumulatedDistance = true;

	updateFireLocation();
}

void Engine::updateFireLocation()
{
	for (unsigned int i = 0; i < ownedModelObjects.size(); ++i) {
		ModelObject *modelObject = ownedModelObjects[i];

		modelObject->setAnchor(fireLocation);
	}
}

void Engine::updateParticleLightPositions()
{
	assert(lightSystem.get());

	std::vector<float> positions(3 * lightSystem->getNumParticles());

	lightSystem->readParticlePositions(&*positions.begin());

	for (unsigned int i = 0; i < lightSystem->getNumParticles(); ++i) {
		Math::Vector3d pos(positions[3*i], positions[3*i+1], positions[3*i+2]);
		ownedLights[i]->setPosition(pos);
	}
}

void Engine::initialize()
{
	initializeConsole();

	std::srand(static_cast<unsigned int>(std::time(0)));

	parameters = autoPtr(new EngineParameters(xmlDoc, *console.get()));

	initializeOpenGL();
	initializeEventHandler();
	initializeRunner();

	if (parameters->getFireSimulationParams().wasRead) {
		initializeFireSimulator();
		if (parameters->getParticleSystemParams().wasRead)
			initializeParticleSystem();

		if (parameters->getSmokeSystemParams().wasRead)
			initializeSmokeSystem();

		if (parameters->getLightSystemParams().wasRead)
			initializeLightSystem();

		if (parameters->getVolumeRendererParams().wasRead)
			initializeVolumeRenderer();
	}

	if (parameters->getFluidSimulationParams().wasRead) 
		initializeFluidSimulator();

	EngineParameters::SetupParams sp = parameters->getSetupParams();

	if (sp.runTests)
		initializeTestSuite();

	bool testGood = true;
	if (sp.runTests) {
		console->output("Running tests...");
		std::stringstream testResult;
		testGood = testSuite->runTests(testResult);
		console->output(testResult.str());
	}

	initializeScene();

	if (!textureContainers.empty())
		initializeContainerCycler();	

	console->loadConfiguration(sp.configFile);	

	camera = autoPtr(new Camera(
		Math::Vector3d(sp.cameraOrigin.x, sp.cameraOrigin.y, sp.cameraOrigin.z)));

	camera->pitch(sp.cameraPitch * Math::pi / 180.0f);
	camera->yaw(sp.cameraYaw * Math::pi / 180.0f);

	const CgContext &context = CgContext::getInstance();
	cgFragmentLightingVS = autoPtr(new CgProgram(context.getCgProgramPath() + "lighting.cg", 
	                               context.getVertexShader(), "fragmentLightingVS", 0));

	cgFragmentLightingFS = autoPtr(new CgProgram(context.getCgProgramPath() + "lighting.cg", 
	                               context.getFragmentShader(), "fragmentLightingFS", 0));

	cgFragmentLightingTexturedFS = autoPtr(new CgProgram(context.getCgProgramPath() + "lighting.cg", 
	                                       context.getFragmentShader(), "fragmentLightingTexturedFS", 0));
}

void Engine::resetParticles(const std::string &particleSources, unsigned int spawnBaseHeight, 
                            unsigned int spawnHeight, unsigned int voxelGridHeight,
							ParticleSystem &particleSystem)
{
	Util::VoxelGrid<float> binaryImage = createBinaryVoxelGrid(particleSources);
	unsigned int voxelGridWidth = binaryImage.getWidth();
	unsigned int voxelGridDepth = binaryImage.getHeight();

	std::vector<Math::Vector3d> spawnPositions;

	for (unsigned int z = 0; z < voxelGridDepth; ++z) {
		for (unsigned int x = 0; x < voxelGridWidth; ++x) {
			if (binaryImage(x, z, 0) == 1.0) {

				float localX = (static_cast<float>(x) - static_cast<float>(voxelGridWidth) / 2.0f) / 
					static_cast<float>(voxelGridWidth);
				float localZ = (static_cast<float>(z) - static_cast<float>(voxelGridDepth) / 2.0f) / 
					static_cast<float>(voxelGridDepth);
				
				for (unsigned int y = spawnBaseHeight; y < spawnBaseHeight + spawnHeight; ++y) {
					float localY = static_cast<float>(y) / static_cast<float>(voxelGridHeight);
					spawnPositions.push_back(Math::Vector3d(localX, localY, localZ));
				}
			}
		}
	}	

	std::vector<float> positions(3 * particleSystem.getNumParticles());
	std::vector<float> velocities(3 * particleSystem.getNumParticles());
	std::vector<float> spawnTimes(particleSystem.getNumParticles());

	float voxelWidth = 1.0f / static_cast<float>(voxelGridWidth);
	float voxelHeight = 1.0f / static_cast<float>(voxelGridHeight);
	float voxelDepth = 1.0f / static_cast<float>(voxelGridDepth);

	for (unsigned int i = 0; i < positions.size(); i += 3) {
		unsigned int index = spawnPositions.size() * Util::randf();

		positions[i + 0] = spawnPositions[index].x + voxelWidth * Util::randf();
		positions[i + 1] = spawnPositions[index].y + voxelHeight * Util::randf();
		positions[i + 2] = spawnPositions[index].z + voxelDepth * Util::randf();
	}	

	/*for (unsigned int i = 0; i < spawnTimes.size(); ++i) {
		spawnTimes[i] = 1.0f + 1.0f * randf();		
	}*/

	particleSystem.setSpawnPositions(positions);
	particleSystem.setSpawnVelocities(velocities);
	particleSystem.setSpawnTimes(spawnTimes);
}


Util::VoxelGrid<float> Engine::createBinaryVoxelGrid(const std::string &imageFilename)
{
	Image::Image image(imageFilename);
	unsigned int width = image.getWidth();
	unsigned int height = image.getHeight();

	Util::VoxelGrid<float> result(width, height, 1);

	unsigned char *imageData = image.getData();

	for (int y = static_cast<int>(height) - 1; y >= 0; --y) {
		for (unsigned int x = 0; x < width; ++x) {
			if (imageData[0] || imageData[1] || imageData[2])
				result(x, y, 0) = 1.0f;

			imageData += 3;
		}
	}

	return result;
}

Util::VoxelGrid<Color> Engine::createVoxelGrid(const std::string &imageFilename)
{
	Image::Image image(imageFilename);
	unsigned int width = image.getWidth();
	unsigned int height = image.getHeight();

	Util::VoxelGrid<Color> result(width, height, 1);

	unsigned char *imageData = image.getData();

	for (int y = static_cast<int>(height) - 1; y >= 0; --y) {
		for (unsigned int x = 0; x < width; ++x) {
			float red = static_cast<float>(*imageData++) / 255.0f;
			float green = static_cast<float>(*imageData++) / 255.0f;
			float blue = static_cast<float>(*imageData++) / 255.0f;

			result(x, y, 0) = Color(red, green, blue, 0);
		}
	}

	return result;
}

Util::VoxelGrid<Color> Engine::createSliceField(const std::string &filename,
												float scale,												
                                                unsigned int sliceWidth,
                                                unsigned int sliceHeight)
{
	Util::VoxelGrid<Color> image = createVoxelGrid(filename);
	unsigned int imageWidth = image.getWidth();
	unsigned int imageHeight = image.getHeight();

	Util::VoxelGrid<Color> result(sliceWidth, sliceHeight, 1);

	float ratioX = static_cast<float>(imageWidth) / static_cast<float>(sliceWidth);
	float ratioY = static_cast<float>(imageHeight) / static_cast<float>(sliceHeight);

	for (unsigned int y = 0; y < sliceHeight; ++y) {
		for (unsigned int x = 0; x < sliceWidth; ++x) {
			Color c = image(static_cast<unsigned int>(x * ratioX),
			                static_cast<unsigned int>(y * ratioY), 0);

			result(x, y, 0) = Color(c.getRed() * scale,
			                        c.getGreen() * scale,
			                        c.getBlue() * scale,
			                        c.getAlpha() * scale);
		}
	}

	return result;
}

Util::VoxelGrid<Color> Engine::createVelocitySliceField(const std::string &filename,
                                                        Math::Vector3d direction,												
                                                        unsigned int sliceWidth,
                                                        unsigned int sliceHeight)
{
	Util::VoxelGrid<float> image = createBinaryVoxelGrid(filename);
	unsigned int imageWidth = image.getWidth();
	unsigned int imageHeight = image.getHeight();

	Util::VoxelGrid<Color> result(sliceWidth, sliceHeight, 1);

	float ratioX = static_cast<float>(imageWidth) / static_cast<float>(sliceWidth);
	float ratioY = static_cast<float>(imageHeight) / static_cast<float>(sliceHeight);

	for (unsigned int y = 0; y < sliceHeight; ++y) {
		for (unsigned int x = 0; x < sliceWidth; ++x) {
			float v = image(static_cast<unsigned int>(x * ratioX),
			                static_cast<unsigned int>(y * ratioY), 0);
			
			result(x, y, 0) = Color(direction.x * v, direction.y * v, direction.z * v, 0);
		}
	}

	return result;
}

Util::VoxelGrid<float> Engine::createRandomSliceField(const std::string &filename,
                                                      float minFieldValue,
                                                      float maxFieldValue,
                                                      unsigned int sliceWidth,
                                                      unsigned int sliceHeight,
                                                      unsigned int numSlices)
{
	Util::VoxelGrid<float> binaryImage(createBinaryVoxelGrid(filename));
	unsigned int imageWidth = binaryImage.getWidth();
	unsigned int imageHeight = binaryImage.getHeight();

	Util::VoxelGrid<float> result(sliceWidth, sliceHeight, numSlices);

	float ratioX = static_cast<float>(imageWidth) / static_cast<float>(sliceWidth);
	float ratioY = static_cast<float>(imageHeight) / static_cast<float>(sliceHeight);

	for (unsigned int z = 0; z < numSlices; ++z) {
		for (unsigned int y = 0; y < sliceHeight; ++y) {
			for (unsigned int x = 0; x < sliceWidth; ++x) {
				result(x, y, z) =
					binaryImage(static_cast<unsigned int>(x * ratioX),
				                static_cast<unsigned int>(y * ratioY), 0) *
				                (minFieldValue + (maxFieldValue - minFieldValue) * 
					                             Util::randf());
			}
		}
	}

	return result;
}

Util::VoxelGrid<float> Engine::createRandom3dField(const std::string &filename,
                                                   float minFieldValue,
                                                   float maxFieldValue,
	                                               unsigned int fieldBaseHeight,
                                                   unsigned int fieldHeight,
                                                   unsigned int voxelGridWidth,
                                                   unsigned int voxelGridHeight,
                                                   unsigned int voxelGridDepth)
{
	Util::VoxelGrid<float> binaryImage(createBinaryVoxelGrid(filename));
	unsigned int imageWidth = binaryImage.getWidth();
	unsigned int imageDepth = binaryImage.getHeight();

	Util::VoxelGrid<float> result(voxelGridWidth, voxelGridHeight, voxelGridDepth);

	float ratioX = static_cast<float>(imageWidth) / static_cast<float>(voxelGridWidth);
	float ratioZ = static_cast<float>(imageDepth) / static_cast<float>(voxelGridDepth);

	for (unsigned int z = 0; z < voxelGridDepth; ++z) {
		for (unsigned int y = fieldBaseHeight; y < fieldBaseHeight + fieldHeight; ++y) {
			for (unsigned int x = 0; x < voxelGridWidth; ++x) {
				result(x, y, z) =
					binaryImage(static_cast<unsigned int>(x * ratioX),
				                static_cast<unsigned int>(z * ratioZ), 0) *
				                (minFieldValue + (maxFieldValue - minFieldValue) *
							                     Util::randf());
			}
		}
	}

	return result;
}

void Engine::addVolumeRendererParameters()
{
	console->output("Adding volume renderer parameters to console.");
	using namespace Console;

	console->addVariable("volume.stepSize",
		createVariableAccessor(&VolumeRenderer::getStepSize,
		                       &VolumeRenderer::setStepSize,
		                       volumeRenderer.get()));

	console->addVariable("volume.drawBoundingBox",
		createVariableAccessor(&VolumeRenderer::getDrawBoundingBox,
		                       &VolumeRenderer::setDrawBoundingBox,
		                       volumeRenderer.get()));

	console->addVariable("volume.boundingBox",
		createVariableAccessor(&VolumeRenderer::getBoundingBox,
		                       &VolumeRenderer::setBoundingBox,
		                       volumeRenderer.get()));
}

void Engine::addFireSimulatorParameters()
{
	console->output("Adding fire simulator parameters to console.");
	using namespace Console;

	FireParameters *fireParameters = fireSimulator.get();

	console->addVariable("fire.buoyancy",
		createVariableAccessor(&FireParameters::getBuoyancy,
		                       &FireParameters::setBuoyancy,
		                       fireParameters));

	console->addVariable("fire.burningRate",
		createVariableAccessor(&FireParameters::getBurningRate,
		                       &FireParameters::setBurningRate,
		                       fireParameters));

	console->addVariable("fire.exhaustDiffusion",
		createVariableAccessor(&FireParameters::getExhaustDiffusion,
		                       &FireParameters::setExhaustDiffusion,
		                       fireParameters));

	console->addVariable("fire.exhaustDissipation",
		createVariableAccessor(&FireParameters::getExhaustDissipation,
		                       &FireParameters::setExhaustDissipation,
		                       fireParameters));

	console->addVariable("fire.fuelDiffusion",
		createVariableAccessor(&FireParameters::getFuelDiffusion,
		                       &FireParameters::setFuelDiffusion,
		                       fireParameters));

	console->addVariable("fire.fuelDissipation",
		createVariableAccessor(&FireParameters::getFuelDissipation,
		                       &FireParameters::setFuelDissipation,
		                       fireParameters));

	console->addVariable("fire.temperatureDiffusion",
		createVariableAccessor(&FireParameters::getTemperatureDiffusion,
		                       &FireParameters::setTemperatureDiffusion,
		                       fireParameters));

	console->addVariable("fire.temperatureDissipation",
		createVariableAccessor(&FireParameters::getTemperatureDissipation,
		                       &FireParameters::setTemperatureDissipation,
		                       fireParameters));

	console->addVariable("fire.gravity",
		createVariableAccessor(&FireParameters::getGravity,
		                       &FireParameters::setGravity,
		                       fireParameters));

	console->addVariable("fire.outputHeat",
		createVariableAccessor(&FireParameters::getOutputHeat,
		                       &FireParameters::setOutputHeat,
		                       fireParameters));

	console->addVariable("fire.stoichiometricMixture",
		createVariableAccessor(&FireParameters::getStoichiometricMixture,
		                       &FireParameters::setStoichiometricMixture,
		                       fireParameters));

	console->addVariable("fire.threshold",
		createVariableAccessor(&FireParameters::getThreshold,
		                       &FireParameters::setThreshold,
		                       fireParameters));

	//console->addVariable("fire.viscosity",
	//	createVariableAccessor(&FireParameters::getViscosity,
	//	                       &FireParameters::setViscosity,
	//	                       fireParameters));

	console->addVariable("fire.vorticity",
		createVariableAccessor(&FireParameters::getVorticity,
		                       &FireParameters::setVorticity,
		                       fireParameters));

	console->addVariable("fire.colorScale",
		createVariableAccessor(&FireParameters::getFireColorScale,
		                       &FireParameters::setFireColorScale,
		                       fireParameters));

	console->addVariable("fire.jacobiSteps",
		createVariableAccessor(&FireParameters::getJacobiSteps,
		                       &FireParameters::setJacobiSteps,
		                       fireParameters));

	console->addVariable("fire.smokeTemperatureScale",
		createVariableAccessor(&FireParameters::getSmokeTemperatureScale,
		                       &FireParameters::setSmokeTemperatureScale,
		                       fireParameters));

	console->addVariable("fire.smokeIntensity",
		createVariableAccessor(&FireParameters::getSmokeDensity,
		                       &FireParameters::setSmokeDensity,
		                       fireParameters));

	console->addVariable("fire.smokeColor",
		createVariableAccessor(&FireParameters::getSmokeColor,
		                       &FireParameters::setSmokeColor,
		                       fireParameters));

	console->addVariable("fire.smokeVisibilityThreshold",
		createVariableAccessor(&FireParameters::getSmokeVisibilityThreshold,
		                       &FireParameters::setSmokeVisibilityThreshold,
		                       fireParameters));

	console->addVariable("fire.vorticityWeight",
		createVariableAccessor(&FireParameters::getVorticityWeight,
		                       &FireParameters::setVorticityWeight,
		                       fireParameters));

	console->addVariable("fire.compute2dColors",
		createVariableAccessor(&FireParameters::getCompute2dColors,
		                       &FireParameters::setCompute2dColors,
		                       fireParameters));

	console->addVariable("fire.massExpansion",
		createVariableAccessor(&FireParameters::getMassExpansion,
		                       &FireParameters::setMassExpansion,
		                       fireParameters));

	console->addVariable("fire.globalWind",
		createVariableAccessor(&FireParameters::getGlobalWind,
		                       &FireParameters::setGlobalWind,
		                       fireParameters));

	console->addVariable("fire.globalWindAmplitude",
		createVariableAccessor(&FireParameters::getGlobalWindAmplitude,
		                       &FireParameters::setGlobalWindAmplitude,
		                       fireParameters));

	console->addVariable("fire.globalWindAngle",
		createVariableAccessor(&FireParameters::getGlobalWindAngle,
		                       &FireParameters::setGlobalWindAngle,
		                       fireParameters));

	console->addVariable("fire.globalWindDynamic",
		createVariableAccessor(&FireParameters::getGlobalWindDynamic,
		                       &FireParameters::setGlobalWindDynamic,
		                       fireParameters));
}

void Engine::addFluidSimulatorParameters()
{
	console->output("Adding fluid simulator parameters to console.");
	using namespace Console;

	FluidParameters *fluidParameters = fluidSimulator.get();

	console->addVariable("fluid.diffusion",
		createVariableAccessor(&FluidParameters::getDiffusion,
		                       &FluidParameters::setDiffusion,
		                       fluidParameters));

	console->addVariable("fluid.viscosity",
		createVariableAccessor(&FluidParameters::getViscosity,
		                       &FluidParameters::setViscosity,
		                       fluidParameters));

	console->addVariable("fluid.vorticity",
		createVariableAccessor(&FluidParameters::getVorticity,
		                       &FluidParameters::setVorticity,
		                       fluidParameters));
}

void Engine::addSceneParameters()
{
	console->output("Adding scene parameters to console.");
	using namespace Console;

	console->addVariable("scene.displayLights",
		createVariableAccessor(&Scene::getDisplayLights,
		                       &Scene::setDisplayLights,
		                       scene.get()));
}

void Engine::initializeOpenGL()
{
	EngineParameters::SetupParams sp = parameters->getSetupParams();	
	glContext = autoPtr(new SDL::OpenGLContext(
		sp.windowWidth, sp.windowHeight, sp.windowCaption, sp.windowFullscreen));

	if (!glContext->good()) {
		console->output("Failed to initialize OpenGL!");
		throw InitializingOpenGLException();		
	}

	console->output("OpenGL context initialized...");

	if (glewInit() != GLEW_OK) {
		console->output("Failed to initialize glew!");
		throw InitializingGlewException();		
	}

	console->output("Glew initialized...");

	glViewport(0, 0, sp.windowWidth, sp.windowHeight);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(75, sp.windowWidth / static_cast<float>(sp.windowHeight), 0.1f, 10000.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void Engine::initializeConsole()
{
	console = autoPtr(new Console::Console(true));

	console->addVariable("paused", new Console::VariableTemplate<bool>(&paused));
	console->addVariable("showFps", new Console::VariableTemplate<bool>(&showFps));	
	console->addVariable("textureContainer.show", new Console::VariableTemplate<bool>(&debug));
	console->addVariable("textureContainer.textColor", new Console::VariableTemplate<Math::Vector3d>(&textColor));
	console->addVariable("textureContainer.gridColor", new Console::VariableTemplate<Math::Vector3d>(&gridColor));
	console->addVariable("textureContainer.separatorColor", new Console::VariableTemplate<Math::Vector3d>(&separatorColor));	
	console->addVariable("textureContainer.smoothDisplay", new Console::VariableTemplate<bool>(&smoothDisplay));
	console->addVariable("backgroundColor", new Console::VariableTemplate<Math::Vector3d>(&background));
	console->addVariable("fixedTimestep", new Console::VariableTemplate<bool>(&fixedTimestep));
	console->addVariable("timestep", new Console::VariableTemplate<float>(&timestep));
}

void Engine::initializeEventHandler()
{
	eventHandler = autoPtr(new EventHandler(*console.get()));

	console->addAction("exit", eventHandler.get());
	console->addAction("quit", eventHandler.get());
}

void Engine::initializeFireSimulator()
{
	console->output("Initializing fire simulator.");
	
	EngineParameters::FireSimulationParams fsp = parameters->getFireSimulationParams();

	std::auto_ptr<Fluid3dSolver::FluidFactory> fluidFactory (new Fluid3dSolver::FluidFactoryImpl(fsp.slice));
	std::auto_ptr<FireFactory> fireFactory(new FireFactory(fsp.slice));		

	fireSimulator = autoPtr(new Fire3dSimulator(fsp.width, fsp.height, fsp.depth, 
	                                            fluidFactory.get(), fireFactory.get(),
												fsp.densityBoundaryType, fsp.velocityBoundaryType));

	console->addAction("fire.ignite",
		Console::createFunctorAction(
			std::bind1st(
				std::mem_fun(&Fire3dSimulator::ignite),
				fireSimulator.get()),
			0.1f,
			"Ignited fire simulation."));


	addFireSimulatorParameters();	

	if (fsp.slice) {
		fireSimulator->setFuelSources(
			createRandomSliceField(fsp.fuelSources, fsp.minFuelValue, fsp.maxFuelValue, 
			                       fsp.width, fsp.height, fsp.depth));
		//fireSimulator->setTemperatureSources(
		//	createRandomSliceField(fsp.tempSources, fsp.minTempValue, fsp.maxTempValue, 
		//	                       fsp.width, fsp.height, fsp.depth));
	} else {
		fireSimulator->setFuelSources(
			createRandom3dField(fsp.fuelSources, fsp.minFuelValue, fsp.maxFuelValue, 
			                    fsp.sourceBaseHeight, fsp.sourceHeight, fsp.width, fsp.height, fsp.depth));
		//fireSimulator->setTemperatureSources(
		//	createRandom3dField(fsp.tempSources, fsp.minTempValue, fsp.maxTempValue,
		//	                    fsp.sourceBaseHeight, fsp.sourceHeight, fsp.width, fsp.height, fsp.depth));
	}
		
	EngineParameters::BlackbodyRadiationParams brp = parameters->getBlackbodyRadiationParams();
	BlackbodyRadiationTable table(brp.dimension, brp.minTemp, brp.maxTemp, brp.exposureLevel);
	fireSimulator->setColorTable(table.getColorPixels(), brp.dimension);

	runner->add(fireSimulator.get());

	TextureContainer *primaryFire = new PrimaryTextureContainer(fireSimulator->getFlat3dTextureSet());
	primaryFire->setDescription("Fire Simulator 3D Primary");
	textureContainers.push_back(primaryFire);

	TextureContainer *secondaryFire = new SecondaryTextureContainer(fireSimulator->getFlat3dTextureSet());
	secondaryFire->setDescription("Fire Simulator 3D Secondary");
	textureContainers.push_back(secondaryFire);

	fireLocation = fsp.fireLocation;
	fireScale = fsp.fireScale;
}

void Engine::initializeLightSystem()
{
	console->output("Initializing light system.");
	
	EngineParameters::ParticleSystemParams lsp = parameters->getLightSystemParams();

	if (parameters->getFireSimulationParams().slice)
		lightSystem = autoPtr(new ParticleSystemSlice(lsp.numParticles, 
													  "",
													  0,
													  fireSimulator->getFlat3dTextureSet(),
													  fireSimulator->getColorTexture(),
													  fireSimulator->getVelocityTexture(),
													  fireSimulator->getVoxelSize(),
													  1.0f / fireSimulator->getFlat3dTextureSet()->getComputationWidth()));
	else
		lightSystem = autoPtr(new ParticleSystem3d(lsp.numParticles, 
												   "",
		                                           0,
												   fireSimulator->getFlat3dTextureSet(),
												   fireSimulator->getColorTexture(),
												   fireSimulator->getVelocityTexture(),
												   fireSimulator->getVoxelSize(),
		                                           1.0f / fireSimulator->getFlat3dTextureSet()->getComputationWidth()));

	console->addVariable("lightSystem.visibilityThreshold",
		Console::createVariableAccessor(&ParticleSystem::getVisibilityThreshold,
		                                &ParticleSystem::setVisibilityThreshold,
		                                lightSystem.get()));

	console->addVariable("lightSystem.minimumLifeTime",
		Console::createVariableAccessor(&ParticleSystem::getMinimumLifeTime,
		                                &ParticleSystem::setMinimumLifeTime,
		                                lightSystem.get()));

	resetParticles(lsp.particleSources, lsp.spawnBaseHeight, lsp.spawnHeight, lsp.sourceHeight, *lightSystem);

	runner->add(lightSystem.get());
}

void Engine::initializeParticleSystem()
{
	console->output("Initializing particle system.");
	
	EngineParameters::ParticleSystemParams psp = parameters->getParticleSystemParams();

	if (parameters->getFireSimulationParams().slice)
		particleSystem = autoPtr(new ParticleSystemSlice(psp.numParticles, 
		                                                 psp.particleSplat,
														 psp.splatThreshold,
		                                                 fireSimulator->getFlat3dTextureSet(),
		                                                 fireSimulator->getColorTexture(),
		                                                 fireSimulator->getVelocityTexture(),
		                                                 fireSimulator->getVoxelSize(),
														 1.0f / fireSimulator->getFlat3dTextureSet()->getComputationWidth()));
	else
		particleSystem = autoPtr(new ParticleSystem3d(psp.numParticles, 
		                                              psp.particleSplat,
													  psp.splatThreshold,
		                                              fireSimulator->getFlat3dTextureSet(),
		                                              fireSimulator->getColorTexture(),
		                                              fireSimulator->getVelocityTexture(),
		                                              fireSimulator->getVoxelSize(),
													  1.0f / fireSimulator->getFlat3dTextureSet()->getComputationWidth()));	

	console->addVariable("particleSystem.particleSize",
		Console::createVariableAccessor(&ParticleSystem::getParticleSize,
		                                &ParticleSystem::setParticleSize,
		                                particleSystem.get()));

	console->addVariable("particleSystem.visibilityThreshold",
		Console::createVariableAccessor(&ParticleSystem::getVisibilityThreshold,
		                                &ParticleSystem::setVisibilityThreshold,
		                                particleSystem.get()));

	console->addVariable("particleSystem.minimumLifeTime",
		Console::createVariableAccessor(&ParticleSystem::getMinimumLifeTime,
		                                &ParticleSystem::setMinimumLifeTime,
		                                particleSystem.get()));

	console->addVariable("particleSystem.blendFunc",
		Console::createVariableAccessor(&ParticleSystem::getBlendFunc,
	                                    &ParticleSystem::setBlendFunc,
	                                    particleSystem.get()));

	console->addVariable("particleSystem.drawBox",
		Console::createVariableAccessor(&ParticleSystem::getDrawBox,
		                                &ParticleSystem::setDrawBox,
		                                particleSystem.get()));

	resetParticles(psp.particleSources, psp.spawnBaseHeight, psp.spawnHeight, psp.sourceHeight, *particleSystem);

	runner->add(particleSystem.get());

	TextureContainer *primaryParticle = particleSystem->getTextureContainer();
	primaryParticle->setGridActive(false);
	textureContainers.push_back(primaryParticle);
}

void Engine::initializeSmokeSystem()
{
	console->output("Initializing smoke system.");
	
	EngineParameters::ParticleSystemParams ssp = parameters->getSmokeSystemParams();

	if (parameters->getFireSimulationParams().slice)
		smokeSystem = autoPtr(new ParticleSystemSlice(ssp.numParticles, 
		                                              ssp.particleSplat,
													  ssp.splatThreshold,
		                                              fireSimulator->getFlat3dTextureSet(),
		                                              fireSimulator->getSmokeTexture(),
		                                              fireSimulator->getVelocityTexture(),
		                                              fireSimulator->getVoxelSize(),
		                                              1.0f / fireSimulator->getFlat3dTextureSet()->getComputationWidth()));
	else
		smokeSystem = autoPtr(new ParticleSystem3d(ssp.numParticles, 
		                                           ssp.particleSplat,
												   ssp.splatThreshold,
		                                           fireSimulator->getFlat3dTextureSet(),
		                                           fireSimulator->getSmokeTexture(),
		                                           fireSimulator->getVelocityTexture(),
		                                           fireSimulator->getVoxelSize(),
		                                           1.0f / fireSimulator->getFlat3dTextureSet()->getComputationWidth()));	

	console->addVariable("smokeSystem.particleSize",
		Console::createVariableAccessor(&ParticleSystem::getParticleSize,
		                                &ParticleSystem::setParticleSize,
		                                smokeSystem.get()));

	console->addVariable("smokeSystem.blendFunc",
		Console::createVariableAccessor(&ParticleSystem::getBlendFunc,
	                                    &ParticleSystem::setBlendFunc,
	                                    smokeSystem.get()));

	console->addVariable("smokeSystem.visibilityThreshold",
		Console::createVariableAccessor(&ParticleSystem::getVisibilityThreshold,
		                                &ParticleSystem::setVisibilityThreshold,
		                                smokeSystem.get()));

	console->addVariable("smokeSystem.minimumLifeTime",
		Console::createVariableAccessor(&ParticleSystem::getMinimumLifeTime,
		                                &ParticleSystem::setMinimumLifeTime,
		                                smokeSystem.get()));

	resetParticles(ssp.particleSources, ssp.spawnBaseHeight, ssp.spawnHeight, ssp.sourceHeight, *smokeSystem);

	runner->add(smokeSystem.get());
}

void Engine::initializeVolumeRenderer()
{	
	EngineParameters::VolumeRendererParams volumePs = parameters->getVolumeRendererParams();
	EngineParameters::FireSimulationParams firePs = parameters->getFireSimulationParams();
	
	if (volumePs.type == EngineParameters::VolumeRendererType::SIMPLE && !firePs.slice) {
		volumeRenderer = autoPtr(new SimpleVolumeRenderer(fireSimulator->getFlat3dTextureSet(),
		                                                  fireSimulator->getColorTexture()));
	} else {
		volumeRenderer = autoPtr(new RayMarcher(fireSimulator->getFlat3dTextureSet(),
												fireSimulator->getColorTexture(),
												firePs.slice,
												BoundingBox(Vector3d(-0.5, 0, -0.5),
												            Vector3d(0.5, 1, 0.5))));
	}

	TextureContainer *primaryVolume = new PrimaryTextureContainer(volumeRenderer->getTextureSet());
	primaryVolume->setDescription("Volume Renderer Primary");
	
	TextureContainer *secondaryVolume = new SecondaryTextureContainer(volumeRenderer->getTextureSet());
	secondaryVolume->setDescription("Volume Renderer Secondary");

	textureContainers.push_back(primaryVolume);
	textureContainers.push_back(secondaryVolume);

	addVolumeRendererParameters();
}

void Engine::initializeScene()
{
	console->output("Initializing scene.");

	bool slice = parameters->getFireSimulationParams().slice;

	const std::vector<EngineParameters::LightParams> lightParams = parameters->getLightParams();

	if (lightSystem.get()) {
		Util::Color lightColor = 
			parameters->getLightSystemParams().color * 
			(parameters->getLightSystemParams().intensity / lightSystem->getNumParticles());

		for (unsigned int i = 0; i < lightSystem->getNumParticles(); ++i) {
			Light *light = new Light();

			light->setDynamic(true);
			light->setColor(lightColor);

			ownedLights.push_back(light);
		}
	}

	for (unsigned int i = 0; i < lightParams.size(); ++i) {
		Light *light = new Light(lightParams[i]);
		ownedLights.push_back(light);
	}

	scene = autoPtr(new Scene(ownedLights, fireSimulator.get(), slice));

	scene->setGlobalAmbient(parameters->getSceneParams().globalAmbient);

	const std::vector<EngineParameters::ModelObjectParams> modelObjectParams = 
		parameters->getModelObjectParams();

	for (unsigned int i = 0; i < modelObjectParams.size(); ++i) {

		Geometry::ModelLoader ml(modelObjectParams[i].filename);		
		Geometry::Model *model = ml.getModel().release();
		ownedModels.push_back(model);

		ModelObject *modelObject = new ModelObject(model, modelObjectParams[i]);
		ownedModelObjects.push_back(modelObject);
		scene->addSceneObject(*modelObject);
	}

	TextureContainer *primaryScene = new PrimaryTextureContainer(scene->getIoTextureSet());
	primaryScene->setDescription("Scene primary");
	textureContainers.push_back(primaryScene);

	addSceneParameters();
	
	const std::vector<EngineParameters::NoisySurfaceParams> noisySurfaceParams = 
		parameters->getNoisySurfaceParams();

	for (unsigned int i = 0; i < noisySurfaceParams.size(); ++i) {
		const EngineParameters::NoisySurfaceParams &p = noisySurfaceParams[i];

		Terrain *terrain = new Terrain(p.width, p.height, p.texture, p.amplitude, p.frequency, p.textureScale, p.scale);
		terrain->setMaterial(p.material);
		terrain->setPosition(p.position);
		terrain->setRotationAngle(p.rotationAngle);
		terrain->setRotationAxis(p.rotationAxis);

		scene->addSceneObject(*terrain);

		ownedTerrains.push_back(terrain);
	}
}

void Engine::updateLightingParameters(float modelMatrix[16])
{
	Matrix4x4 model(modelMatrix, false);
	Matrix4x4 modelInv = model.getInverse();

	Vector3d cameraOS = modelInv.multiplyVec3Pos(camera->getPosition());
	Vector3d lightOS = modelInv.multiplyVec3Pos(lightPosition);

	float eyePos[4] = { cameraOS.x, cameraOS.y, cameraOS.z, 0.0f };

	float lightPos[4] = { lightOS.x, lightOS.y, lightOS.z, 0.0f };

	cgFragmentLightingFS->setParameter("eyePosition", eyePos);
	cgFragmentLightingFS->setParameter("lightPosition", lightPos);

	cgFragmentLightingTexturedFS->setParameter("eyePosition", eyePos);
	cgFragmentLightingTexturedFS->setParameter("lightPosition", lightPos);
}

void Engine::initializeFluidSimulator()
{
	console->output("Initializing fluid simulator.");

	EngineParameters::FluidSimulationParams fsp = parameters->getFluidSimulationParams();

	std::auto_ptr<Fluid3dSolver::FluidFactory> fluidFactory(new Fluid3dSolver::FluidFactoryImpl(true));
	fluidSimulator = autoPtr(new Fluid3dSimulator(fsp.width, fsp.height, 1, 
	                         fluidFactory.get(), fsp.densityBoundaryType, fsp.velocityBoundaryType));	

	addFluidSimulatorParameters();

	OpenGL::Flat3dTextureSet *textures = fluidSimulator->getFlat3dTextureSet();

	Util::VoxelGrid<Color> dfs = createSliceField(fsp.densitySources, fsp.densityScale, fsp.width, fsp.height);
	textures->fillTexture(fluidSimulator->getDensityForceTexture(), dfs);

	Util::VoxelGrid<Color> vfs = createVelocitySliceField(fsp.velocitySources, fsp.velocityDirection, fsp.width, fsp.height);
	textures->fillTexture(fluidSimulator->getVelocityForceTexture(), vfs);	

	runner->add(fluidSimulator.get());

	OpenGL::TextureContainer *fluidSimulatorContainer = 
		new OpenGL::PrimaryTextureContainer(fluidSimulator->getFlat3dTextureSet());
	fluidSimulatorContainer->setGridActive(false);
	fluidSimulatorContainer->setDescription("Fluid 3d simulator");
	textureContainers.push_back(fluidSimulatorContainer);
}

void Engine::initializeTestSuite()
{
	console->output("Initializing tests.");

	testSuite = autoPtr(new Test::TestSuite);

	advectionSliceTest = autoPtr(new Test::AdvectionSliceTest);
	diffusionSliceTest = autoPtr(new Test::DiffusionSliceTest);
	divergenceSliceTest = autoPtr(new Test::DivergenceSliceTest);
	volumetricExtrusionTest = autoPtr(new Test::VolumetricExtrusionTest);

	advection3dTest = autoPtr(new Test::Advection3dTest);
	projection3dTest = autoPtr(new Test::Projection3dTest);

	testSuite->addTest(advectionSliceTest.get());
	testSuite->addTest(diffusionSliceTest.get());
	testSuite->addTest(divergenceSliceTest.get());
	testSuite->addTest(volumetricExtrusionTest.get());

	testSuite->addTest(advection3dTest.get());
	testSuite->addTest(projection3dTest.get());

	runner->add(projection3dTest.get());

	TextureContainer *projection3dPrimary = new PrimaryTextureContainer(projection3dTest->getIoTextureSet());
	projection3dPrimary->setDescription("Test::Projection3dTest");

	textureContainers.push_back(projection3dPrimary);

	runner->add(advectionSliceTest.get());
	runner->add(diffusionSliceTest.get());
	runner->add(divergenceSliceTest.get());
	runner->add(volumetricExtrusionTest.get());

	TextureContainer *primaryAdvectionSlice = new PrimaryTextureContainer(advectionSliceTest->getIoTextureSet());
	primaryAdvectionSlice->setDescription("Test::AdvectionSliceTest");

	TextureContainer *primaryDiffusionSlice = new PrimaryTextureContainer(diffusionSliceTest->getIoTextureSet());
	primaryDiffusionSlice->setDescription("Test::DiffusionSliceTest");

	TextureContainer *primaryDivergenceSlice = new PrimaryTextureContainer(divergenceSliceTest->getIoTextureSet());
	primaryDivergenceSlice->setDescription("Test::DivergenceSliceTest");

	TextureContainer *primaryVolumetric = new PrimaryTextureContainer(volumetricExtrusionTest->getIoTextureSet());
	primaryVolumetric->setDescription("Test::VolumetricExtrusionTest Primary");
	TextureContainer *secondaryVolumetric = new SecondaryTextureContainer(volumetricExtrusionTest->getIoTextureSet());
	secondaryVolumetric->setDescription("Test::VolumetricExtrusionTest Secondary");

	textureContainers.push_back(primaryAdvectionSlice);
	textureContainers.push_back(primaryDiffusionSlice);
	textureContainers.push_back(primaryDivergenceSlice);
	textureContainers.push_back(primaryVolumetric);
	textureContainers.push_back(secondaryVolumetric);
}

void Engine::initializeRunner()
{
	runner = autoPtr(new Runner);

	console->addAction("run", runner.get());
}

void Engine::initializeContainerCycler()
{
	containerCycler = autoPtr(new Console::CycleAction<OpenGL::TextureContainer *>(textureContainers));

	console->addAction("cycleTextureContainer", containerCycler.get());

	textureSelector = autoPtr(new TextureSelector(containerCycler->getCurrent()));
	channelSelector = autoPtr(new ChannelSelector(containerCycler->getCurrent()));
	gridToggler = autoPtr(new GridToggler(containerCycler->getCurrent()));
	textureDumper = autoPtr(new TextureDumper(containerCycler->getCurrent()));
	textureId = autoPtr(new TextureId(containerCycler->getCurrent()));

	console->addAction("textureselect", textureSelector.get());
	console->addAction("channelselect", channelSelector.get());
	console->addAction("gridtoggle", gridToggler.get());
	console->addAction("texturedump", textureDumper.get());
	console->addAction("textureid", textureId.get());
}

void Engine::updateTextureContainerActions()
{
	textureSelector->setContainer(containerCycler->getCurrent());
	channelSelector->setContainer(containerCycler->getCurrent());
	textureDumper->setContainer(containerCycler->getCurrent());
	gridToggler->setContainer(containerCycler->getCurrent());
	textureId->setContainer(containerCycler->getCurrent());
}

struct Variance {
public:
	Variance(float average) : average(average) {}

	float operator()(float variance, float fps) {
		return variance + (fps - average) * (fps - average);
	}

private:
	float average;
};

void Engine::appendTimerLog(const std::vector<float> &fps)
{
	float avgfps = std::accumulate(fps.begin(), fps.end(), 0.0f) / fps.size();

	float stddev = std::sqrt(std::accumulate(fps.begin(), fps.end(), 0.0f, Variance(avgfps)) / fps.size());

	EngineParameters::TimerParams tp = parameters->getTimerParams();

	std::ofstream file(tp.logFile.c_str(), std::ios::out | std::ios::app);

	file << tp.logDesc << ' ' << avgfps << ' ' << stddev << '\n';
}

int Engine::run()
{
	SDL::Timer timer;
	OpenGL::FpsCounter fpsCounter;

	float accumulate = 0.0f;

	console->output("Ready to roll...");

	EngineParameters::TimerParams tp = parameters->getTimerParams();
	EngineParameters::ParticleSystemParams psp = parameters->getParticleSystemParams();
	EngineParameters::ParticleSystemParams ssp = parameters->getSmokeSystemParams();

	unsigned int run = 0;

	std::vector<float> fps(tp.numRuns);

	unsigned int frames = 0;

	float lastTime = timer.seconds();

	bool doReset = false;

	if (tp.wasRead) {
		runner->reset();
		if (fireSimulator.get()) {
			doReset = true;
		}
	}

	bool firstRun = true;

	do {
		if (tp.wasRead && timer.seconds() > tp.seconds) {
			fps[run] = frames / timer.seconds();
			frames = 0;
			
			runner->reset();
			if (fireSimulator.get()) {
				doReset = true;
			}

			timer.reset();
			lastTime = timer.seconds();

			if (++run >= tp.numRuns)
				break;
		}

		float currentTime = timer.seconds();

		float timeDelta = currentTime - lastTime;

		lastTime = currentTime;

		fpsCounter.update(timeDelta);

		lightPosition = fireLocation + Vector3d(0, 50, 0);

		if (containerCycler.get())
			updateTextureContainerActions();

		if (timeDelta > 0.1f)
			timeDelta = 0.1f;
		else if (timeDelta < 0.0001f)
			timeDelta = 0.0001f;

		if (glContext->lock()) {
			if (!paused) {
				scene->setFireLocation(fireLocation);
				scene->setFireScale(fireScale);

				if (lightSystem.get())
					updateParticleLightPositions();

				if (fixedTimestep) {
					accumulate += timeDelta;

					while (accumulate >= timestep) {
						if (fireSimulator.get() && hasAccumulatedDistance) {
							fireSimulator->setUniformAdvection(accumulatedDistance);

							accumulatedDistance = Math::Vector3d();

							hasAccumulatedDistance = false;
						}

						accumulate -= timestep;

						if (fireSimulator.get()) {
							if (doReset) {
								fireSimulator->run(1.0f);
								fireSimulator->ignite(0.1f);
								doReset = false;
							} else
								fireSimulator->run(timestep);
						}

						if (particleSystem.get())
							particleSystem->run(timestep);

						if (lightSystem.get())
							lightSystem->run(timestep);

						if (smokeSystem.get())
							smokeSystem->run(timestep);

						if (fluidSimulator.get())
							fluidSimulator->run(timestep);

						scene->compute();

						//break;
					}

				} else if (!firstRun) {
					if (fireSimulator.get() && hasAccumulatedDistance) {
						fireSimulator->setUniformAdvection(accumulatedDistance);

						accumulatedDistance = Math::Vector3d();

						hasAccumulatedDistance = false;
					}

					if (fireSimulator.get()) {
						if (doReset) {
							fireSimulator->run(1.0f);
							fireSimulator->ignite(0.1f);
							doReset = false;
						} else
							fireSimulator->run(timeDelta);
					}

					if (particleSystem.get())
						particleSystem->run(timeDelta);

					if (lightSystem.get())
						lightSystem->run(timeDelta);

					if (smokeSystem.get())
						smokeSystem->run(timeDelta);

					if (fluidSimulator.get())
						fluidSimulator->run(timeDelta);

					scene->compute();
				}
			}

			if (volumeRenderer.get()) {
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();

				camera->transform();

				glTranslatef(fireLocation.x, fireLocation.y, fireLocation.z);
				glScalef(fireScale, fireScale, fireScale);

				volumeRenderer->compute();

				glLoadIdentity();
			}

			IoTextureSet::disableFbo();

			glClearColor(background.x, background.y, background.z, 1.0f);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

			if (debug && containerCycler.get()) {				
				containerCycler->getCurrent()->setGridColor(gridColor.x, gridColor.y, gridColor.z);
				containerCycler->getCurrent()->setSeparatorColor(separatorColor.x, separatorColor.y, separatorColor.z);
				containerCycler->getCurrent()->setTextColor(textColor.x, textColor.y, textColor.z);
				containerCycler->getCurrent()->render();	
				TextureContainer::setSmoothDisplay(smoothDisplay);
			} else {
				// CAMERA
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();

				camera->transform();

				glGetFloatv(GL_MODELVIEW_MATRIX, viewMatrix);

				glEnable(GL_DEPTH_TEST);

				scene->render(*camera.get());

				// PARTICLE SYSTEM
				float baseHeight = -10.0f;

				glPushMatrix();
				glTranslatef(fireLocation.x, fireLocation.y, fireLocation.z);
				glScalef(fireScale, fireScale, fireScale);

				glDepthMask(GL_FALSE);

				if (smokeSystem.get() && ssp.render)
					smokeSystem->renderParticles();

				if (particleSystem.get() && psp.render)
					particleSystem->renderParticles();

				if (volumeRenderer.get())
					volumeRenderer->render();
			
				glDepthMask(GL_TRUE);

				glDisable(GL_DEPTH_TEST);

				glPopMatrix();

				if (showFps)
					fpsCounter.render(10, 10);
			}

			if (eventHandler->consoleActive())
				console->render();

			handleInput(timeDelta);

			glContext->flip();
			glContext->unlock();
		}

		firstRun = false;

		eventHandler->handleEvents(timeDelta);

		++frames;
	} while (!eventHandler->done());

	std::stringstream fpsStream;

	if (tp.wasRead) {
		appendTimerLog(fps);
		fpsStream << "FPS: " << std::accumulate(fps.begin(), fps.end(), 0.0f) / fps.size() << '\n';
	} else 
		fpsStream << "FPS: " << frames / timer.seconds() << '\n';
	
	console->output(fpsStream.str());

	return 0;
}

}

