#ifndef GPUTHESIS_ENGINE_H
#define GPUTHESIS_ENGINE_H

#include <memory>
#include <string>
#include <vector>

#include "SceneObject.h"
#include "console/Action.h"
#include "math/Vector3d.h"
#include "util/Color.h"
#include "util/VoxelGrid.h"

//class TiXmlDocument;

namespace Console {

class Console;
}

namespace Geometry {

class Model;
}

namespace OpenGL {

class Camera;
class CgProgram;
class TextureContainer;
class TextureSelector;
class ChannelSelector;
class GridToggler;
class TextureDumper;
class TextureId;
}

namespace SDL {

class OpenGLContext;
}

namespace SimpleXml {

class XmlDocument;
class XmlNode;
class XmlNodeWrapper;
}

namespace GpuThesis {

class EngineParameters;
class EventHandler;
class Fire3dSimulator;
class FireParameters;
class Fluid3dSimulator;
class Light;
class ModelObject;
class ParticleSystem;
class Runner;
class Scene;
class VolumeRenderer;
class Terrain;

namespace Test {

class TestSuite;
class AdvectionSliceTest;
class DiffusionSliceTest;
class DivergenceSliceTest;
class VolumetricExtrusionTest;

class Advection3dTest;
class Projection3dTest;
}

/**
 * The main class responsible for initializing and running everything.
 */
class Engine {
public:	
	/**
	 * Initialize the Engine based on the given XML document.
	 * @param xmlDoc the XML configuration document.
     */
	Engine(SimpleXml::XmlDocument &xmlDoc);
	~Engine();

	/**
	 * Start running the simulation.
	 * @return the return code to be returned by main().
	 */
	int run();

	class InitializingOpenGLException : public std::exception {
	public:
		const char *what() const { return "OpenGL initialization failed."; }
	};

	class InitializingGlewException : public std::exception {
	public:
		const char *what() const { return "Glew initialization failed."; }
	};

	class LoadingXmlfileException : public std::exception {
	public:
		LoadingXmlfileException(const std::string &xmlFile) :
		  errorStr("Failed loading xml file: " + xmlFile) {}

		const char *what() const { return errorStr.c_str(); }
	private:
		std::string errorStr;
	};

private:
	void handleInput(float timestep);
	void moveFireSimulation(const Math::Vector3d &distance);
	void resetParticles(const std::string &particleSources, unsigned int spawnBaseHeight,
	                    unsigned int spawnHeight, unsigned int voxelGridHeight,
	                    ParticleSystem &particleSystem);
	
	Util::VoxelGrid<float> createBinaryVoxelGrid(const std::string &imageFilename);
	Util::VoxelGrid<Util::Color> createVoxelGrid(const std::string &imageFilename);
	Util::VoxelGrid<Util::Color> createSliceField(const std::string &filename,
	                                              float scale,												  
	                                              unsigned int sliceWidth,
	                                              unsigned int sliceHeight);
	Util::VoxelGrid<Util::Color> createVelocitySliceField(const std::string &filename,
	                                                      Math::Vector3d direction,												
	                                                      unsigned int sliceWidth,
	                                                      unsigned int sliceHeight);
	Util::VoxelGrid<float> createRandomSliceField(const std::string &filename, 
	                                              float minFieldValue, 
	                                              float maxFieldValue, 
	                                              unsigned int sliceWidth, 
	                                              unsigned int sliceHeight,
	                                              unsigned int numSlices);
	Util::VoxelGrid<float> createRandom3dField(const std::string &filename, 
	                                           float minFieldValue,
	                                           float maxFieldValue, 
	                                           unsigned int fieldBaseHeight,
	                                           unsigned int fieldHeight,
	                                           unsigned int voxelGridWidth,
	                                           unsigned int voxelGridHeight,
	                                           unsigned int voxelGridDepth);

	void addFireSimulatorParameters();
	void addFluidSimulatorParameters();
	void addSceneParameters();
	void addVolumeRendererParameters();
	
	void initialize();
	void initializeOpenGL();	
	void initializeConsole();
	void initializeEventHandler();
	void initializeFireSimulator();	
	void initializeLightSystem();
	void initializeParticleSystem();
	void initializeSmokeSystem();
	void initializeVolumeRenderer();
	void initializeScene();
	void initializeFluidSimulator();
	void initializeTestSuite();
	void initializeRunner();
	void initializeContainerCycler();
	void updateTextureContainerActions();	

	void updateFireLocation();
	void updateParticleLightPositions();

	void updateLightingParameters(float modelMatrix[16]);

	void appendTimerLog(const std::vector<float> &fps);

	SimpleXml::XmlDocument &xmlDoc;	
	std::auto_ptr<Console::Console> console;
	std::auto_ptr<GpuThesis::EngineParameters> parameters;
	std::auto_ptr<EventHandler> eventHandler;

	std::auto_ptr<SDL::OpenGLContext> glContext;

	std::auto_ptr<OpenGL::Camera> camera;
	const float deltaDist; // = 1 * 24;
	const float deltaAngle; // = 24 * Math::pi / 60;

	std::auto_ptr<GpuThesis::Fire3dSimulator> fireSimulator;	
	std::auto_ptr<GpuThesis::ParticleSystem> particleSystem, smokeSystem, lightSystem;
	std::auto_ptr<GpuThesis::VolumeRenderer> volumeRenderer;

	std::auto_ptr<GpuThesis::Fluid3dSimulator> fluidSimulator;

	// Scene objects
	std::vector<Geometry::Model*> ownedModels;
	std::vector<ModelObject*> ownedModelObjects;
	std::vector<Terrain*> ownedTerrains;
	std::vector<Light*> ownedLights;
	std::auto_ptr<Scene> scene;

	// Test suite
	std::auto_ptr<Test::TestSuite> testSuite;
	std::auto_ptr<Test::AdvectionSliceTest> advectionSliceTest;
	std::auto_ptr<Test::DiffusionSliceTest> diffusionSliceTest;
	std::auto_ptr<Test::DivergenceSliceTest> divergenceSliceTest;
	std::auto_ptr<Test::VolumetricExtrusionTest> volumetricExtrusionTest;

	std::auto_ptr<Test::Advection3dTest> advection3dTest;
	std::auto_ptr<Test::Projection3dTest> projection3dTest;

	std::auto_ptr<Runner> runner;

	// Texture Containers
	std::vector<OpenGL::TextureContainer *> textureContainers;
	typedef Console::CycleAction<OpenGL::TextureContainer *> ContainerCycler;
	std::auto_ptr<ContainerCycler> containerCycler;
	std::auto_ptr<OpenGL::TextureSelector> textureSelector;
	std::auto_ptr<OpenGL::ChannelSelector> channelSelector;
	std::auto_ptr<OpenGL::GridToggler> gridToggler;
	std::auto_ptr<OpenGL::TextureDumper> textureDumper;
	std::auto_ptr<OpenGL::TextureId> textureId;

	bool paused; // = false;
	bool showFps; // = false;
	bool debug; // = false;

	Math::Vector3d background;

	// TextureContainer variables
	Math::Vector3d textColor, separatorColor, gridColor;
	bool smoothDisplay;

	Math::Vector3d fireLocation;
	float fireScale;

	Math::Vector3d accumulatedDistance;
	bool hasAccumulatedDistance;

	Math::Vector3d lightPosition;

	float viewMatrix[16];

	// Per-fragment lighting
	std::auto_ptr<OpenGL::CgProgram> cgFragmentLightingVS;
	std::auto_ptr<OpenGL::CgProgram> cgFragmentLightingFS;
	std::auto_ptr<OpenGL::CgProgram> cgFragmentLightingTexturedFS;

	bool fixedTimestep;
	float timestep;
};

}

#endif