#include "EngineParameters.h"

#include <sstream>

#include "Fluid3dSolver.h"

#include "console/Console.h"

#include "simple-xml/XmlDocument.h"
#include "simple-xml/XmlNode.h"
#include "simple-xml/XmlNodeWrapper.h"

#include "util/Util.h"

using namespace SimpleXml;
using namespace Util;

namespace GpuThesis {


EngineParameters::EngineParameters(SimpleXml::XmlDocument &xmlDoc, Console::Console &console) :
	xmlDoc(xmlDoc), console(console)
{
	setupPs = initSetupParams();
	firePs = initFireSimulationParams();
	blackbodyPs = initBlackbodyRadiationParams();
	particlePs = initParticleSystemParams("particle-system");
	smokePs = initParticleSystemParams("smoke-system");
	lightSystemPs = initLightSystemParams();
	volumePs = initVolumeRendererParams();
	fluidPs = initFluidSimulationParams();
	timerPs = initTimerParams();	
	scenePs = initSceneParams();
	initModelObjects();
	initLights();
	initNoisySurfaces();
}
	
EngineParameters::FireSimulationParams EngineParameters::getFireSimulationParams()
{
	return firePs;
}

EngineParameters::BlackbodyRadiationParams EngineParameters::getBlackbodyRadiationParams()
{
	return blackbodyPs;
}

EngineParameters::ParticleSystemParams EngineParameters::getParticleSystemParams()
{
	return particlePs;
}

EngineParameters::ParticleSystemParams EngineParameters::getSmokeSystemParams()
{
	return smokePs;
}

EngineParameters::ParticleSystemParams EngineParameters::getLightSystemParams()
{
	return lightSystemPs;
}

EngineParameters::VolumeRendererParams EngineParameters::getVolumeRendererParams()
{
	return volumePs;
}

EngineParameters::SetupParams EngineParameters::getSetupParams()
{
	return setupPs;
}

EngineParameters::FluidSimulationParams EngineParameters::getFluidSimulationParams()
{
	return fluidPs;
}

EngineParameters::TimerParams EngineParameters::getTimerParams()
{
	return timerPs;
}

EngineParameters::SceneParams EngineParameters::getSceneParams()
{
	return scenePs;
}

const std::vector<EngineParameters::ModelObjectParams> 
EngineParameters::getModelObjectParams() const
{
	return modelObjectPs;
}

const std::vector<EngineParameters::LightParams> 
EngineParameters::getLightParams() const
{
	return lightPs;
}

const std::vector<EngineParameters::NoisySurfaceParams> 
EngineParameters::getNoisySurfaceParams() const
{
	return noisySurfacePs;
}

EngineParameters::FireSimulationParams EngineParameters::initFireSimulationParams()
{
	FireSimulationParams result;
	result.slice = true;
	result.width = 32;
	result.height = 32;
	result.depth = 2;
	result.velocityBoundaryType = Fluid3dSolver::VELOCITY_SMOOTH;
	result.densityBoundaryType = Fluid3dSolver::DENSITY_CLOSED;
	result.fuelSources = "content/fuelSourcesSlice.png";
	//result.tempSources = "content/temperatureSourcesSlice.png";
	result.minFuelValue = 5.0f;
	result.maxFuelValue = 5.0f;
	//result.minTempValue = 1.0f;
	//result.maxTempValue = 1.0f;
	result.sourceBaseHeight = 3;
	result.sourceHeight = 4;
	result.fireLocation = Math::Vector3d(0, -10, 0);
	result.fireScale = 100;

	std::string nodeName = "fire-simulation";

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild("setup").getFirstChild(nodeName);

	if (wrapper.getNode()) {
		readingNode(nodeName);
		result.wasRead = true;

		if (wrapper.getFirstChild("slice").getNode()) {
			result.slice = true;
			console.output("Using slice simulation.");
			readXmlValue(wrapper.getFirstChild("slice"), "count", result.depth);
		} else if (wrapper.getFirstChild("full-3d").getNode()) {
			result.slice = false;
			console.output("Using full 3d simulation.");
			readXmlValue(wrapper.getFirstChild("full-3d"), "computation-depth", result.depth);
			readXmlValue(wrapper.getFirstChild("full-3d"), "source-base-height", result.sourceBaseHeight);
			readXmlValue(wrapper.getFirstChild("full-3d"), "source-height", result.sourceHeight);
		}

		readXmlValue(wrapper, "computation-width", result.width);
		readXmlValue(wrapper, "computation-height", result.height);
				
		readXmlBoundaryValue(wrapper.getFirstChild("velocity-boundary"), "type", result.velocityBoundaryType, true);
		readXmlBoundaryValue(wrapper.getFirstChild("density-boundary"), "type", result.densityBoundaryType, false);		
		
		// No longer necessary. The fire should burn without explicit temperature sources.
		//readXmlValue(wrapper.getFirstChild("temperature-sources"), "image", result.tempSources);
		//readXmlValue(wrapper.getFirstChild("temperature-sources"), "min-value", result.minTempValue);
		//readXmlValue(wrapper.getFirstChild("temperature-sources"), "max-value", result.maxTempValue);

		readXmlValue(wrapper.getFirstChild("fuel-sources"), "image", result.fuelSources);
		readXmlValue(wrapper.getFirstChild("fuel-sources"), "min-value", result.minFuelValue);
		readXmlValue(wrapper.getFirstChild("fuel-sources"), "max-value", result.maxFuelValue);

		readXmlValue(wrapper, "position", result.fireLocation);
		readXmlValue(wrapper, "scale", result.fireScale);
	} else {
		nodeNotFound(nodeName);
		result.wasRead = false;
	}

	return result;
}

EngineParameters::BlackbodyRadiationParams EngineParameters::initBlackbodyRadiationParams()
{
	BlackbodyRadiationParams result;
	result.dimension = 256;
	result.exposureLevel = 16 * 2.669e-11;
	result.maxTemp = 2200;
	result.minTemp = 1240;

	std::string nodeName = "blackbody-radiation";

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).getFirstChild("setup").
		getFirstChild("fire-simulation").getFirstChild("visualization").getFirstChild(nodeName);

	if (wrapper.getNode()) {
		readingNode(nodeName);

		readXmlValue(wrapper, "tablesize", result.dimension);
		readXmlValue(wrapper, "min-temperature", result.minTemp);
		readXmlValue(wrapper, "max-temperature", result.maxTemp);
		readXmlValue(wrapper, "exposure", result.exposureLevel);
	} else {
		nodeNotFound(nodeName);
	}

	return result;
}

EngineParameters::ParticleSystemParams
EngineParameters::initLightSystemParams()
{
	ParticleSystemParams result = initParticleSystemParams("light-system");

	result.color = Util::Color(1, 1, 1, 0);
	result.intensity = 1.0f;

	if (result.wasRead) {
		XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).getFirstChild("setup").
			getFirstChild("fire-simulation").getFirstChild("visualization").getFirstChild("light-system");

		readXmlValue(wrapper, "color", result.color);
		readXmlValue(wrapper, "intensity", result.intensity);
	}

	return result;
}


EngineParameters::ParticleSystemParams
EngineParameters::initParticleSystemParams(const std::string &nodeName)
{
	ParticleSystemParams result;
	result.numParticles = 45 * 45;
	result.particleSplat = "content/splat2-256.bmp";
	result.splatThreshold = 20;
	result.particleSources = "content/particleSources.png";
	result.spawnBaseHeight = 10;
	result.spawnHeight = 4;
	result.sourceHeight = 64;
	result.render = true;

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).getFirstChild("setup").
		getFirstChild("fire-simulation").getFirstChild("visualization").getFirstChild(nodeName);

	if (wrapper.getNode()) {
		readingNode(nodeName);
		result.wasRead = true;

		readXmlValue(wrapper, "particle-count", result.numParticles);
		readXmlValue(wrapper, "particle-texture", result.particleSplat);
		readXmlValue(wrapper, "particle-texture-threshold", result.splatThreshold);
		readXmlValue(wrapper, "render", result.render);

		readXmlValue(wrapper.getFirstChild("particle-sources"), "image", result.particleSources);
		readXmlValue(wrapper.getFirstChild("particle-sources"), "spawn-base-height", result.spawnBaseHeight);
		readXmlValue(wrapper.getFirstChild("particle-sources"), "spawn-height", result.spawnHeight);
		readXmlValue(wrapper.getFirstChild("particle-sources"), "source-height", result.sourceHeight);
	} else {
		nodeNotFound(nodeName);
		result.wasRead = false;
	}

	return result;
}

EngineParameters::VolumeRendererParams EngineParameters::initVolumeRendererParams()
{
	VolumeRendererParams result;
	result.type = VolumeRendererType::RAYMARCHER;
	std::string defaultType = "RAYMARCHER";

	std::string nodeName = "volume-renderer";

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).getFirstChild("setup").
		getFirstChild("fire-simulation").getFirstChild("visualization").getFirstChild(nodeName);

	if (wrapper.getNode()) {
		readingNode(nodeName);
		result.wasRead = true;

		std::string type;
		readXmlValue(wrapper, "type", type);

		if (type == "SIMPLE")
			result.type = VolumeRendererType::SIMPLE;
		else if (type == "RAYMARCHER")
			result.type = VolumeRendererType::RAYMARCHER;
		else
			console.output("Unsupported volume renderer type: " + type + ", using default: " + defaultType);

	} else {
		nodeNotFound(nodeName);
		result.wasRead = false;
	}

	return result;
}

EngineParameters::NoisySurfaceParams 
EngineParameters::initNoisySurfaceParams(SimpleXml::XmlNodeWrapper &wrapper)
{
	NoisySurfaceParams result;

	readingNode(wrapper.getNode()->getName());		

	readXmlValue(wrapper, "width", result.width);
	readXmlValue(wrapper, "height", result.height);

	readXmlValue(wrapper, "position", result.position);
	readXmlValue(wrapper, "rotation-angle", result.rotationAngle);
	readXmlValue(wrapper, "rotation-axis", result.rotationAxis);
	readXmlValue(wrapper, "scale", result.scale);

	readXmlValue(wrapper, "texture", result.texture);
	readXmlValue(wrapper, "amplitude", result.amplitude);
	readXmlValue(wrapper, "frequency", result.frequency);
	readXmlValue(wrapper, "texture-scale", result.textureScale);

	XmlNodeWrapper material = wrapper.getFirstChild("material");
	if (material.getNode()) {
		readingNode("material");
		result.material = initMaterialParams(material);	
	} else {
		nodeNotFound("material");
		result.material = MaterialParams();
	}

	return result;
}

EngineParameters::SetupParams EngineParameters::initSetupParams()
{
	SetupParams result;
	result.configFile = "config.cfg";
	result.runTests = false;
	result.windowFullscreen = false;
	result.windowWidth = 1024;
	result.windowHeight = 768;
	result.cameraOrigin = Math::Vector3d(0, 30, -80);
	result.cameraYaw = 0;
	result.cameraPitch = 0;

	std::string nodeName = "setup";

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild(nodeName);

	if (wrapper.getNode()) {
		readingNode(nodeName);

		readXmlValue(wrapper, "config", result.configFile);		
		readXmlValue(wrapper, "run-tests", result.runTests);
		
		readXmlValue(wrapper.getFirstChild("window"), "fullscreen", result.windowFullscreen);		
		readXmlValue(wrapper.getFirstChild("window"), "width", result.windowWidth);
		readXmlValue(wrapper.getFirstChild("window"), "height", result.windowHeight);
		readXmlValue(wrapper.getFirstChild("window"), "caption", result.windowCaption);

		readXmlValue(wrapper.getFirstChild("camera"), "origin", result.cameraOrigin);
		readXmlValue(wrapper.getFirstChild("camera"), "pitch", result.cameraPitch);
		readXmlValue(wrapper.getFirstChild("camera"), "yaw", result.cameraYaw);
	} else {
		nodeNotFound(nodeName);
	}

	return result;
}

EngineParameters::FluidSimulationParams EngineParameters::initFluidSimulationParams()
{
	FluidSimulationParams result;
	result.width = 128;
	result.height = 128;
	result.velocityBoundaryType = Fluid3dSolver::VELOCITY_SMOOTH;
	result.densityBoundaryType = Fluid3dSolver::DENSITY_CLOSED;
	result.velocitySources = "content/fluidVelocitySources.png";
	result.velocityDirection = Math::Vector3d(0, 1, 0);
	result.densitySources = "content/fluidDensitySources.png";
	result.densityScale = 1.0f;	

	std::string nodeName = "fluid-simulation";

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild("setup").getFirstChild(nodeName);

	if (wrapper.getNode()) {
		readingNode(nodeName);
		result.wasRead = true;

		readXmlValue(wrapper, "computation-width", result.width);
		readXmlValue(wrapper, "computation-height", result.height);
				
		readXmlBoundaryValue(wrapper.getFirstChild("velocity-boundary"), "type", result.velocityBoundaryType, true);
		readXmlBoundaryValue(wrapper.getFirstChild("density-boundary"), "type", result.densityBoundaryType, false);		
		
		readXmlValue(wrapper.getFirstChild("velocity-sources"), "image", result.velocitySources);
		readXmlValue(wrapper.getFirstChild("velocity-sources"), "direction", result.velocityDirection);

		readXmlValue(wrapper.getFirstChild("density-sources"), "image", result.densitySources);
		readXmlValue(wrapper.getFirstChild("density-sources"), "scale", result.densityScale);
		
	} else {
		nodeNotFound(nodeName);
		result.wasRead = false;
	}

	return result;
}

EngineParameters::TimerParams EngineParameters::initTimerParams()
{
	TimerParams result;
	result.numRuns = 1;
	result.seconds = 1.0f;
	result.logFile = "timer.log";
	result.logDesc = "timer.log";

	std::string nodeName = "timer";

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild(nodeName);

	if (wrapper.getNode()) {
		readingNode(nodeName);
		result.wasRead = true;

		readXmlValue(wrapper, "runs", result.numRuns);
		readXmlValue(wrapper, "seconds", result.seconds);		
		readXmlValue(wrapper, "log-file", result.logFile);
		readXmlValue(wrapper, "log-description", result.logDesc);

	} else {
		nodeNotFound(nodeName);
		result.wasRead = false;
	}

	return result;
}

EngineParameters::SceneParams
EngineParameters::initSceneParams() 
{
	SceneParams result;

	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild("setup").getFirstChild("scene");

	if (wrapper.getNode()) {
		readingNode("scene");

		readXmlValue(wrapper, "global-ambient", result.globalAmbient);
	} else {
		nodeNotFound("scene");	
	}

	return result;
}

EngineParameters::MaterialParams 
EngineParameters::initMaterialParams(SimpleXml::XmlNodeWrapper &wrapper)
{
	MaterialParams result;
	
	readXmlValue(wrapper, "ka", result.ka);
	readXmlValue(wrapper, "kd", result.kd);
	readXmlValue(wrapper, "ks", result.ks);
	readXmlValue(wrapper, "shininess", result.shininess);
	
	return result;
}

EngineParameters::ModelObjectParams
EngineParameters::initModelObjectParams(SimpleXml::XmlNodeWrapper &wrapper)
{
	ModelObjectParams result;

	readXmlValue(wrapper, "position", result.position);
	readXmlValue(wrapper, "rotation-angle", result.rotationAngle);
	readXmlValue(wrapper, "rotation-axis", result.rotationAxis);
	readXmlValue(wrapper, "scale", result.scale);
	readXmlValue(wrapper, "relative", result.relative);
	readXmlValue(wrapper, "filename", result.filename);
	
	std::string ignoreMeshes;
	std::string faceNormalMeshes;

	readXmlValue(wrapper, "ignore-meshes", ignoreMeshes);
	readXmlValue(wrapper, "face-normal-meshes", faceNormalMeshes);

	result.ignoreMeshes = readSet<unsigned int>(std::istringstream(ignoreMeshes));
	result.faceNormalMeshes = readSet<unsigned int>(std::istringstream(faceNormalMeshes));

	XmlNodeWrapper material = wrapper.getFirstChild("material");
	if (material.getNode()) {
		readingNode("material");
		result.material = initMaterialParams(material);	
	} else {
		nodeNotFound("material");
		result.material = MaterialParams();
	}

	/*XmlNodeWrapper light = XmlNodeWrapper(wrapper.getFirstChild("light").getNode());

	if (light.getNode()) {
		readXmlValue(light, "position", result.lightPosition);
		readXmlValue(light, "diffuse", result.lightDiffuse);
		readXmlValue(light, "ambient", result.lightAmbient);
	}*/

	return result;
}

EngineParameters::LightParams 
EngineParameters::initLightParams(SimpleXml::XmlNodeWrapper &wrapper)
{
	LightParams result;

	readXmlValue(wrapper, "position", result.position);
	readXmlValue(wrapper, "color", result.color);
	readXmlValue(wrapper, "intensity", result.intensity);
	readXmlValue(wrapper, "dynamic", result.dynamic);

	return result;
}

void EngineParameters::initModelObjects()
{
	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild("setup").getFirstChild("scene").
		getFirstChild("model-objects").getFirstChild("model-object");

	while (wrapper.getNode()) {
		readingNode("model-object");

		modelObjectPs.push_back(initModelObjectParams(wrapper));

		wrapper = XmlNodeWrapper(wrapper.getNode()->getNextSibling("model-object"));
	}
}

void EngineParameters::initLights()
{
	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild("setup").getFirstChild("scene").
		getFirstChild("lights").getFirstChild("light");

	while (wrapper.getNode()) {
		readingNode("light");

		lightPs.push_back(initLightParams(wrapper));

		wrapper = XmlNodeWrapper(wrapper.getNode()->getNextSibling("light"));
	}
}

void EngineParameters::initNoisySurfaces()
{
	XmlNodeWrapper wrapper = XmlNodeWrapper(xmlDoc.getRootNode()).
		getFirstChild("setup").getFirstChild("scene").
		getFirstChild("noisy-surfaces").getFirstChild("noisy-surface");

	while (wrapper.getNode()) {
		readingNode("noisy-surface");

		noisySurfacePs.push_back(initNoisySurfaceParams(wrapper));		

		wrapper = XmlNodeWrapper(wrapper.getNode()->getNextSibling("noisy-surface"));
	}
}

void EngineParameters::readXmlBoundaryValue(SimpleXml::XmlNodeWrapper &wrapper, 
                                            const std::string &childNode, 											
                                            Fluid3dSolver::BoundaryType &type,
                                            bool velocityBoundary)
{
	std::string value;
	readXmlValue(wrapper, childNode, value);

	if (velocityBoundary) {
		if (value == "OPEN")
			type = Fluid3dSolver::VELOCITY_OPEN;
		else if (value == "SMOOTH")
			type = Fluid3dSolver::VELOCITY_SMOOTH;
		else if (value == "CONCRETE")
			type = Fluid3dSolver::VELOCITY_CONCRETE;
		else if (value == "ROUGH")
			type = Fluid3dSolver::VELOCITY_ROUGH;
		else if (value == "WIND")
			type = Fluid3dSolver::VELOCITY_WIND;
		else
			console.output("Unsupported velocity boundary: " + value + ". Using default value.");
	} else {
		if (value == "OPEN") 
			type = Fluid3dSolver::DENSITY_OPEN;
		else if (value == "CLOSED")
			type = Fluid3dSolver::DENSITY_CLOSED;
		else 
			console.output("Unsupported density boundary: " + value + ". Using default value.");
	}	
}

void EngineParameters::readingNode(const std::string &nodeName)
{
	console.output("Reading xml node <" + nodeName + ">.");	
}

void EngineParameters::nodeNotFound(const std::string &nodeName)
{
	console.output("Xml node <" + nodeName + "> not found.");
}

}
