#ifndef GPUTHESIS_ENGINEPARAMETERS_H
#define GPUTHESIS_ENGINEPARAMETERS_H

#include <set>
#include <string>
#include <vector>

#include "math/Vector3d.h"
#include "util/Color.h"
#include "util/Util.h"


namespace Console {

class Console;
}

namespace Fluid3dSolver {

enum BoundaryType;
}

namespace SimpleXml {

class XmlDocument;
class XmlNode;
class XmlNodeWrapper;
}

namespace GpuThesis {

/**
 * Reads all parameters from an XML configuration document.
 */
class EngineParameters {
public:
	EngineParameters(SimpleXml::XmlDocument &xmlDoc, Console::Console &console);

	/**
	 * Corresponds to "fire-simulation" tag.
	 */
	struct FireSimulationParams {
		bool wasRead;
		bool slice;
		unsigned int width;
		unsigned int height;
		unsigned int depth;
		Fluid3dSolver::BoundaryType velocityBoundaryType;
		Fluid3dSolver::BoundaryType densityBoundaryType;		
		std::string fuelSources;
		std::string tempSources;
		float minFuelValue;
		float maxFuelValue;
		float minTempValue;
		float maxTempValue;
		unsigned int sourceBaseHeight;
		unsigned int sourceHeight;
		Math::Vector3d fireLocation;
		float fireScale;
	};

	/**
	 * Corresponds to "blackbody-radiation" tag.
	 */
	struct BlackbodyRadiationParams {
		unsigned int dimension; 
		double minTemp; 
		double maxTemp; 
		double exposureLevel;
	};

	/**
	 * Corresponds to "particle-system", "smoke-system", and "light-system" tags.
	 */
	struct ParticleSystemParams {
		bool wasRead;
		unsigned int numParticles;
		std::string particleSplat;
		float splatThreshold;
		std::string particleSources;
		unsigned int spawnBaseHeight;
		unsigned int spawnHeight;
		unsigned int sourceHeight;
		bool render;
		Util::Color color;
		float intensity;
	};

	enum VolumeRendererType {
		SIMPLE,
		RAYMARCHER,
	};

	/**
	 * Corresponds to "volume-renderer" tag.
	 */
	struct VolumeRendererParams {
		bool wasRead;
		VolumeRendererType type;
	};	

	/**
	 * Corresponds to "light" tag.
	 */
	struct LightParams {
		LightParams() : dynamic(false), intensity(1.0) {}
		Math::Vector3d position;
		Util::Color color;
		float intensity;
		bool dynamic;
	};

	/**
	 * Corresponds to "material" tag.
	 */
	struct MaterialParams {
		MaterialParams() : ka(0), kd(0), ks(0), shininess(0) {}
		float ka;
		float kd;
		float ks;
		float shininess;		
	};

	/**
	 * Corresponds to "scene" tag.
	 */ 
	struct SceneParams {
		Util::Color globalAmbient;	
	};	

	/**
	 * Corresponds to "noisy-surface" tag.
	 */
	struct NoisySurfaceParams {
		NoisySurfaceParams() : width(64), height(64), rotationAngle(0), 
			scale(1), amplitude(1), frequency(1), textureScale(1) {}
		
		unsigned int width;
		unsigned int height;

		Math::Vector3d position;
		float rotationAngle;
		Math::Vector3d rotationAxis;
		float scale;

		std::string texture;

		float amplitude;
		float frequency;
		float textureScale;

		MaterialParams material;
	};

	/**
	 * Corresponds to "model-object" tag.
	 */
	struct ModelObjectParams {
		ModelObjectParams() : rotationAngle(0), scale(0), relative(false) {}
		Math::Vector3d position;
		float rotationAngle;
		Math::Vector3d rotationAxis;
		float scale;
		std::set<unsigned int> ignoreMeshes;
		std::set<unsigned int> faceNormalMeshes;
		bool relative;
		std::string filename;

		MaterialParams material;
	};

	/**
	 * Corresponds to "setup" tag.
	 */
	struct SetupParams {
		std::string configFile;
		bool runTests;
		bool windowFullscreen;
		unsigned int windowWidth;
		unsigned int windowHeight;
		std::string windowCaption;
		Math::Vector3d cameraOrigin;
		float cameraPitch;
		float cameraYaw;
	};

	/**
	 * Corresponds to "fluid-simulation" tag.
	 */
	struct FluidSimulationParams {
		bool wasRead;		
		unsigned int width;
		unsigned int height;		
		Fluid3dSolver::BoundaryType velocityBoundaryType;
		Fluid3dSolver::BoundaryType densityBoundaryType;		
		std::string densitySources;
		float densityScale;
		std::string velocitySources;
		Math::Vector3d velocityDirection;
	};

	/**
	 * Corresponds to "timer" tag.
	 */
	struct TimerParams {
		bool wasRead;
		unsigned int numRuns;
		float seconds;
		std::string logFile;
		std::string logDesc;
	};
	
	FireSimulationParams getFireSimulationParams();
	BlackbodyRadiationParams getBlackbodyRadiationParams();
	ParticleSystemParams getParticleSystemParams();
	ParticleSystemParams getSmokeSystemParams();
	ParticleSystemParams getLightSystemParams();
	VolumeRendererParams getVolumeRendererParams();	
	SetupParams getSetupParams();
	FluidSimulationParams getFluidSimulationParams();
	TimerParams getTimerParams();
	SceneParams getSceneParams();
	const std::vector<ModelObjectParams> getModelObjectParams() const;
	const std::vector<LightParams> getLightParams() const;
	const std::vector<NoisySurfaceParams> getNoisySurfaceParams() const;

private:
	FireSimulationParams initFireSimulationParams();
	BlackbodyRadiationParams initBlackbodyRadiationParams();
	ParticleSystemParams initLightSystemParams();
	ParticleSystemParams initParticleSystemParams(const std::string &nodeName);
	VolumeRendererParams initVolumeRendererParams();	
	SetupParams initSetupParams();
	FluidSimulationParams initFluidSimulationParams();
	TimerParams initTimerParams();	
	SceneParams initSceneParams();
	ModelObjectParams initModelObjectParams(SimpleXml::XmlNodeWrapper &wrapper);
	MaterialParams initMaterialParams(SimpleXml::XmlNodeWrapper &wrapper);
	LightParams initLightParams(SimpleXml::XmlNodeWrapper &wrapper);
	NoisySurfaceParams initNoisySurfaceParams(SimpleXml::XmlNodeWrapper &wrapper);
	void initModelObjects();
	void initLights();
	void initNoisySurfaces();

	void readXmlBoundaryValue(SimpleXml::XmlNodeWrapper &wrapper, 
	                          const std::string &childNode, 
							  Fluid3dSolver::BoundaryType &type, 
	                          bool velocityBoundary);
	void readingNode(const std::string &nodeName);
	void nodeNotFound(const std::string &nodeName);

	template <typename Source>
	void readXmlValue(SimpleXml::XmlNodeWrapper &wrapper, const std::string &childNode, Source &source)
	{
		SimpleXml::XmlNode *root = wrapper.getNode();
		
		std::string nodeString;
		if (root)
			nodeString += "<" + root->getName() + ">::";
		nodeString += "<" + childNode + ">";

		SimpleXml::XmlNode *node = wrapper.getFirstChild(childNode).getNode();
		if (node) {
			Source newValue;
			if (!Util::lexical_cast(node->getValue(), newValue))
				console.output("Failed casting " + nodeString + ": " + node->getValue());
			else {
				source = newValue;
				console.output("Read " + nodeString + ": " + Util::toString(source));
			}
		} else {
			console.output("Failed reading " + nodeString + ", using default: " + Util::toString(source));
		}
	}

	SimpleXml::XmlDocument &xmlDoc;
	Console::Console &console;

	FireSimulationParams firePs;
	BlackbodyRadiationParams blackbodyPs;
	ParticleSystemParams particlePs, smokePs, lightSystemPs;
	VolumeRendererParams volumePs;	
	SetupParams setupPs;
	FluidSimulationParams fluidPs;
	TimerParams timerPs;	
	SceneParams scenePs;
	std::vector<ModelObjectParams> modelObjectPs;
	std::vector<LightParams> lightPs;
	std::vector<NoisySurfaceParams> noisySurfacePs;
};

}

#endif