#include "Fire3dSimulator.h"

#include "opengl/CgContext.h"
#include "opengl/CgFactory.h"
#include "opengl/CgProgram.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/Flat3dTextureOperation.h"
#include "opengl/OrthonormalProjection.h"

#include "Fluid3dSolver.h"

#include "FluidFactory.h"

#include "util/VoxelGrid.h"

using namespace OpenGL;
using namespace Math;
using namespace std;

namespace {

std::auto_ptr<OpenGL::CgProgram>
createFragmentProgram(const std::string &entry,
                      OpenGL::IoTextureSet *textures)
{
	return createFragmentProgram("fire3d.cg", entry, textures);
}

}

namespace GpuThesis {

FireFactory::FireFactory(bool slice) : slice(slice)
{
}

std::auto_ptr<OpenGL::Flat3dTextureOperation>
FireFactory::createFlat3dTextureOperation(const OpenGL::Flat3dTextureSet *textures,
                                          OpenGL::CgProgram *interior)
{
	return OpenGL::createFlat3dTextureOperation(textures, interior, 
		slice ? Flat3dTextureOperation::INTERIOR_2D :
		        Flat3dTextureOperation::INTERIOR_3D);
}

Fire3dSimulator::Fire3dSimulator(unsigned int width, unsigned int height, unsigned int depth,
                                 Fluid3dSolver::FluidFactory *factory,
                                 FireFactory *fireFactory,
                                 Fluid3dSolver::BoundaryType densityBoundary,
                                 Fluid3dSolver::BoundaryType velocityBoundary) :
	textures(new OpenGL::Flat3dTextureSet(width, height, depth, 
	                                      CgContext::getInstance().getTexRectTarget(),
	                                      CgContext::getInstance().getTexRectInternalFormat())),
	voxelSize(1.0f),
	velocityBoundary(velocityBoundary),
	densityBoundary(densityBoundary),
	perlin(3, 0.45f, 3141592),
	time(0.0f)
{
	initializeTextures();
	initializeFireOperations(fireFactory);
	initializeFluidOperations(factory);

	setJacobiSteps(20);
	setMassExpansion(0);
}

Fire3dSimulator::~Fire3dSimulator()
{
}

OpenGL::Flat3dTextureSet *Fire3dSimulator::getFlat3dTextureSet()
{
	return textures.get();
}

unsigned int Fire3dSimulator::getVelocityTexture() const
{
	return velocityTexture;
}

unsigned int Fire3dSimulator::getColorTexture() const
{
	return colorTexture;
}

unsigned int Fire3dSimulator::getSmokeTexture() const
{
	return smokeTexture;
}

void Fire3dSimulator::setFuelSources(const Util::VoxelGrid<float> &values)
{
	unsigned int width = values.getWidth();
	unsigned int height = values.getHeight();
	unsigned int depth = values.getDepth();

	Util::VoxelGrid<Util::Color> sources = textures->getVoxelGrid(densitySources);	

	for (unsigned int z = 0; z < depth; ++z) {
		for (unsigned int y = 0; y < height; ++y) {
			for (unsigned int x = 0; x < width; ++x) {
				sources(x, y, z).setGreen(values(x, y, z));
			}
		}
	}

	textures->fillTexture(densitySources, sources);
}

void Fire3dSimulator::setTemperatureSources(const Util::VoxelGrid<float> &values)
{
	unsigned int width = values.getWidth();
	unsigned int height = values.getHeight();
	unsigned int depth = values.getDepth();

	Util::VoxelGrid<Util::Color> sources = textures->getVoxelGrid(densitySources);	

	for (unsigned int z = 0; z < depth; ++z) {
		for (unsigned int y = 0; y < height; ++y) {
			for (unsigned int x = 0; x < width; ++x) {
				sources(x, y, z).setBlue(values(x, y, z));
			}
		}
	}

	textures->fillTexture(densitySources, sources);
}

float Fire3dSimulator::getVoxelSize() const
{
	return voxelSize;
}

void Fire3dSimulator::setColorTable(const std::vector<float> &values, unsigned int dimension)
{
	textures->resizeSecondaryTexture(colorTable, dimension, 1);
	textures->fillTexture(colorTable, GL_RGBA, GL_FLOAT, &*values.begin());
}

void Fire3dSimulator::ignite(float timestep)
{
	float oldThreshold = getThreshold();

	setThreshold(-0.1f);

	run(timestep);

	setThreshold(oldThreshold);
}

void Fire3dSimulator::setUniformAdvection(const Math::Vector3d &distance)
{
	velocityAdvect->setUniformAdvection(distance);
	densityAdvect->setUniformAdvection(distance);
}

void Fire3dSimulator::doRun(float timestep)
{
	if (getGlobalWindDynamic())
		updateDynamicGlobalWind(time += timestep);

	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();
	
	vorticity->run();
	calculateVelocityForces(timestep);
	velocityForceAdder->run(timestep);

	calculateDensityForces(timestep);
	densityForceAdder->run(timestep);

	calculateDivergenceForces(timestep);

	//velocityDiffuse->run(timestep);

	//project->run(); // uncomment for increased accuracy in simulation
	velocityAdvect->run(timestep);
	project->run();
	
	densityAdvect->run(timestep);	
	densityDiffuse->run(timestep);

	calculateColors();

	textures->disableFbo();

	velocityAdvect->setUniformAdvection(Math::Vector3d(0.0f, 0.0f, 0.0f));
	densityAdvect->setUniformAdvection(Math::Vector3d(0.0f, 0.0f, 0.0f));
}

void Fire3dSimulator::doReset()
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	textures->clearTexture(colorTexture);
	textures->clearTexture(velocityTexture);
	textures->clearTexture(densityTexture);
	textures->clearTexture(velocityForceTexture);
	textures->clearTexture(densityForceTexture);
				
	textures->disableFbo();
}

void Fire3dSimulator::clearTextures()
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	textures->clearTexture(colorTexture);
	textures->clearTexture(velocityTexture);
	textures->clearTexture(densityTexture);
	textures->clearTexture(velocityForceTexture);
	textures->clearTexture(densityForceTexture);
	textures->clearTexture(densitySources);
				
	textures->disableFbo();
}

void Fire3dSimulator::calculateColors()
{
	if (getCompute2dColors())
		calculateColors2dOp->compute();
	else
		calculateColorsOp->compute();

	calculateSmokeColorsOp->compute();
}

void Fire3dSimulator::calculateDensityForces(float timestep)
{
	calculateDensityForcesOp->compute();
}

void Fire3dSimulator::calculateVelocityForces(float timestep)
{
	calculateVelocityForcesOp->compute();
}

void Fire3dSimulator::calculateDivergenceForces(float timestep)
{
	calculateDivergenceForcesOp->compute();
}

void Fire3dSimulator::initializeTextures()
{	
	colorTexture = textures->addPrimaryTexture("Color");
	smokeTexture = textures->addPrimaryTexture("Smoke");
	densityTexture = textures->addPrimaryTexture("Density");
	velocityTexture = textures->addPrimaryTexture("Velocity");

	densityForceTexture = textures->addPrimaryTexture("Density force");
	velocityForceTexture = textures->addPrimaryTexture("Velocity force");
	divergenceForceTexture = textures->addPrimaryTexture("Divergence force");

	densitySources = textures->addPrimaryTexture("Density sources");

	colorTable = textures->addSecondaryTexture(128, 1, 
	                                           GL_TEXTURE_RECTANGLE_NV,
	                                           GL_FLOAT_RGBA32_NV,
	                                           "Color table");
}

void Fire3dSimulator::initializeFireOperations(FireFactory *fireFactory)
{
	cgCalculateDensityForces = createFragmentProgram("calculateDensityForces", textures.get());		
	cgCalculateDensityForces->setParameter("burningRate", getBurningRate());
	cgCalculateDensityForces->setParameter("outputHeat", getOutputHeat());
	cgCalculateDensityForces->setParameter("stoichiometricMixture", getStoichiometricMixture());
	cgCalculateDensityForces->setParameter("threshold", getThreshold());
	cgCalculateDensityForces->setParameter("dissipation", dissipation);
	cgCalculateDensityForces->setTextureParameter("densityField", densityTexture);
	cgCalculateDensityForces->setTextureParameter("sourceField", densitySources);

	cgCalculateVelocityForces = createFragmentProgram("calculateVelocityForces", textures.get());
	cgCalculateVelocityForces->setParameter("gravity", getGravity());
	cgCalculateVelocityForces->setParameter("buoyancy", getBuoyancy());
	cgCalculateVelocityForces->setParameter("reactionThreshold", getThreshold());
	setVorticityWeight(1.0f);
	cgCalculateVelocityForces->setTextureParameter("densityField", densityTexture);
	cgCalculateVelocityForces->setTextureParameter("oldForces", velocityForceTexture);

	cgCalculateDivergenceForces = createFragmentProgram("calculateDivergenceForces", textures.get());
	cgCalculateDivergenceForces->setParameter("burningRate", getBurningRate());	
	cgCalculateDivergenceForces->setParameter("stoichiometricMixture", getStoichiometricMixture());
	cgCalculateDivergenceForces->setParameter("threshold", getThreshold());	
	cgCalculateDivergenceForces->setParameter("massExpansion", getMassExpansion());
	cgCalculateDivergenceForces->setTextureParameter("densityField", densityTexture);	

	cgCalculateColors = createFragmentProgram("calculateColors", textures.get());
	cgCalculateColors->setParameter("scale", 256.0f / 4.0f);
	cgCalculateColors->setTextureParameter("densityField", densityTexture);
	cgCalculateColors->setTextureParameter("colorTable", colorTable);

	cgCalculateColors2d = createFragmentProgram("calculateColors2d", textures.get());
	cgCalculateColors2d->setParameter("scale", 256.0f / 4.0f);
	cgCalculateColors2d->setTextureParameter("densityField", densityTexture);
	cgCalculateColors2d->setTextureParameter("colorTable", colorTable);

	setCompute2dColors(false);

	cgCalculateSmokeColors = createFragmentProgram("calculateSmokeColors", textures.get());	
	cgCalculateSmokeColors->setTextureParameter("fireDensities", densityTexture);	

	calculateDensityForcesOp = 
		fireFactory->createFlat3dTextureOperation(textures.get(), cgCalculateDensityForces.get());	
	calculateDensityForcesOp->setOutputTexture(densityForceTexture);

	calculateVelocityForcesOp = 
		fireFactory->createFlat3dTextureOperation(textures.get(), cgCalculateVelocityForces.get());
	calculateVelocityForcesOp->setOutputTexture(velocityForceTexture);

	calculateDivergenceForcesOp =
		fireFactory->createFlat3dTextureOperation(textures.get(), cgCalculateDivergenceForces.get());
	calculateDivergenceForcesOp->setOutputTexture(divergenceForceTexture);

	calculateColorsOp = 
		fireFactory->createFlat3dTextureOperation(textures.get(), cgCalculateColors.get());
	calculateColorsOp->setOutputTexture(colorTexture);

	calculateColors2dOp = 
		fireFactory->createFlat3dTextureOperation(textures.get(), cgCalculateColors2d.get());
	calculateColors2dOp->setOutputTexture(colorTexture);

	calculateSmokeColorsOp = 
		fireFactory->createFlat3dTextureOperation(textures.get(), cgCalculateSmokeColors.get());
	calculateSmokeColorsOp->setOutputTexture(smokeTexture);
}

void Fire3dSimulator::initializeFluidOperations(Fluid3dSolver::FluidFactory *factory)
{
	cgVelocityBoundary = createBoundaryProgram(velocityBoundary, textures.get(), factory->isSliceSimulation());

	velocityAdvect = 
		factory->createAdvection(textures.get(), velocityTexture, velocityTexture, cgVelocityBoundary.get());

	densityAdvect =
		factory->createAdvection(textures.get(), velocityTexture, densityTexture, densityBoundary);

	project =
		factory->createProjection(textures.get(), velocityTexture, cgVelocityBoundary.get());

	//velocityDiffuse =
	//	factory->createDiffusion(textures.get(), velocityTexture, velocityBoundary);

	densityDiffuse =
		factory->createDiffusion(textures.get(), densityTexture, densityBoundary);

	densityForceAdder =
		factory->createForceAdder(textures.get(), densityTexture, densityForceTexture, densityBoundary);

	velocityForceAdder =
		factory->createForceAdder(textures.get(), velocityTexture, velocityForceTexture, cgVelocityBoundary.get());

	vorticity =
		factory->createVorticityConfinement(textures.get(), velocityTexture, velocityForceTexture);

	velocityAdvect->setVoxelSize(voxelSize);
	densityAdvect->setVoxelSize(voxelSize);

	project->setVoxelSize(voxelSize);
	project->setDivergenceForceTexture(divergenceForceTexture);

	if (velocityBoundary != Fluid3dSolver::VELOCITY_WIND)
		cgVelocityBoundary->setTextureParameter("value", velocityTexture);
	else if (factory->isSliceSimulation())
		cgVelocityBoundary->setParameter("numSlices", static_cast<float>(textures->getComputationDepth()));

	//velocityDiffuse->setVoxelSize(voxelSize);
	//velocityDiffuse->setDiffusionRates(getViscosity(), getViscosity(), getViscosity(), getViscosity());

	densityDiffuse->setVoxelSize(voxelSize);	
	densityDiffuse->setDiffusionRates(getExhaustDiffusion(), 
	                                  getFuelDiffusion(),
	                                  getTemperatureDiffusion(), 
	                                  0);

	vorticity->setVoxelSize(voxelSize);
	vorticity->setVorticityScale(getVorticity());

	clearTextures();
}

void Fire3dSimulator::updateBurningRate(float burningRate)
{
	cgCalculateDensityForces->setParameter("burningRate", burningRate);
	cgCalculateDivergenceForces->setParameter("burningRate", burningRate);
}

void Fire3dSimulator::updateOutputHeat(float outputHeat)
{
	cgCalculateDensityForces->setParameter("outputHeat", outputHeat);
}

void Fire3dSimulator::updateStoichiometricMixture(float stoichiometricMixture)
{
	cgCalculateDensityForces->setParameter("stoichiometricMixture", stoichiometricMixture);	
	cgCalculateDivergenceForces->setParameter("stoichiometricMixture", stoichiometricMixture);	
}

void Fire3dSimulator::updateThreshold(float threshold)
{
	cgCalculateDensityForces->setParameter("threshold", threshold);	
	cgCalculateVelocityForces->setParameter("reactionThreshold", threshold);	
	cgCalculateDivergenceForces->setParameter("threshold", threshold);	
}

void Fire3dSimulator::updateExhaustDissipation(float exhaustDissipation)
{
	dissipation[0] = exhaustDissipation;
	cgCalculateDensityForces->setParameter("dissipation", dissipation);
}

void Fire3dSimulator::updateFuelDissipation(float fuelDissipation)
{
	dissipation[1] = fuelDissipation;
	cgCalculateDensityForces->setParameter("dissipation", dissipation);
}

void Fire3dSimulator::updateTemperatureDissipation(float temperatureDissipation)
{
	dissipation[2] = temperatureDissipation;
	cgCalculateDensityForces->setParameter("dissipation", dissipation);
}

void Fire3dSimulator::updateExhaustDiffusion(float exhaustDiffusion)
{	
	densityDiffuse->setDiffusionRates(exhaustDiffusion, 
	                                  getFuelDiffusion(),
	                                  getTemperatureDiffusion(),
	                                  0);
}

void Fire3dSimulator::updateTemperatureDiffusion(float temperatureDiffusion)
{
	densityDiffuse->setDiffusionRates(getExhaustDiffusion(), 
	                                  getFuelDiffusion(),
	                                  temperatureDiffusion,
	                                  0);
}

void Fire3dSimulator::updateFuelDiffusion(float fuelDiffusion)
{
	densityDiffuse->setDiffusionRates(getExhaustDiffusion(), 
	                                  fuelDiffusion,
	                                  getTemperatureDiffusion(),
	                                  0);
}

void Fire3dSimulator::updateGravity(float gravity)
{
	cgCalculateVelocityForces->setParameter("gravity", gravity);
}

void Fire3dSimulator::updateBuoyancy(float buoyancy)
{
	cgCalculateVelocityForces->setParameter("buoyancy", buoyancy);
}

void Fire3dSimulator::updateVorticityWeight(float vorticityWeight)
{
	cgCalculateVelocityForces->setParameter("vorticityWeight", vorticityWeight);
}

void Fire3dSimulator::updateViscosity(float viscosity)
{
	//velocityDiffuse->setDiffusionRates(viscosity, viscosity, viscosity, viscosity);
}

void Fire3dSimulator::updateVorticity(float vorticity)
{
	this->vorticity->setVorticityScale(vorticity);
}

void Fire3dSimulator::updateJacobiSteps(unsigned int jacobiSteps)
{
	if (jacobiSteps > 1000)
		setJacobiSteps(1000);
	else {
		densityDiffuse->setNumJacobiSteps(jacobiSteps);
		//velocityDiffuse->setNumJacobiSteps(jacobiSteps);
		project->setNumIterations(jacobiSteps);
	}
}

void Fire3dSimulator::updateFireColorScale(float fireColorScale)
{
	cgCalculateColors->setParameter("scale", fireColorScale);
	cgCalculateColors2d->setParameter("scale", fireColorScale);
}

void Fire3dSimulator::updateSmokeTemperatureScale(float smokeTemperatureScale)
{
	cgCalculateSmokeColors->setParameter("temperatureScale", smokeTemperatureScale);
}

void Fire3dSimulator::updateSmokeDensity(float smokeDensity)
{
	cgCalculateSmokeColors->setParameter("smokeDensity", smokeDensity);
}

void Fire3dSimulator::updateSmokeColor(const Math::Vector3d &smokeColor)
{
	float colors[4] = {smokeColor.x, smokeColor.y, smokeColor.z, 0};
	cgCalculateSmokeColors->setParameter("smokeColor", colors);
}


void Fire3dSimulator::updateSmokeVisibilityThreshold(float smokeVisibilityThreshold)
{
	cgCalculateSmokeColors->setParameter("visibilityThreshold", smokeVisibilityThreshold);
}

void Fire3dSimulator::updateCompute2dColors(bool compute2dColors)
{
}

void Fire3dSimulator::updateMassExpansion(float massExpansion)
{
	cgCalculateDivergenceForces->setParameter("massExpansion", massExpansion);
}

void Fire3dSimulator::updateGlobalWind(const Math::Vector3d &globalWind)
{
	float wind[4] = { globalWind.x, globalWind.y, globalWind.z, 0.0f };

	if (velocityBoundary == Fluid3dSolver::VELOCITY_WIND)
		cgVelocityBoundary->setParameter("globalWind", wind);
}

void Fire3dSimulator::updateGlobalWindAmplitude(const Math::Vector3d &globalWindAmplitude)
{

}

void Fire3dSimulator::updateGlobalWindAngle(const Math::Vector3d &globalWindAngle)
{

}

void Fire3dSimulator::updateGlobalWindDynamic(bool globalWindDynamic)
{
	if (!globalWindDynamic)
		updateGlobalWind(getGlobalWind());
}

void Fire3dSimulator::updateDynamicGlobalWind(float time)
{
	Vector3d globalWindAmplitude = getGlobalWindAmplitude();
	Vector3d globalWindAngle = getGlobalWindAngle();

	float amplitudeFactor = perlin.sample(time * globalWindAmplitude.z);
	float angleFactor = perlin.sample(time * globalWindAngle.z);

	float amplitude = globalWindAmplitude.x + 
	                  amplitudeFactor * (globalWindAmplitude.y - globalWindAmplitude.x);

	float angle = globalWindAngle.x + 
	              angleFactor * (globalWindAngle.y - globalWindAngle.x);

	angle *= Math::pi / 180.0f;

	Vector3d wind = Vector3d(cos(angle), 0.0f, sin(angle)) * amplitude;

	updateGlobalWind(wind);
}

}
