#ifndef GPUTHESIS_FIRE3DSIMULATOR_H
#define GPUTHESIS_FIRE3DSIMULATOR_H

#include <memory>
#include <vector>

#include "Runnable.h"

#include "FireParameters.h"
#include "Fluid3dSolver.h"
#include "util/VoxelGrid.h"
#include "util/Perlin.h"

namespace OpenGL {
	class CgProgram;
	class IoTextureSet;
	class Flat3dTextureSet;
	class TextureContainer;
	class Flat3dTextureOperation;
};

namespace Fluid3dSolver {
	class Projection;
	class ForceAdder;
	class Diffusion;
	class Advection;
	class VorticityConfinement;	

	class FluidFactory;
};

namespace GpuThesis {

template <typename T>
class VoxelGrid;

/**
 * The FireFactory class is an abstract factory used to create fire specific instances 
 * of Flat3dTextureOperation to be used in a Fire3dSimulator. 
 * The abstract factory can separate between full 3d
 * simulation and slice (volumetric extrusion) simulation. 
 */
class FireFactory {
public:
	/**
	 * Constructs a new FireFactory for either a full 3d simulation 
	 * or a slice (volumetric extrusion) simulation. 
	 * @param slice if the FireFactory should be used with a slice simulation. 
	 */
	FireFactory(bool slice);

	/**
	 * Creates a new instance of Flat3dTextureOperation.
	 * @param textures the texture set associated with the created instance.
	 * @param interior the interior cg program encapsulated by the created instance.
	 * @return an auto pointer to the created instance. 
	 */
	std::auto_ptr<OpenGL::Flat3dTextureOperation>
	createFlat3dTextureOperation(const OpenGL::Flat3dTextureSet *textures,
	                             OpenGL::CgProgram *interior);

private:
	bool slice;
};

/**
 * The Fire3dSimulator class is used to perform a complete fire simulation 
 * in a specified domain. The simulation will use a fluid solver in combination
 * with fire specific combustion modeling. The Fire3dSimulator class can represent 
 * various simulation types by specifying different abstract factories
 * in the constructor.
 */
class Fire3dSimulator : public FireParameters, public Runnable {
public:
	/**
	 * Constructs a Fire3dSimulator.
	 * @param width the computation width of the simulation domain.
	 * @param height the computation height of the simulation domain.
	 * @param depth the computation depth of the simulation domain.
	 * @param factory the fluid factory used to initialize the fluid solver.
	 * @param fireFactory the fire factory used to initialize the fire specific combustion modeling.
	 * @param densityBoundary the boundary type used for the fire specific density field.
	 * @param velocityBoundary the boundary type used for the velocity field.
	 */
	Fire3dSimulator(unsigned int width, unsigned int height, unsigned int depth,
	                Fluid3dSolver::FluidFactory *factory,
					FireFactory *fireFactory,
                    Fluid3dSolver::BoundaryType densityBoundary,
                    Fluid3dSolver::BoundaryType velocityBoundary);
	~Fire3dSimulator();

	/**
	 * Returns the texture set containing all the simulation textures. 
	 * @return the texture set.
	 */
	OpenGL::Flat3dTextureSet *getFlat3dTextureSet();

	/**
	 * Returns the texture id of the velocity texture.
	 * @return the texture id.
	 */
	unsigned int getVelocityTexture() const;

	/**
	 * Returns the texture id of the color texture. 
	 * This texture contains the visual results of the fire simulation. 
	 * @return the texture id.
	 */
	unsigned int getColorTexture() const;

	/**
	 * Returns the texture id of the smoke texture. 
	 * This texture contains the smoke of the fire simulation. 
	 * @return the texture id.
	 */
	unsigned int getSmokeTexture() const;

	/**
	 * Sets the fuel sources to be used in the fire simulation.
	 * @param values a voxel grid containing the fuel sources.
	 */
	void setFuelSources(const Util::VoxelGrid<float> &values);

	/**
	 * Sets the temperature sources to be used in the fire simulation.
	 * @param values a voxel grid containing the temperature sources. 
	 */
	void setTemperatureSources(const Util::VoxelGrid<float> &values);

	/**
	 * Sets the color table to be used when computing the final color texture. 
	 * The color table should contain the range of colors expected to be seen in the fire, 
	 * from dark to bright colors. 
	 * The values in the color table vector should be in the range 0-1, 
	 * and stored on the form (red, green, blue) * dimension.
	 * @param values the color table values.
	 * @param dimension the dimension (length) of the color table.
	 */
	void setColorTable(const std::vector<float> &values, unsigned int dimension);

	/**
	 * Returns the voxel size used in the simulation domain. 
	 * @return the voxel size.
	 */
	float getVoxelSize() const;
	
	/**
	 * Ignites the fire simulation by setting the fire threshold below zero for timestep seconds.
	 * @param timestep the duration of the ignition. 
	 */
	void ignite(float timestep);

	/**
	 * Advects the contents of the fire simulation (density and velocity).
	 * Note that distance must be in local simulation scale.
	 * @param distance the distance to advect.
	 */
	void setUniformAdvection(const Math::Vector3d &distance);

private:	
	void doRun(float timestep);
	void doReset();

	void clearTextures();

	void calculateColors();
	void calculateDensityForces(float timestep);
	void calculateVelocityForces(float timestep);
	void calculateDivergenceForces(float timestep);

	void initializeTextures();
	void initializeFireOperations(FireFactory *fireFactory);
	void initializeFluidOperations(Fluid3dSolver::FluidFactory *factory);

	void updateBurningRate(float burningRate);
	void updateOutputHeat(float outputHeat);
	void updateStoichiometricMixture(float stoichiometricMixture);
	void updateThreshold(float threshold);
	void updateExhaustDissipation(float exhaustDissipation);
	void updateFuelDissipation(float fuelDissipation);
	void updateTemperatureDissipation(float temperatureDissipation);
	void updateExhaustDiffusion(float exhaustDiffusion);
	void updateTemperatureDiffusion(float temperatureDiffusion);
	void updateFuelDiffusion(float fuelDiffusion);
	void updateGravity(float gravity);
	void updateBuoyancy(float buoyancy);
	void updateVorticityWeight(float vorticityWeight);
	void updateViscosity(float viscosity);
	void updateVorticity(float vorticity);
	void updateJacobiSteps(unsigned int jacobiSteps);
	void updateFireColorScale(float fireColorScale);
	void updateSmokeTemperatureScale(float smokeTemperatureScale);
    void updateSmokeDensity(float smokeDensity);
    void updateSmokeVisibilityThreshold(float smokeVisibilityThreshold);
	void updateCompute2dColors(bool compute2dColors);
	void updateMassExpansion(float massExpansion);
	void updateSmokeColor(const Math::Vector3d &smokeColor);
	void updateGlobalWind(const Math::Vector3d &globalWind);
	void updateGlobalWindAmplitude(const Math::Vector3d &globalWindAmplitude);
	void updateGlobalWindAngle(const Math::Vector3d &globalWindAngle);
	void updateGlobalWindDynamic(bool globalWindDynamic);

	void updateDynamicGlobalWind(float time);

	std::auto_ptr<OpenGL::Flat3dTextureSet> textures;

	std::auto_ptr<Fluid3dSolver::Advection> velocityAdvect;
	std::auto_ptr<Fluid3dSolver::Advection> densityAdvect;

	std::auto_ptr<Fluid3dSolver::Projection> project;

	std::auto_ptr<Fluid3dSolver::Diffusion> velocityDiffuse;
	std::auto_ptr<Fluid3dSolver::Diffusion> densityDiffuse;

	std::auto_ptr<Fluid3dSolver::ForceAdder> densityForceAdder;
	std::auto_ptr<Fluid3dSolver::ForceAdder> velocityForceAdder;	

	std::auto_ptr<Fluid3dSolver::VorticityConfinement> vorticity;

	std::auto_ptr<OpenGL::CgProgram> cgCalculateDensityForces;
	std::auto_ptr<OpenGL::CgProgram> cgCalculateVelocityForces;
	std::auto_ptr<OpenGL::CgProgram> cgCalculateColors, cgCalculateColors2d;
	std::auto_ptr<OpenGL::CgProgram> cgCalculateSmokeColors;
	std::auto_ptr<OpenGL::CgProgram> cgCalculateDivergenceForces;

	std::auto_ptr<OpenGL::Flat3dTextureOperation> calculateDensityForcesOp;
	std::auto_ptr<OpenGL::Flat3dTextureOperation> calculateVelocityForcesOp;
	std::auto_ptr<OpenGL::Flat3dTextureOperation> calculateColorsOp, calculateColors2dOp;
	std::auto_ptr<OpenGL::Flat3dTextureOperation> calculateSmokeColorsOp;
	std::auto_ptr<OpenGL::Flat3dTextureOperation> calculateDivergenceForcesOp;

	std::auto_ptr<OpenGL::CgProgram> cgVelocityBoundary;

	unsigned int densityTexture;
	unsigned int velocityTexture;

	unsigned int densityForceTexture;
	unsigned int velocityForceTexture;
	unsigned int divergenceForceTexture;

	unsigned int densitySources;

	unsigned int colorTexture;
	unsigned int smokeTexture;
	unsigned int colorTable; // black body radiation table

	float voxelSize; // = 1.0f;
	Fluid3dSolver::BoundaryType velocityBoundary; // = Fluid3dSolver::VELOCITY_SLIP;
	Fluid3dSolver::BoundaryType densityBoundary; // = Fluid3dSolver::DENSITY_CLOSED;

	// Used while calculating density forces
	float dissipation[4];

	Perlin1D perlin;

	float time;
};

}

#endif