#ifndef GPUTHESIS_FIREPARAMETERS_H
#define GPUTHESIS_FIREPARAMETERS_H

#include "math/Vector3d.h"

namespace GpuThesis {

/**
 * Contains all the parameters governing the fire simulation.
 */
class FireParameters {
public:
	/**
	 * Burning rate governs how fast fuel can be combusted.
	 * @return the burning rate.
	 */
	float getBurningRate() const;

	/**
	 * Set the burning rate.
	 * @see getBurningRate
	 * @param burningRate the new burning rate.
	 */
	void setBurningRate(float burningRate);

	/**
	 * Output heat governs how much heat is created in the combustion reaction.
	 * @return the output heat.
	 */
	float getOutputHeat() const;

	/**
	 * Set the output heat.
	 * @see getOutputHeat
	 * @param outputHeat the new output heat.
	 */
	void setOutputHeat(float outputHeat);

	/**
	 * Stoichiometric mixture governs how much oxygen is involved in the combustion reaction.
	 * Higher stoichiometric mixture causes more exhaust gas to be produced by one unit of fuel.
	 * @return the stoichiometric mixture.
	 */
	float getStoichiometricMixture() const;

	/**
	 * Set the stoichiometric mixture.
	 * @see getStoichiometricMixture
	 * @param stoichiometricMixture the new stoichiometricMixture.
	 */
	void setStoichiometricMixture(float stoichiometricMixture);

	/**
	 * Threshold is the lowest temperature at which the combustion reaction takes place.
	 * @return the temperature threshold.
	 */
	float getThreshold() const;

	/**
	 * Set the threshold.
	 * @see getThreshold
	 * @param threshold the new threshold.
	 */
	void setThreshold(float threshold);

	/**
	 * Exhaust dissipation is the rate at which exhaust dissipates.
	 * @return the exhaust dissipation.
	 */
	float getExhaustDissipation() const;

	/**
	 * Set the exhaust dissipation.
	 * @see getExhaustDissipation
	 * @param exhaustDissipation the exhaust dissipation.
	 */
	void setExhaustDissipation(float exhaustDissipation);

	/**
	 * Temperature dissipation is the rate at which temperature dissipates.
	 * @return the temperature dissipation.
	 */
	float getTemperatureDissipation() const;

	/**
	 * Set the temperature dissipation.
	 * @see getTemperatureDissipation
	 * @param temperatureDissipation the temperature dissipation.
	 */
	void setTemperatureDissipation(float temperatureDissipation);

	/**
	 * Fuel dissipation is the rate at which fuel dissipates.
	 * @return the fuel dissipation.
	 */
	float getFuelDissipation() const;

	/**
	 * Set the fuel dissipation.
	 * @see getFuelDissipation
	 * @param fuelDissipation the fuel dissipation.
	 */
	void setFuelDissipation(float fuelDissipation);

	/**
	 * Exhaust diffusion is the rate at which exhaust diffuses.
	 * @return the exhaust diffusion.
	 */
	float getExhaustDiffusion() const;

	/**
	 * Set the exhaust diffusion.
	 * @see getExhaustDiffusion
	 * @param exhaustDiffusion the exhaust diffusion.
	 */
	void setExhaustDiffusion(float exhaustDiffusion);

	/**
	 * Temperature diffusion is the rate at which temperature diffuses.
	 * @return the temperature diffusion.
	 */
	float getTemperatureDiffusion() const;

	/**
	 * Set the temperature diffusion.
	 * @see getTemperatureDiffusion
	 * @param temperatureDiffusion the temperature diffusion.
	 */
	void setTemperatureDiffusion(float temperatureDiffusion);

	/**
	 * Fuel diffusion is the rate at which fuel diffuses.
	 * @return the fuel diffusion.
	 */
	float getFuelDiffusion() const;

	/**
	 * Set the fuel diffusion.
	 * @see getFuelDiffusion
	 * @param fuelDiffusion the fuel diffusion.
	 */
	void setFuelDiffusion(float fuelDiffusion);

	/**
	 * Gravity controls the force with which fuel and exhaust are pulled down.
	 * @return the gravity.
	 */
	float getGravity() const;

	/**
	 * Set the gravity.
	 * @see getGravity
	 * @param gravity the new gravity.
	 */
	void setGravity(float gravity);

	/**
	 * Buoyancy controls the force which makes air rise due to heat.
	 * @return the buoyancy.
	 */
	float getBuoyancy() const;

	/**
	 * Set the buoyancy.
	 * @see getBuoyancy
	 * @param buoyancy the new buoyancy.
	 */
	void setBuoyancy(float buoyancy);

	/**
	 * The vorticity weight controls how the vorticity is increased outside the flame.
	 * @return the vorticity weight.
	 */
	float getVorticityWeight() const;

	/**
	 * Set the vorticity weight.
	 * @see getVorticityWeight
	 * @param vorticityWeight the new vorticity weight.
	 */
	void setVorticityWeight(float vorticityWeight);

	/**
	 * Viscosity controls the "thickness" of the air.
	 * @return the viscosity.
	 */
	float getViscosity() const;

	/**
	 * Set the viscosity.
	 * @see getViscosity
	 * @param viscosity the new viscosity.
	 */
	void setViscosity(float viscosity);

	/**
	 * Vorticity controls the rate at which vortices in the velocity field are enhanced.
	 * @return the vorticity.
	 */
	float getVorticity() const;

	/**
	 * Set the vorticity.
	 * @see getVorticity
	 * @param vorticity the vorticity.
	 */
	void setVorticity(float vorticity);

	/**
	 * Jacobi steps controls how many steps are used in the diffusion and projection solvers.
	 * The more Jacobi steps the more accurate diffusion and projection.
	 * @return the number of jacobi steps.
	 */
	unsigned int getJacobiSteps() const;
	
	/**
	 * Set the number of jacobi steps.
	 * @see getJacobiSteps
	 * @param jacobiSteps the new number of jacobi steps.
	 */
	void setJacobiSteps(unsigned int jacobiSteps);

	/**
	 * The fire color scale controls the brightness of the color field.
	 * @return the fire color scale.
	 */
	float getFireColorScale() const;

	/**
	 * Set the fire color scale.
	 * @see getFireColorScale
	 * @param fireColorScale the new fire color scale.
	 */
	void setFireColorScale(float fireColorScale);

	/**
	 * The smoke temperature scale controls how cool the exhaust gas must be to be treated as smoke.
	 * @return the smoke temperature scale.
	 */
    float getSmokeTemperatureScale() const;

	/**
	 * Set the smoke temperature scale.
	 * @see getSmokeTemperatureScale
	 * @param smokeTemperatureScale the new smoke temperature scale.
	 */
	void setSmokeTemperatureScale(float smokeTemperatureScale);

	/**
	 * The smoke density controls how dense the smoke appears.
	 * @return the smoke density.
	 */
    float getSmokeDensity() const;

	/**
	 * Set the smoke density.
	 * @see getSmokeDensity
	 * @param smokeDensity the new smoke density.
	 */
	void setSmokeDensity(float smokeDensity);

	/**
	 * The color of the smoke.
	 * @return the smoke's color as a Vector3d.
	 */
	Math::Vector3d getSmokeColor() const;

	/**
	 * Set the smoke color.
	 * @see getSmokeColor
	 * @param smokeColor the new smoke color.
	 */
	void setSmokeColor(Math::Vector3d smokeColor);

	/**
	 * The smoke visibility threshold controls the amount of exhaust gas below which smoke particles are respawned.
	 * @return the smoke visibility threshold.
	 */
    float getSmokeVisibilityThreshold() const;

	/**
	 * Set the smoke visibility threshold.
	 * @see getSmokeVisibilityThreshold
	 * @param smokeVisibilityThreshold the new smoke visibility threshold.
	 */
	void setSmokeVisibilityThreshold(float smokeVisibilityThreshold);	

	/**
	 * Compute 2d colors controls whether to compute the color field using the approach from the 2d gpu fire.
	 * @return whether to compute 2d colors.
	 */
	bool getCompute2dColors() const;

	/**
	 * Set whether to compute 2d colors.
	 * @see getCompute2dColors
	 * @param compute2dColors whether to compute 2d colors.
	 */
	void setCompute2dColors(bool compute2dColors);

	/**
	 * Mass expansion controls how fuel gas expands when being converted to exhaust gas.
	 * @return the mass expansion.
	 */
	float getMassExpansion() const;

	/**
	 * Set the mass expansion.
	 * @see getMassExpansion
	 * @param massExpansion the new mass expansion.
	 */
	void setMassExpansion(float massExpansion);

	/**
	 * The static global wind of the fire simulation (the wind at the boundaries).
	 * @return the static global wind.
	 */
	Math::Vector3d getGlobalWind() const;

	/**
	 * Set the static global wind.
	 * @see getGlobalWind
	 * @param globalWind the new static global wind.
	 */
	void setGlobalWind(Math::Vector3d globalWind);

	/**
	 * The amplitude parameters of the dynamic global wind.
	 * The x, y, and z components control the min, max, and frequency of the amplitude.
	 * @return the dynamic global wind amplitude parameters.
	 */
	Math::Vector3d getGlobalWindAmplitude() const;

	/**
	 * Set the dynamic global wind amplitude parameters.
	 * @see getGlobalWindAmplitude
	 * @param globalWindAmplitude the new dynamic global wind amplitude parameters.
	 */
	void setGlobalWindAmplitude(Math::Vector3d globalWindAmplitude);

	/**
	 * The angle parameters of the dynamic global wind.
	 * The x, y, and z components control the min, max, and frequency of the angle.
	 * @return the dynamic global wind angle parameters.
	 */
	Math::Vector3d getGlobalWindAngle() const;

	/**
	 * Set the dynamic global wind angle parameters.
	 * @see getGlobalWindAngle
	 * @param globalWindAngle the new dynamic global wind angle parameters.
	 */
	void setGlobalWindAngle(Math::Vector3d globalWindAngle);

	/**
	 * Global wind dynamic controls whether to have static or dynamic global wind.
	 * Static global wind is constant, whereas dynamic global wind is governed by angle and amplitude parameters.
	 * @see getGlobalWind
	 * @see getGlobalWindAmplitude
	 * @see getGlobalWindAngle
	 * @return whether to use dynamic global wind.
	 */
	bool getGlobalWindDynamic() const;

	/**
	 * Set whether to use dynamic global wind.
	 * @see getGlobalWindDynamic
	 * @param globalWindDynamic whether to use dynamic global wind.
	 */
	void setGlobalWindDynamic(bool globalWindDynamic);

private:
	virtual void updateBurningRate(float burningRate) = 0;
	virtual void updateOutputHeat(float outputHeat) = 0;
	virtual void updateStoichiometricMixture(float stoichiometricMixture) = 0;
	virtual void updateThreshold(float threshold) = 0;
	virtual void updateExhaustDissipation(float exhaustDissipation) = 0;
	virtual void updateFuelDissipation(float fuelDissipation) = 0;
	virtual void updateTemperatureDissipation(float temperatureDissipation) = 0;
	virtual void updateExhaustDiffusion(float exhaustDiffusion) = 0;
	virtual void updateTemperatureDiffusion(float temperatureDiffusion) = 0;
	virtual void updateFuelDiffusion(float fuelDiffusion) = 0;
	virtual void updateGravity(float gravity) = 0;
	virtual void updateBuoyancy(float buoyancy) = 0;
	virtual void updateVorticityWeight(float vorticityWeight) = 0;
	virtual void updateViscosity(float viscosity) = 0;
	virtual void updateVorticity(float vorticity) = 0;
	virtual void updateJacobiSteps(unsigned int jacobiSteps) = 0;
	virtual void updateFireColorScale(float fireColorScale) = 0;
	virtual void updateSmokeColor(const Math::Vector3d &smokeColor) = 0;
	virtual void updateSmokeTemperatureScale(float smokeTemperatureScale) = 0;
    virtual void updateSmokeDensity(float smokeDensity) = 0;	
    virtual void updateSmokeVisibilityThreshold(float smokeVisibilityThreshold) = 0;
	virtual void updateCompute2dColors(bool compute2dColors) = 0;
	virtual void updateMassExpansion(float massExpansion) = 0;
	virtual void updateGlobalWind(const Math::Vector3d &globalWind) = 0;
	virtual void updateGlobalWindAmplitude(const Math::Vector3d &globalWindAmplitude) = 0;
	virtual void updateGlobalWindAngle(const Math::Vector3d &globalWindAngle) = 0;
	virtual void updateGlobalWindDynamic(bool globalWindDynamic) = 0;

	float burningRate;
	float outputHeat;
	float stoichiometricMixture;
	float threshold;
	float exhaustDissipation;
	float fuelDissipation;
	float temperatureDissipation;
	float exhaustDiffusion;
	float temperatureDiffusion;
	float fuelDiffusion;
	float gravity;
	float buoyancy;
	float vorticityWeight;
	float viscosity;
	float vorticity;
	float fireColorScale;		
	unsigned int jacobiSteps;
	Math::Vector3d smokeColor;
	float smokeTemperatureScale;
    float smokeDensity;	
    float smokeVisibilityThreshold;
	bool compute2dColors;
	float massExpansion;
	Math::Vector3d globalWind;
	Math::Vector3d globalWindAmplitude;
	Math::Vector3d globalWindAngle;
	bool globalWindDynamic;
};

}

#endif