#include <algorithm>

#include "Fluid3dSimulator.h"

#include "Fluid3dSolver.h"

#include "FluidFactory.h"

#include "opengl/CgContext.h"
#include "opengl/CgProgram.h"
#include "opengl/Flat3dTextureOperation.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/OrthonormalProjection.h"
#include "opengl/PrimaryTextureContainer.h"
#include "opengl/TextureContainer.h"

using std::min;
using std::max;

using namespace Fluid3dSolver;

namespace GpuThesis {

Fluid3dSimulator::Fluid3dSimulator(unsigned int width,
                                   unsigned int height,
								   unsigned int depth,
                                   FluidFactory *factory,
                                   Fluid3dSolver::BoundaryType densityBoundary,
                                   Fluid3dSolver::BoundaryType velocityBoundary) :
	textures(new OpenGL::Flat3dTextureSet(width, height, depth, 
	                                      GL_TEXTURE_RECTANGLE_NV, 
	                                      GL_FLOAT_RGBA32_NV))
{
	densityTexture = textures->addPrimaryTexture("Density");
	densityForceTexture = textures->addPrimaryTexture("Density force");
	velocityTexture = textures->addPrimaryTexture("Velocity");
	velocityForceTexture = textures->addPrimaryTexture("Velocity force");
	//vorticityTexture = textures->addPrimaryTexture("Vorticity");
	vorticityForceTexture = textures->addPrimaryTexture("Vorticity force");

	float voxelSize = 1.0f;

	velocityAdvect = 
		factory->createAdvection(textures.get(), velocityTexture, velocityTexture, velocityBoundary);

	densityAdvect = 
		factory->createAdvection(textures.get(), velocityTexture, densityTexture, densityBoundary);

	project =
		factory->createProjection(textures.get(), velocityTexture, velocityBoundary);

	velocityDiffuse =
		factory->createDiffusion(textures.get(), velocityTexture, velocityBoundary);

	densityDiffuse = 
		factory->createDiffusion(textures.get(), densityTexture, densityBoundary);

	densityForceAdder =
		factory->createForceAdder(textures.get(), densityTexture, densityForceTexture, densityBoundary);

	velocityForceAdder =
		factory->createForceAdder(textures.get(), velocityTexture, velocityForceTexture, velocityBoundary);

	vorticity =
		factory->createVorticityConfinement(textures.get(), velocityTexture, vorticityForceTexture);

	vorticityForceAdder = 
		factory->createForceAdder(textures.get(), velocityTexture, vorticityForceTexture, velocityBoundary);

	velocityAdvect->setVoxelSize(voxelSize);
	densityAdvect->setVoxelSize(voxelSize);
	project->setVoxelSize(voxelSize);
	densityDiffuse->setVoxelSize(voxelSize);	
	velocityDiffuse->setVoxelSize(voxelSize);
	vorticity->setVoxelSize(voxelSize);

	setDiffusion(1.0f);
	setViscosity(0.0f);
	setVorticity(1.0f);

	clearTextures();
}



Fluid3dSimulator::~Fluid3dSimulator()
{
}



OpenGL::Flat3dTextureSet *Fluid3dSimulator::getFlat3dTextureSet()
{
	return textures.get();
}



void Fluid3dSimulator::clearTextures()
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	textures->clearTexture(densityTexture);
	//textures->clearTexture(densityForceTexture);
	textures->clearTexture(velocityTexture);
	//textures->clearTexture(velocityForceTexture);
	textures->clearTexture(vorticityForceTexture);

	textures->disableFbo();
}


unsigned int Fluid3dSimulator::getDensityTexture() const
{
	return densityTexture;
}


unsigned int Fluid3dSimulator::getDensityForceTexture() const
{
	return densityForceTexture;
}


unsigned int Fluid3dSimulator::getVelocityTexture() const
{
	return velocityTexture;
}


unsigned int Fluid3dSimulator::getVelocityForceTexture() const
{
	return velocityForceTexture;
}


void Fluid3dSimulator::doRun(float timestep)
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();
	
	vorticity->run();

	densityForceAdder->run(timestep);

	vorticityForceAdder->run(timestep);

	velocityForceAdder->run(timestep);

	if (getVorticity() > 0)
		velocityDiffuse->run(timestep);

	//project->run();
	velocityAdvect->run(timestep);
	project->run();

	densityAdvect->run(timestep);
	densityDiffuse->run(timestep);

	textures->disableFbo();
}

void Fluid3dSimulator::updateDiffusion(float diffusion)
{
	densityDiffuse->setDiffusionRates(diffusion, diffusion, diffusion, diffusion);
}

void Fluid3dSimulator::updateViscosity(float viscosity)
{
	velocityDiffuse->setDiffusionRates(viscosity, viscosity, viscosity, viscosity);
}

void Fluid3dSimulator::updateVorticity(float vorticity)
{
	this->vorticity->setVorticityScale(vorticity);
}

void Fluid3dSimulator::updateJacobiSteps(unsigned int jacobiSteps)
{

}

}
