#ifndef GPUTHESIS_FLUID3DSIMULATOR_H
#define GPUTHESIS_FLUID3DSIMULATOR_H

#include <memory>
#include <vector>

#include "opengl/OpenGL.h"

#include "Fluid3dSolver.h"
#include "FluidParameters.h"

#include "Runnable.h"

namespace OpenGL {
	class CgProgram;
	class IoTextureSet;
	class Flat3dTextureSet;
	class TextureContainer;
	class Flat3dTextureOperation;
};

namespace Fluid3dSolver {
	class Projection;
	class ForceAdder;
	class Diffusion;
	class Advection;

	class FluidFactory;
};

namespace GpuThesis {

/**
 * Performs uncoupled fluid simulations.
 */
class Fluid3dSimulator : public FluidParameters, public Runnable {
public:
	/**
	 * Create a fluid simulation at the given computation dimensions.
	 * @param width the width of the simulation domain.
	 * @param height the height of the simulation domain.
	 * @param depth the depth of the simulation domain.
	 * @param factory the factory used to create fluid solver operations.
	 * @param densityBoundary the density boundary type.
	 * @param velocityBoundary the velocity boundary type.
	 */

	Fluid3dSimulator(unsigned int width,
	                 unsigned int height,
					 unsigned int depth,
	                 Fluid3dSolver::FluidFactory *factory,
	                 Fluid3dSolver::BoundaryType densityBoundary,
	                 Fluid3dSolver::BoundaryType velocityBoundary);

	~Fluid3dSimulator();

	/**
	 * Get the Flat3dTextureSet instance used.
	 * @return the Flat3dTextureSet instance.
	 */
	OpenGL::Flat3dTextureSet *getFlat3dTextureSet();

	/**
	 * Get the identifier of the density texture.
	 * @return the density texture identifier.
	 */
	unsigned int getDensityTexture() const;

	/**
	 * Get the identifier of the density force texture.
	 * @return the density force texture identifier.
	 */
	unsigned int getDensityForceTexture() const;

	/**
	 * Get the identifier of the velocity texture.
	 * @return the velocity texture identifier.
	 */
	unsigned int getVelocityTexture() const;

	/**
	 * Get the identifier of the velocity force texture.
	 * @return the velocity force texture identifier.
	 */
	unsigned int getVelocityForceTexture() const;

private:
	void doRun(float timestep);
	void doReset() { clearTextures(); }

	void clearTextures();

	void updateDiffusion(float diffusion);
	void updateViscosity(float viscosity);
	void updateVorticity(float vorticity);
	void updateJacobiSteps(unsigned int jacobiSteps);	

	std::auto_ptr<OpenGL::Flat3dTextureSet> textures;

	std::auto_ptr<Fluid3dSolver::Advection> velocityAdvect, densityAdvect;
	std::auto_ptr<Fluid3dSolver::Projection> project;
	std::auto_ptr<Fluid3dSolver::Diffusion> velocityDiffuse, densityDiffuse;
	std::auto_ptr<Fluid3dSolver::ForceAdder> densityForceAdder, velocityForceAdder, vorticityForceAdder;
	std::auto_ptr<Fluid3dSolver::VorticityConfinement> vorticity;

	unsigned int densityTexture;
	unsigned int densityForceTexture;
	unsigned int velocityTexture;
	unsigned int velocityForceTexture;	
	unsigned int vorticityForceTexture;
};

}

#endif