#include <cassert>

#include "Fluid3dSolver.h"

#include "opengl/CgContext.h"
#include "opengl/CgProgram.h"
#include "opengl/Flat3dTextureOperation.h"
#include "opengl/Flat3dTextureSet.h"

#include "util/Util.h"

namespace OpenGL {
	class IoTextureSet;
}

using namespace OpenGL;

namespace Fluid3dSolver {

namespace {

std::auto_ptr<OpenGL::CgProgram>
createFragmentProgram(const std::string &entry,
                      OpenGL::IoTextureSet *textures,
                      bool slice)
{
	const OpenGL::CgContext &context = OpenGL::CgContext::getInstance();

	return std::auto_ptr<OpenGL::CgProgram>(
		new OpenGL::CgProgram(context.getCgProgramPath() + (slice ? "fluid2d.cg" : "fluid3d.cg"), 
		                      context.getFragmentShader(), 
		                      entry, 
		                      textures));
}

std::auto_ptr<OpenGL::Flat3dTextureOperation>
createFlat3dTextureOperation(const OpenGL::Flat3dTextureSet *textures,
                             OpenGL::CgProgram *interior,
                             OpenGL::CgProgram *boundary,
                             bool slice)
{
	return std::auto_ptr<OpenGL::Flat3dTextureOperation>(
		new OpenGL::Flat3dTextureOperation(textures->getComputationWidth(),
		                                   textures->getComputationHeight(),
		                                   textures->getLayoutWidth(),
		                                   textures->getLayoutHeight(),
		                                   textures->getComputationDepth(),
		                                   interior,
		                                   boundary,
										   slice ? Flat3dTextureOperation::INTERIOR_2D :
		                                           Flat3dTextureOperation::INTERIOR_3D));
}


std::auto_ptr<OpenGL::Flat3dTextureOperation>
createFlat3dTextureOperation(const OpenGL::Flat3dTextureSet *textures,
                             OpenGL::CgProgram *interior,
                             bool slice)
{
	return std::auto_ptr<OpenGL::Flat3dTextureOperation>(
		new OpenGL::Flat3dTextureOperation(textures->getComputationWidth(),
		                                   textures->getComputationHeight(),
		                                   textures->getLayoutWidth(),
		                                   textures->getLayoutHeight(),
		                                   textures->getComputationDepth(),
		                                   interior,
										   slice ? Flat3dTextureOperation::INTERIOR_2D :
		                                           Flat3dTextureOperation::INTERIOR_3D));
}

}

std::auto_ptr<OpenGL::CgProgram> 
createBoundaryProgram(BoundaryType boundaryType,
					  OpenGL::Flat3dTextureSet *textures,
                      bool slice)
{
	std::auto_ptr<OpenGL::CgProgram> program;

	std::string boundary = slice ? "boundary2d" : "boundary3d";
	std::string velocityBoundary = slice ? "boundaryVelocity2d" : "boundaryVelocity3d";
	std::string windBoundary = slice ? "boundaryWindSlice" : "boundaryWind3d";

	switch (boundaryType) {
	case DENSITY_CLOSED:
		program = createFragmentProgram(boundary, textures, slice);
		program->setParameter("scale", 1.0f);
		break;
	case NEUMANN:
		program = createFragmentProgram(boundary, textures, slice);
		program->setParameter("scale", 1.0f);
		break;
	case DENSITY_OPEN: {
		program = createFragmentProgram("boundaryOpen", textures, slice);
		float v[4] = {0};
		program->setParameter("v", v);
		break; }
	case VELOCITY_ROUGH:
		program = createFragmentProgram(velocityBoundary, textures, slice);
		program->setParameter("normScale", -1.0f);
		program->setParameter("perpScale", -1.0f);
		break;
	case VELOCITY_CONCRETE:
		program = createFragmentProgram(velocityBoundary, textures, slice);
		program->setParameter("normScale", -1.0f);
		program->setParameter("perpScale", 0.0f);
		break;
	case VELOCITY_SMOOTH:
		program = createFragmentProgram(velocityBoundary, textures, slice);
		program->setParameter("normScale", -1.0f);
		program->setParameter("perpScale", 1.0f);
		break;
	case VELOCITY_OPEN:
		program = createFragmentProgram(velocityBoundary, textures, slice);
		program->setParameter("normScale", 0.0f);
		program->setParameter("perpScale", 0.0f);
		break;
	case VELOCITY_WIND:
		program = createFragmentProgram(windBoundary, textures, slice);
		break;
	default:
		assert(false);
	}

	if (!slice && boundaryType != VELOCITY_WIND && boundaryType != DENSITY_OPEN) {
		program->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		program->setTextureParameter("domainLookup", textures->getDomainLookup());
	}

	if (slice && boundaryType == VELOCITY_WIND) {
		program->setTextureParameter("domainLookup", textures->getDomainLookup());
	}

	return program;
}

Advection::~Advection()
{
}

void Advection::run(float timestep)
{
	doRun(timestep);
}

void Advection::setVoxelSize(float voxelSize)
{
	doSetVoxelSize(voxelSize);
}

void Advection::setUniformAdvection(const Math::Vector3d &uniformAdvection)
{
	doSetUniformAdvection(uniformAdvection);
}

Diffusion::~Diffusion()
{
}

void Diffusion::run(float timestep)
{
	doRun(timestep);
}

void Diffusion::setDiffusionRates(float r, float g, float b, float a)
{
	doSetDiffusionRates(r, g, b, a);
}

void Diffusion::setNumJacobiSteps(unsigned int numJacobiSteps)
{
	doSetNumJacobiSteps(numJacobiSteps);
}

void Diffusion::setVoxelSize(float voxelSize)
{
	doSetVoxelSize(voxelSize);
}

ForceAdder::~ForceAdder()
{
}

void ForceAdder::run(float timestep)
{
	doRun(timestep);
}

Projection::~Projection()
{
}

void Projection::run()
{
	doRun();
}

void Projection::setNumIterations(unsigned int numIterations)
{
	doSetNumIterations(numIterations);
}

void Projection::setVoxelSize(float voxelSize)
{
	doSetVoxelSize(voxelSize);
}

unsigned int Projection::getDivergenceTexture() const
{
	return doGetDivergenceTexture();
}

void Projection::setDivergenceForceTexture(unsigned int textureId)
{
	doSetDivergenceForceTexture(textureId);
}

VorticityConfinement::~VorticityConfinement()
{
}

void VorticityConfinement::run()
{
	doRun();
}

void VorticityConfinement::setVorticityScale(float epsilon)
{
	doSetVorticityScale(epsilon);
}

void VorticityConfinement::setVoxelSize(float voxelSize)
{
	doSetVoxelSize(voxelSize);
}

AdvectionImpl::AdvectionImpl(OpenGL::Flat3dTextureSet *textures,
                             unsigned int velocityTexture, unsigned int valueTexture, 
                             BoundaryType boundaryType,
							 bool slice) :
	slice(slice)
{
	cgAdvect = createFragmentProgram(slice ? "advectSlice" : "advect3d", textures, slice);

	timestepParameter = cgAdvect->getParameterId("timestep");

	if (!slice)
		uniformAdvectParameter = cgAdvect->getParameterId("uniformAdvect");

	voxelSizeParameter = cgAdvect->getParameterId("voxelSize");

	cgAdvect->setTextureParameter("velocity", velocityTexture);
	cgAdvect->setTextureParameter("value", valueTexture);

	cgAdvect->setTextureParameter("domainLookup", textures->getDomainLookup());

	cgAdvect->setParameter("width", textures->getComputationWidth());
	cgAdvect->setParameter("height", textures->getComputationHeight());

	if (!slice) {
		cgAdvect->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgAdvect->setParameter("depth", textures->getComputationDepth());
	}

	cgBoundary = createBoundaryProgram(boundaryType, textures, slice);
	cgBoundary->setTextureParameter("value", valueTexture);

	advectOp = createFlat3dTextureOperation(textures, cgAdvect.get(), cgBoundary.get(), slice);
	advectOp->setOutputTexture(valueTexture);
}

AdvectionImpl::AdvectionImpl(OpenGL::Flat3dTextureSet *textures,
                             unsigned int velocityTexture, unsigned int valueTexture, 
                             CgProgram *cgBoundary,
							 bool slice) :
	slice(slice)
{
	cgAdvect = createFragmentProgram(slice ? "advectSlice" : "advect3d", textures, slice);

	timestepParameter = cgAdvect->getParameterId("timestep");

	if (!slice)
		uniformAdvectParameter = cgAdvect->getParameterId("uniformAdvect");

	voxelSizeParameter = cgAdvect->getParameterId("voxelSize");

	cgAdvect->setTextureParameter("velocity", velocityTexture);
	cgAdvect->setTextureParameter("value", valueTexture);

	cgAdvect->setTextureParameter("domainLookup", textures->getDomainLookup());

	cgAdvect->setParameter("width", textures->getComputationWidth());
	cgAdvect->setParameter("height", textures->getComputationHeight());

	if (!slice) {
		cgAdvect->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgAdvect->setParameter("depth", textures->getComputationDepth());
	}

	advectOp = createFlat3dTextureOperation(textures, cgAdvect.get(), cgBoundary, slice);
	advectOp->setOutputTexture(valueTexture);
}

void AdvectionImpl::doSetVoxelSize(float voxelSize)
{
	cgAdvect->setParameter(voxelSizeParameter, voxelSize);
}

void AdvectionImpl::doSetUniformAdvection(const Math::Vector3d &uniformAdvection)
{
	if (!slice) {
		float uniformAdvect[4] = { uniformAdvection.x, uniformAdvection.y, uniformAdvection.z, 0.0f };
		cgAdvect->setParameter(uniformAdvectParameter, uniformAdvect);
	}
}

void AdvectionImpl::doRun(float timestep)
{
	cgAdvect->setParameter(timestepParameter, timestep);
	advectOp->compute();
}


DiffusionImpl::DiffusionImpl(OpenGL::Flat3dTextureSet *textures,
                             unsigned int valueTexture, unsigned int tempTexture,
                             BoundaryType boundaryType,
                             bool slice) : 
	textures(textures),
	valueTexture(valueTexture),
	tempTexture(tempTexture),
	numJacobiSteps(20)
{
	initialize(boundaryType, slice);
}

DiffusionImpl::DiffusionImpl(OpenGL::Flat3dTextureSet *textures,
					         unsigned int valueTexture, BoundaryType boundaryType,
					         bool slice) :
	textures(textures),
	valueTexture(valueTexture),
	numJacobiSteps(20),
	slice(slice)
{
	tempTexture = textures->addPrimaryTexture("DiffusionImpl temp texture");

	initialize(boundaryType, slice);
}

void DiffusionImpl::doSetDiffusionRates(float r, float g, float b, float a)
{
	diffusionRates[0] = r;
	diffusionRates[1] = g;
	diffusionRates[2] = b;
	diffusionRates[3] = a;
}

void DiffusionImpl::doSetVoxelSize(float voxelSize)
{
	this->voxelSize = voxelSize;
}

void DiffusionImpl::doSetNumJacobiSteps(unsigned int numJacobiSteps)
{
	this->numJacobiSteps = numJacobiSteps;
}

void DiffusionImpl::doRun(float timestep)
{
	float alpha[4], rBeta[4];

	float neighbors = slice ? 4.0f : 6.0f;

	for (unsigned int i = 0; i < 4; ++i) {
		alpha[i] = diffusionRates[i] * timestep / (voxelSize * voxelSize * (slice ? 1.0f : voxelSize));
		rBeta[i] = 1.0f / (1.0f + neighbors * alpha[i]);
	}

	cgDiffuse->setParameter(alphaParameter, alpha);
	cgDiffuse->setParameter(rBetaParameter, rBeta);

	if (numJacobiSteps > 0) {
		textures->clearTexture(tempTexture);

		diffuseOp->setOutputTexture(tempTexture);
		for (unsigned int i = 0; i < numJacobiSteps - 1; ++i)
			diffuseOp->compute();

		diffuseOp->setOutputTexture(valueTexture);
		diffuseOp->compute();
	}
}

void DiffusionImpl::initialize(BoundaryType boundaryType, bool slice) 
{
	cgDiffuse = createFragmentProgram(slice ? "jacobi2d" : "jacobi3d", textures, slice);

	alphaParameter = cgDiffuse->getParameterId("alpha");
	rBetaParameter = cgDiffuse->getParameterId("rBeta");

	cgDiffuse->setTextureParameter("x", tempTexture);
	cgDiffuse->setTextureParameter("b", valueTexture);

	if (!slice) {
		cgDiffuse->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgDiffuse->setTextureParameter("domainLookup", textures->getDomainLookup());
		cgDiffuse->setTextureParameter("neighborLookup", textures->getNeighborLookup());
	}

	cgBoundary = createBoundaryProgram(boundaryType, textures, slice);
	cgBoundary->setTextureParameter("value", valueTexture);

	diffuseOp = createFlat3dTextureOperation(textures, cgDiffuse.get(), cgBoundary.get(), slice);
	diffuseOp->setOutputTexture(tempTexture);
}

ProjectionImpl::ProjectionImpl(OpenGL::Flat3dTextureSet *textures,
                               unsigned int velocityTexture,
                               BoundaryType velocityBoundary,
                               bool slice) :
	cgVelocityBoundary(createBoundaryProgram(velocityBoundary, textures, slice)),
	iterations(20),
	pressureTexture(textures->addPrimaryTexture("Pressure")),
	textures(textures),
	addDivergenceForce(false),
	slice(slice)
{
	divergenceTexture = textures->addPrimaryTexture("Divergence");

	cgCalculateDivergence = 
		createFragmentProgram(slice ? "divergence2d" : "divergence3d", textures, slice);

	cgPressureSolver = 
		createFragmentProgram(slice ? "jacobi2d" : "jacobi3d", textures, slice);

	cgSubtractPressureGradient = 
		createFragmentProgram(slice ? "subtractGradient2d" : "subtractGradient3d", textures, slice);

	cgDivergenceBoundary = createBoundaryProgram(NEUMANN, textures, slice);
	cgPressureBoundary = createBoundaryProgram(NEUMANN, textures, slice);

	if (!slice) {
		cgSubtractPressureGradient->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgSubtractPressureGradient->setTextureParameter("domainLookup", textures->getDomainLookup());

		cgCalculateDivergence->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgCalculateDivergence->setTextureParameter("domainLookup", textures->getDomainLookup());

		cgPressureSolver->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgPressureSolver->setTextureParameter("domainLookup", textures->getDomainLookup());
		cgPressureSolver->setTextureParameter("neighborLookup", textures->getNeighborLookup());
	}

	calculateDivergenceOperation =
		createFlat3dTextureOperation(textures,
		                             cgCalculateDivergence.get(),
		                             cgDivergenceBoundary.get(),
		                             slice);

	solvePressureOperation =
		createFlat3dTextureOperation(textures,
		                             cgPressureSolver.get(),
		                             cgPressureBoundary.get(),
		                             slice);

	subtractPressureGradientOperation =
		createFlat3dTextureOperation(textures,
		                             cgSubtractPressureGradient.get(),
		                             cgVelocityBoundary.get(),
		                             slice);

	cgCalculateDivergence->setTextureParameter("velocity", velocityTexture);

	cgPressureSolver->setTextureParameter("x", pressureTexture);
	cgPressureSolver->setTextureParameter("b", divergenceTexture);

	float rBetaComponent = slice ? 0.25f : 1.0f / 6.0f;

	float alpha[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	float rBeta[4] = {rBetaComponent, rBetaComponent, rBetaComponent, rBetaComponent};

	cgPressureSolver->setParameter("alpha", alpha);
	cgPressureSolver->setParameter("rBeta", rBeta);

	cgSubtractPressureGradient->setTextureParameter("pressure", pressureTexture);
	cgSubtractPressureGradient->setTextureParameter("velocity", velocityTexture);

	cgDivergenceBoundary->setTextureParameter("value", divergenceTexture);
	cgPressureBoundary->setTextureParameter("value", pressureTexture);
	cgVelocityBoundary->setTextureParameter("value", velocityTexture);

	calculateDivergenceOperation->setOutputTexture(divergenceTexture);
	solvePressureOperation->setOutputTexture(pressureTexture);
	subtractPressureGradientOperation->setOutputTexture(velocityTexture);
}

ProjectionImpl::ProjectionImpl(OpenGL::Flat3dTextureSet *textures,
                               unsigned int velocityTexture,
							   OpenGL::CgProgram *cgVelocityBoundary,
                               bool slice) :
	iterations(20),
	pressureTexture(textures->addPrimaryTexture("Pressure")),
	textures(textures),
	addDivergenceForce(false),
	slice(slice)
{
	divergenceTexture = textures->addPrimaryTexture("Divergence");

	cgCalculateDivergence = 
		createFragmentProgram(slice ? "divergence2d" : "divergence3d", textures, slice);

	cgPressureSolver = 
		createFragmentProgram(slice ? "jacobi2d" : "jacobi3d", textures, slice);

	cgSubtractPressureGradient = 
		createFragmentProgram(slice ? "subtractGradient2d" : "subtractGradient3d", textures, slice);

	cgDivergenceBoundary = createBoundaryProgram(NEUMANN, textures, slice);
	cgPressureBoundary = createBoundaryProgram(NEUMANN, textures, slice);

	if (!slice) {
		cgSubtractPressureGradient->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgSubtractPressureGradient->setTextureParameter("domainLookup", textures->getDomainLookup());

		cgCalculateDivergence->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgCalculateDivergence->setTextureParameter("domainLookup", textures->getDomainLookup());

		cgPressureSolver->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgPressureSolver->setTextureParameter("domainLookup", textures->getDomainLookup());
		cgPressureSolver->setTextureParameter("neighborLookup", textures->getNeighborLookup());
	}

	calculateDivergenceOperation =
		createFlat3dTextureOperation(textures,
		                             cgCalculateDivergence.get(),
		                             cgDivergenceBoundary.get(),
		                             slice);

	solvePressureOperation =
		createFlat3dTextureOperation(textures,
		                             cgPressureSolver.get(),
		                             cgPressureBoundary.get(),
		                             slice);

	subtractPressureGradientOperation =
		createFlat3dTextureOperation(textures,
		                             cgSubtractPressureGradient.get(),
		                             cgVelocityBoundary,
		                             slice);

	cgCalculateDivergence->setTextureParameter("velocity", velocityTexture);

	cgPressureSolver->setTextureParameter("x", pressureTexture);
	cgPressureSolver->setTextureParameter("b", divergenceTexture);

	float rBetaComponent = slice ? 0.25f : 1.0f / 6.0f;

	float alpha[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	float rBeta[4] = {rBetaComponent, rBetaComponent, rBetaComponent, rBetaComponent};

	cgPressureSolver->setParameter("alpha", alpha);
	cgPressureSolver->setParameter("rBeta", rBeta);

	cgSubtractPressureGradient->setTextureParameter("pressure", pressureTexture);
	cgSubtractPressureGradient->setTextureParameter("velocity", velocityTexture);

	cgDivergenceBoundary->setTextureParameter("value", divergenceTexture);
	cgPressureBoundary->setTextureParameter("value", pressureTexture);

	calculateDivergenceOperation->setOutputTexture(divergenceTexture);
	solvePressureOperation->setOutputTexture(pressureTexture);
	subtractPressureGradientOperation->setOutputTexture(velocityTexture);
}

void ProjectionImpl::doSetNumIterations(unsigned int numIterations)
{
	this->iterations = numIterations;
}

void ProjectionImpl::doSetVoxelSize(float voxelSize)
{
	cgCalculateDivergence->setParameter("halfVoxelSize", voxelSize * 0.5f);
	cgSubtractPressureGradient->setParameter("halfVoxelSize", voxelSize * 0.5f);
}

void ProjectionImpl::doRun()
{
	calculateDivergenceOperation->compute();

	if (addDivergenceForce)
		divergenceForceAdder->run(1.0);

	textures->clearTexture(pressureTexture);

	for (unsigned int i = 0; i < iterations; ++i)
		solvePressureOperation->compute();

	subtractPressureGradientOperation->compute();
}

unsigned int ProjectionImpl::doGetDivergenceTexture() const
{
	return divergenceTexture;
}

void ProjectionImpl::doSetDivergenceForceTexture(unsigned int textureId)
{
	divergenceForceTexture = textureId;
	addDivergenceForce = true;

	divergenceForceAdder = Util::autoPtr(new ForceAdderImpl(textures, divergenceTexture, 
		divergenceForceTexture, NEUMANN, slice));
}

ForceAdderImpl::ForceAdderImpl(OpenGL::Flat3dTextureSet *textures,
                               unsigned int valueTexture,
					           unsigned int forceTexture,
	                           BoundaryType boundaryType,
                               bool slice) :
	cgAddForce(createFragmentProgram(slice ? "addForce2d" : "addForce3d", textures, slice)),
	cgBoundary(createBoundaryProgram(boundaryType, textures, slice))
{
	addForceOperation = 
		createFlat3dTextureOperation(textures, 
		                             cgAddForce.get(),
		                             cgBoundary.get(),
		                             slice);

	cgBoundary->setTextureParameter("value", valueTexture);

	cgAddForce->setTextureParameter("force", forceTexture);
	cgAddForce->setTextureParameter("value", valueTexture);

	timestepParameter = cgAddForce->getParameterId("timestep");

	addForceOperation->setOutputTexture(valueTexture);
}

ForceAdderImpl::ForceAdderImpl(OpenGL::Flat3dTextureSet *textures,
                               unsigned int valueTexture,
					           unsigned int forceTexture,
							   OpenGL::CgProgram *cgBoundary,
                               bool slice) :
	cgAddForce(createFragmentProgram(slice ? "addForce2d" : "addForce3d", textures, slice))
{
	addForceOperation = 
		createFlat3dTextureOperation(textures, 
		                             cgAddForce.get(),
		                             cgBoundary,
		                             slice);

	cgAddForce->setTextureParameter("force", forceTexture);
	cgAddForce->setTextureParameter("value", valueTexture);

	timestepParameter = cgAddForce->getParameterId("timestep");

	addForceOperation->setOutputTexture(valueTexture);
}

void ForceAdderImpl::doRun(float timestep)
{
	cgAddForce->setParameter(timestepParameter, timestep);

	addForceOperation->compute();
}


VorticityConfinementImpl::VorticityConfinementImpl(OpenGL::Flat3dTextureSet *textures,
                                                   unsigned int velocityTexture, 
                                                   unsigned int velocityForceTexture,
										           unsigned int vorticityTexture,
                                                   bool slice) :
	textures(textures), velocityTexture(velocityTexture), 
	velocityForceTexture(velocityForceTexture), vorticityTexture(vorticityTexture)
{
	initialize(slice);
}


VorticityConfinementImpl::VorticityConfinementImpl(OpenGL::Flat3dTextureSet *textures,
                                                   unsigned int velocityTexture, 
										           unsigned int velocityForceTexture,
                                                   bool slice) :
	textures(textures), velocityTexture(velocityTexture), 
	velocityForceTexture(velocityForceTexture)
{
	vorticityTexture = textures->addPrimaryTexture("Vorticity");

	initialize(slice);
}


void VorticityConfinementImpl::doSetVorticityScale(float epsilon)
{
	cgAddVorticity->setParameter("epsilon", epsilon);
}


void VorticityConfinementImpl::doSetVoxelSize(float voxelSize)
{
	cgGenerateVorticity->setParameter("halfVoxelSize", voxelSize * 0.5f);
	cgAddVorticity->setParameter("halfVoxelSize", voxelSize * 0.5f);
}


void VorticityConfinementImpl::doRun()
{
	generateVorticityOp->compute();
	addVorticityOp->compute();
}

void VorticityConfinementImpl::initialize(bool slice)
{
	cgGenerateVorticity = 
		createFragmentProgram(slice ? "generateVorticity2d" : "generateVorticity3d", textures, slice);
	
	cgGenerateVorticity->setTextureParameter("velocity", velocityTexture);

	if (!slice) {
		cgGenerateVorticity->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgGenerateVorticity->setTextureParameter("domainLookup", textures->getDomainLookup());
	}

	generateVorticityOp = createFlat3dTextureOperation(textures, cgGenerateVorticity.get(), slice);
	generateVorticityOp->setOutputTexture(vorticityTexture);

	cgAddVorticity = createFragmentProgram(slice ? "addVorticity2d" : "addVorticity3d", textures, slice);

	cgAddVorticity->setTextureParameter("vorticity", vorticityTexture);

	if (!slice) {
		cgAddVorticity->setTextureParameter("offsetLookup", textures->getOffsetLookup());
		cgAddVorticity->setTextureParameter("domainLookup", textures->getDomainLookup());
	}

	addVorticityOp = createFlat3dTextureOperation(textures, cgAddVorticity.get(), slice);
	addVorticityOp->setOutputTexture(velocityForceTexture);
}



}