#ifndef FLUID3DSOLVER_H
#define FLUID3DSOLVER_H

#include <memory>
#include <vector>

#include "math/Vector3d.h"

#include "opengl/OpenGL.h"

namespace OpenGL {
	class CgProgram;
	class Flat3dTextureSet;
	class Flat3dTextureOperation;
}

/**
 * Contains fluid solver operations.
 */
namespace Fluid3dSolver {

/**
 * Boundary types.
 */
enum BoundaryType {
	DENSITY_CLOSED,     /// equal to interior value
	DENSITY_OPEN,       /// boundary set to zero
	NEUMANN,            /// equal to interior value
	VELOCITY_OPEN,      /// velocity equals zero at the boundaries
	VELOCITY_SMOOTH,    /// velocity flows smoothly at the boundaries
	VELOCITY_CONCRETE,  /// velocity slows at the boundaries
	VELOCITY_ROUGH,     /// velocity gets turbulent at the boundaries
	VELOCITY_WIND       /// boundary set to global wind field
};

/** 
 * Creates a boundary program with the given boundary type and texture set.
 * Uses a slice or 3d boundary program.
 * @param boundaryType the boundary type.
 * @param textures the Flat3dTextureSet the boundary belongs to.
 * @param slice whether to create a slice or 3d boundary program.
 * @return an auto_ptr containing the boundary CgProgram.
 */
std::auto_ptr<OpenGL::CgProgram> 
createBoundaryProgram(BoundaryType boundaryType,
					  OpenGL::Flat3dTextureSet *textures,
                      bool slice);

/**
 * Advects (transports) a value field along a velocity field. 
 * The value field and velocity field are each represented by a 
 * flat 3d texture. 
 */
class Advection {
public:
	virtual ~Advection() = 0;

	/**
	 * Sets the voxel size of the computation domain.
	 * @param voxelSize the voxel size to be set.
	 */
	void setVoxelSize(float voxelSize);

	/**
	 * Sets the uniform advection (same for all cells).
	 * @param uniformAdvection the uniform advection.
	 */
	void setUniformAdvection(const Math::Vector3d &uniformAdvection);

	/**
	 * Performs one computation step of the advection.
	 * @param timestep the time step to be used in the computation step.
	 */
	void run(float timestep);

private:
	virtual void doSetVoxelSize(float voxelSize) = 0;
	virtual void doSetUniformAdvection(const Math::Vector3d &uniformAdvection) = 0;
	virtual void doRun(float timestep) = 0;
};

/**
 * Diffuses a value field (represented by a flat 3d texture) by solving a set of linear equations. 
 */
class Diffusion {
public:
	virtual ~Diffusion() = 0;

	/**
	 * Sets the diffusion rates to be used on the different channels.
	 * @param r the diffusion rate used for texture channel red.
	 * @param g the diffusion rate used for texture channel green.
	 * @param b the diffusion rate used for texture channel blue.
	 * @param a the diffusion rate used for texture channel alpha.
	 */
	void setDiffusionRates(float r, float g, float b, float a);

	/**
	 * Sets the number of jacobi steps to be used.
	 * @param numJacobiSteps the number of jacobi steps.
	 */
	void setNumJacobiSteps(unsigned int numJacobiSteps);

	/**
	 * Sets the voxel size of the computation domain.
	 * @param voxelSize the voxel size to be set.
	 */
	void setVoxelSize(float voxelSize);

	/**
	 * Performs one computation step of the diffusion.
	 * @param timestep the time step to be used in the computation step.
	 */
	void run(float timestep);

private:
	virtual void doSetDiffusionRates(float r, float g, float b, float a) = 0;
	virtual void doSetNumJacobiSteps(unsigned int numJacobiSteps) = 0;
	virtual void doSetVoxelSize(float voxelSize) = 0;
	virtual void doRun(float timestep) = 0;
};

/**
 * Adds a force texture to a value texture.
 */
class ForceAdder {
public:
	virtual ~ForceAdder() = 0;

	/**
	 * Runs the ForceAdder over a single timestep.
	 */
	void run(float timestep);

private:
	virtual void doRun(float timestep) = 0;
};

/**
 * Projection is used for making a velocity field mass conserving,
 * by solving a system of linear equations, using an iterative approach.
 */
class Projection {
public:
	virtual ~Projection() = 0;

	/**
	 * Sets the number of iterations to be used when solving the linear system.
	 * @param numIterations the number of iterations.
	 */
	void setNumIterations(unsigned int numIterations);

	/**
	 * Sets the voxel size of the computation domain.
	 * @param voxelSize the voxel size to be set.
	 */
	void setVoxelSize(float voxelSize);

	/**
	 * Runs the projection.	 
	 */
	void run();

	unsigned int getDivergenceTexture() const;

	/**
	 * Sets the divergence force texture. 
	 * This texture is added to the divergence texture 
	 * before solving the Poisson equation in run().
 	 */
	void setDivergenceForceTexture(unsigned int textureId);

private:
	virtual void doSetNumIterations(unsigned int numIterations) = 0;
	virtual void doSetVoxelSize(float voxelSize) = 0;
	virtual void doRun() = 0;

	virtual unsigned int doGetDivergenceTexture() const = 0;
	virtual void doSetDivergenceForceTexture(unsigned int textureId) = 0;
};

/**
 * Adds turbulence to a velocity field by means of vorticity confinement.
 * Writes the resulting vorticity confinement force to a given force texture.
 */
class VorticityConfinement {
public:
	virtual ~VorticityConfinement() = 0;

	/**
	 * Sets the vorticity scale to be used.
	 * @param epsilon the scale of the vorticity confinement.
	 */
	void setVorticityScale(float epsilon);

	/**
	 * Sets the voxel size of the computation domain.
	 * @param voxelSize the voxel size to be set.
	 */
	void setVoxelSize(float voxelSize);

	/**
	 * Performs one computation step of the vorticity confinement.
	 */
	void run();

private:
	virtual void doSetVorticityScale(float epsilon) = 0;
	virtual void doSetVoxelSize(float voxelSize) = 0;
	virtual void doRun() = 0;
};


/**
 * Implementation of Advection.
 */
class AdvectionImpl : public Advection {
public:
	/**
	 * Default constructor.
	 * @param textures the texture set containing the velocity and value texture.
	 * @param velocityTexture the identifier of the velocity texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param boundaryType enum specifying which boundary condition to be used. 
	 * @param slice use independent slices.
	 */
	AdvectionImpl(OpenGL::Flat3dTextureSet *textures,
	              unsigned int velocityTexture, unsigned int valueTexture,
	              BoundaryType boundaryType,
	              bool slice);

	/**
	 * Default constructor.
	 * @param textures the texture set containing the velocity and value texture.
	 * @param velocityTexture the identifier of the velocity texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param cgBoundary the boundary's CgProgram. 
	 * @param slice use independent slices.
	 */
	AdvectionImpl(OpenGL::Flat3dTextureSet *textures,
	              unsigned int velocityTexture, unsigned int valueTexture,
				  OpenGL::CgProgram *cgBoundary,
	              bool slice);

private:
	void doSetVoxelSize(float voxelSize);
	void doSetUniformAdvection(const Math::Vector3d &uniformAdvection);
	void doRun(float timestep);

	std::auto_ptr<OpenGL::CgProgram> cgAdvect, cgBoundary;
	std::auto_ptr<OpenGL::Flat3dTextureOperation> advectOp;

	bool slice;

	unsigned int timestepParameter;
	unsigned int uniformAdvectParameter;
	unsigned int voxelSizeParameter;
};

/**
 * Implementation of Diffusion.
 */
class DiffusionImpl : public Diffusion {
public:
	/**
	 * Default constructor.
	 * @param textures the texture set containing the value and temporary texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param tempTexture the identifier of the temporary texture used when solving 
	 *                    the set of linear equations.
	 * @param boundaryType enum specifying which boundary condition to be used. 	
	 * @param slice use independent slices.
	 */
	DiffusionImpl(OpenGL::Flat3dTextureSet *textures,
	              unsigned int valueTexture, unsigned int tempTexture,
	              BoundaryType boundaryType,
	              bool slice);

	/**
	 * Constructor where the temporary texture is created and added to the texture set.
	 * @param textures the texture set containing the value texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param boundaryType enum specifying which boundary condition to be used. 	
	 * @param slice use independent slices.
	 */
	DiffusionImpl(OpenGL::Flat3dTextureSet *textures,
	              unsigned int valueTexture, BoundaryType boundaryType,
	              bool slice);

private:
	void doSetDiffusionRates(float r, float g, float b, float a);
	void doSetNumJacobiSteps(unsigned int numJacobiSteps);
	void doSetVoxelSize(float voxelSize);
	void doRun(float timestep);

	void initialize(BoundaryType boundaryType, bool slice);
	
	std::auto_ptr<OpenGL::CgProgram> cgDiffuse, cgBoundary;
	std::auto_ptr<OpenGL::Flat3dTextureOperation> diffuseOp;
	OpenGL::Flat3dTextureSet *textures;

	unsigned int alphaParameter, rBetaParameter;
	float diffusionRates[4];
	float voxelSize;

	unsigned int valueTexture, tempTexture;
	unsigned int numJacobiSteps;

	float slice;
};

/**
 * Implementation of Projection.
 */
class ProjectionImpl : public Projection {
public:
	/**
	 * Initializes Projection.
	 * @param textures The texture set.
	 * @param velocityTexture The velocity texture.
	 * @param velocityBoundary The velocity boundary conditions.
	 * @param slice use independent slices.
	 */
	ProjectionImpl(OpenGL::Flat3dTextureSet *textures,
	               unsigned int velocityTexture,
			       BoundaryType velocityBoundary,
			       bool slice);

	/**
	 * Initializes Projection.
	 * @param textures The texture set.
	 * @param velocityTexture The velocity texture.
	 * @param cgVelocityBoundary The velocity boundary's CgProgram.
	 * @param slice use independent slices.
	 */
	ProjectionImpl(OpenGL::Flat3dTextureSet *textures,
	               unsigned int velocityTexture,
				   OpenGL::CgProgram *cgVelocityBoundary,
			       bool slice);

private:
	void doSetNumIterations(unsigned int numIterations);
	void doSetVoxelSize(float voxelSize);
	void doRun();

	unsigned int doGetDivergenceTexture() const;
	void doSetDivergenceForceTexture(unsigned int textureId);

	std::auto_ptr<OpenGL::CgProgram> 
		cgCalculateDivergence, cgDivergenceBoundary, cgPressureSolver, 
		cgPressureBoundary, cgSubtractPressureGradient, cgVelocityBoundary;

	std::auto_ptr<OpenGL::Flat3dTextureOperation> 
		calculateDivergenceOperation, solvePressureOperation,
		subtractPressureGradientOperation;

	unsigned int iterations;

	unsigned int divergenceTexture, pressureTexture;

	unsigned int divergenceForceTexture;
	bool addDivergenceForce;
	std::auto_ptr<Fluid3dSolver::ForceAdder> divergenceForceAdder;
	bool slice;

	OpenGL::Flat3dTextureSet *textures;
};

/**
 * Implementation of ForceAdder.
 */
class ForceAdderImpl : public ForceAdder {
public:
	/**
	 * Initializes the ForceAdder.
	 * @param textures The texture set.
	 * @param valueTexture The value texture.
	 * @param forceTexture The force texture.
	 * @param boundaryType The boundary condition.
	 * @param slice use independent slices.
	 */
	ForceAdderImpl(OpenGL::Flat3dTextureSet *textures,
	               unsigned int valueTexture,
                   unsigned int forceTexture,
			       BoundaryType boundaryType,
			       bool slice);

	/**
	 * Initializes the ForceAdder.
	 * @param textures The texture set.
	 * @param valueTexture The value texture.
	 * @param forceTexture The force texture.
	 * @param cgBoundary The boundary's CgProgram.
	 * @param slice use independent slices.
	 */
	ForceAdderImpl(OpenGL::Flat3dTextureSet *textures,
	               unsigned int valueTexture,
                   unsigned int forceTexture,
				   OpenGL::CgProgram *cgBoundary,
			       bool slice);

private:
	void doRun(float timestep);

	std::auto_ptr<OpenGL::CgProgram> cgAddForce, cgBoundary;

	std::auto_ptr<OpenGL::Flat3dTextureOperation> addForceOperation;

	unsigned int timestepParameter;
};

/**
 * Implementation of VorticityConfinent.
 */
class VorticityConfinementImpl : public VorticityConfinement {
public:
	/**
	 * Default constructor.
	 * @param textures the texture set containing the velocity, velocity force, and vorticity texture.
	 * @param velocityTexture the identifier of the velocity texture.	
	 * @param velocityForceTexture the identifier of the velocity force texture.
	 * @param vorticityTexture the identifier of the temporary vorticity texture used. 
	 * @param slice use independent slices.
	 */
	VorticityConfinementImpl(OpenGL::Flat3dTextureSet *textures,
	                         unsigned int velocityTexture, 
			                 unsigned int velocityForceTexture,
			                 unsigned int vorticityTexture,
						     bool slice);

	/**
	 * Constructor where the vorticity texture is created and added to the texture set.
	 * @param textures the texture set containing the velocity and velocity force texture.
	 * @param velocityTexture the identifier of the velocity texture.	
	 * @param velocityForceTexture the identifier of the velocity force texture.  
	 * @param slice use independent slices.
	 */
	VorticityConfinementImpl(OpenGL::Flat3dTextureSet *textures,
	                     unsigned int velocityTexture, 
			             unsigned int velocityForceTexture,
						 bool slice);

private:
	void initialize(bool slice);

	void doSetVorticityScale(float epsilon);
	void doSetVoxelSize(float voxelSize);
	void doRun();

	std::auto_ptr<OpenGL::CgProgram> cgGenerateVorticity, cgAddVorticity;
	std::auto_ptr<OpenGL::Flat3dTextureOperation> generateVorticityOp, addVorticityOp;
	OpenGL::Flat3dTextureSet *textures;

	unsigned int velocityTexture, velocityForceTexture, vorticityTexture;
};

}

#endif