#ifndef FLUID3DSOLVER_FLUIDFACTORY_H
#define FLUID3DSOLVER_FLUIDFACTORY_H

#include <memory>

#include "Fluid3dSolver.h"

namespace Fluid3dSolver {

/**
 * The FluidFactory class is an abstract factory used to create various
 * abstract Fluid3dSolver objects.
 */
class FluidFactory {
public:
	/**
	 * Check if the fluid factory is for slice or 3d fluid simulations.
	 * @return if the fluid factory generates slice or 3d fluid operations.
	 */
	virtual bool isSliceSimulation() const = 0;

	/**
	 * Creates an Advection object.	 	 
	 * @param textures the texture set containing the velocity and value texture.
	 * @param velocityTexture the identifier of the velocity texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param boundaryType enum specifying which boundary condition to be used. 	 
	 * @return an auto pointer to the Advection object.
	 */
	virtual std::auto_ptr<Advection> createAdvection(OpenGL::Flat3dTextureSet *textures,
	                                                 unsigned int velocityTexture, 
	                                                 unsigned int valueTexture,
	                                                 BoundaryType boundaryType) = 0;

	/**
	 * Creates an Advection object.	 	 
	 * @param textures the texture set containing the velocity and value texture.
	 * @param velocityTexture the identifier of the velocity texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param cgBoundary the boundary's CgProgram. 	 
	 * @return an auto pointer to the Advection object.
	 */
	virtual std::auto_ptr<Advection> createAdvection(OpenGL::Flat3dTextureSet *textures,
	                                                 unsigned int velocityTexture, 
	                                                 unsigned int valueTexture,
													 OpenGL::CgProgram *cgBoundary) = 0;


	/**
	 * Creates a Diffusion object.
	 * @param textures the texture set containing the value and temporary texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param tempTexture the identifier of the temporary texture used when solving 
	 *                    the set of linear equations.
	 * @param boundaryType enum specifying which boundary condition to be used. 	
	 * @return an auto pointer to the Diffusion object.
	 */
	virtual std::auto_ptr<Diffusion> createDiffusion(OpenGL::Flat3dTextureSet *textures,
	                                                 unsigned int valueTexture, 
	                                                 unsigned int tempTexture,
	                                                 BoundaryType boundaryType) = 0;

	/**
	 * Creates a Diffusion object where the temporary texture is created and added to the texture set.
	 * @param textures the texture set containing the value texture.
	 * @param valueTexture the identifier of the value texture.	
	 * @param boundaryType enum specifying which boundary condition to be used. 	
	 * @return an auto pointer to the Diffusion object.
	 */
	virtual std::auto_ptr<Diffusion> createDiffusion(OpenGL::Flat3dTextureSet *textures,
	                                                 unsigned int valueTexture,
													 BoundaryType boundaryType) = 0;

	/**
	 * Creates a ForceAdder object.
	 * @param textures the texture set containing the value and force texture.
	 * @param valueTexture the identifier of the value texture.
	 * @param forceTexture the identifier of the force texture.
	 * @param boundaryType enum specifying which boundary condition to be used.
	 * @return an auto pointer to the ForceAdder object.
	 */
	virtual std::auto_ptr<ForceAdder> createForceAdder(OpenGL::Flat3dTextureSet *textures,
	                                                   unsigned int valueTexture,
	                                                   unsigned int forceTexture,
	                                                   BoundaryType boundaryType) = 0;

	/**
	 * Creates a ForceAdder object.
	 * @param textures the texture set containing the value and force texture.
	 * @param valueTexture the identifier of the value texture.
	 * @param forceTexture the identifier of the force texture.
	 * @param cgBoundary the boundary's CgProgram.
	 * @return an auto pointer to the ForceAdder object.
	 */
	virtual std::auto_ptr<ForceAdder> createForceAdder(OpenGL::Flat3dTextureSet *textures,
	                                                   unsigned int valueTexture,
	                                                   unsigned int forceTexture,
													   OpenGL::CgProgram *cgBoundary) = 0;

	/**
	 * Creates a Projection object.
	 * @param textures the texture set containing the velocity texture. 
	 * @param velocityTexture the identifier of the velocity texture.
	 * @param velocityBoundary enum specifying which boundary condition to be used.
	 * @return an auto pointer to the Projection object.
	 */
	virtual std::auto_ptr<Projection> createProjection(OpenGL::Flat3dTextureSet *textures,
	                                                   unsigned int velocityTexture,
	                                                   BoundaryType velocityBoundary) = 0;

	/**
	 * Creates a Projection object.
	 * @param textures the texture set containing the velocity texture. 
	 * @param velocityTexture the identifier of the velocity texture.
	 * @param cgVelocityBoundary the velocity boundary's CgProgram.
	 * @return an auto pointer to the Projection object.
	 */
	virtual std::auto_ptr<Projection> createProjection(OpenGL::Flat3dTextureSet *textures,
	                                                   unsigned int velocityTexture,
													   OpenGL::CgProgram *cgVelocityBoundary) = 0;


	/**
	 * Creates a VorticityConfinement object.
	 * @param textures the texture set containing the velocity, velocity force, and vorticity texture.
	 * @param velocityTexture the identifier of the velocity texture.	
	 * @param velocityForceTexture the identifier of the velocity force texture.
	 * @param vorticityTexture the identifier of the temporary vorticity texture used. 
	 * @return an auto pointer to the VorticityConfinement object.
	 */
	virtual std::auto_ptr<VorticityConfinement> createVorticityConfinement(OpenGL::Flat3dTextureSet *textures,
	                                                                       unsigned int velocityTexture, 
	                                                                       unsigned int velocityForceTexture,
	                                                                       unsigned int vorticityTexture) = 0;

	/**
	 * Creates a VorticityConfinement object where the vorticity texture is created and added to the texture set.
	 * @param textures the texture set containing the velocity and velocity force texture.
	 * @param velocityTexture the identifier of the velocity texture.	
	 * @param velocityForceTexture the identifier of the velocity force texture.  
	 * @return an auto pointer to the VorticityConfinement object.
	 */
	virtual std::auto_ptr<VorticityConfinement> createVorticityConfinement(OpenGL::Flat3dTextureSet *textures,
	                                                                       unsigned int velocityTexture, 
	                                                                       unsigned int velocityForceTexture) = 0;
};

/**
 * An implementation of FluidFactory which can separate between full 3d simulation 
 * and slice (volumetric extrusion) simulation.
 */
class FluidFactoryImpl : public FluidFactory {
public:
	/** 
	 * Constructs a FluidFactoryImpl for either a full 3d simulation
	 * or a slice (volumetric extrusion) simulation.
	 * @param slice if the FluidFactoryImpl should be used with a slice simulation.
	 */
	FluidFactoryImpl(bool slice);

	bool isSliceSimulation() const;

	std::auto_ptr<Advection> createAdvection(OpenGL::Flat3dTextureSet *textures,
	                                         unsigned int velocityTexture, 
	                                         unsigned int valueTexture,
	                                         BoundaryType boundaryType);

	std::auto_ptr<Advection> createAdvection(OpenGL::Flat3dTextureSet *textures,
	                                         unsigned int velocityTexture, 
	                                         unsigned int valueTexture,
											 OpenGL::CgProgram *cgBoundary);

	std::auto_ptr<Diffusion> createDiffusion(OpenGL::Flat3dTextureSet *textures,
	                                         unsigned int valueTexture, 
	                                         unsigned int tempTexture,
	                                         BoundaryType boundaryType);

	std::auto_ptr<Diffusion> createDiffusion(OpenGL::Flat3dTextureSet *textures,
	                                         unsigned int valueTexture, 	                                         
	                                         BoundaryType boundaryType);

	std::auto_ptr<ForceAdder> createForceAdder(OpenGL::Flat3dTextureSet *textures,
	                                           unsigned int valueTexture,
	                                           unsigned int forceTexture,
	                                           BoundaryType boundaryType);

	std::auto_ptr<ForceAdder> createForceAdder(OpenGL::Flat3dTextureSet *textures,
	                                           unsigned int valueTexture,
	                                           unsigned int forceTexture,
											   OpenGL::CgProgram *cgBoundary);

	std::auto_ptr<Projection> createProjection(OpenGL::Flat3dTextureSet *textures,
	                                           unsigned int velocityTexture,
	                                           BoundaryType velocityBoundary);

	std::auto_ptr<Projection> createProjection(OpenGL::Flat3dTextureSet *textures,
	                                           unsigned int velocityTexture,
											   OpenGL::CgProgram *cgVelocityBoundary);

	std::auto_ptr<VorticityConfinement> 
		createVorticityConfinement(OpenGL::Flat3dTextureSet *textures,
		                           unsigned int velocityTexture, 
		                           unsigned int velocityForceTexture,
		                           unsigned int vorticityTexture);

	std::auto_ptr<VorticityConfinement> 
		createVorticityConfinement(OpenGL::Flat3dTextureSet *textures,
		                           unsigned int velocityTexture, 
		                           unsigned int velocityForceTexture);

private:
	bool slice;
};

}

#endif