#ifndef GPUTHESIS_FLUIDPARAMETERS_H
#define GPUTHESIS_FLUIDPARAMETERS_H

namespace GpuThesis {

/**
 * Contains all the parameters governing a fluid simulation.
 */
class FluidParameters {
public:
	/**
	 * Diffusion is the rate at which the fluid density diffuses.
	 * @return the diffusion.
	 */
	float getDiffusion() const;

	/**
	 * Set the diffusion.
	 * @see getDiffusion
	 * @param diffusion the new diffusion
	 */
	void setDiffusion(float diffusion);

	/**
	 * Viscosity is similar to diffusion, but for the velocity field.
	 * Higher viscosity makes the fluid feel more thick, like syrup, glue, etc.
	 * @return the viscosity.
	 */
	float getViscosity() const;

	/**
	 * Set the viscosity.
	 * @see getViscosity
	 * @param viscosity the new viscosity.
	 */
	void setViscosity(float viscosity);

	/**
	 * Vorticity is the rate at which vortices in the velocity field are enhanced.
	 * Higher vorticity causes more turbulence.
	 * @return the vorticity.
	 */
	float getVorticity() const;

	/**
	 * Set the vorticity.
	 * @see getVorticity
	 * @param vorticity the new vorticity.
	 */
	void setVorticity(float vorticity);

	/**
	 * Jacobi steps controls how many steps are used in the diffusion and projection solvers.
	 * The more Jacobi steps the more accurate diffusion and projection.
	 * @return the number of jacobi steps.
	 */
	unsigned int getJacobiSteps() const;

	/**
	 * Set the number of jacobi steps.
	 * @see getJacobiSteps
	 * @param jacobiSteps the new number of jacobi steps.
	 */
	void setJacobiSteps(unsigned int jacobiSteps);

private:
	virtual void updateDiffusion(float diffusion) = 0;
	virtual void updateViscosity(float viscosity) = 0;
	virtual void updateVorticity(float vorticity) = 0;
	virtual void updateJacobiSteps(unsigned int jacobiSteps) = 0;	

	float diffusion;
	float viscosity;
	float vorticity;
	unsigned int jacobiSteps;
};

}

#endif