#ifndef GPUTHESIS_MATERIAL_H
#define GPUTHESIS_MATERIAL_H

#include "EngineParameters.h"
#include "opengl/OpenGL.h"

namespace GpuThesis {

/**
 * Material properties for a surface.
 */
class Material {
public:
	Material();

	/**
	 * Initialize the material with the given parameters.
	 * @param material the material parameters.
	 */
	Material(EngineParameters::MaterialParams material);

	/**
	 * Get the ambient coefficient of the material.
	 * @return the ambient coefficient.
	 */
	float getKa() const;

	/**
	 * Set the ambient coefficient of the material.
	 * @param ka the new ambient coefficient.
	 */
	void setKa(float ka);

	/**
	 * Get the diffuse coefficient of the material.
	 * @return the diffuse coefficient.
	 */
	float getKd() const;

	/**
	 * Set the diffuse coefficient of the material.
	 * @param kd the new diffuse coefficient.
	 */
	void setKd(float kd);

	/**
	 * Get the specular coefficient of the material.
	 * @return the specular coefficient.
	 */
	float getKs() const;

	/**
	 * Set the specular coefficient of the material.
	 * @param ks the new specular coefficient.
	 */
	void setKs(float ks);

	/**
	 * Get the shininess of the material.
	 * Higher shininess causes sharper highlights.
	 * @return the shininess.
	 */
	float getShininess() const;

	/**
	 * Set the shininess of the material.
	 * @see getShininess
	 * @param shininess the new shininess.
	 */
	void setShininess(float shininess);

	/**
	 * Get the texture id of the material.
	 * @return the OpenGL texture id.
	 */
	GLuint getTexture() const;

	/**
	 * Set the texture id of the material.
	 * @return the OpenGL texture id.
	 */
	void setTexture(GLuint texture);

	/**
	 * Check if the material has a texture.
	 * @return whether the material has a texture.
	 */
	bool getHasTexture() const;

	/** 
	 * Set whether the material has a texture.
	 * @param hasTexture whether the material has a texture.
	 */
	void setHasTexture(bool hasTexture);

private:
	float ka;
	float kd;
	float ks;
	float shininess;
	GLuint texture;
	bool hasTexture;
};

}

#endif