#include <algorithm>
#include <cmath>
#include <sstream>

#include "ParticleSystem.h"

#include "opengl/CgContext.h"
#include "opengl/CgProgram.h"
#include "opengl/IoTextureSet.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/OrthoNormalProjection.h"
#include "opengl/PrimaryTextureContainer.h"
#include "opengl/Texture2dOperation.h"
#include "opengl/TextureContainer.h"

#include "image/Image.h"

#include "util/Util.h"

using namespace OpenGL;

namespace GpuThesis {

unsigned int calculateTextureDimension(unsigned int numPixels)
{
	unsigned int textureDimension = 1;

	while (textureDimension * textureDimension < numPixels)
		textureDimension *= 2;

	return textureDimension;
}

ParticleSystem::ParticleSystem(unsigned int numParticles, const std::string &particleSplat, float splatThreshold,
							   float domainScale,
							   bool useVBO) :
	textureDimension(calculateTextureDimension(numParticles)),
	textures(new OpenGL::IoTextureSet(textureDimension, textureDimension,
	                                  CgContext::getInstance().getTex2dTarget(),
	                                  CgContext::getInstance().getTex2dInternalFormat())),
	positions(textures->addPrimaryTexture("Particle positions")),
	spawnPositions(textures->addPrimaryTexture("Particle spawn positions")),
	spawnVelocities(textures->addPrimaryTexture("Particle spawn velocities")),
	velocities(textures->addPrimaryTexture("Particle velocities")),
	particleColors(textures->addPrimaryTexture("Particle colors")),
	useVBO(useVBO),
	time(0.0f),
	particleSize(0.08f),
	maxParticleSize(256.0f),
	gravity(0.4f),
	visibilityThreshold(0.004f),
	minimumLifeTime(0.5f),
	numParticles(numParticles),
	spawnPositionsVector(4 * textureDimension * textureDimension),
	spawnVelocitiesVector(3 * textureDimension * textureDimension),
	blendFunc(0),
	box(false)
{
	initializeCgPrograms();

	if (!particleSplat.empty()) {
		initializeParticleTexture(particleSplat, splatThreshold);
		hasParticleTexture = true;
	} else
		hasParticleTexture = false;

	initializeParticles();
	

	textureContainer = new OpenGL::PrimaryTextureContainer(textures);

	textureContainer->setDescription("Particle system.");
}

ParticleSystem::~ParticleSystem()
{
	if (hasParticleTexture)
		glDeleteTextures(1, &particleTexture);

	delete textureContainer;
	delete updatePositionsOp;
	delete updateVelocitiesOp;
	delete advectParticlesOp;
	delete cgRenderParticles;
	delete cgUpdatePositions;
	delete cgUpdateVelocities;
	delete cgAdvectParticles;
	delete textures;
}

OpenGL::TextureContainer *ParticleSystem::getTextureContainer() const
{
	return textureContainer;
}

void ParticleSystem::doRun(float timestep) 
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 1, 0, 1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	textures->bindFbo();

	updateVelocities(timestep);
	updatePositions(timestep);
	updateColors();

	//textures->disableFbo();	

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}

void ParticleSystem::readParticlePositions(float *positions) const
{
	GLuint texture = textures->lookup(getPositions());

	unsigned int width = getTextureDimension();
	unsigned int height = getTextureDimension();

	std::vector<float> v(3 * width * height);

	glBindTexture(CgContext::getInstance().getTex2dTarget(), texture);
	glGetTexImage(CgContext::getInstance().getTex2dTarget(), 0, GL_RGB, GL_FLOAT, &*v.begin());

	for (unsigned int i = 0; i < 3 * numParticles; ++i) {
		float pos = v[i];
		positions[i] = pos;
	}
}

void ParticleSystem::advect(const Math::Vector3d &distance)
{
	OrthoNormalProjection ortho;

	textures->bindFbo();

	float dist[4] = { distance.x, distance.y, distance.z, 0.0f };

	cgAdvectParticles->setParameter("distance", dist);

	advectParticlesOp->compute();

	textures->disableFbo();
}

void ParticleSystem::doReset()
{
	time = 0.0f;

	textures->fillTexture(this->positions, GL_RGBA, GL_FLOAT, &*spawnPositionsVector.begin());
	textures->fillTexture(this->velocities, GL_RGB, GL_FLOAT, &*spawnVelocitiesVector.begin());
}

#ifdef near
#define REDEFINE_near near
#undef near
#endif

#ifdef far
#define REDEFINE_far far
#undef far
#endif

void ParticleSystem3d::drawBox() const
{
	glColor3f(1, 1, 1);

	float left = -static_cast<float>(fluidSet->getComputationWidth()) * scale / 2.0f;
	float right = static_cast<float>(fluidSet->getComputationWidth()) * scale / 2.0f;
	float bottom = 0.0f;
	float top = static_cast<float>(fluidSet->getComputationHeight()) * scale;
	float near = -static_cast<float>(fluidSet->getComputationDepth()) * scale / 2.0f;
	float far = static_cast<float>(fluidSet->getComputationDepth()) * scale / 2.0f;

	glBegin(GL_LINE_LOOP);
	glVertex3f(left, bottom, near);
	glVertex3f(right, bottom, near);
	glVertex3f(right, bottom, far);
	glVertex3f(left, bottom, far);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3f(left, top, near);
	glVertex3f(right, top, near);
	glVertex3f(right, top, far);
	glVertex3f(left, top, far);
	glEnd();

	glBegin(GL_LINES);
	glVertex3f(left, bottom, near);
	glVertex3f(left, top, near);
	glVertex3f(right, bottom, near);
	glVertex3f(right, top, near);
	glVertex3f(right, bottom, far);
	glVertex3f(right, top, far);
	glVertex3f(left, bottom, far);
	glVertex3f(left, top, far);
	glEnd();
}

void ParticleSystemSlice::drawBox() const
{
	glColor3f(1, 1, 1);

	float left = -static_cast<float>(fluidSet->getComputationWidth()) * scale / 2.0f;
	float right = static_cast<float>(fluidSet->getComputationWidth()) * scale / 2.0f;
	float bottom = 0.0f;
	float top = static_cast<float>(fluidSet->getComputationHeight()) * scale;
	float near = -static_cast<float>(fluidSet->getComputationWidth()) * scale / 2.0f;
	float far = static_cast<float>(fluidSet->getComputationWidth()) * scale / 2.0f;

	glBegin(GL_LINE_LOOP);
	glVertex3f(left, bottom, near);
	glVertex3f(right, bottom, near);
	glVertex3f(right, bottom, far);
	glVertex3f(left, bottom, far);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3f(left, top, near);
	glVertex3f(right, top, near);
	glVertex3f(right, top, far);
	glVertex3f(left, top, far);
	glEnd();

	glBegin(GL_LINES);
	glVertex3f(left, bottom, near);
	glVertex3f(left, top, near);
	glVertex3f(right, bottom, near);
	glVertex3f(right, top, near);
	glVertex3f(right, bottom, far);
	glVertex3f(right, top, far);
	glVertex3f(left, bottom, far);
	glVertex3f(left, top, far);
	glEnd();
}

#ifdef REDEFINE_near
#define near REDEFINE_near
#undef REDEFINE_near
#endif

#ifdef REDEFINE_far
#define far REDEFINE_far
#undef REDEFINE_far
#endif

void ParticleSystem::updateVelocities(float timestep)
{	
	time += timestep;		
	
	computeVelocities(timestep);
}

void ParticleSystem::computeVelocities(float timestep)
{
	cgUpdateVelocities->setParameter("gravity", gravity);
	cgUpdateVelocities->setParameter("time", time);
	cgUpdateVelocities->setParameter("timestep", timestep);

	updateVelocitiesOp->compute();
}

void ParticleSystem::updateColors()
{
	computeColors();	
}

void ParticleSystem::computeColors()
{
//	setColorsOp->compute();
}

void ParticleSystem::updatePositions(float timestep)
{		
	cgUpdatePositions->setParameter("gravity", gravity);
	cgUpdatePositions->setParameter("time", time);
	cgUpdatePositions->setParameter("timestep", timestep);
	
	updatePositionsOp->compute();
}

void ParticleSystem::setViewport()
{
	glGetIntegerv(GL_VIEWPORT, oldViewport);
	glViewport(0, 0, textureDimension, textureDimension);

}

void ParticleSystem::resetViewport()
{
	glViewport(oldViewport[0], oldViewport[1], oldViewport[2], oldViewport[3]);
}

void ParticleSystem::setBlendFunc(unsigned int value)
{
	blendFunc = value;
}

unsigned int ParticleSystem::getBlendFunc() const
{
	return blendFunc;
}

bool ParticleSystem::getDrawBox() const
{
	return box;
}

void ParticleSystem::setDrawBox(bool drawBox)
{
	box = drawBox;
}

void ParticleSystem::renderParticles() const
{
	cgRenderParticles->loadModelViewProjection(
		cgRenderParticles->getParameterId("modelViewProjection"));

	cgRenderParticles->setParameter("particleSize", particleSize);
	cgRenderParticles->setParameter("time", time);

	cgRenderParticles->activate();

	glEnable(GL_BLEND);

	switch (blendFunc) {
	case 0:
		glBlendFunc(GL_SRC_COLOR, GL_ONE);
		break;
	case 1:
		glBlendFunc(GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR);
		break;
	case 2:
		glBlendFunc(GL_ONE, GL_DST_ALPHA);
		break;
	case 3:
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
		break;
	case 4:
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
		break;
	case 5:
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		break;
	case 6:
		glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
		break;
	default:
		glBlendFunc(GL_SRC_ALPHA, GL_DST_ALPHA);
		break;
	}

	if (hasParticleTexture) {
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, particleTexture);
	}

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	if (useVBO) {
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexVBO);
		glVertexPointer(4, GL_FLOAT, 0, 0);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, textureVBO);
		glTexCoordPointer(2, GL_FLOAT, 0, 0);
	} else {
		glVertexPointer(4, GL_FLOAT, 0, &*vertices.begin());
		glTexCoordPointer(2, GL_FLOAT, 0, &*texCoords.begin());
	}

	glDrawArrays(GL_QUADS, 0, 4 * numParticles);

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	if (hasParticleTexture)
		glDisable(GL_TEXTURE_2D);

	glDisable(GL_BLEND);

	cgRenderParticles->deactivate();

	if (box)
		drawBox();
}

void ParticleSystem::setSpawnPositions(const std::vector<float> &positions)
{
	std::copy(positions.begin(), positions.end(), spawnPositionsVector.begin());

	for (unsigned int i = 0; i < positions.size(); ++i) {
		float pos = positions[i];
		spawnPositionsVector[(i / 3) * 4 + (i % 3)] = pos;
	}

	textures->fillTexture(this->spawnPositions, GL_RGBA, GL_FLOAT, &*spawnPositionsVector.begin());
	textures->fillTexture(this->positions, GL_RGBA, GL_FLOAT, &*spawnPositionsVector.begin());
}

void ParticleSystem::setSpawnVelocities(const std::vector<float> &velocities)
{
	std::copy(velocities.begin(), velocities.end(), spawnVelocitiesVector.begin());

	textures->fillTexture(this->spawnVelocities, GL_RGB, GL_FLOAT, &*spawnVelocitiesVector.begin());
	textures->fillTexture(this->velocities, GL_RGB, GL_FLOAT, &*spawnVelocitiesVector.begin());
}

void ParticleSystem::setSpawnTimes(const std::vector<float> &times)
{
	std::vector<float> spawnTimes(4 * textureDimension * textureDimension);

	GLuint texture = textures->lookup(positions);	

	glBindTexture(GL_TEXTURE_2D, texture);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, &*spawnTimes.begin());
	
	for (unsigned int i = 0; i < times.size(); ++i) {		

		spawnTimes[4*i+3] = times[i];
		spawnPositionsVector[4*i+3] = times[i];
	}

	textures->fillTexture(positions, GL_RGBA, GL_FLOAT, &*spawnTimes.begin());
}

void ParticleSystem::setGravity(float gravity)
{
	this->gravity = gravity;
	cgUpdateVelocities->setParameter("gravity", gravity);
}

void ParticleSystem::setMinimumLifeTime(float minimumLifeTime)
{
	this->minimumLifeTime = minimumLifeTime;
	cgUpdatePositions->setParameter("minimumLifeTime", minimumLifeTime);
}

void ParticleSystem::setVisibilityThreshold(float visibilityThreshold)
{
	this->visibilityThreshold = visibilityThreshold;
	cgUpdatePositions->setParameter("visibilityThreshold", visibilityThreshold);
}

float ParticleSystem::getMinimumLifeTime() const
{
	return minimumLifeTime;
}

float ParticleSystem::getVisibilityThreshold() const
{
	return visibilityThreshold;
}

void ParticleSystem::setParticleSize(float size)
{
	this->particleSize = size;
}

float ParticleSystem::getParticleSize() const
{
	return particleSize;
}

unsigned int ParticleSystem::getNumParticles() const
{
	return numParticles;
}

void ParticleSystem::initializeCgPrograms()
{
	cgUpdateVelocities =
		new OpenGL::CgProgram("src/cg/particles.cg",
		                      OpenGL::CgContext::getInstance().getFragmentShader(),
		                      "updateVelocities", textures);

	cgUpdateVelocities->setTextureParameter("positions", positions);
	cgUpdateVelocities->setTextureParameter("spawnVelocities", spawnVelocities);
	cgUpdateVelocities->setTextureParameter("velocities", velocities);

	cgAdvectParticles =
		new OpenGL::CgProgram("src/cg/particles.cg",
		                      OpenGL::CgContext::getInstance().getFragmentShader(),
		                      "advectParticles", textures);

	cgAdvectParticles->setTextureParameter("positions", positions);

	advectParticlesOp = new OpenGL::Texture2dOperation(textureDimension, textureDimension, 
		                                               cgAdvectParticles, false, true);
	advectParticlesOp->setOutputTexture(positions);

	updateVelocitiesOp = new OpenGL::Texture2dOperation(textureDimension, textureDimension, 
		                                                cgUpdateVelocities, false, true);
	updateVelocitiesOp->setOutputTexture(velocities);


	cgUpdatePositions =
		new OpenGL::CgProgram("src/cg/particles.cg",
		                      OpenGL::CgContext::getInstance().getFragmentShader(),
		                      "updatePositions", textures);

	cgUpdatePositions->setTextureParameter("positions", positions);
	cgUpdatePositions->setTextureParameter("velocities", velocities);
	cgUpdatePositions->setTextureParameter("spawnPositions", spawnPositions);
	cgUpdatePositions->setTextureParameter("particleColors", particleColors);
	cgUpdatePositions->setParameter("visibilityThreshold", visibilityThreshold);
	cgUpdatePositions->setParameter("minimumLifeTime", minimumLifeTime);

	updatePositionsOp = new OpenGL::Texture2dOperation(textureDimension, textureDimension, 
		                                               cgUpdatePositions, false, true);
	updatePositionsOp->setOutputTexture(positions);
	
	cgRenderParticles =
		new OpenGL::CgProgram("src/cg/particles.cg",
		                      OpenGL::CgContext::getInstance().getVertexShader(),
		                      "renderParticles", textures);

	cgRenderParticles->setTextureParameter("positions", positions);
	cgRenderParticles->setTextureParameter("colors", particleColors);
}

void ParticleSystem::initializeParticles()
{
	float texDelta = 1.0f / static_cast<float>(textureDimension);

	float halfPixelSize = texDelta / 2.0f;

	for (unsigned int i = 0; i < numParticles; ++i) {
		
		float x = (i % textureDimension) * texDelta;
		float y = (i / textureDimension) * texDelta;

		/*for (unsigned int j = 0; j < 4; ++j) {
			vertices.push_back(x + halfPixelSize);
			vertices.push_back(y + halfPixelSize);
			vertices.push_back(static_cast<float>((j + 1) & 2) - 1.0f);
			vertices.push_back(static_cast<float>(j & 2) - 1.0f);

			texCoords.push_back(static_cast<float>(((j + 1) & 2) / 2));
			texCoords.push_back(static_cast<float>(((j + 2) & 2) / 2));
		}*/

		vertices.push_back(x + halfPixelSize);
		vertices.push_back(y + halfPixelSize);
		vertices.push_back(particleXLeft);
		vertices.push_back(particleYBottom);

		vertices.push_back(x + halfPixelSize);
		vertices.push_back(y + halfPixelSize);
		vertices.push_back(particleXRight);
		vertices.push_back(particleYBottom);

		vertices.push_back(x + halfPixelSize);
		vertices.push_back(y + halfPixelSize);
		vertices.push_back(particleXRight);
		vertices.push_back(particleYTop);

		vertices.push_back(x + halfPixelSize);
		vertices.push_back(y + halfPixelSize);
		vertices.push_back(particleXLeft);
		vertices.push_back(particleYTop);

		texCoords.push_back(particleSLeft);
		texCoords.push_back(particleTTop);

		texCoords.push_back(particleSRight);
		texCoords.push_back(particleTTop);

		texCoords.push_back(particleSRight);
		texCoords.push_back(particleTBottom);

		texCoords.push_back(particleSLeft);
		texCoords.push_back(particleTBottom);

	}

	if (useVBO) {
		glGenBuffersARB(1, &vertexVBO);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexVBO);
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertices.size() * sizeof(float),
						&*vertices.begin(), GL_STATIC_DRAW_ARB);

		glGenBuffersARB(1, &textureVBO);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, textureVBO);
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, texCoords.size() * sizeof(float),
						&*texCoords.begin(), GL_STATIC_DRAW_ARB );
	}
}

#ifdef max
#undef max
#endif

#ifdef min
#undef min
#endif

void ParticleSystem::initializeParticleTexture(const std::string &particleSplat, float splatThreshold)
{
	Image::Image image(particleSplat);

	std::vector<float> data(4 * image.getWidth() * image.getHeight());

	unsigned char *imageData = image.getData();

	float maxColor = 0.0f;

	float maxAlpha = 0.0f;

	// Find max color values and compute alpha values
	for (unsigned int y = 0; y < image.getHeight(); ++y) {
		for (unsigned int x = 0; x < image.getWidth(); ++x) {

			unsigned int red = imageData[(y * image.getWidth() + x) * image.getBpp()];
			unsigned int green = imageData[(y * image.getWidth() + x) * image.getBpp() + 1];
			unsigned int blue = imageData[(y * image.getWidth() + x) * image.getBpp() + 2];

			float redF = static_cast<float>(red) / 255.0f;
			float greenF = static_cast<float>(green) / 255.0f;
			float blueF = static_cast<float>(blue) / 255.0f;

			float alphaF = redF * redF + greenF * greenF + blueF * blueF;

			data[(y * image.getWidth() + x) * 4] = redF;
			data[(y * image.getWidth() + x) * 4 + 1] = greenF;
			data[(y * image.getWidth() + x) * 4 + 2] = blueF;
			data[(y * image.getWidth() + x) * 4 + 3] = alphaF;

			maxColor = std::max(maxColor, std::max(redF, std::max(greenF, blueF)));

			maxAlpha = std::max(maxAlpha, alphaF);
		}
	}
	
	// Histogram Normalization
	for (unsigned int i = 0; i < data.size(); ++i) {
		if (i % 4 == 3)
			data[i] /= maxAlpha;
		else
			data[i] /= maxColor;
	}


	unsigned int firstX = image.getWidth() - 1;
	unsigned int lastX = 0;
	unsigned int firstY = image.getHeight() - 1;
	unsigned int lastY = 0;
	unsigned int colorThreshold = static_cast<unsigned int>(splatThreshold);

	// Find crop positions
	for (unsigned int y = 0; y < image.getHeight(); ++y) {
		for (unsigned int x = 0; x < image.getWidth(); ++x) {

			unsigned int red = imageData[(y * image.getWidth() + x) * image.getBpp()];
			unsigned int green = imageData[(y * image.getWidth() + x) * image.getBpp() + 1];
			unsigned int blue = imageData[(y * image.getWidth() + x) * image.getBpp() + 2];

			if (red > colorThreshold || green > colorThreshold || blue > colorThreshold) {
				if (x < firstX)
					firstX = x;
				if (x > lastX)
					lastX = x;
				if (y < firstY)
					firstY = y;
				if (y > lastY)
					lastY = y;
			}
		}
	}

	particleSLeft = 0;
	particleSRight = 0;
	particleTTop = 0;
	particleTBottom = 0;

	if (firstX < lastX) {
		particleSLeft = firstX / static_cast<float>(image.getWidth());
		particleSRight = (image.getWidth() - 1 - lastX) / static_cast<float>(image.getWidth());			
	}

	if (firstY < lastY) {
		particleTTop = (image.getHeight() - 1 - lastY) / static_cast<float>(image.getHeight());
		particleTBottom = firstY / static_cast<float>(image.getHeight());
	}

	particleXLeft = (particleSLeft - 0.5) * 2;
	particleXRight = (0.5 - particleSRight) * 2;
	particleYTop = (0.5 - particleTTop) * 2;
	particleYBottom = (particleTBottom - 0.5) * 2;

	particleSRight = 1 - particleSRight;
	particleTTop = 1 - particleTTop;


	glGenTextures(1, &particleTexture);

	glBindTexture(GL_TEXTURE_2D, particleTexture);

	glTexImage2D(GL_TEXTURE_2D,
	             0,
	             GL_RGBA, 
	             image.getWidth(),
	             image.getHeight(),
	             0,
	             GL_RGBA,
	             GL_FLOAT,
				 &*data.begin()); // pointer to data

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}

unsigned int ParticleSystem::getPositions() const
{
	return positions;
}

unsigned int ParticleSystem::getVelocities() const
{
	return velocities;
}

unsigned int ParticleSystem::getParticleColors() const
{
	return particleColors;
}

unsigned int ParticleSystem::getTextureDimension() const
{
	return textureDimension;
}

OpenGL::IoTextureSet *ParticleSystem::getTextureSet() const
{
	return textures;
}

ParticleSystem3d::ParticleSystem3d(unsigned int numParticles, const std::string &particleSplat, float splatThreshold,
                                   OpenGL::Flat3dTextureSet *fluidSet,
                                   unsigned int fluidColors,
                                   unsigned int velocityTexture,
                                   float voxelSize,
                                   float domainScale,
                                   bool useVBO) :
	ParticleSystem(numParticles, particleSplat, splatThreshold, domainScale, useVBO),
	fluidSet(fluidSet),
	scale(domainScale)
{
	initialize(fluidColors, velocityTexture, voxelSize);
}

void ParticleSystem3d::computeVelocities(float timestep)
{
	setVelocitiesOp->compute();
}

void ParticleSystem3d::computeColors()
{
	setColorsOp->compute();
}

void ParticleSystem3d::initialize(unsigned int fluidColors,
                                  unsigned int velocityTexture,
                                  float voxelSize)
{
	using Util::autoPtr;

	OpenGL::CgContext &context = OpenGL::CgContext::getInstance();
	cgSetVelocities = 
		autoPtr(new OpenGL::CgProgram(context.getCgProgramPath() + "particles.cg", 
	                                  context.getFragmentShader(),
		                              "setVelocities",
		                              getTextureSet()));

	cgSetColors = 
		autoPtr(new OpenGL::CgProgram(context.getCgProgramPath() + "particles.cg",
		                              context.getFragmentShader(),
		                              "setColors",
		                              getTextureSet()));

	unsigned int fluidOffsetLookup =
		getTextureSet()->importSecondaryTexture(fluidSet, fluidSet->getOffsetLookup(), 
		                                        "Fluid offset lookup");

	cgSetVelocities->setTextureParameter("positions", getPositions());
	cgSetVelocities->setTextureParameter("fluidVelocity",
		getTextureSet()->importSecondaryTexture(fluidSet, velocityTexture, "Fluid velocity"));
	cgSetVelocities->setTextureParameter("offsetLookup", fluidOffsetLookup);
	cgSetVelocities->setParameter("computationWidth", fluidSet->getComputationWidth());
	cgSetVelocities->setParameter("computationHeight", fluidSet->getComputationHeight());
	cgSetVelocities->setParameter("computationDepth", fluidSet->getComputationDepth());
	cgSetVelocities->setParameter("scale", scale);
	cgSetVelocities->setParameter("voxelSize", voxelSize);

	cgSetColors->setTextureParameter("positions", getPositions());
	cgSetColors->setTextureParameter("fluidColors", 
		getTextureSet()->importSecondaryTexture(fluidSet, fluidColors, "Fluid colors"));
	cgSetColors->setTextureParameter("offsetLookup", fluidOffsetLookup);
	cgSetColors->setParameter("computationWidth", fluidSet->getComputationWidth());
	cgSetColors->setParameter("computationHeight", fluidSet->getComputationHeight());
	cgSetColors->setParameter("computationDepth", fluidSet->getComputationDepth());
	cgSetColors->setParameter("scale", scale);

	setVelocitiesOp =
		autoPtr(new OpenGL::Texture2dOperation(getTextureDimension(), getTextureDimension(), 
		                                       cgSetVelocities.get(), false, true));
	setVelocitiesOp->setOutputTexture(getVelocities());

	setColorsOp = 
		autoPtr(new OpenGL::Texture2dOperation(getTextureDimension(), getTextureDimension(),
		                                       cgSetColors.get(), false, true));
	setColorsOp->setOutputTexture(getParticleColors());
}

ParticleSystemSlice::ParticleSystemSlice(unsigned int numParticles, const std::string &particleSplat, float splatThreshold,
                                         OpenGL::Flat3dTextureSet *fluidSet, 
                                         unsigned int fluidColors,
                                         unsigned int velocityTexture,
                                         float voxelSize,
										 float domainScale,
										 bool useVBO) :
	ParticleSystem(numParticles, particleSplat, splatThreshold, domainScale, useVBO),
	fluidSet(fluidSet),
	scale(domainScale)
{
	initialize(fluidColors, velocityTexture, voxelSize);
}

void ParticleSystemSlice::computeVelocities(float timestep)
{
	setVelocitiesOp->compute();
}

void ParticleSystemSlice::computeColors()
{
	setColorsOp->compute();
}

void ParticleSystemSlice::initialize(unsigned int fluidColors,
                                     unsigned int velocityTexture,
                                     float voxelSize)
{
	using Util::autoPtr;

	OpenGL::CgContext &context = OpenGL::CgContext::getInstance();
	cgSetVelocities = 
		autoPtr(new OpenGL::CgProgram(context.getCgProgramPath() + "particles.cg", 
	                                  context.getFragmentShader(),
		                              "setVelocitiesSlice",
		                              getTextureSet()));

	cgSetColors = 
		autoPtr(new OpenGL::CgProgram(context.getCgProgramPath() + "particles.cg",
		                              context.getFragmentShader(),
		                              "setColorsSlice",
		                              getTextureSet()));

	unsigned int fluidSliceLookup =
		getTextureSet()->importSecondaryTexture(fluidSet, fluidSet->getSliceLookup(), 
		                                        "Fluid slice lookup");

	cgSetVelocities->setTextureParameter("positions", getPositions());
	cgSetVelocities->setTextureParameter("fluidVelocity",
		getTextureSet()->importSecondaryTexture(fluidSet, velocityTexture, "Fluid velocity"));
	cgSetVelocities->setTextureParameter("sliceLookup", fluidSliceLookup);
	cgSetVelocities->setParameter("computationWidth", fluidSet->getComputationWidth());
	cgSetVelocities->setParameter("computationHeight", fluidSet->getComputationHeight());
	cgSetVelocities->setParameter("sliceCount", fluidSet->getComputationDepth());
	cgSetVelocities->setParameter("scale", scale);
	cgSetVelocities->setParameter("voxelSize", voxelSize);

	cgSetColors->setTextureParameter("positions", getPositions());
	cgSetColors->setTextureParameter("fluidColors", 
		getTextureSet()->importSecondaryTexture(fluidSet, fluidColors, "Fluid colors"));
	cgSetColors->setTextureParameter("sliceLookup", fluidSliceLookup);
	cgSetColors->setParameter("computationWidth", fluidSet->getComputationWidth());
	cgSetColors->setParameter("computationHeight", fluidSet->getComputationHeight());
	cgSetColors->setParameter("sliceCount", fluidSet->getComputationDepth());
	cgSetColors->setParameter("scale", scale);

	setVelocitiesOp =
		autoPtr(new OpenGL::Texture2dOperation(getTextureDimension(), getTextureDimension(), 
		                                       cgSetVelocities.get(), false, true));
	setVelocitiesOp->setOutputTexture(getVelocities());

	setColorsOp = 
		autoPtr(new OpenGL::Texture2dOperation(getTextureDimension(), getTextureDimension(),
		                                       cgSetColors.get(), false, true));
	setColorsOp->setOutputTexture(getParticleColors());
}

}
