#ifndef GPUTHESIS_PARTICLESYSTEM_H
#define GPUTHESIS_PARTICLESYSTEM_H

#include <string>
#include <vector>

#include "opengl/OpenGL.h"

#include "math/Vector3d.h"

#include "Runnable.h"

namespace OpenGL {
	class IoTextureSet;
	class CgProgram;
	class Flat3dTextureSet;
	class Texture2dOperation;
	class TextureContainer;
}

namespace GpuThesis {

/**
 * The ParticleSystem class.
 */
class ParticleSystem : public Runnable {
public:
	/**
	 * Constructs a standard particle system. The particle velocities are
	 * determined by the fluid velocity texture, and the particle colors 
	 * by the fluid color texture. 
	 * @param numParticles the number of particles in the particle system.
	 * @param splatThreshold the lowest intensity of the splat that should be visible.
	 * @param domainScale the scale of the particle domain with respect to the simulation domain.
	 * @param particleSplat the image to be used as texture for each particle.
	 * @param useVBO if the implementation shall use Vertex Buffer Object.
	 */
	ParticleSystem(unsigned int numParticles, const std::string &particleSplat,
	               float splatThreshold,
				   float domainScale,
				   bool useVBO = true);
	virtual ~ParticleSystem();
	
	/**
	 * Renders the particles to the screen based on their positions. 	 
	 */
	void renderParticles() const;

	/**
	 * Reads the particle positions into a float array.
	 * @param positions float array of size at least 3 * numParticles
	 */
	void readParticlePositions(float *positions) const;

	/**
	 * Sets the positions where the particles spawn.
	 * The position vector should be on the form (x, y, z) * numParticles.
	 * @param positions the spawn positions.
	 */
	void setSpawnPositions(const std::vector<float> &positions);

	/**
	 * Sets the velocities the particles should have when spawn.
	 * The velocity vector should be on the form (x, y, z) * numParticles.
	 * @param velocities the spawn velocities.
	 */
	void setSpawnVelocities(const std::vector<float> &velocities);

	/**
	 * Sets the particle spawn times. These times indicate the time from 
	 * the particle system is initialized to the particles spawn. 
	 * The time vector should be on the form (time) * numParticles.
	 * @param times the spawn times in seconds. 
	 */
	void setSpawnTimes(const std::vector<float> &times);

	/**
	 * Sets the gravity constant to be used in the particle system.
	 * @param gravity the new gravity constant. 
	 */
	void setGravity(float gravity);

	/**
	 * Sets the alpha threshold at which the particles will respawn.
	 * @param visibilityThreshold the visibility threshold of the particle's alpha color.
	 */
	void setVisibilityThreshold(float visibilityThreshold);

	/**
	 * Sets the minimum lifetime of the particles.
	 * @param minimumLifeTime the minimum particle lifetime.
	 */
	void setMinimumLifeTime(float minimumLifeTime);

	/**
	 * Gets the alpha threshold at which the particles will respawn.
	 * @return the visibility threshold of the particle's alpha color.
	 */
	float getVisibilityThreshold() const;

	/**
	 * Gets the minimum lifetime of the particles.
	 * @return the minimum particle lifetime.
	 */
	float getMinimumLifeTime() const;


	/**
	 * Sets the particle size.
	 * @param size the new particle size.
	 */
	void setParticleSize(float size);

	/**
	 * Returns the particle size.
	 * @return the particle size.
	 */
	float getParticleSize() const;

	/**
	 * Returns the number of particles in the particle system.
	 * @return the number of particles. 
	 */
	unsigned int getNumParticles() const;

	/**
	 * Returns the texture container containing the textures 
	 * used by the particle system.
	 * @return the texture container.
	 */
	OpenGL::TextureContainer *getTextureContainer() const;

	/**
	 * Returns the type of blend function used when rendering the particles.
	 * @return the type of blend function. 
	 */
	unsigned int getBlendFunc() const;

	/**
	 * Sets the type of blend function used when rendering the particles.
	 * @param value the new type of blend function. 
	 */
	void setBlendFunc(unsigned int value);

	/**
	 * Returns whether a box is drawn around the domain
	 * where the particle system operates. 
	 * @return whether a box is drawn. 
	 */
	bool getDrawBox() const;

	/**
	 * Sets whether a box should be drawn around the domain
	 * where the particle system operates.
	 * @param drawBox whether a box should be drawn. 
	 */
	void setDrawBox(bool drawBox);

	/**
	 * Advect all particles.
	 * @param distance the distance to advect (in local space).
	 */
	void advect(const Math::Vector3d &distance);

protected:
	unsigned int getPositions() const;
	unsigned int getVelocities() const;
	unsigned int getParticleColors() const;

	unsigned int getTextureDimension() const;

	OpenGL::IoTextureSet *getTextureSet() const;

private:
	void doRun(float timestep);
	void doReset();

	unsigned int textureDimension;

	OpenGL::IoTextureSet *textures;

	unsigned int positions;
	unsigned int spawnPositions;
	unsigned int spawnVelocities;
	unsigned int velocities;
	unsigned int particleColors;

	OpenGL::TextureContainer *textureContainer;

	std::vector<float> vertices;  // position texture coordinates for each particle
	std::vector<float> texCoords; // texture coordinates for each particle

	bool useVBO;

	float time;
	float particleSize;
	float maxParticleSize;
	float gravity;
	float visibilityThreshold;
	float minimumLifeTime;

	unsigned int numParticles;

	GLuint particleTexture; // particle texture splat
	GLuint vertexVBO; // vertex buffer object
	GLuint textureVBO; // texture buffer object

	bool hasParticleTexture;

	OpenGL::CgProgram *cgUpdateVelocities, *cgUpdatePositions;
	OpenGL::CgProgram *cgRenderParticles;
	OpenGL::CgProgram *cgAdvectParticles;

	OpenGL::Texture2dOperation *updateVelocitiesOp, *updatePositionsOp, *advectParticlesOp;

	//float scale;

	void initializeCgPrograms();
	void initializeParticles();
	void initializeParticleTexture(const std::string &particleSplat, float splatThreshold);

	virtual void drawBox() const = 0;

	void updateVelocities(float timestep);
	void updatePositions(float timestep);
	void updateColors();

	virtual void computeVelocities(float timestep);
	virtual void computeColors();

	void setViewport();
	void resetViewport();

	GLint oldViewport[4];

	std::vector<float> spawnPositionsVector, spawnVelocitiesVector;

	unsigned int blendFunc;

	bool box;

	float particleSLeft, particleSRight;
	float particleTTop, particleTBottom;
	float particleXLeft, particleXRight;
	float particleYTop, particleYBottom;
};

/**
 * The class ParticleSystem3d is a particle system used
 * to visualize the results from a full 3d fluid simulation. 
 */
class ParticleSystem3d : public ParticleSystem {
public:
	/**
	 * Constructs a particle system used to visualize the results from a full 3d fluid simulation. 
	 * The particle velocities are determined by the fluid velocity texture, 
	 * and the particle colors by the fluid color texture. 
	 * @param numParticles the number of particles in the particle system.
	 * @param particleSplat the image to be used as texture for each particle.
	 * @param splatThreshold the lowest intensity of the splat that should be visible.
	 * @param fluidSet the texture set containing the fluid color and fluid velocity texture.
	 * @param fluidColors the identifier of the fluid color texture.
	 * @param velocityTexture the identifier of the fluid velocity texture.
	 * @param voxelSize the size of the voxels in the domain where the particle system operates. 
	 * @param domainScale the scale of the particle system domain.
	 * @param useVBO if the implementation shall use Vertex Buffer Object.
	 */
	ParticleSystem3d(unsigned int numParticles, const std::string &particleSplat, float splatThreshold,
				     OpenGL::Flat3dTextureSet *fluidSet, 
	                 unsigned int fluidColors,
				     unsigned int velocityTexture,
	                 float voxelSize,
					 float domainScale,
				     bool useVBO = true);

private:
	void computeVelocities(float timestep);
	void computeColors();

	void initialize(unsigned int fluidColors,
	                unsigned int velocityTexture,
	                float voxelSize);

	void drawBox() const;

	OpenGL::Flat3dTextureSet *fluidSet;

	float scale;

	std::auto_ptr<OpenGL::CgProgram> cgSetVelocities, cgSetColors;

	std::auto_ptr<OpenGL::Texture2dOperation> setVelocitiesOp, setColorsOp;
};

/**
 * The class ParticleSystemSlice is a particle system used to visualize
 * the results from a slice (volumetric extrusion) fluid simulation. 
 */
class ParticleSystemSlice : public ParticleSystem {
public:
	/**
	 * Constructs a particle system used to visualize the results from a slice fluid simulation. 
	 * The particle velocities are determined by the fluid velocity texture, 
	 * and the particle colors by the fluid color texture. 
	 * @param numParticles the number of particles in the particle system.
	 * @param particleSplat the image to be used as texture for each particle.
	 * @param splatThreshold the lowest intensity of the splat that should be visible.
	 * @param fluidSet the texture set containing the fluid color and fluid velocity texture.
	 * @param fluidColors the identifier of the fluid color texture.
	 * @param velocityTexture the identifier of the fluid velocity texture.
	 * @param voxelSize the size of the voxels in the domain where the particle system operates. 
	 * @param domainScale the scale of the particle system domain.
	 * @param useVBO if the implementation shall use Vertex Buffer Object.
	 */
	ParticleSystemSlice(unsigned int numParticles, const std::string &particleSplat, float splatThreshold,
				        OpenGL::Flat3dTextureSet *fluidSet, 
	                    unsigned int fluidColors,
				        unsigned int velocityTexture,
	                    float voxelSize,
						float domainScale,
				        bool useVBO = true);

private:
	void computeVelocities(float timestep);
	void computeColors();

	void initialize(unsigned int fluidColors,
	                unsigned int velocityTexture,
	                float voxelSize);

	void drawBox() const;

	OpenGL::Flat3dTextureSet *fluidSet;

	float scale;

	std::auto_ptr<OpenGL::CgProgram> cgSetVelocities, cgSetColors;

	std::auto_ptr<OpenGL::Texture2dOperation> setVelocitiesOp, setColorsOp;	
};

}

#endif