#include "RayMarcher.h"

#include <string>

#include "opengl/CgContext.h"
#include "opengl/CgProgram.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/IoTextureSet.h"

#include "util/Util.h"

using namespace Math;
using namespace OpenGL;
using namespace Util;

namespace GpuThesis {

RayMarcher::RayMarcher(OpenGL::Flat3dTextureSet *textureSet,
                       unsigned int dataVolume,
                       bool slice,
					   Math::BoundingBox box) :
	VolumeRenderer(box),
	textureSet(textureSet),
	dataVolume(dataVolume),
	stepSize(1.0f),
	slice(slice)
{
	int viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	unsigned int width = viewport[2], height = viewport[3];

	glGenTextures(1, &volumeTexture);

	CgContext &context = CgContext::getInstance();

	rmTextureSet = autoPtr(new IoTextureSet(width, height, 
	                                        context.getTexRectTarget(), 
	                                        context.getTexRectInternalFormat()));

	tiEntryPoints = rmTextureSet->addPrimaryTexture("Entry Points");
	tiRayDirections = rmTextureSet->addPrimaryTexture("Ray Directions");

	unsigned int tiDataVolume = rmTextureSet->importSecondaryTexture(textureSet, dataVolume, "Imported Data Volume");
	unsigned int tiOffsetLookup = rmTextureSet->importSecondaryTexture(textureSet, textureSet->getOffsetLookup(), "Imported Offset Lookup");	
	unsigned int tiSliceLookup = rmTextureSet->importSecondaryTexture(textureSet, textureSet->getSliceLookup(), "Imported Slice Lookup");

	std::string programFile = context.getCgProgramPath() + "volume.cg";

	cgRayEntryPoints = autoPtr(new CgProgram(programFile,
	                                         context.getFragmentShader(),
	                                         "entryPointDetermination",
	                                         rmTextureSet.get()));

	cgRayDirections	= autoPtr(new CgProgram(programFile,
	                                        context.getFragmentShader(),
	                                        "rayDirectionDetermination",
	                                        rmTextureSet.get()));

	cgRayDirections->setTextureParameter("entryPoints", tiEntryPoints);

	std::string entryPoint = slice ? "rayMarcherSlice" : "rayMarcher";
	
	cgRayMarcher = autoPtr(new CgProgram(programFile,
	                                     context.getFragmentShader(),
	                                     entryPoint,
	                                     rmTextureSet.get()));

	cgRayMarcher->setParameter("domainWidth", textureSet->getComputationWidth());
	cgRayMarcher->setParameter("domainHeight", textureSet->getComputationHeight());
	cgRayMarcher->setParameter("domainDepth", textureSet->getComputationDepth());
	cgRayMarcher->setParameter("stepSize", stepSize);
	cgRayMarcher->setTextureParameter("flat3dTexture", tiDataVolume);
	
	if (slice)
		cgRayMarcher->setTextureParameter("sliceLookup", tiSliceLookup);
	else
		cgRayMarcher->setTextureParameter("flat3dOffset", tiOffsetLookup);		

	cgRayMarcher->setTextureParameter("rayDirections", tiRayDirections);

	updateBoundingBox();
}

void RayMarcher::setStepSize(float stepSize)
{
	this->stepSize = stepSize;

	cgRayMarcher->setParameter("stepSize", stepSize);
}

float RayMarcher::getStepSize() const
{
	return stepSize;
}

OpenGL::IoTextureSet *RayMarcher::getTextureSet()
{
	return rmTextureSet.get();
}

void RayMarcher::updateBoundingBox()
{
	BoundingBox box = getBoundingBox();

	// positive z towards viewer
	vertices[0] = Vector3d(box.getMin().x, box.getMax().y, box.getMax().z);
	vertices[1] = Vector3d(box.getMax().x, box.getMax().y, box.getMax().z);
	vertices[2] = Vector3d(box.getMax().x, box.getMax().y, box.getMin().z);
	vertices[3] = Vector3d(box.getMin().x, box.getMax().y, box.getMin().z);

	vertices[4] = Vector3d(box.getMin().x, box.getMin().y, box.getMax().z);
	vertices[5] = Vector3d(box.getMax().x, box.getMin().y, box.getMax().z);
	vertices[6] = Vector3d(box.getMax().x, box.getMin().y, box.getMin().z);
	vertices[7] = Vector3d(box.getMin().x, box.getMin().y, box.getMin().z);	

	for (unsigned int i = 0; i < 8; ++i)
		texCoords[i] =
			Vector3d((vertices[i].x + 0.5) * textureSet->getComputationWidth(),
			         vertices[i].y * textureSet->getComputationHeight(),
					 slice ? (vertices[i].z + 0.5) * textureSet->getComputationWidth() :
	                         (vertices[i].z + 0.5) * textureSet->getComputationDepth());

	// counter-clockwise for front faces
	for (unsigned int i = 0; i < 4; ++i) {
		indices[i][3] = i;
		indices[i][2] = (i + 1) % 4;
		indices[i][1] = 4 + (i + 1) % 4;
		indices[i][0] = 4 + i;
	}

	// counter-clockwise for front faces
	for (unsigned int i = 0; i < 4; ++i) {
		indices[4][i] = i;
		indices[5][i] = 7 - i;
	}

	for (unsigned int i = 0; i < 6; ++i) {
		normals[i] = cross((vertices[indices[i][1]] - vertices[indices[i][0]]),
		                   (vertices[indices[i][3]] - vertices[indices[i][0]]));

		normals[i] = normalize(normals[i]);
	}
}

void RayMarcher::compute()
{
	rmTextureSet->bindFbo();
	rmTextureSet->clearTexture(tiEntryPoints);
	rmTextureSet->clearTexture(tiRayDirections);

	rmTextureSet->setActiveTexture(tiEntryPoints);
	cgRayEntryPoints->activate();

	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	drawBox(); // draw front faces
	
	cgRayEntryPoints->deactivate();
	rmTextureSet->setActiveTexture(tiRayDirections);
	cgRayDirections->activate();

	glCullFace(GL_FRONT);
	drawBox(); // draw back faces

	glDisable(GL_CULL_FACE);
	
	cgRayDirections->deactivate();
}

void RayMarcher::doRender()
{
	cgRayMarcher->activate();

	glEnable(GL_BLEND);
	//glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glBlendFunc(GL_ONE, GL_ONE);

	glEnable(GL_CULL_FACE);

	glCullFace(GL_BACK);
	drawBox(); // draw front faces

	glDisable(GL_BLEND);

	cgRayMarcher->deactivate();

	glDisable(GL_CULL_FACE);
}

void RayMarcher::drawBox() const
{
	glBegin(GL_QUADS);

	for (unsigned int i = 0; i < 6; ++i) {
		const Vector3d &normal = normals[i];
		glNormal3f(normal.x, normal.y, normal.z);

		for (unsigned int j = 0; j < 4; ++j) {
			const Vector3d &texCoord = texCoords[indices[i][j]];
			glTexCoord3f(texCoord.x, texCoord.y, texCoord.z);
			glColor3f(texCoord.x, texCoord.y, texCoord.z);

			const Vector3d &vertex = vertices[indices[i][j]];
			glVertex3f(vertex.x, vertex.y, vertex.z);
		}
	}

	glEnd();
}

}

