#ifndef GPUTHESIS_RAYMARCHER_H
#define GPUTHESIS_RAYMARCHER_H

#include <memory>

#include "VolumeRenderer.h"

#include "math/BoundingBox.h"
#include "math/Vector3d.h"

namespace OpenGL {

class CgProgram;
class Flat3dTextureSet;
class IoTextureSet;
}

namespace GpuThesis {

/**
 * Specialization of VolumeRenderer which uses ray marching on the GPU.
 */
class RayMarcher : public VolumeRenderer {
public:
	/**
	 * Initializes the RayMarcher.
	 * @param textureSet the texture set which contains the data to render.
	 * @param dataVolume the identifier of the flat 3d texture containing the data.
	 * @param slice whether the data is stored as cylindrical slices (as opposed to 3d).
	 * @param box the bounding box of the interesting data.
	 */
	RayMarcher(OpenGL::Flat3dTextureSet *textureSet,
	           unsigned int dataVolume,
	           bool slice,
			   Math::BoundingBox box);

	/**
	 * Get the texture set used internally by the RayMarcher.
	 * The texture set contains ray entry points and ray directions.
	 * @return the texture set.
	 */
	OpenGL::IoTextureSet *getTextureSet();

	/**
	 * Set the step size to use when stepping through the volume.
	 * Larger step size is faster but less accurate.
	 * @param stepSize the step size to use.
	 */
	void setStepSize(float stepSize);

	/**
	 * Get the current step size.
	 * @see setStepSize
	 * @return the step size.
	 */
	float getStepSize() const;

	/**
	 * Perform per-frame computations.
	 */
	void compute();

private:
	void doRender();	
	void updateBoundingBox();

	void drawBox() const;

	std::auto_ptr<OpenGL::IoTextureSet> rmTextureSet;
	std::auto_ptr<OpenGL::CgProgram> cgRayEntryPoints;
	std::auto_ptr<OpenGL::CgProgram> cgRayDirections;
	std::auto_ptr<OpenGL::CgProgram> cgRayMarcher;

	OpenGL::Flat3dTextureSet *textureSet;

	unsigned int dataVolume;
	unsigned int volumeTexture;

	float stepSize;

	bool slice;

	// texture ids
	unsigned int tiEntryPoints;
	unsigned int tiRayDirections;

	// Box variables
	Math::Vector3d vertices[8];
	Math::Vector3d texCoords[8];
	Math::Vector3d normals[6];
	unsigned int indices[6][4];
};

}

#endif