#include "Scene.h"

#include <algorithm>
#include <cassert>
#include <functional>

#include "Fire3dSimulator.h"
#include "Material.h"
#include "SceneObject.h"

#include "math/Matrix4x4.h"
#include "math/Vector3d.h"

#include "opengl/Camera.h"
#include "opengl/CgContext.h"
#include "opengl/CgProgram.h"
#include "opengl/CgOperation.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/IoTextureSet.h"
#include "opengl/OrthoNormalProjection.h"
#include "opengl/Rendering.h"
#include "opengl/Texture2dOperation.h"

#include "util/Util.h"

using namespace Geometry;
using namespace Math;
using namespace OpenGL;
using namespace Util;

namespace GpuThesis {

namespace {

struct LightComparison : public std::binary_function<Light*, Light*, bool>
{
public:
	bool operator()(const Light *l1, const Light *l2) const
	{
		return  l1->getDynamic() && !l2->getDynamic();
	}
};

struct LightDynamic : public std::unary_function<Light*, bool>
{
public:
	bool operator()(const Light *light) const
	{
		return light->getDynamic();
	}
};

}

Scene::Scene(const std::vector<Light *> &lights, Fire3dSimulator *fireSimulation, bool slice) :
	lights(lights),
	fireSimulation(fireSimulation),
	slice(slice),
	textureSet(new IoTextureSet(static_cast<unsigned int>(lights.size()), 1, GL_TEXTURE_RECTANGLE_NV, GL_FLOAT_RGBA32_NV)),
	lightPositionsWs(lights.size()),
	lightData(4 * lights.size()),
	displayLights(false)
{
	std::sort(this->lights.begin(), this->lights.end(), LightComparison());

	lightColorTexture = textureSet->addPrimaryTexture("Light colors");
	lightColorOriginalTexture = textureSet->addPrimaryTexture("Original light colors");
	lightPositionTexture = textureSet->addPrimaryTexture("Light positions");

	fireColorField = textureSet->importSecondaryTexture(fireSimulation->getFlat3dTextureSet(), fireSimulation->getColorTexture(), "Fire colors");
	sliceLookup = textureSet->importSecondaryTexture(fireSimulation->getFlat3dTextureSet(), fireSimulation->getFlat3dTextureSet()->getSliceLookup(), "Slice lookup");
	offsetLookup = textureSet->importSecondaryTexture(fireSimulation->getFlat3dTextureSet(), fireSimulation->getFlat3dTextureSet()->getOffsetLookup(), "Offset lookup");

	numDynamicLights = std::count_if(lights.begin(), lights.end(), LightDynamic());

	initializeCgOperations();

	generateSphere(sphere, 0.75f, 2);
	sphere.generateVertexNormals();
	sphere.diffuse = Util::Color(1.0f, 1.0f, 0.0f, 0.0f);

	for (unsigned int i = 0; i < lights.size(); ++i) {
		float red = lights[i]->getColor().getRed();
		float green = lights[i]->getColor().getGreen();
		float blue = lights[i]->getColor().getBlue();

		lightData[4 * i] =     red;
		lightData[4 * i + 1] = green;
		lightData[4 * i + 2] = blue;
	}

	textureSet->fillTexture(lightColorTexture, GL_RGBA, GL_FLOAT, &*lightData.begin());
	textureSet->fillTexture(lightColorOriginalTexture, GL_RGBA, GL_FLOAT, &*lightData.begin());

	updateLights(false);
}


Scene::~Scene()
{
}


void Scene::setDisplayLights(bool displayLights)
{
	this->displayLights = displayLights;
}


bool Scene::getDisplayLights() const
{
	return displayLights;
}

void Scene::compute()
{
	updateLights();
}

void Scene::render(const Camera &cam)
{
	float modelMatrix[16];

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	cam.transform();

	glGetFloatv(GL_MODELVIEW_MATRIX, viewMatrix);

	if (displayLights)
		drawLights();

	for (unsigned int i = 0; i < sceneObjects.size(); ++i) {
		glLoadIdentity();

		sceneObjects[i]->transform();

		glGetFloatv(GL_MODELVIEW_MATRIX, modelMatrix);

		glLoadMatrixf(viewMatrix);
		glMultMatrixf(modelMatrix);

		updateLightingParameters(modelMatrix, cam.getPosition(), sceneObjects[i]->getMaterial(0));

		cgFragmentLightingVS->loadModelViewProjection(cgFragmentLightingVS->getParameterId("modelViewProjection"));
		cgFragmentLightingVS->activate();

		if (sceneObjects[i]->getMaterial(0).getHasTexture())
			cgFragmentLightingTexturedFS->activate();
		else
			cgFragmentLightingFS->activate();

		sceneObjects[i]->render();

		if (sceneObjects[i]->getMaterial(0).getHasTexture())
			cgFragmentLightingTexturedFS->deactivate();
		else
			cgFragmentLightingFS->deactivate();

		cgFragmentLightingVS->deactivate();
	}

	glPopMatrix();
}


void Scene::updateLights(bool sample)
{
	for (unsigned int i = 0; i < lights.size(); ++i) {
		lightPositionsWs[i] = lights[i]->getPosition();

		lightData[4 * i] =     lights[i]->getPosition().x;
		lightData[4 * i + 1] = lights[i]->getPosition().y;
		lightData[4 * i + 2] = lights[i]->getPosition().z;
	}

	textureSet->fillTexture(lightPositionTexture, GL_RGBA, GL_FLOAT, &*lightData.begin());

	if (sample && numDynamicLights > 0)
		sampleLightColors();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glTranslatef(fireLocation.x, fireLocation.y, fireLocation.z);
	glScalef(fireScale, fireScale, fireScale);

	float fireMatrixf[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, fireMatrixf);

	Matrix4x4 fireMatrix(fireMatrixf, false);

	for (unsigned int i = 0; i < numDynamicLights; ++i)
		lightPositionsWs[i] = fireMatrix.multiplyVec3Pos(lightPositionsWs[i]);
}


void Scene::drawLights()
{
	glPushMatrix();
	glLoadIdentity();
	glLoadMatrixf(viewMatrix);

	for (unsigned int i = 0; i < lights.size(); ++i) {
		glPushMatrix();
		glTranslatef(lightPositionsWs[i].x, lightPositionsWs[i].y, lightPositionsWs[i].z);
		sphere.diffuse = lights[i]->getColor();
		displayMesh(sphere);
		glPopMatrix();
	}

	glPopMatrix();
}


void Scene::sampleLightColors()
{
	assert(numDynamicLights > 0);

	OrthoNormalProjection ortho;

	textureSet->bindFbo();
	sampleLightColorsOp->compute();
}


void Scene::setFireLocation(const Math::Vector3d &location)
{
	fireLocation = location;
}


void Scene::setFireScale(float scale)
{
	fireScale = scale;
}

void Scene::setGlobalAmbient(const Util::Color &globalAmbient)
{
	this->globalAmbient = globalAmbient;

	float ga[4] = {globalAmbient.getRed(), globalAmbient.getGreen(), 
	               globalAmbient.getBlue(), globalAmbient.getAlpha() };

	cgFragmentLightingFS->setParameter("globalAmbient", ga);
	cgFragmentLightingTexturedFS->setParameter("globalAmbient", ga);
}

OpenGL::IoTextureSet *Scene::getIoTextureSet()
{
	return textureSet.get();
}


void Scene::updateLightingParameters(float modelMatrix[16], const Math::Vector3d &viewPos, const Material &mat)
{
	Matrix4x4 model(modelMatrix, false);
	Matrix4x4 modelInv = model.getInverse();

	Vector3d viewPosOS = modelInv.multiplyVec3Pos(viewPos);

	float viewPosOSf[4] = { viewPosOS.x, viewPosOS.y, viewPosOS.z, 0.0f };

	//float modelInvf[16];
	//modelInv.getMatrix(modelInvf, false);

	for (unsigned int i = 0; i < lights.size(); ++i) {
		Math::Vector3d position = modelInv.multiplyVec3Pos(lightPositionsWs[i]);

		lightData[4 * i] =     position.x;
		lightData[4 * i + 1] = position.y;
		lightData[4 * i + 2] = position.z;
	}

	textureSet->fillTexture(lightPositionTexture, GL_RGBA, GL_FLOAT, &*lightData.begin());

	cgFragmentLightingFS->setParameter("eyePosition", viewPosOSf);
	cgFragmentLightingTexturedFS->setParameter("eyePosition", viewPosOSf);

	cgFragmentLightingFS->setParameter("ka", mat.getKa());
	cgFragmentLightingFS->setParameter("kd", mat.getKd());
	cgFragmentLightingFS->setParameter("ks", mat.getKs());
	cgFragmentLightingFS->setParameter("shininess", mat.getShininess());

	cgFragmentLightingTexturedFS->setParameter("ka", mat.getKa());
	cgFragmentLightingTexturedFS->setParameter("kd", mat.getKd());
	cgFragmentLightingTexturedFS->setParameter("ks", mat.getKs());
	cgFragmentLightingTexturedFS->setParameter("shininess", mat.getShininess());

	if (mat.getHasTexture())
		cgFragmentLightingTexturedFS->setFreeTextureParameter("texture", mat.getTexture());
}


void Scene::addSceneObject(SceneObject &sceneObject)
{
	sceneObjects.push_back(&sceneObject);
}


void Scene::initializeCgOperations()
{
	CgContext &context = CgContext::getInstance();

	cgFragmentLightingVS = autoPtr(new CgProgram(context.getCgProgramPath() + "lighting.cg", 
	                               context.getVertexShader(), "fragmentLightingVS", 0));

	cgFragmentLightingFS = autoPtr(new CgProgram(context.getCgProgramPath() + "lighting.cg", 
	                               context.getFragmentShader(), "fragmentLightingFS", textureSet.get()));

	cgFragmentLightingFS->setTextureParameter("lightColors", lightColorTexture);
	cgFragmentLightingFS->setTextureParameter("lightPositions", lightPositionTexture);
	cgFragmentLightingFS->setParameter("numLights", lights.size());

	cgFragmentLightingTexturedFS = autoPtr(new CgProgram(context.getCgProgramPath() + "lighting.cg", 
	                                       context.getFragmentShader(), "fragmentLightingTexturedFS", textureSet.get()));

	cgFragmentLightingTexturedFS->setTextureParameter("lightColors", lightColorTexture);
	cgFragmentLightingTexturedFS->setTextureParameter("lightPositions", lightPositionTexture);
	cgFragmentLightingTexturedFS->setParameter("numLights", lights.size());

	if (numDynamicLights > 0) {
		std::string sampleLightColorsEntry = slice ? "sampleLightColorsSlice" : "sampleLightColors";

		cgSampleLightColors = autoPtr(new CgProgram(context.getCgProgramPath() + "lighting.cg", 
									context.getFragmentShader(), sampleLightColorsEntry, textureSet.get()));
		cgSampleLightColors->setTextureParameter("lightPositions", lightPositionTexture);
		cgSampleLightColors->setTextureParameter("fireColorField", fireColorField);
		cgSampleLightColors->setTextureParameter("lightColors", lightColorOriginalTexture);
		cgSampleLightColors->setParameter("computationWidth", fireSimulation->getFlat3dTextureSet()->getComputationWidth());

		if (slice) {
			cgSampleLightColors->setTextureParameter("sliceLookup", sliceLookup);
			cgSampleLightColors->setParameter("computationHeight", fireSimulation->getFlat3dTextureSet()->getComputationHeight());
			cgSampleLightColors->setParameter("sliceCount", fireSimulation->getFlat3dTextureSet()->getComputationDepth());
		} else
			cgSampleLightColors->setTextureParameter("offsetLookup", offsetLookup);

		sampleLightColorsOp = autoPtr(new Texture2dOperation(numDynamicLights, 1, cgSampleLightColors.get(), false, false));
		sampleLightColorsOp->setOutputTexture(lightColorTexture);
	}
}

}