#ifndef GPUTHESIS_SCENE_H
#define GPUTHESIS_SCENE_H

#include <vector>

#include "Light.h"

#include "math/Vector3d.h"

#include "geometry/Mesh.h"

#include "util/Color.h"

namespace OpenGL {
class Camera;
class CgProgram;
class CgOperation;
class IoTextureSet;
}

namespace GpuThesis {

class Fire3dSimulator;
class Material;
class SceneObject;

/**
 * Represents a Scene with contained lights and objects.
 */
class Scene {
public:
	/**
	 * Initialize the scene.
	 * The Fire3dSimulation is needed for dynamic lights.
	 * @param lights the lights in the scene.
	 * @param fireSimulation the fire simulation used for lighting.
	 * @param slice whether the fire simulation is a slice simulation.
	 */
	Scene(const std::vector<Light *> &lights, Fire3dSimulator *fireSimulation, bool slice);
	~Scene();

	/**
	 * Perform per-frame computations (light sampling).
	 */
	void compute();

	/**
	 * Render the scene from the given viewpoint.
	 * @param cam a Camera specifying the viewpoint.
	 */
	void render(const OpenGL::Camera &cam);

	/**
	 * Add a scene object to the Scene.
	 * @param sceneObject the scene object to add.
	 */
	void addSceneObject(SceneObject &sceneObject);

	/**
	 * Set the location of the fire.
	 * Used for dynamic lights.
	 * @param location the location of the fire.
	 */
	void setFireLocation(const Math::Vector3d &location);

	/**
	 * Set the scale of the fire.
	 * Used for positioning of dynamic lights.
	 * @param scale The scale of the fire.
	 */
	void setFireScale(float scale);

	/**
	 * Set the global ambient color.
	 * The global ambient affects all scene objects based on their material properties.
	 * @param globalAmbient the global ambient color.
	 */
	void setGlobalAmbient(const Util::Color &globalAmbient);

	/**
	 * Set whether to display the point lights as spheres.
	 * Used mainly for debugging.
	 * @param displayLights whether to display lights as spheres.
	 */
	void setDisplayLights(bool displayLights);

	/**
	 * Check whether lights are displayed as spheres.
	 * @return whether lights are displayed as spheres.
	 */
	bool getDisplayLights() const;

	/**
	 * Get the texture set used for light positions and colors.
	 * @return the internal texture set.
	 */
	OpenGL::IoTextureSet *getIoTextureSet();

private:
	void initializeCgOperations();
	void drawLights();
	void sampleLightColors();

	void updateLights(bool sample = true);
	void updateLightingParameters(float modelMatrix[16], const Math::Vector3d &viewPos, const Material &mat);

	std::vector<Light *> lights;

	std::vector<SceneObject *> sceneObjects;

	std::auto_ptr<OpenGL::CgProgram> cgFragmentLightingFS,
	                                 cgFragmentLightingTexturedFS,
	                                 cgFragmentLightingVS,
	                                 cgSampleLightColors;

	std::auto_ptr<OpenGL::CgOperation> sampleLightColorsOp;

	unsigned int lightColorOriginalTexture;

	unsigned int lightColorTexture;
	unsigned int lightPositionTexture;

	unsigned int fireColorField;
	unsigned int sliceLookup;
	unsigned int offsetLookup;

	Fire3dSimulator *fireSimulation;

	bool slice;

	std::auto_ptr<OpenGL::IoTextureSet> textureSet;

	std::vector<Math::Vector3d> lightPositionsWs;

	std::vector<float> lightData;

	unsigned int numDynamicLights;

	Math::Vector3d fireLocation;
	float fireScale;

	Util::Color globalAmbient;

	Geometry::Mesh sphere;

	float viewMatrix[16];

	bool displayLights;
};

}

#endif