#include "Terrain.h"

#include "util/Perlin.h"
#include "image/Image.h"

#include <cassert>
#include <ctime>

#include "GL/glew.h"

namespace GpuThesis {

Terrain::Terrain(unsigned int width, unsigned int height, const std::string &texture,
                 float amplitude, float frequency, float textureScale, float scale, bool useVBO)
	: heightMap(width * height),
	  normalMap(width * height),
	  width(width),
	  height(height),
	  perlin(new Perlin2D(3, 0.45f, 110000)),
	  amplitude(amplitude),
	  frequency(frequency),
	  textureScale(textureScale),
	  useVBO(useVBO),
	  rotationAngle(0),
	  scale(scale)
{
	for (unsigned int y = 0; y < height; ++y) {
		for (unsigned int x = 0; x < width; ++x) {

			float fX = static_cast<float>(x);
			float fY = static_cast<float>(y);

			float hCentre = static_cast<float>(calculateHeight(fX, fY));

			float delta = 0.01f;

			float dx = calculateHeight(fX + delta, fY) - hCentre;
			float dy = calculateHeight(fX, fY + delta) - hCentre;

			Math::Vector3d vDx(delta, 0.0f, dx);
			Math::Vector3d vDy(0.0f, delta, dy);

			heightMap[y * width + x] = hCentre;

			normalMap[y * width + x] = Math::normalize(cross(vDx, vDy));
		}
	}

	if (useVBO)
		makeVBO();
	else
		makeDisplayList();
		
	loadTexture(texture);
}


Terrain::~Terrain()
{
	glDeleteTextures(1, &texture);
}


void Terrain::setMaterial(const Material &material)
{
	mat = material;

	mat.setHasTexture(true);
	mat.setTexture(getTexture());
}


float Terrain::getHeight(float x, float y) const
{
	assert(x > 0 && x < width && y > 0 && y < width);

	unsigned int iX = static_cast<unsigned int>(x);
	unsigned int iY = static_cast<unsigned int>(y);

	return heightMap[iX + iY * width];
}



float Terrain::calculateHeight(float x, float y) const
{
	return amplitude * perlin->sample(x * frequency, y * frequency);
}



void Terrain::makeDisplayList()
{
	displayList = glGenLists(1);

	glNewList(displayList, GL_COMPILE);

	glBegin(GL_TRIANGLES);

	for (unsigned int y = 0; y < height - 1; ++y) {
		for (unsigned int x = 0; x < width - 1; ++x) {
			float upperLeft  = heightMap[y * width + x];
			float upperRight = heightMap[y * width + x + 1];
			float lowerLeft  = heightMap[(y + 1) * width + x];
			float lowerRight = heightMap[(y + 1) * width + x + 1];

			const Math::Vector3d &upperLeftNormal  = normalMap[y * width + x];
			const Math::Vector3d &upperRightNormal = normalMap[y * width + x + 1];
			const Math::Vector3d &lowerLeftNormal  = normalMap[(y + 1) * width + x];
			const Math::Vector3d &lowerRightNormal = normalMap[(y + 1) * width + x + 1];

			float fX = static_cast<GLfloat>(x) - width * 0.5f;
			float fY = static_cast<GLfloat>(y) - height * 0.5f;

			glNormal3f(upperLeftNormal.x, upperLeftNormal.y, upperLeftNormal.z);
			glVertex3f(scale * fX, scale * fY, scale * upperLeft);

			glNormal3f(lowerRightNormal.x, lowerRightNormal.y, lowerRightNormal.z);
			glVertex3f(scale * (fX + 1.0f), scale * (fY + 1.0f), scale * lowerRight);

			glNormal3f(lowerLeftNormal.x, lowerLeftNormal.y, lowerLeftNormal.z);
			glVertex3f(scale * fX, scale * (fY + 1.0f), scale * lowerLeft);

			glNormal3f(upperLeftNormal.x, upperLeftNormal.y, upperLeftNormal.z);
			glVertex3f(scale * fX, scale * fY, scale * upperLeft);

			glNormal3f(upperRightNormal.x, upperRightNormal.y, upperRightNormal.z);
			glVertex3f(scale * (fX + 1.0f), scale * fY, scale * upperRight);

			glNormal3f(lowerRightNormal.x, lowerRightNormal.y, lowerRightNormal.z);
			glVertex3f(scale * (fX + 1.0f), scale * (fY + 1.0f), scale * lowerRight);
		}
	}

	glEnd();

	glEndList();
}

void Terrain::makeVBO() 
{
	std::vector<float> vertices;
	std::vector<float> normals;	
	std::vector<float> texCoords;

	float dX = textureScale / static_cast<float>(width);
	float dY = textureScale / static_cast<float>(height);

	for (unsigned int y = 0; y < height - 1; ++y) {
		for (unsigned int x = 0; x < width - 1; ++x) {
			float upperLeft  = heightMap[y * width + x];
			float upperRight = heightMap[y * width + x + 1];
			float lowerLeft  = heightMap[(y + 1) * width + x];
			float lowerRight = heightMap[(y + 1) * width + x + 1];

			const Math::Vector3d &upperLeftNormal  = normalMap[y * width + x];
			const Math::Vector3d &upperRightNormal = normalMap[y * width + x + 1];
			const Math::Vector3d &lowerLeftNormal  = normalMap[(y + 1) * width + x];
			const Math::Vector3d &lowerRightNormal = normalMap[(y + 1) * width + x + 1];

			float fX = static_cast<GLfloat>(x) - width * 0.5f;
			float fY = static_cast<GLfloat>(y) - height * 0.5f;
			
			float tX = x * dX;
			float tY = y * dY;

			normals.push_back(upperLeftNormal.x);
			normals.push_back(upperLeftNormal.y);
			normals.push_back(upperLeftNormal.z);
			vertices.push_back(scale * (fX));
			vertices.push_back(scale * (fY));
			vertices.push_back(scale * (upperLeft));
			texCoords.push_back(tX);
			texCoords.push_back(-tY);

			normals.push_back(lowerRightNormal.x);
			normals.push_back(lowerRightNormal.y);
			normals.push_back(lowerRightNormal.z);
			vertices.push_back(scale * (fX + 1.0f));
			vertices.push_back(scale * (fY + 1.0f));
			vertices.push_back(scale * (lowerRight));		
			texCoords.push_back(tX + dX);
			texCoords.push_back(-(tY + dY));

			normals.push_back(lowerLeftNormal.x);
			normals.push_back(lowerLeftNormal.y);
			normals.push_back(lowerLeftNormal.z);
			vertices.push_back(scale * (fX));
			vertices.push_back(scale * (fY + 1.0f));
			vertices.push_back(scale * (lowerLeft));	
			texCoords.push_back(tX);
			texCoords.push_back(-(tY + dY));

			normals.push_back(upperLeftNormal.x);
			normals.push_back(upperLeftNormal.y);
			normals.push_back(upperLeftNormal.z);
			vertices.push_back(scale * (fX));
			vertices.push_back(scale * (fY));
			vertices.push_back(scale * (upperLeft));	
			texCoords.push_back(tX);
			texCoords.push_back(-tY);

			normals.push_back(upperRightNormal.x);
			normals.push_back(upperRightNormal.y);
			normals.push_back(upperRightNormal.z);
			vertices.push_back(scale * (fX + 1.0f));
			vertices.push_back(scale * (fY));
			vertices.push_back(scale * (upperRight));
			texCoords.push_back(tX + dX);
			texCoords.push_back(-tY);

			normals.push_back(lowerRightNormal.x);
			normals.push_back(lowerRightNormal.y);
			normals.push_back(lowerRightNormal.z);
			vertices.push_back(scale * (fX + 1.0f));
			vertices.push_back(scale * (fY + 1.0f));
			vertices.push_back(scale * (lowerRight));
			texCoords.push_back(tX + dX);
			texCoords.push_back(-(tY + dY));
		}
	}

	glGenBuffersARB(1, &vertexVBO);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexVBO);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertices.size() * sizeof(float),
					&*vertices.begin(), GL_STATIC_DRAW_ARB);

	glGenBuffersARB(1, &normalVBO);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, normalVBO);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, normals.size() * sizeof(float),
					&*normals.begin(), GL_STATIC_DRAW_ARB );

	glGenBuffersARB(1, &texCoordsVBO);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, texCoordsVBO);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, texCoords.size() * sizeof(float),
	                &*texCoords.begin(), GL_STATIC_DRAW_ARB);
}


void Terrain::transform()
{
	glTranslatef(position.x, position.y, position.z);
	glRotatef(rotationAngle, rotationAxis.x, rotationAxis.y, rotationAxis.z);
}


void Terrain::render(unsigned int)
{
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

	glEnable(GL_COLOR_MATERIAL);

	if (useVBO) {
		glColor4f(1, 1, 1, 1);

		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, texture);

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexVBO);
		glVertexPointer(3, GL_FLOAT, 0, 0);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, normalVBO);
		glNormalPointer(GL_FLOAT, 0, 0);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, texCoordsVBO);
		glTexCoordPointer(2, GL_FLOAT, 0, 0);

		glDrawArrays(GL_TRIANGLES, 0, (height - 1) * (width - 1) * 6);

		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);

		glDisable(GL_TEXTURE_2D);

	} else {
		
		glCallList(displayList);
	}

	glDisable(GL_COLOR_MATERIAL);
}

bool Terrain::isLit()
{
	return true;
}

unsigned int Terrain::getMeshCount()
{
	return 1;
}

const Material& Terrain::getMaterial(unsigned int meshNumber)
{
	return mat;
}

void Terrain::setPosition(const Math::Vector3d &position)
{
	this->position = position;
}

void Terrain::setRotationAxis(const Math::Vector3d &rotationAxis)
{
	this->rotationAxis = rotationAxis;
}

void Terrain::setRotationAngle(float angle)
{
	this->rotationAngle = angle;
}

GLuint Terrain::getTexture() const
{
	return static_cast<GLuint>(texture);
}

void Terrain::loadTexture(const std::string &fileName)
{
	Image::Image image(fileName);

	glGenTextures(1, &texture);

	glBindTexture(GL_TEXTURE_2D, texture);

	glTexImage2D(GL_TEXTURE_2D,
	             0,
	             GL_RGB, 
	             image.getWidth(),
	             image.getHeight(),
	             0,
	             GL_BGR,
	             GL_UNSIGNED_BYTE,
	             image.getData()); // pointer to data

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}



}