#ifndef GPUTHESIS_TERRAIN_H
#define GPUTHESIS_TERRAIN_H

#include "math/Vector3D.h"
#include "opengl/OpenGL.h"
#include "SceneObject.h"
#include "Material.h"

#include <memory>
#include <vector>

class Perlin2D;

namespace GpuThesis {

/**
 * Terrain class.
 * Represents a noisy surface.
 */
class Terrain : public SceneObject {
public:
	/**
	 * Initialize the terrain.
	 * @param width the width in vertices.
	 * @param height the height in vertices.
	 * @param texture the texture to use.
	 * @param amplitude the amplitude of the terrain.
	 * @param frequency the frequency of the terrain (the smaller the smoother).
	 * @param textureScale the scale of the texture.
	 * @param scale the scale of the terrain.
	 * @param useVBO whether to use vertex buffer objects (currently required for texturing).
	 */
	Terrain(unsigned int width, unsigned int height, const std::string &texture,
	        float amplitude, float frequency, float textureScale, float scale, bool useVBO = true);
	~Terrain();

	/**
	 * Set the material.
	 * @param material the material.
	 */
	void setMaterial(const Material &material);

	/**
	 * Get the height of the terrain at the given coordinates.
	 * @return the terrain height.
	 */
	float getHeight(float x, float y) const;

	/**
	 * Get the texture used by the terrain.
	 * @return the OpenGL texture id.
	 */
	GLuint getTexture() const;

	// From SceneObject
	/**
	 * Transform the terrain.
	 */
	void transform();

	/**
	 * Render the terrain.
	 */
	void render(unsigned int meshNumber = 0);

	/**
	 * Check whether the terrain should be lit.
	 * @return whether the terrain should be lit.
	 */
	bool isLit();

	/**
	 * Get the mesh count of the terrain.
	 * @return the terrain's mesh count.
	 */
	unsigned int getMeshCount(); 

	/**
	 * Get the material of the given mesh.
	 * @param meshNumber the given mesh.
	 * @return the material.
	 */
	const Material& getMaterial(unsigned int meshNumber);

	/**
	 * Set the terrain position.
	 * @param position the new position.
	 */
	void setPosition(const Math::Vector3d &position);

	/**
	 * Set the rotation axis.
	 * @param rotationAxis the new rotation axis.
	 */
	void setRotationAxis(const Math::Vector3d &rotationAxis);

	/**
	 * Set the rotation angle.
	 * @param angle the new rotation angle.
	 */
	void setRotationAngle(float angle);

private:
	void makeDisplayList();
	void makeVBO();

	void loadTexture(const std::string &fileName);

	float calculateHeight(float x, float y) const;

	std::vector<float> heightMap;
	std::vector<Math::Vector3d> normalMap;

	unsigned int width, height;

	std::auto_ptr<Perlin2D> perlin;
	float amplitude;
	float frequency;

	float textureScale;

	unsigned int displayList;
	
	unsigned int texture;

	unsigned int vertexVBO;
	unsigned int normalVBO;
	unsigned int texCoordsVBO;

	bool useVBO;

	Material mat;

	Math::Vector3d position;
	Math::Vector3d rotationAxis;
	float rotationAngle;
	float scale;
};

}

#endif