#include "VolumetricExtrusion.h"

#include "opengl/CgContext.h"
#include "opengl/CgFactory.h"
#include "opengl/CgProgram.h"
#include "opengl/Flat3dTextureOperation.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/OrthoNormalProjection.h"
#include "opengl/PrimaryTextureContainer.h"
#include "opengl/SecondaryTextureContainer.h"

using namespace OpenGL;

namespace GpuThesis {

VolumetricExtrusion::VolumetricExtrusion(OpenGL::Flat3dTextureSet *importTextures,
                                                     unsigned int sliceTexture)
{
	initialize(importTextures, sliceTexture);
}

VolumetricExtrusion::~VolumetricExtrusion() 
{	
}

TextureContainer *VolumetricExtrusion::getTextureContainer()
{
	return textureContainer.get();
}

TextureContainer *VolumetricExtrusion::getSecondaryTextureContainer()
{
	return secondaryContainer.get();
}

Flat3dTextureSet *VolumetricExtrusion::getTextureSet()
{
	return textures.get();
}

void VolumetricExtrusion::initialize(Flat3dTextureSet *importTextures, unsigned int sliceTexture)
{		
	unsigned int sliceWidth = importTextures->getComputationWidth();
	unsigned int sliceHeight = importTextures->getComputationHeight();
	unsigned int numSlices = importTextures->getComputationDepth();

	CgContext &context = CgContext::getInstance();
	textures = std::auto_ptr<Flat3dTextureSet>(
		new Flat3dTextureSet(sliceWidth,
		                     sliceHeight, 
		                     sliceWidth,
		                     context.getTexRectTarget(),
		                     context.getTexRectInternalFormat()));

	volumetricTexture = textures->addPrimaryTexture("Volumetric extrusion");	

	cgVolumetricExtrusion = OpenGL::createFragmentProgram("slice.cg", 
	                                                      "volumetricExtrusion", 
	                                                      textures.get());

	float computationDimensions[2] = {static_cast<float>(sliceWidth), 
	                                  static_cast<float>(sliceHeight)};
	cgVolumetricExtrusion->setParameter("computationDimensions", computationDimensions);
	cgVolumetricExtrusion->setParameter("sliceCount", numSlices);
	cgVolumetricExtrusion->setParameter("scale", 1.0f);
	cgVolumetricExtrusion->setTextureParameter("sliceTexture", 
		textures->importSecondaryTexture(importTextures, sliceTexture, "Imported slice texture"));
	cgVolumetricExtrusion->setTextureParameter("sliceLookup", 
		textures->importSecondaryTexture(importTextures, importTextures->getSliceLookup(),
		                                 "Imported slice lookup"));
	cgVolumetricExtrusion->setTextureParameter("domainLookup", textures->getDomainLookup());


	volumetricExtrusionOp = OpenGL::createFlat3dTextureOperation(textures.get(), 
	                                                             cgVolumetricExtrusion.get(), 
	                                                             Flat3dTextureOperation::INTERIOR_FULL);
	volumetricExtrusionOp->setOutputTexture(volumetricTexture);	

	textureContainer = std::auto_ptr<TextureContainer>(new PrimaryTextureContainer(textures.get()));
	textureContainer->setDescription("Volumetric fire extrusion");

	secondaryContainer = std::auto_ptr<TextureContainer>(new SecondaryTextureContainer(textures.get()));
	secondaryContainer->setDescription("Secondary volumetric fire extrusion");
}

void VolumetricExtrusion::doRun(float timestep) 
{
	OrthoNormalProjection ortho;

	textures->bindFbo();
	volumetricExtrusionOp->compute();
	textures->disableFbo();
}

void VolumetricExtrusion::doReset()
{

}


}
