#include "Fluid3dTests.h"

#include <cmath>

#include "test-suite/GpuTest.h"

#include "Fluid3dSolver.h"
#include "FluidFactory.h"

#include "util/VoxelGrid.h"
#include "util/Color.h"
#include "util/Util.h"

#include "opengl/CgContext.h"
#include "opengl/CgFactory.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/OrthoNormalProjection.h"

using namespace Fluid3dSolver;
using namespace GpuThesis;
using namespace OpenGL;
using namespace Util;

namespace {

std::auto_ptr<Fluid3dSolver::FluidFactory> createFluidFactory() 
{
	return std::auto_ptr<Fluid3dSolver::FluidFactory>(
		new Fluid3dSolver::FluidFactoryImpl(false));
}

std::auto_ptr<Flat3dTextureSet> createTextureSet(unsigned int width,
                                                 unsigned int height,
                                                 unsigned int depth)
{
	CgContext &context = CgContext::getInstance();

	return std::auto_ptr<Flat3dTextureSet>(
		new Flat3dTextureSet(width, height, depth, 
		                     context.getTexRectTarget(), 
		                     context.getTexRectInternalFormat()));
}

}

namespace GpuThesis {

namespace Test {

Advection3dTest::Advection3dTest() :
	width(5), height(5), depth(5), factory(createFluidFactory())
{
	textures = createTextureSet(width, height, depth);

	// Advection Interior
	velocity = textures->addPrimaryTexture("Advection 3d velocity");
	value = textures->addPrimaryTexture("Advection 3d value");
	boundary = Fluid3dSolver::DENSITY_CLOSED;
	advect = factory->createAdvection(textures.get(), 
	                                  velocity, 
	                                  value, 
	                                  boundary);	
	advect->setVoxelSize(1.0f);

	setIoTextureSet(textures.get());	

	setInput();
	setOuput();
}

Advection3dTest::~Advection3dTest()
{
}

void Advection3dTest::doRun(float timestep)
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();
	
	advect->run(timestep);
	
	textures->disableFbo();
}

void Advection3dTest::doReset()
{
	setInput();
}

void Advection3dTest::setInput()
{
	VoxelGrid<Color> values(width, height, depth);
	VoxelGrid<Color> velocities(width, height, depth);

	values(3, 3, 3) = Color(1, 1, 1, 1);
	velocities(2, 2, 2) = Color(-1, -1, -1, 0);

	textures->fillTexture(value, values);
	textures->fillTexture(velocity, velocities);
}

void Advection3dTest::setOuput()
{
	ColorMap output;

	output[textures->getGlobalCoordinates(3, 3, 3)] = Color(1, 1, 1, 1);
	output[textures->getGlobalCoordinates(2, 2, 2)] = Color(1, 1, 1, 1);

	setExpectedOutput(value, output, 1.0f);

	output[textures->getGlobalCoordinates(2, 2, 2)] = Color(0.125f, 0.125f, 0.125f, 0.125f);

	setExpectedOutput(value, output, 0.5f);
}

Projection3dTest::Projection3dTest() :
	width(7), height(7), depth(5), factory(createFluidFactory())
{
	textures = createTextureSet(width, height, depth);

	velocity = textures->addPrimaryTexture("Projection3dTest::Velocity");
	boundary = Fluid3dSolver::VELOCITY_ROUGH;
	project = factory->createProjection(textures.get(),
	                                    velocity, 
		                                boundary);

	project->setVoxelSize(1.0f);
	project->setNumIterations(2);

	setIoTextureSet(textures.get());	

	setInput();
	setOutput();
}

Projection3dTest::~Projection3dTest()
{
}

void Projection3dTest::doRun(float timestep)
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();
	
	project->run();
	
	textures->disableFbo();
}

void Projection3dTest::doReset()
{
	setInput();
}

void Projection3dTest::setInput()
{
	VoxelGrid<Color> velocities(7, 7, 5);

	velocities(3, 3, 2) = Color(12, 0, 0, 0);

	textures->fillTexture(velocity, velocities);
}

void Projection3dTest::setOutput()
{
	ColorMap output;

	output[textures->getGlobalCoordinates(3, 3, 2)] = Color(11, 0, 0, 0);
	output[textures->getGlobalCoordinates(4, 3, 2)] = Color(-1.0f/12.0f, 0, 0, 0);
	output[textures->getGlobalCoordinates(5, 3, 2)] = Color(0.5f, 0, 0, 0); //Color(5.0f/12.0f, 0, 0, 0);
	output[textures->getGlobalCoordinates(5, 2, 2)] = Color(1.0f/12.0f, -1.0f/12.0f, 0, 0);
	output[textures->getGlobalCoordinates(4, 3, 1)] = Color(0.0f, 0, -0.5f, 0); //Color(0, 0, -5.0f/12.0f, 0);

	setExpectedOutput(velocity, output, 0.0f);
}

}

}