#include "FluidSliceTests.h"

#include <cmath>

#include "test-suite/GpuTest.h"

#include "Fluid3dSolver.h"
#include "FluidFactory.h"
#include "VolumetricExtrusion.h"

#include "util/VoxelGrid.h"
#include "util/Color.h"
#include "util/Util.h"

#include "opengl/CgContext.h"
#include "opengl/CgFactory.h"
#include "opengl/Flat3dTextureSet.h"
#include "opengl/OrthoNormalProjection.h"

using namespace Fluid3dSolver;
using namespace GpuThesis;
using namespace OpenGL;
using namespace Util;

namespace {

std::auto_ptr<Fluid3dSolver::FluidFactory> createFluidFactory() 
{
	return std::auto_ptr<Fluid3dSolver::FluidFactory>(
		new Fluid3dSolver::FluidFactoryImpl(true));
}

std::auto_ptr<Flat3dTextureSet> createTextureSet(unsigned int width,
                                                 unsigned int height,
                                                 unsigned int numSlices)
{
	CgContext &context = CgContext::getInstance();

	return std::auto_ptr<Flat3dTextureSet>(
		new Flat3dTextureSet(width, height, numSlices, 
		                     context.getTexRectTarget(), 
		                     context.getTexRectInternalFormat()));
}

}

namespace GpuThesis {
namespace Test {

AdvectionSliceTest::~AdvectionSliceTest()
{
}

AdvectionSliceTest::AdvectionSliceTest() :
	width(5), height(5), numSlices(2), factory(createFluidFactory())
{
	textures = createTextureSet(width, height, numSlices);


	// Advection Interior
	interiorVelocity = textures->addPrimaryTexture("Advection Interior Velocity");
	interiorValue = textures->addPrimaryTexture("Advection Interior Value");
	interiorBoundary = Fluid3dSolver::DENSITY_CLOSED;
	interior = factory->createAdvection(textures.get(), 
	                                     interiorVelocity, 
	                                     interiorValue, 
	                                     interiorBoundary);	
	interior->setVoxelSize(1.0f);


	// Advection Boundary
	boundaryVelocity = textures->addPrimaryTexture("Advection Boundary Velocity");
	boundaryValue = textures->addPrimaryTexture("Advection Boundary Value");
	boundaryBoundary = Fluid3dSolver::DENSITY_CLOSED;
	boundary = factory->createAdvection(textures.get(), 
	                                    boundaryVelocity, 
	                                    boundaryValue, 
	                                    boundaryBoundary);	
	boundary->setVoxelSize(1.0f);


	// Velocity Slip Boundary
	slipVelocity = textures->addPrimaryTexture("Velocity Slip Boundary");
	slipBoundary = Fluid3dSolver::VELOCITY_SMOOTH;
	slip = factory->createAdvection(textures.get(),
	                                slipVelocity,
	                                slipVelocity,
	                                slipBoundary);
	slip->setVoxelSize(1.0f);


	// Velocity No Slip Boundary
	noSlipVelocity = textures->addPrimaryTexture("Velocity No Slip Boundary");
	noSlipBoundary = Fluid3dSolver::VELOCITY_ROUGH;
	noSlip = factory->createAdvection(textures.get(),
	                                  noSlipVelocity,
	                                  noSlipVelocity,
	                                  noSlipBoundary);
	noSlip->setVoxelSize(1.0f);


	setIoTextureSet(textures.get());	

	doReset();
	setOuput();
}

void AdvectionSliceTest::doRun(float timestep)
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();
	
	interior->run(timestep);
	boundary->run(timestep);
	slip->run(timestep);
	noSlip->run(timestep);
	
	textures->disableFbo();
}

void AdvectionSliceTest::doReset()
{
	textures->clearTexture(interiorValue);	
	textures->clearTexture(interiorVelocity);	

	textures->clearTexture(boundaryValue);
	textures->clearTexture(boundaryVelocity);

	textures->clearTexture(slipVelocity);

	textures->clearTexture(noSlipVelocity);

	setInput();
}

void AdvectionSliceTest::setInput()
{
	// Advection Interior
	VoxelGrid<Color> interiorValues(width, height, numSlices);
	VoxelGrid<Color> interiorVelocities(width, height, numSlices);

	for (unsigned int s = 0; s < numSlices; ++s) {
		interiorValues(1, 3, s) = Color(1, 1, 1, 1);
		interiorValues(3, 1, s) = Color(1, 1, 1, 1);
	}

	for (unsigned int s = 0; s < numSlices; ++s) {
		interiorVelocities(2, 3, s) = Color(1, 0, 0, 0);
		interiorVelocities(3, 2, s) = Color(0, 1, 0, 0);
	}

	textures->fillTexture(interiorValue, interiorValues);
	textures->fillTexture(interiorVelocity, interiorVelocities);


	// Advection Boundary
	VoxelGrid<Color> boundaryValues(width, height, numSlices);
	VoxelGrid<Color> boundaryVelocities(width, height, numSlices);

	for (unsigned int s = 0; s < numSlices; ++s) {
		boundaryValues(0, 0, s) = Color(1, 1, 1, 1);
		boundaryValues(0, 2, s) = Color(1, 1, 1, 1);
		boundaryValues(2, 0, s) = Color(1, 1, 1, 1);
		boundaryValues(2, 4, s) = Color(1, 1, 1, 1);
		boundaryValues(4, 2, s) = Color(1, 1, 1, 1);
		boundaryValues(4, 4, s) = Color(1, 1, 1, 1);
	}

	for (unsigned int s = 0; s < numSlices; ++s) {
		boundaryVelocities(1, 1, s) = Color(1, 1, 0, 0);
		boundaryVelocities(1, 2, s) = Color(1, 0, 0, 0);
		boundaryVelocities(2, 1, s) = Color(0, 1, 0, 0);
		boundaryVelocities(2, 3, s) = Color(0, -1, 0, 0);
		boundaryVelocities(3, 2, s) = Color(-1, 0, 0, 0);
		boundaryVelocities(3, 3, s) = Color(-1, -1, 0, 0);
	}

	textures->fillTexture(boundaryValue, boundaryValues);
	textures->fillTexture(boundaryVelocity, boundaryVelocities);


	// Velocity Slip & No Slip Boundary
	VoxelGrid<Color> slipVelocities(width, height, numSlices);
	VoxelGrid<Color> noSlipVelocities(width, height, numSlices);


	for (unsigned int s = 0; s < numSlices; ++s) {

		slipVelocities(1, 1, s) = Color(0, 1, 0, 0);
		slipVelocities(2, 1, s) = Color(1, 0, 0, 0);
		slipVelocities(3, 1, s) = Color(0, 1, 0, 0);
		slipVelocities(1, 2, s) = Color(1, 0, 0, 0);
		slipVelocities(3, 2, s) = Color(1, 0, 0, 0);
		slipVelocities(1, 3, s) = Color(0, 1, 0, 0);
		slipVelocities(2, 3, s) = Color(1, 0, 0, 0);
		slipVelocities(3, 3, s) = Color(0, 1, 0, 0);

		noSlipVelocities(1, 1, s) = Color(0, 1, 0, 0);
		noSlipVelocities(2, 1, s) = Color(1, 0, 0, 0);
		noSlipVelocities(3, 1, s) = Color(0, 1, 0, 0);
		noSlipVelocities(1, 2, s) = Color(1, 0, 0, 0);
		noSlipVelocities(3, 2, s) = Color(1, 0, 0, 0);
		noSlipVelocities(1, 3, s) = Color(0, 1, 0, 0);
		noSlipVelocities(2, 3, s) = Color(1, 0, 0, 0);
		noSlipVelocities(3, 3, s) = Color(0, 1, 0, 0);
	}

	textures->fillTexture(slipVelocity, slipVelocities);
	textures->fillTexture(noSlipVelocity, noSlipVelocities);


}

void AdvectionSliceTest::setOuput()
{
	// Advection Interior
	ColorMap interiorOutput;

	for (unsigned int s = 0; s < numSlices; ++s) {
		interiorOutput[textures->getGlobalCoordinates(2, 3, s)] = Color(1, 1, 1, 1);
		interiorOutput[textures->getGlobalCoordinates(3, 2, s)] = Color(1, 1, 1, 1);
	}

	setExpectedOutput(interiorValue, interiorOutput, 1.0f);


	// Advection Boundary
	ColorMap boundaryOutput;

	for (unsigned int s = 0; s < numSlices; ++s) {
		boundaryOutput[textures->getGlobalCoordinates(1, 1, s)] = Color(1, 1, 1, 1);
		boundaryOutput[textures->getGlobalCoordinates(1, 2, s)] = Color(1, 1, 1, 1);
		boundaryOutput[textures->getGlobalCoordinates(2, 1, s)] = Color(1, 1, 1, 1);
		boundaryOutput[textures->getGlobalCoordinates(2, 3, s)] = Color(1, 1, 1, 1);
		boundaryOutput[textures->getGlobalCoordinates(3, 2, s)] = Color(1, 1, 1, 1);
		boundaryOutput[textures->getGlobalCoordinates(3, 3, s)] = Color(1, 1, 1, 1);
	}

	setExpectedOutput(boundaryValue, boundaryOutput, 1.0f);
	setExpectedOutput(boundaryValue, boundaryOutput, 3.0f);
	setExpectedOutput(boundaryValue, boundaryOutput, 10.0f);


	// Velocity Slip Boundary
	ColorMap slipOutput;

	for (unsigned int s = 0; s < numSlices; ++s) {

		slipOutput[textures->getGlobalCoordinates(1, 0, s)] = Color(0, -1, 0, 0);
		slipOutput[textures->getGlobalCoordinates(2, 0, s)] = Color(1, 0, 0, 0);
		slipOutput[textures->getGlobalCoordinates(3, 0, s)] = Color(0, -1, 0, 0);
		slipOutput[textures->getGlobalCoordinates(0, 1, s)] = Color(0, 1, 0, 0);
		slipOutput[textures->getGlobalCoordinates(4, 1, s)] = Color(0, 1, 0, 0);
		slipOutput[textures->getGlobalCoordinates(0, 2, s)] = Color(-1, 0, 0, 0);
		slipOutput[textures->getGlobalCoordinates(4, 2, s)] = Color(-1, 0, 0, 0);
		slipOutput[textures->getGlobalCoordinates(0, 3, s)] = Color(0, 1, 0, 0);
		slipOutput[textures->getGlobalCoordinates(4, 3, s)] = Color(0, 1, 0, 0);
		slipOutput[textures->getGlobalCoordinates(1, 4, s)] = Color(0, -1, 0, 0);
		slipOutput[textures->getGlobalCoordinates(2, 4, s)] = Color(1, 0, 0, 0);
		slipOutput[textures->getGlobalCoordinates(3, 4, s)] = Color(0, -1, 0, 0);
	}

	setExpectedOutput(slipVelocity, slipOutput, 0.0f);

	// Velocity No Slip Boundary
	ColorMap noSlipOutput;

	for (unsigned int s = 0; s < numSlices; ++s) {

		noSlipOutput[textures->getGlobalCoordinates(1, 0, s)] = Color(0, -1, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(2, 0, s)] = Color(-1, 0, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(3, 0, s)] = Color(0, -1, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(0, 1, s)] = Color(0, -1, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(4, 1, s)] = Color(0, -1, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(0, 2, s)] = Color(-1, 0, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(4, 2, s)] = Color(-1, 0, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(0, 3, s)] = Color(0, -1, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(4, 3, s)] = Color(0, -1, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(1, 4, s)] = Color(0, -1, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(2, 4, s)] = Color(-1, 0, 0, 0);
		noSlipOutput[textures->getGlobalCoordinates(3, 4, s)] = Color(0, -1, 0, 0);
	}

	setExpectedOutput(noSlipVelocity, noSlipOutput, 0.0f);
}

DiffusionSliceTest::DiffusionSliceTest() :
	width(5), height(5), numSlices(2), factory(createFluidFactory())
{
	textures = createTextureSet(width, height, numSlices);

	valueTexture = textures->addPrimaryTexture("Diffusion Texture");
	diffusion = factory->createDiffusion(textures.get(), valueTexture, Fluid3dSolver::DENSITY_CLOSED);

	diffusion->setVoxelSize(1.0f);
	diffusion->setDiffusionRates(1.0f, 2.0f, 0.25f, 0.5f);
	diffusion->setNumJacobiSteps(2);

	setIoTextureSet(textures.get());	

	setInput();
	setOuput();
}

DiffusionSliceTest::~DiffusionSliceTest()
{
}

void DiffusionSliceTest::doRun(float timestep)
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();
	diffusion->run(timestep);
	textures->disableFbo();
}

void DiffusionSliceTest::doReset()
{
	setInput();
}

void DiffusionSliceTest::setInput()
{
	VoxelGrid<Color> values(width, height, numSlices);

	for (unsigned int s = 0; s < 2; ++s) {
		values(2, 2, s) = Color(1, 1, 1, 1);
	}

	textures->fillTexture(valueTexture, values);
}

void DiffusionSliceTest::setOuput()
{
	ColorMap output;

	for (unsigned int s = 0; s < 2; ++s) {
		output[textures->getGlobalCoordinates(2, 2, s)] = Color(1.0f/5.0f, 1.0f/9.0f, 1.0f/2.0f, 1.0f/3.0f);
		output[textures->getGlobalCoordinates(1, 2, s)] = Color(1.0f/25.0f, 2.0f/81.0f, 1.0f/16.0f, 1.0f/18.0f);
		output[textures->getGlobalCoordinates(1, 1, s)] = Color(0.0f, 0.0f, 0.0f, 0.0f);
	}

	setExpectedOutput(valueTexture, output, 1.0f);
}

DivergenceSliceTest::DivergenceSliceTest() :
	width(5), height(5), numSlices(2), factory(createFluidFactory())
{
	textures = createTextureSet(width, height, numSlices);

	velocityTexture = textures->addPrimaryTexture("Velocity texture");

	projection = factory->createProjection(textures.get(), velocityTexture, Fluid3dSolver::VELOCITY_SMOOTH);

	projection->setNumIterations(0);
	projection->setVoxelSize(1.0f);

	setIoTextureSet(textures.get());

	setInput();
	setOutput();
}

DivergenceSliceTest::~DivergenceSliceTest()
{
}

void DivergenceSliceTest::doRun(float timestep)
{
	OpenGL::OrthoNormalProjection ortho;

	textures->bindFbo();
	projection->run();
	textures->disableFbo();
}

void DivergenceSliceTest::doReset()
{
	textures->clearTexture(projection->getDivergenceTexture());

	setInput();
}

void DivergenceSliceTest::setInput()
{
	VoxelGrid<Color> velocities(width, height, numSlices);

	// slice 1
	velocities(2, 2, 0) = Color(0, 1, 0, 0);

	// slice 2
	velocities(1, 2, 1) = Color(1, 0, 0, 0);
	velocities(3, 2, 1) = Color(-1, 0, 0, 0);

	textures->fillTexture(velocityTexture, velocities);
}

void DivergenceSliceTest::setOutput()
{
	ColorMap output;

	output[textures->getGlobalCoordinates(2, 2, 0)] = Color(0, 0, 0, 0);
	output[textures->getGlobalCoordinates(1, 2, 0)] = Color(0, 0, 0, 0);
	output[textures->getGlobalCoordinates(2, 3, 0)] = Color(0.5, 0.5, 0.5, 0.5);
	output[textures->getGlobalCoordinates(2, 1, 0)] = Color(-0.5, -0.5, -0.5, -0.5);

	output[textures->getGlobalCoordinates(2, 2, 1)] = Color(1, 1, 1, 1);
	output[textures->getGlobalCoordinates(1, 2, 1)] = Color(0, 0, 0, 0);
	output[textures->getGlobalCoordinates(2, 3, 1)] = Color(0, 0, 0, 0);
	output[textures->getGlobalCoordinates(2, 1, 1)] = Color(0, 0, 0, 0);

	setExpectedOutput(projection->getDivergenceTexture(), output, 0.0f);
}


SubtractGradientSliceTest::SubtractGradientSliceTest() :
	width(5), height(5), numSlices(2)
{
	textures = createTextureSet(width, height, numSlices);

	velocityTexture = textures->addPrimaryTexture("Subtract Gradient Velocity");
	pressureTexture = textures->addPrimaryTexture("Subtract Gradient Pressure");
	cgSubtractGradient = OpenGL::createFragmentProgram("fluid2d.cg", "subtractGradient2d", textures.get());
	cgSubtractGradient->setParameter("halfVoxelSize", 0.5f);
	cgSubtractGradient->setTextureParameter("pressure", pressureTexture);
	cgSubtractGradient->setTextureParameter("velocity", velocityTexture);

	subtractGradientOp = OpenGL::createFlat3dTextureOperation(textures.get(), 
	                                                          cgSubtractGradient.get(),
	                                                          Flat3dTextureOperation::INTERIOR_2D);
	subtractGradientOp->setOutputTexture(velocityTexture);

	setIoTextureSet(textures.get());	

	doReset();
	setOutput();
}

SubtractGradientSliceTest::~SubtractGradientSliceTest()
{
}


void SubtractGradientSliceTest::doRun(float timestep)
{
	OrthoNormalProjection ortho;
	textures->bindFbo();
	
	subtractGradientOp->compute();
	
	textures->disableFbo();
}

void SubtractGradientSliceTest::doReset()
{
	textures->bindFbo();

	textures->clearTexture(velocityTexture);
	textures->clearTexture(pressureTexture);

	textures->disableFbo();

	setInput();
}

void SubtractGradientSliceTest::setInput()
{
	VoxelGrid<Color> velocities(width, height, numSlices);
	VoxelGrid<Color> pressures(width, height, numSlices);

	for (unsigned int s = 0; s < numSlices; ++s) {
		
		velocities(2, 1, s) = Color(1, 0, 0, 0);
		velocities(1, 2, s) = Color(0, 1, 0, 0);
		velocities(3, 2, s) = Color(0, 1, 0, 0);
		velocities(2, 3, s) = Color(1, 0, 0, 0);

		pressures(1, 1, s) = Color(2, 0, 0, 0);				
		pressures(3, 3, s) = Color(2, 0, 0, 0);
	}

	textures->fillTexture(velocityTexture, velocities);
	textures->fillTexture(pressureTexture, pressures);
}

void SubtractGradientSliceTest::setOutput()
{
	ColorMap output;

	for (unsigned int s = 0; s < numSlices; ++s) {
		
		output[textures->getGlobalCoordinates(2, 1, s)] = Color(2, 0, 0, 0);
		output[textures->getGlobalCoordinates(1, 2, s)] = Color(0, 2, 0, 0);
		output[textures->getGlobalCoordinates(3, 2, s)] = Color(0, 0, 0, 0);
		output[textures->getGlobalCoordinates(2, 3, s)] = Color(0, 0, 0, 0);
	}

	setExpectedOutput(velocityTexture, output, 0.0f);
}

VorticityConfinementSliceTest::VorticityConfinementSliceTest() :
	width(7), height(7), numSlices(2)
{
	textures = createTextureSet(width, height, numSlices);

	factory = createFluidFactory();

	velocityTexture = textures->addPrimaryTexture("Vorticity Confinement Velocity");
	velocityForceTexture = textures->addPrimaryTexture("Vorticity Confinement Velocity Force");
	vorticityTexture = textures->addPrimaryTexture("Vorticity Confinement Vorticity");	
		
	vorticity = factory->createVorticityConfinement(textures.get(), 
	                                                velocityTexture, 
	                                                velocityForceTexture, 
	                                                vorticityTexture);
	vorticity->setVoxelSize(1.0f);
	vorticity->setVorticityScale(1.0f);


	setIoTextureSet(textures.get());	

	doReset();
	setOutput();
}

VorticityConfinementSliceTest::~VorticityConfinementSliceTest()
{
}


void VorticityConfinementSliceTest::doRun(float timestep)
{
	OrthoNormalProjection ortho;
	textures->bindFbo();
	
	vorticity->run();
	
	textures->disableFbo();
}

void VorticityConfinementSliceTest::doReset()
{
	textures->bindFbo();

	textures->clearTexture(velocityTexture);
	textures->clearTexture(velocityForceTexture);
	textures->clearTexture(vorticityTexture);	

	textures->disableFbo();

	setInput();
}

void VorticityConfinementSliceTest::setInput()
{
	VoxelGrid<Color> velocities(width, height, numSlices);	
	
	// First slice
	velocities(2, 1, 0) = Color(1, 0, 0, 0);
	velocities(4, 1, 0) = Color(1, 0, 0, 0);
	
	velocities(1, 2, 0) = Color(0, -1, 0, 0);
	velocities(3, 2, 0) = Color(0, 1, 0, 0);
	velocities(5, 2, 0) = Color(0, 1, 0, 0);

	velocities(2, 3, 0) = Color(-1, 0, 0, 0);
	velocities(4, 3, 0) = Color(-1, 0, 0, 0);

	velocities(1, 4, 0) = Color(0, 1, 0, 0);
	velocities(3, 4, 0) = Color(0, -1, 0, 0);
	velocities(5, 4, 0) = Color(0, -1, 0, 0);

	velocities(2, 5, 0) = Color(1, 0, 0, 0);
	velocities(4, 5, 0) = Color(1, 0, 0, 0);
	

	// Second slice
	for (unsigned int y = 1; y < 3; ++y) {
		for (unsigned int x = 1; x < width - 1; ++x) {			
			velocities(x, y, 1) = Color(1, 0, 0, 0);
		}
	}

	for (unsigned int y = 4; y < height - 1; ++y) {
		for (unsigned int x = 1; x < width - 1; ++x) {			
			velocities(x, y, 1) = Color(-1, 0, 0, 0);
		}
	}

	textures->fillTexture(velocityTexture, velocities);	
}

void VorticityConfinementSliceTest::setOutput()
{
	ColorMap vorticityOutput;
	
	// First slice
	vorticityOutput[textures->getGlobalCoordinates(2, 2, 0)] = Color(2, 2, 0, 0);
	vorticityOutput[textures->getGlobalCoordinates(4, 2, 0)] = Color(1, 1, 0, 0);
	vorticityOutput[textures->getGlobalCoordinates(2, 4, 0)] = Color(-2, 2, 0, 0);
	vorticityOutput[textures->getGlobalCoordinates(4, 4, 0)] = Color(-1, 1, 0, 0);
	

	// Second slice
	for (unsigned int x = 1; x < width - 1; ++x) 
		vorticityOutput[textures->getGlobalCoordinates(x, 1, 1)] = Color(-0.5, 0.5, 0, 0);
	
	for (unsigned int x = 1; x < width - 1; ++x) 
		vorticityOutput[textures->getGlobalCoordinates(x, 2, 1)] = Color(0.5, 0.5, 0, 0);
	
	for (unsigned int x = 1; x < width - 1; ++x) 
		vorticityOutput[textures->getGlobalCoordinates(x, 3, 1)] = Color(1, 1, 0, 0);
	
	for (unsigned int x = 1; x < width - 1; ++x) 
		vorticityOutput[textures->getGlobalCoordinates(x, 4, 1)] = Color(0.5, 0.5, 0, 0);

	for (unsigned int x = 1; x < width - 1; ++x) 
		vorticityOutput[textures->getGlobalCoordinates(x, 5, 1)] = Color(-0.5, 0.5, 0, 0);
	

	setExpectedOutput(vorticityTexture, vorticityOutput, 0.0f);


	ColorMap velocityForces;

	// Row 3
	velocityForces[textures->getGlobalCoordinates(1, 3, 1)] = Color(0, -1, 0, 0);
	velocityForces[textures->getGlobalCoordinates(5, 3, 1)] = Color(0, 1, 0, 0);
	for (int x = 2; x < 5; ++x) 
		velocityForces[textures->getGlobalCoordinates(x, 3, 1)] = Color(0, 0, 0, 0);

	// Row 4
	float v = std::sqrt(0.5f) / 2.0f;
	velocityForces[textures->getGlobalCoordinates(1, 4, 1)] = Color(-v, -v, 0, 0);
	velocityForces[textures->getGlobalCoordinates(5, 4, 1)] = Color(-v, v, 0, 0);
	for (int x = 2; x < 5; ++x) 
		velocityForces[textures->getGlobalCoordinates(x, 4, 1)] = Color(-0.5, 0, 0, 0);

	setExpectedOutput(velocityForceTexture, velocityForces, 0.0f);
	
}



VolumetricExtrusionTest::VolumetricExtrusionTest() :
	sliceWidth(34), sliceHeight(34), numSlices(2)
{
	CgContext &context = CgContext::getInstance();
	importTextures = autoPtr(new Flat3dTextureSet(sliceWidth, 
		sliceHeight, 
		numSlices, 
		context.getTexRectTarget(),
		context.getTexRectInternalFormat()));

	sliceTexture = importTextures->addPrimaryTexture("Slice texture");

	volumetricExtrusion = autoPtr(new VolumetricExtrusion(importTextures.get(), sliceTexture));

	setIoTextureSet(volumetricExtrusion->getTextureSet());

	setInput();
}

VolumetricExtrusionTest::~VolumetricExtrusionTest()
{
}


void VolumetricExtrusionTest::doRun(float timestep)
{
	volumetricExtrusion->run(timestep);
}

void VolumetricExtrusionTest::doReset()
{

}

void VolumetricExtrusionTest::setInput()
{
	VoxelGrid<Color> sliceValues(sliceWidth, sliceHeight, numSlices);

	for (unsigned int y = 0; y < sliceHeight; ++y) {
		for (unsigned int x = 0; x < sliceWidth; ++x) {
			
			if (y % 4 < 2)				
				sliceValues(x, y, 1) = Color(0, 1, 0, 0);

			if (x % 4 < 2)
				sliceValues(x, y, 0) = Color(1, 0, 0, 0);							
		}
	}

	importTextures->fillTexture(sliceTexture, sliceValues);
}

void VolumetricExtrusionTest::setOutput()
{

}




}
}

