#include "GpuTest.h"

#include <algorithm>
#include <cassert>
#include <cmath>

#include "util/Util.h"

#include "opengl/IoTextureSet.h"

using namespace OpenGL;

namespace GpuThesis {
namespace Test {

namespace {

class TextureEvaluator : public std::unary_function<GpuTest::TextureMapping, TestResult *> {
public:
	TextureEvaluator(const GpuTest *tester, float timestep) : tester(tester), timestep(timestep) {}

	TestResult *operator()(const GpuTest::TextureMapping &textureMapping) {
		return tester->evaluate(textureMapping, timestep);
	}

private:
	const GpuTest *tester;
	float timestep;
};

}

GpuTest::GpuTest() :
	textureSet(0)
{
}

GpuTest::~GpuTest()
{
}


std::auto_ptr<TestResultContainer> GpuTest::runTest()
{
	TestCaseMap::const_iterator iter = testCaseMap.begin(), end = testCaseMap.end();

	std::auto_ptr<TestResultContainer> result(new TestResultContainer());

	for (; iter != end; ++iter) {
		reset();
		run(iter->first);

		std::transform(iter->second.begin(), iter->second.end(),
		               std::back_inserter(*result.get()),
					   TextureEvaluator(this, iter->first));
	}

	return result;
}

TestResult *GpuTest::evaluate(const TextureMapping &m, float timestep) const
{
	const ColorMap &colorMap = m.second;

	FailedTest::DiffMap difference;

	std::vector<Util::Color> texture = textureSet->getTexture(m.first);

	ColorMap::const_iterator iter = colorMap.begin(), end = colorMap.end();

	float epsilon = 1e-3;

	for (; iter != end; ++iter) {
		unsigned int x = iter->first.first, y = iter->first.second;

		assert(x < textureSet->getWidth(m.first) && y < textureSet->getHeight(m.first));

		const Util::Color &expected = iter->second;
		const Util::Color &actual = texture[y * textureSet->getWidth(m.first) + x];

		if (expected.diff(actual) > epsilon)
			difference[iter->first] = std::make_pair(expected, actual);
	}

	if (difference.empty())
		return new SucceededTest(textureSet->getDescription(m.first), timestep);
	else
		return new FailedTest(textureSet->getDescription(m.first), timestep, difference);
}

OpenGL::IoTextureSet *GpuTest::getIoTextureSet() const
{
	return textureSet;
}

void GpuTest::setExpectedOutput(unsigned int textureId, 
	                            const ColorMap &colorMap,
	                            float timestep)
{
	testCaseMap[timestep][textureId] = colorMap;
}

void GpuTest::setIoTextureSet(IoTextureSet *textures)
{
	textureSet = textures;
}

TestResultContainer::TestResultContainer()
{
}

TestResultContainer::~TestResultContainer()
{
	std::for_each(testResults.begin(), testResults.end(), Util::destroy<TestResult>);
}

TestResultContainer::const_iterator TestResultContainer::begin() const
{
	return testResults.begin();
}

TestResultContainer::const_iterator TestResultContainer::end() const
{
	return testResults.end();
}

TestResultContainer::iterator TestResultContainer::begin()
{
	return testResults.begin();
}

TestResultContainer::iterator TestResultContainer::end()
{
	return testResults.end();
}

unsigned int TestResultContainer::size() const
{
	return static_cast<unsigned int>(testResults.size());
}

void TestResultContainer::push_back(TestResult *testResult)
{
	testResults.push_back(testResult);
}

std::string TestResult::getDescription() const
{
	return toString();
}

bool TestResult::wasSuccessful() const
{
	return success();
}

TestResult::~TestResult()
{
}

FailedTest::FailedTest(std::string textureDesc, float timestep,
					   const DiffMap &diff) :
	textureDesc(textureDesc),
	timestep(timestep),
	diff(diff)
{
}

std::string FailedTest::toString() const
{
	std::stringstream ss;

	ss << "FAILED: Texture \"" << textureDesc << "\" at timestep " << timestep << ".\n";

	DiffMap::const_iterator iter = diff.begin(), end = diff.end();

	for (; iter != end; ++iter)
		ss << '(' << iter->first.first << ',' << iter->first.second << ") Expected: "
		   << iter->second.first << " Actual: " << iter->second.second << '\n';

	return ss.str();
}

bool FailedTest::success() const
{
	return false;
}

SucceededTest::SucceededTest(std::string textureDesc, float timestep) :
	textureDesc(textureDesc),
	timestep(timestep)
{
}

std::string SucceededTest::toString() const
{
	std::stringstream ss;

	ss << "Correct: Texture \"" << textureDesc << "\" at timestep " << timestep << '.';

	return ss.str();
}

bool SucceededTest::success() const
{
	return true;
}

}
}
