#ifndef GPUTHESIS_TEST_GPUTEST_H
#define GPUTHESIS_TEST_GPUTEST_H

#include <map>
#include <memory>
#include <string>
#include <vector>
#include <utility>

#include "Runnable.h"

#include "util/Color.h"

namespace OpenGL {
	class IoTextureSet;
}

namespace GpuThesis {
namespace Test {

typedef std::pair<unsigned int, unsigned int> Coordinate;

class TestResult;
class TestResultContainer;

/**
 * GpuTest is a base class for Gpu operation tests.
 */
class GpuTest : public Runnable {
public:
	GpuTest();
	virtual ~GpuTest() = 0;

	/**
	 * Runs the test, returning a TestResultContainer.
	 * @return a TestResultContainer with the test result.
	 */
	std::auto_ptr<TestResultContainer> runTest();

	typedef std::map<Coordinate, Util::Color> ColorMap;

	typedef std::map<unsigned int, ColorMap> TextureMap;
	typedef std::map<float, TextureMap> TestCaseMap;

	typedef TextureMap::value_type TextureMapping;

	/**
	 * Evaluate the given texture mapping.
	 * Compares the texture mapping to the actual texture to see if they correspond.
	 * @param m the texture mapping.
	 * @param timestep the timestep.
	 * @return the TestResult.
	 */
	TestResult *evaluate(const TextureMapping &m, float timestep) const;

	/**
	 * Get the test's IoTextureSet.
	 * @return the test's IoTextureSet.
	 */
	OpenGL::IoTextureSet *getIoTextureSet() const;

protected:
	/**
	 * Set the expected output of the given texture after running with the given timestep.
	 * Used by subclasses.
	 * @param textureId the identifier of the texture.
	 * @param colorMap the expected output.
	 * @param timestep the timestep to run to get the expected output.
	 */
	void setExpectedOutput(unsigned int textureId, 
	                       const ColorMap &colorMap,
	                       float timestep);

	/**
	 * Called by subclasses to set the IoTextureSet.
	 * @param textures the IoTextureSet.
	 */
	void setIoTextureSet(OpenGL::IoTextureSet *textures);

private:
	TestCaseMap testCaseMap;

	OpenGL::IoTextureSet *textureSet;
};

/**
 * Contains test results.
 * Acts as a vector of TestResult, except it takes ownership and deletes at destruction.
 */
class TestResultContainer {
public:
	TestResultContainer();
	~TestResultContainer();

	typedef std::vector<TestResult *>::const_iterator const_iterator;
	typedef std::vector<TestResult *>::iterator iterator;

	typedef TestResult *value_type;

	typedef value_type &reference;
	typedef const value_type &const_reference;

	const_iterator begin() const;
	const_iterator end() const;

	iterator begin();
	iterator end();

	unsigned int size() const;

	void push_back(TestResult *testResult);

private:
	std::vector<TestResult *> testResults;

	TestResultContainer(const TestResultContainer&);
	TestResultContainer &operator=(const TestResultContainer&);
};

/**
 * A class representing the result of a test.
 */
class TestResult {
public:
	/**
	 * Get the test's description.
	 * @return the test's description.
	 */
	std::string getDescription() const;

	/**
	 * Returns whether the test was successful.
	 * @return whether the test was successful.
	 */
	bool wasSuccessful() const;

	virtual ~TestResult() = 0;

private:
	virtual std::string toString() const = 0;
	virtual bool success() const = 0;
};

/**
 * Specialization of TestResult for failed tests.
 */
class FailedTest : public TestResult {
public:
	typedef std::map<Coordinate, std::pair<Util::Color, Util::Color> > DiffMap;

	/**
	 * Initializes FailedTest.
	 * @param textureDesc the texture description.
	 * @param timestep the timestep for the test operation.
	 * @param diff a map containing the values where the expected and actual differed.
	 */
	FailedTest(std::string textureDesc, float timestep,
	           const DiffMap &diff);

private:
	std::string toString() const;
	bool success() const;

	std::string textureDesc;
	float timestep;

	DiffMap diff;
};

/**
 * Specialization of TestResult for successful tests.
 */
class SucceededTest : public TestResult {
public:
	/**
	 * Initializes SucceededTest.
	 * @param textureDesc the texture description.
	 * @param timestep the timestep for the test operation.
	 */
	SucceededTest(std::string textureDesc, float timestep);

private:
	std::string toString() const;
	bool success() const;

	std::string textureDesc;
	float timestep;
};

}
}

#endif