#ifndef ACTION_H
#define ACTION_H

#include <string>
#include <vector>

#include "util/Util.h"

/**
 * The Console namespace.
 */
namespace Console {

/**
 * Action class.
 * Represents an action which takes a variable number of parameters.
 */
class Action {
public:
	/**
	 * Performs the action based on the given parameters.
	 * @param parameters the action's parameters.
	 * @return the result of the action.
	 */
	std::string handle(const std::vector<std::string> &parameters) {
		return run(parameters);
	}

	virtual ~Action() {}

private:
	/**
	 * Implementation of handle, must be supplied by concrete subclasses.
	 * @param parameters the action's parameters.
	 * @return the result of the action.
	 */
	virtual std::string run(const std::vector<std::string> &parameters) = 0;
};

/**
 * Specialization of Action for functors, which performs an action by calling
 * the functor, and returns a constant result description.
 */
template <typename Func, typename T>
class FunctorAction : public Action {
public:
	/**
	 * Initialize a FunctorAction with the given functor, arguments, and
	 * description.
	 * @param functor the functor to be called by the action.
	 * @param arg the argument to be passed to the functor.
	 * @param description the description to be returned by the action.
	 */
	FunctorAction(Func functor, const T &arg, const std::string &description) :
	  functor(functor), arg(arg), description(description) {}

	/**
	 * Calls the given functor, ignoring the passed parameters.
	 * @param parameters the action's parametes, ignored by this specialization.
	 * @return the result of the action (the description given to the constructor).
	 */
	std::string run(const std::vector<std::string> &parameters) {
		functor(arg);
		return description;
	}

private:
	Func functor;
	T arg;
	std::string description;
};

/**
 * Helper function template to create functor actions.
 * Used to automatically deduce template parameters.
 * @param functor the functor to be called by the action.
 * @param arg the argument to be passed to the functor.
 * @param description the description to be returned by the action.
 * @return a pointer to the newly created Action specialization.
 */
template <typename Func, typename T>
Action *createFunctorAction(Func functor, const T &arg, const std::string& description)
{
	return new FunctorAction<Func, T>(functor, arg, description);
}

/**
 * Specialization of Action which cycles an index to a vector,
 * offering functionality to retrieve the object at the given index.
 */
template <typename T>
class CycleAction : public Action {
public:
	/**
	 * Initialize the CycleAction with the given vector.
	 * @param v the vector to cycle through.
	 */
	CycleAction(const std::vector<T> &v) : v(v), pos(0) {}

	/**
	 * Get the object at the current vector index.
	 * @return the object at the current vector index.
	 */
	T getCurrent() { return v[pos]; }

	/**
	 * Cycle the index.
	 * @return a string describing the current index.
	 */
	std::string run(const std::vector<std::string> &/*parameters*/) {
		if (++pos >= static_cast<unsigned int>(v.size()))
			pos = 0;

		return "Current index: " + Util::toString(pos);
	}

private:
	std::vector<T> v;

	unsigned int pos;
};

}

#endif
