#include <algorithm>
#include <cassert>
#include <fstream>
#include <iostream>
#include <limits>
#include <sstream>

#include "opengl/OpenGL.h"
#include "opengl/Rendering.h"

#include "Console.h"

#include "Action.h"
#include "Variable.h"

namespace Console {

ReservedKeywordException::ReservedKeywordException(const std::string &keyword) :
	errorStr("Console: Reserved keyword " + keyword)
{
}

const char *ReservedKeywordException::what() const
{
	return errorStr.c_str();
}

Console::Console(bool autolog) :
	tabbed(false),
	browsePos(0),
	autolog(autolog),
	transparency(0.5f)
{
	initialize();

	addVariable("consoleTransparency", new VariableTemplate<float>(&transparency));
}

Console::~Console()
{
	if (autolog)
		saveLog("auto.log");
}



void Console::addAction(const std::string &name, Action *action)
{
	if (reserved.find(name) != reserved.end())
		throw ReservedKeywordException(name);

	actions[name] = action;

	keywords.insert(name);
}



void Console::addVariable(const std::string &name, Variable *variable)
{
	if (reserved.find(name) != reserved.end())
		throw ReservedKeywordException(name);

	variables[name] = variable;

	keywords.insert(name);
}



void Console::removeAction(const std::string &name)
{
	ActionMap::iterator iter = actions.find(name);

	if (iter != actions.end())
		actions.erase(iter);

	StringSet::iterator keywordIter = keywords.find(name);

	assert(keywordIter != keywords.end());

	keywords.erase(keywordIter);
}



void Console::removeVariable(const std::string &name)
{
	VariableMap::iterator iter = variables.find(name);

	if (iter != variables.end())
		variables.erase(iter);

	StringSet::iterator keywordIter = keywords.find(name);

	assert(keywordIter != keywords.end());

	keywords.erase(keywordIter);
}



void Console::output(const std::string &str)
{
	std::string::size_type pos = str.find_first_of("\n");

	if (pos != std::string::npos) {
		std::string prefix(str.begin(), str.begin() + pos);
		std::string suffix(str.begin() + pos + 1, str.end());

		if (!prefix.empty())
			output(prefix);

		if (!suffix.empty())
			output(suffix);
	} else {
		outputBuffer.push_back(str);

		std::cout << str << '\n';
	}
}


void Console::render() const
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 1, 0, 1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glEnable(GL_BLEND);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glColor4f(0.2f, 0.2f, 0.2f, 1.0f - transparency);

	const float border = 0.05f;
	const float textDelta = 0.0025f;

	glBegin(GL_QUADS);

	glVertex2f(border, border);
	glVertex2f(1.0f - border, border);
	glVertex2f(1.0f - border, 1.0f - border);
	glVertex2f(border, 1.0f - border);

	glEnd();

	glDisable(GL_BLEND);

	glColor4f(0.8f, 0.8f, 0.8f, 0.5f);

	void *font = GLUT_BITMAP_HELVETICA_12;

	const unsigned int fontHeight = 12;
	const unsigned int fontSpacing = 5;

	float viewport[4];
	glGetFloatv(GL_VIEWPORT, viewport);

	const float separatorHeight = border + (fontHeight + fontSpacing * 1.5f) / viewport[3];

	glBegin(GL_LINE_STRIP);

	glVertex2f(border, border);
	glVertex2f(1.0f - border, border);
	glVertex2f(1.0f - border, 1.0f - border);
	glVertex2f(border, 1.0f - border);
	glVertex2f(border, border);

	glEnd();

	glBegin(GL_LINES);

	glVertex2f(border, separatorHeight);
	glVertex2f(1.0f - border, separatorHeight);

	glEnd();

	const float textBorder = border + textDelta;

	const unsigned int maxWidth = viewport[2] * (1.0f - textBorder * 2.0f);

	glRasterPos2f(textBorder, textBorder + fontSpacing / viewport[3]);
	OpenGL::displayString(font, "> " + currentInput, maxWidth);

	unsigned int numLines = (viewport[3] * (1.0f - border - separatorHeight)) / (fontSpacing + fontHeight);

	if (outputBuffer.size() < numLines)
		numLines = static_cast<unsigned int>(outputBuffer.size());

	for (unsigned int i = 0; i < numLines; ++i) {
		glRasterPos2f(textBorder, textBorder + (fontSpacing + (i + 1) * (fontHeight + fontSpacing)) / viewport[3]);
		OpenGL::displayString(font, outputBuffer[outputBuffer.size() - i - 1], maxWidth);
	}

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}



void Console::parseStream(std::istream &is)
{
	std::string line;

	while (std::getline(is, line))
		parseLine(line);
}



bool Console::handleBinding(char c)
{
	BindingMap::const_iterator iter = bindings.find(c);
	if (iter != bindings.end()) {
		parseLine(iter->second, true);
		return true;
	} else {
		return false;
	}
}



void Console::handleChar(char c)
{
	currentInput.push_back(c);

	tabbed = false;
}



void Console::handleBackspace()
{
	if (!currentInput.empty())
		currentInput.resize(currentInput.size() - 1);

	tabbed = false;
}



void Console::handleEnter()
{
	inputBuffer.push_back(currentInput);

	parseLine(currentInput);

	currentInput.clear();

	tabbed = false;
	browsePos = 0;
}



void Console::setBrowseInput()
{
	if (browsePos == 0)
		currentInput.clear();
	else
		currentInput = inputBuffer[inputBuffer.size() - browsePos];
}



void Console::handleUpArrow()
{
	if (browsePos < inputBuffer.size()) {
		++browsePos;
		setBrowseInput();
		
		tabbed = false;
	}
}



void Console::handleDownArrow()
{
	if (browsePos > 0) {
		--browsePos;
		setBrowseInput();

		tabbed = false;
	}
}



bool isPrefix(const std::string &a, const std::string &b)
{
	return a.empty() || b.find(a) == 0;
}



void Console::handleTab()
{
	if (!tabbed) {
		tabbed = true;
		baseInput = currentInput;
	}

	StringSet::const_iterator iter = keywords.upper_bound(currentInput);

	if (iter == keywords.end() || !isPrefix(baseInput, *iter))
		iter = keywords.upper_bound(baseInput);

	if (iter != keywords.end() && isPrefix(baseInput, *iter))
		currentInput = *iter;
}



std::vector<std::string> tokenize(const std::string &line)
{
	std::vector<std::string> tokens;

	std::string::size_type pos = 0;

	while (pos != std::string::npos) {
		pos = line.find_first_not_of(" ", pos);

		if (pos != std::string::npos) {
			std::string::size_type current = pos;

			pos = line.find_first_of(" ", pos);

			tokens.push_back(line.substr(current, pos == std::string::npos ? pos : pos - current));
		}
	}

	return tokens;
}



void Console::parseLine(const std::string &line, bool silent)
{
	assert(line.find('\n') == std::string::npos);

	std::vector<std::string> parameters = tokenize(line);

	if (!parameters.empty()) {
		if (parameters[0] == "help")
			listActions();
		else if (parameters[0] == "list")
			listVariables();
		else if (parameters[0] == "save") {
			if (parameters.size() > 1)
				saveVariables(parameters[1]);
			else
				output("Missing filename");
		} else if (parameters[0] == "load") {
			if (parameters.size() > 1)
				loadConfiguration(parameters[1]);
			else
				output("Missing filename");
		} else if (parameters[0] == "bind") {
			if (parameters.size() == 1) 
				listBindings();
			else if (parameters.size() == 2) 
				outputBinding(parameters[1]);
			else {
				std::stringstream ss;
				std::copy(parameters.begin() + 2, parameters.end(), 
				          std::ostream_iterator<std::string>(ss, " "));

				createBinding(parameters[1], ss.str());
			}			
		} else {
			ActionMap::iterator actionIter = actions.find(parameters[0]);
			VariableMap::iterator variableIter = variables.find(parameters[0]);

			if (actionIter != actions.end()) {
				parameters.erase(parameters.begin());

				std::string result = actionIter->second->handle(parameters);

				if (!result.empty() && !silent)
					output(result);
			} else if (variableIter != variables.end()) {
				parameters.erase(parameters.begin());

				std::string result;

				if (!parameters.empty()) {
					std::string parameterString;

					std::vector<std::string>::const_iterator iter = parameters.begin();

					while (iter != parameters.end()) {
						parameterString += *iter;

						if (++iter != parameters.end())
							parameterString += ' ';
					}

					result = variableIter->first + ": " + variableIter->second->setValue(parameterString);
				} else
					result = variableIter->first + ": " + variableIter->second->getValue();

				if (!result.empty() && !silent)
					output(result);
			} else if (!silent)
				output("Unknown command/variable " + parameters[0]);
		}
	}
}



void Console::listActions()
{
	std::vector<std::string> v(reserved.begin(), reserved.end());

	for (ActionMap::const_iterator iter = actions.begin(); iter != actions.end(); ++iter)
		v.push_back(iter->first);

	std::sort(v.begin(), v.end());

	for (std::vector<std::string>::iterator iter = v.begin(); iter != v.end(); ++iter)
		output(*iter);

	std::stringstream ss;

	ss << "Number of commands: " << static_cast<unsigned int>(v.size());

	output(ss.str());
}



void Console::listVariables()
{
	std::vector<std::string> v;

	for (VariableMap::const_iterator iter = variables.begin(); iter != variables.end(); ++iter)
		v.push_back(iter->first);

	std::sort(v.begin(), v.end());

	for (std::vector<std::string>::iterator iter = v.begin(); iter != v.end(); ++iter)
		output(*iter);

	std::stringstream ss;

	ss << "Number of variables: " << static_cast<unsigned int>(v.size());

	output(ss.str());
}



void Console::saveLog(const std::string &filename)
{
	std::ofstream file;
	file.open(filename.c_str());

	for (std::vector<std::string>::const_iterator iter = outputBuffer.begin();
		 iter != outputBuffer.end(); ++iter)
	{
		file << *iter << '\n';
	}

	file.close();
}



void Console::saveVariables(std::string &filename)
{
	std::ofstream file;
	file.open(filename.c_str());

	for (VariableMap::const_iterator iter = variables.begin(); iter != variables.end(); ++iter)
		file << iter->first << " " << iter->second->getValue() << '\n';

	output("Saved variables");

	file.close();
}



void Console::loadConfiguration(const std::string& filename) 
{
	std::ifstream file;
	file.open(filename.c_str());

	if (file) {
		parseStream(file);

		output("Loaded configuration from " + filename + '.');
	} else
		output("Failed opening " + filename + '.');

	file.close();
}

void Console::listBindings()
{
	for (BindingMap::const_iterator iter = bindings.begin(); iter != bindings.end(); ++iter) {

		output(iter->first + std::string(": ") + iter->second);
	}

	std::stringstream ss;
	ss << "Number of bindings: " << static_cast<unsigned int>(bindings.size());
	output(ss.str());
}

void Console::outputBinding(const std::string &key) 
{
	if (key.size() != 1)
		output("Binding should be a single character.");
	else {

		BindingMap::const_iterator iter = bindings.find(key[0]);
		if (iter != bindings.end())
			output(iter->first + std::string(": ") + iter->second);
		else
			output("Binding " + key + " not found.");
	}
}


void Console::createBinding(const std::string &key, const std::string &parameters)
{
	if (key.size() != 1)
		output("Binding should be a single character.");
	else {
		bindings[key[0]] = parameters;
		output(key[0] + std::string(": ") + parameters);
	}
}

void Console::initialize()
{
	reserved.insert("help");
	reserved.insert("list");
	reserved.insert("save");
	reserved.insert("load");
	reserved.insert("bind");

	std::copy(reserved.begin(), reserved.end(), inserter(keywords, keywords.end()));
}


}