#ifndef CONSOLE_H
#define CONSOLE_H

#include <iosfwd>
#include <map>
#include <set>
#include <string>
#include <vector>

namespace Console {

class Action;
class Variable;

class ReservedKeywordException : public std::exception {
public:
	ReservedKeywordException(const std::string &keyword);
	const char *what() const;
private:
	std::string errorStr;
};

/**
 * The Console class represent a console where specified commands can execute actions 
 * (represented by an Action object) and set variables (represented by a Variable object).
 * The console has a text based interface, and it can be rendered to the screen using OpenGL.
 * The console window rendered consists of two parts: a large output area and an input
 * text area at the bottom. 
 *
 * The console has 5 built-in commands: <br>
 * "help" - lists all action commands currently present in the console. <br>
 * "list" - lists all variables commands currently present in the console. <br>
 * "save <filename>" - saves all variables currently present in the console and their values 
 * to the specified file. <br>
 * "load <filename>" - loads a specified configuration file, and executes
 * all lines in the configuration as a command. <br>
 * "bind" - lists all bindings (between a character and a command) currently present in the console. <br>
 * "bind <char>" - outputs the command associated with the specified character. <br>
 * "bind <char> <command>" - make a binding between the specified character and the specified command.
 * <command> can consists of a list of command parameters in addition to the command itself. 
 */
class Console {
public:
	/**
	 * Constructs a console and executes the commands specified in the config file. 
	 * @param configFile the file with the commands to be executed.
	 * @param autolog whether the console shall log all outputs to a log file.
	 */
	Console(std::string configFile, bool autolog = false);

	/**
	 * Constructs a console.
	 * @param autolog whether the console shall log all outputs to a log file.
	 */
	Console(bool autolog = false);
	~Console();

	/**
	 * Adds an action to the console. 
	 * @param name the command name associated with the added action.
	 * @param action the action to be added.
	 */
	void addAction(const std::string &name, Action *action);

	/**
	 * Adds a variable to the console.
	 * @param name the command name associated with the added variable.
	 * @param variable the variable to be added.
	 */
	void addVariable(const std::string &name, Variable *variable);

	/**
	 * Removes an action from the console.
	 * @param name the command name associated with the action to remove.
	 */
	void removeAction(const std::string &name);

	/**
	 * Removes a variable from the console.
	 * @param name the command name associated with the variable to remove.
	 */	
	void removeVariable(const std::string &name);

	/**
	 * Outputs a string in the console window.
	 * @param str the string to output.
	 */
	void output(const std::string &str);

	/**
	 * Renders the console to the screen using OpenGL.
	 */
	void render() const;

	/**
	 * Parses an input stream into lines, and executes 
	 * all lines as a command. 
	 * @param is the input stream to be parsed. 
	 */
	void parseStream(std::istream &is);

	/**
	 * Handles a binding for a specified character.
	 * The binding acts as a mapping between a single character and a command. 
	 * A binding can be activated using the built-in console command "bind".
	 * @param c the bind character.
	 * @return true if there exists a binding for the specified character.
	 */
	bool handleBinding(char c);

	/**
	 * Handles a single character keystroke.
	 * @param c the input character.
	 */
	void handleChar(char c);

	/**
	 * Handles a backspace keystroke.
	 */
	void handleBackspace();

	/**
	 * Handles an enter keystroke.
	 */
	void handleEnter();

	/**
	 * Handles a tab keystroke.
	 */
	void handleTab();

	/**
	 * Handles a down arrow keystroke.
	 */
	void handleDownArrow();

	/**
	 * Handles an up arrow keystroke.
	 */
	void handleUpArrow();

	/**
	 * Saves the log to a file.
	 * @param filename the name of the file to save the log.
	 */
	void saveLog(const std::string &filename);

	/**
	 * Loads a configuration from a specified file, and executes all lines in the configuration as a command.
	 * @param filename the name of the file to load the configuration. 
	 */
	void loadConfiguration(const std::string &filename);

private:
	typedef std::map<std::string, Action *> ActionMap;
	typedef std::map<std::string, Variable *> VariableMap;

	typedef std::set<std::string> StringSet;

	typedef std::map<char, std::string> BindingMap;

	ActionMap actions;
	BindingMap bindings;
	StringSet reserved;
	VariableMap variables;
	
	StringSet keywords;

	std::vector<std::string> inputBuffer;
	std::vector<std::string> outputBuffer;

	std::string currentInput;
	std::string baseInput;

	bool tabbed;
	unsigned int browsePos;

	bool autolog;

	float transparency;

	void initialize();

	void parseLine(const std::string &line, bool silent = false);

	void listActions();
	void listVariables();
	void listBindings();

	void saveVariables(std::string &filename);

	void setBrowseInput();

	void outputBinding(const std::string &key); 
	void createBinding(const std::string &key, const std::string &parameters);

};

}

#endif
