#ifndef VARIABLE_ACTION_H
#define VARIABLE_ACTION_H

#include <sstream>
#include <string>

#include "Action.h"

#include "util/Util.h"

namespace Console {

/**
 * Represents a variable whose value can be set and retrieved.
 */
class Variable {
public:
	/**
	 * Gets the value of the variable as a string.
	 * @return the value of the variable as a string.
	 */
	virtual std::string getValue() const = 0;

	/**
	 * Tries to set the value of the variable as a string.
	 * @param parameter the value to attempt to set the variable to.
	 * @return the new value of the variable, or an error string if the parameter was illegal.
	 */
	virtual std::string setValue(const std::string &parameter) = 0;

	virtual ~Variable() {}
};

template <typename T, class C, typename Getter, typename Setter>
class VariableAccessor : public Variable {
public:
	//typedef T (C::*Getter)() const;
	//typedef void (C::*Setter)(T);

	/**
	 * A specialization of Variable for classes which offer getters and setters.
	 * @param getter a pointer to a member function which sets the variable.
	 * @param setter a pointer to a member function which gets the variable.
	 * @param object the object to manipulate.
	 */
	VariableAccessor(Getter getter, Setter setter, C *object);

	/**
	 * Gets the value using the object's getter.
	 * @return the value retrieved from the getter as a string.
	 */
	std::string getValue() const;

	/**
	 * Tries to set the value using the object's setter.
	 * @param parameter the value to pass to the object's setter.
	 * @return the value that was set, or "Parse error" if the parameter could not be converted to the type expected by the setter.
	 */
	std::string setValue(const std::string &parameter);

private:
	Getter getter;
	Setter setter;

	C *object;
};

template <typename T, class C, typename Getter, typename Setter>
VariableAccessor<T, C, Getter, Setter>::VariableAccessor(Getter getter, Setter setter, C *object) :
	object(object),
	getter(getter),
	setter(setter)
{
}


template <typename T, class C, typename Getter, typename Setter>
std::string VariableAccessor<T, C, Getter, Setter>::getValue() const
{
	return Util::toString((object->*getter)());
}

template <typename T, class C, typename Getter, typename Setter>
std::string VariableAccessor<T, C, Getter, Setter>::setValue(const std::string &parameter)
{
	T temp;

	if (!Util::lexical_cast(parameter, temp))
		return "Parse error";

	(object->*setter)(temp);

	return Util::toString(temp);
}

/**
 * Helper function template to create a VariableAccessor for const getters.
 * @param getter a pointer to a member function which sets the variable.
 * @param setter a pointer to a member function which gets the variable.
 * @param object the object to manipulate.
 * @return a pointer to the newly created Variable specialization.
 */
template <typename T, class C>
Variable *createVariableAccessor(T (C::*getter)() const, void (C::*setter)(T), C *object)
{
	return new VariableAccessor<T, C, T (C::*)() const, void (C::*)(T)>(getter, setter, object);
}

/**
 * Helper function template to create a VariableAccessor for non-const getters.
 * @param getter a pointer to a member function which sets the variable.
 * @param setter a pointer to a member function which gets the variable.
 * @param object the object to manipulate.
 * @return a pointer to the newly created Variable specialization.
 */
template <typename T, class C>
Variable *createVariableAccessor(T (C::*getter)(), void (C::*setter)(T), C *object)
{
	return new VariableAccessor<T, C, T (C::*)(), void (C::*)(T)>(getter, setter, object);
}

/**
 * A specialization of Variable for simple types which can be written and read using iostreams.
 */
template <typename T>
class VariableTemplate : public Variable {
public:
	/**
	 * Initializes the VariableTemplate with the given variable pointer.
	 * @param variable a pointer to the variable to manipulate.
	 */
	VariableTemplate(T *variable) : variable(variable) {}

	/**
	 * Gets the value of the variable represented as a string.
	 * @return a string representing the value of the variable.
	 */
	std::string getValue() const {
		return Util::toString(*variable);
	}

	/**
	 * Sets the value of the variable from the parameter string.
	 * @param parameter the value to set the variable with.
	 * @return the newly set value or "Parse error".
	 */
	std::string setValue(const std::string& parameter) {

		std::stringstream temp;

		if (!(temp << parameter && temp >> *variable))
			return "Parse error";
		
		return Util::toString(*variable);
	}

private:
	T *variable;
};

/**
 * A specialization of VariableTemplate for string, to allow spaces.
 */
template <>
class VariableTemplate<std::string> : public Variable {
public:
	/**
	 * Initializes the VariableTemplate with the given string pointer.
	 * @param variable a pointer to the string to manipulate.
	 */
	VariableTemplate(std::string *variable) : variable(variable) {}

	/**
	 * Gets the string value.
	 * @return the string value.
	 */
	std::string getValue() const {

		return *variable;
	}

	/**
	 * Sets the string value.
	 * @param parameter the new string value.
	 * @return the string value.
	 */
	std::string setValue(const std::string& parameter) {

		*variable = parameter;

		return *variable;
	}

private:
	std::string *variable;
};

/**
 * A specialization of VariableTemplate for bool.
 * Allows toggling of bools.
 */
template <>
class VariableTemplate<bool> : public Variable {
public:
	/**
	 * Initializes the VariableTemplate with the given bool pointer.
	 * @param variable a pointer to the bool to manipulate.
	 */
	VariableTemplate(bool *variable) : variable(variable) {}

	/**
	 * Gets the bool value represented as a string.
	 * @return a string value representing the bool.
	 */
	std::string getValue() const {
		return Util::toString(*variable);
	}

	/**
	 * Sets the bool value.
	 * If the passed parameter is "toggle", the bool will be set to its inverse.
	 * @param parameter a string representing the new bool value, or "toggle".
	 * @return a string representing the bool, or "Parse error".
	 */
	std::string setValue(const std::string& parameter) {
		std::stringstream temp;

		if (parameter == "toggle")
			*variable = !*variable;
		else if (!Util::lexical_cast(parameter, *variable))
			return "Parse error";

		return Util::toString(*variable);
	}

private:
	bool *variable;
};

}

#endif
