#include "Mesh.h"

#include "opengl/OpenGL.h"

#include <algorithm>

namespace Geometry
{

Mesh::Mesh() : useVBO(false), useVertexNormals(true)
{
}

void Mesh::generateFaceNormals()
{
  faceNormals.clear();

  for (unsigned int i = 0; i < indices.size(); i += 3) {
    Math::Vector3d edgeA = vertices[indices[i+1]] - vertices[indices[i]];
    Math::Vector3d edgeB = vertices[indices[i+2]] - vertices[indices[i]];

    Math::Vector3d normal = cross(edgeA, edgeB);

    faceNormals.push_back(normalize(normal));
  }
}


void Mesh::generateVertexNormals()
{
  generateFaceNormals();

  vertexNormals = std::vector<Math::Vector3d>(vertices.size());

	for (unsigned int i = 0; i < indices.size(); i += 3) {
		Math::Vector3d &faceNormal = faceNormals[i / 3];

		for (unsigned int j = 0; j < 3; ++j)
		vertexNormals[indices[i + j]] += faceNormal;
	}

	std::transform(vertexNormals.begin(), vertexNormals.end(), vertexNormals.begin(),
		Math::normalize);
}


void Mesh::generateVBOs()
{
	glGenBuffersARB(1, &vertexVBO);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, vertexVBO);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertices.size() * sizeof(float),
					&*vertices.begin(), GL_STATIC_DRAW_ARB);

	glGenBuffersARB(1, &normalVBO);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, normalVBO);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, vertexNormals.size() * sizeof(float),
					&*vertexNormals.begin(), GL_STATIC_DRAW_ARB );

	useVBO = false;
}


const Mesh refineSphere(Mesh &source, float radius)
{
  Mesh temp;

  temp.vertices = source.vertices;

  for (unsigned int i = 0; i < source.indices.size(); i += 3) {
    unsigned int a = source.indices[i], 
                 b = source.indices[i + 1], 
                 c = source.indices[i + 2],
                 d = static_cast<unsigned int>(temp.vertices.size()),
                 e = static_cast<unsigned int>(temp.vertices.size() + 1),
                 f = static_cast<unsigned int>(temp.vertices.size() + 2);

    for (unsigned int j = 0; j < 3; ++j) {
      unsigned int current = source.indices[i + j];
      unsigned int next = source.indices[i + ((j + 1) % 3)];
      
      Math::Vector3d centre = (source.vertices[current] + source.vertices[next]) / 2;

      centre *= radius / Math::abs(centre);

      temp.vertices.push_back(centre);
    }

    temp.indices.push_back(a);
    temp.indices.push_back(d);
    temp.indices.push_back(f);

    temp.indices.push_back(b);
    temp.indices.push_back(e);
    temp.indices.push_back(d);

    temp.indices.push_back(c);
    temp.indices.push_back(f);
    temp.indices.push_back(e);

    temp.indices.push_back(d);
    temp.indices.push_back(e);
    temp.indices.push_back(f);    
  }

  return temp;
}


void generateSphere(Mesh &target, float radius, unsigned int detail)
{
  target = Mesh();

  Math::fp sqrt3 = static_cast<Math::fp>(std::sqrt(3.0));

  target.vertices.push_back(Math::Vector3d(0, radius, 0));
  target.vertices.push_back(Math::Vector3d( sqrt3 * radius/2, -radius/2,  0));
  target.vertices.push_back(Math::Vector3d(-sqrt3 * radius/4, -radius/2, -3*radius/4));
  target.vertices.push_back(Math::Vector3d(-sqrt3 * radius/4, -radius/2,  3*radius/4));

  target.indices.push_back(0);
  target.indices.push_back(1);
  target.indices.push_back(2);

  target.indices.push_back(0);
  target.indices.push_back(2);
  target.indices.push_back(3);

  target.indices.push_back(0);
  target.indices.push_back(3);
  target.indices.push_back(1);

  target.indices.push_back(3);
  target.indices.push_back(2);
  target.indices.push_back(1);

  for (unsigned int i = 0; i < detail; ++i)
    target = refineSphere(target, radius);
}
  
}
