#include "ModelLoader.h"

#include <lib3ds/file.h>
#include <lib3ds/mesh.h>
#include <lib3ds/material.h>

namespace Geometry {

ModelLoader::ModelLoader(const std::string &filename) :
	model(new Model)
{
	Lib3dsFile *file = lib3ds_file_load(filename.c_str());

	if (!file)
		throw LoadModelFailedException();

	loadMaterials(file);
	loadMeshes(file);

	lib3ds_file_free(file);
}

void ModelLoader::loadMaterials(Lib3dsFile *file)
{
	Lib3dsMaterial *current = file->materials;

	while (current) {
		Material mat;

		mat.ambient.setRed(current->ambient[0]);
		mat.ambient.setGreen(current->ambient[1]);
		mat.ambient.setBlue(current->ambient[2]);

		mat.diffuse.setRed(current->diffuse[0]);
		mat.diffuse.setGreen(current->diffuse[1]);
		mat.diffuse.setBlue(current->diffuse[2]);

		mat.specular.setRed(current->specular[0]);
		mat.specular.setGreen(current->specular[1]);
		mat.specular.setBlue(current->specular[2]);

		materials[current->name] = mat;

		current = current->next;
	}
}

void ModelLoader::loadMeshes(Lib3dsFile *file)
{
	Lib3dsMesh* current = file->meshes;

	while (current) {
		model->meshes.resize(model->meshes.size() + 1);

		Mesh &mesh = model->meshes[model->meshes.size() - 1];

		mesh.vertices.resize(current->points);
		mesh.indices.resize(3 * current->faces);

		for (unsigned int i = 0; i < current->points; ++i) {
			mesh.vertices[i].x = current->pointL[i].pos[0];
			mesh.vertices[i].y = current->pointL[i].pos[1];
			mesh.vertices[i].z = current->pointL[i].pos[2];

			/*for (unsigned int j = 0; j < 3; ++j) {
				unsigned int point = current->faceL[i].points[j];

				mesh.vertices[3 * i + j].x = current->pointL[point].pos[0];
				mesh.vertices[3 * i + j].y = current->pointL[point].pos[1];
				mesh.vertices[3 * i + j].z = current->pointL[point].pos[2];
			}*/
		}

		std::map<std::string, unsigned int> materialCount;

		for (unsigned int i = 0; i < current->faces; ++i) {
			for (unsigned int j = 0; j < 3; ++j)
				mesh.indices[3 * i + j] = current->faceL[i].points[j];

			//mesh.indices[i] = i;
		}

		for (unsigned int i = 0; i < current->faces; ++i)
			++materialCount[current->faceL[i].material];

		std::string highest = materialCount.begin()->first;

		for (std::map<std::string, unsigned int>::iterator iter = materialCount.begin();
		     iter != materialCount.end();
			 ++iter)
		{
			if (iter->second > materialCount[highest])
				highest = iter->first;
		}

		const Material &mat = materials[highest];

		mesh.ambient = mat.ambient;
		mesh.diffuse = mat.diffuse;
		mesh.specular = mat.specular;

		mesh.generateVertexNormals();
		//mesh.generateVBOs();

		current = current->next;
	}
}

std::auto_ptr<Model> ModelLoader::getModel()
{
	return model;
}

}