// Matrix4x4.cpp

#include "Matrix4x4.h" // class definition is found here

#include <cmath> // for sin, cos
#include <iostream>

namespace Math {

Matrix4x4::Matrix4x4() {
	setIdentity();
}

Matrix4x4::Matrix4x4(const Matrix4x4& m) {
	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			elements[i][j] = m.elements[i][j];
		}
	}
}

Matrix4x4::Matrix4x4(const Matrix3x3& m) {
	setIdentity();
	for(int i = 0; i < 3; ++i) {
		for(int j = 0; j < 3; ++j) {
			elements[i][j] = m.elements[i][j];
		}
	}
}

Matrix4x4::Matrix4x4(fp m[16], bool rowMajor) {
	int index = 0;

	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			if (rowMajor)
				elements[i][j] = m[index++];
			else
				elements[j][i] = m[index++];
		}
	}
}

void Matrix4x4::getMatrix(float *m, bool rowMajor)
{
	int index = 0;

	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			if (rowMajor)
				 m[index++] = elements[i][j];
			else
				m[index++] = elements[j][i];
		}
	}
}

Matrix4x4 Matrix4x4::operator*(const Matrix4x4& m) const {
	Matrix4x4 result;

	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			fp temp = 0;

			for(int k = 0; k < 4; ++k) {
				temp += elements[i][k] * m.elements[k][j];
			}

			result.elements[i][j] = temp;
		}
	}

	return result;
}

Matrix4x4 Matrix4x4::operator*(fp f) const {
	Matrix4x4 result;

	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			result.elements[i][j] = elements[i][j] * f;
		}
	}

	return result;
}

Matrix4x4 Matrix4x4::operator/(fp f) const {
	Matrix4x4 result;
	fp invf = 1.0f / f;

	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			result.elements[i][j] = elements[i][j] * invf;
		}
	}

	return result;
}

Vector3d Matrix4x4::multiplyVec3Pos(const Vector3d& v) const {
	float result[3];

	for(int i = 0; i < 3; i++) {
		result[i] = elements[i][0] * v.x + elements[i][1] * v.y + elements[i][2] * v.z + elements[i][3]; // translation
	}

	return Vector3d(result[0], result[1], result[2]);
}

Vector3d Matrix4x4::multiplyVec3Dir(const Vector3d& v) const {
	float result[3];

	for(int i = 0; i < 3; i++) {
		result[i] = elements[i][0] * v.x + elements[i][1] * v.y + elements[i][2] * v.z; // no translation
	}

	return Vector3d(result[0], result[1], result[2]);
}

void Matrix4x4::setIdentity() {
	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			elements[i][j] = i==j ? (fp)1.0 : (fp)0.0;
		}
	}
}

Matrix3x3 Matrix4x4::getSubMatrix(int row, int column) const {
	Matrix3x3 temp;
	int r = 0;
	for(int i = 0; i < 4; ++i) {
		if(i != row) {
			int c = 0;
			for(int j = 0; j < 4; ++j) {
				if(j != column) {
					temp.elements[r][c++] = elements[i][j];
				}
			}
			++r;
		}
	}
	return temp;
}

fp Matrix4x4::getDeterminant() const {
	fp result = 0;
	for(int k = 0; k < 4; ++k) {
		Matrix3x3 temp = getSubMatrix(0, k);

		result = result - (2*(k & 1)-1) * temp.getDeterminant();
	}

	return result;
}

Matrix4x4 Matrix4x4::getInverse() const {
	Matrix4x4 inverse;

	Matrix3x3 m = getSubMatrix(3,3).getInverse();
	Vector3d v(elements[0][3], elements[1][3], elements[2][3]);

	Vector3d nv = m * v;

	float nvf[3] = { nv.x, nv.y, nv.z };

	for(int i = 0; i < 3; ++i) {
		for(int j = 0; j < 3; ++j) {
			inverse.elements[i][j] = m.elements[i][j];
		}
		inverse.elements[i][3] = -nvf[i];
	}

	return inverse;
}


std::ostream& operator<<(std::ostream& out, const Matrix4x4& m) {
	m.output(out);
	return out;
}

void Matrix4x4::output(std::ostream& out) const {
	for(int i = 0; i < 4; ++i) {
		for(int j = 0; j < 4; ++j) {
			out << elements[i][j] << " ";
		}
		out << std::endl;
	}
}

}
