#ifndef MATH_VECTOR3D_H
#define MATH_VECTOR3D_H

#include <cmath>
#include <iosfwd>

#include "Definitions.h"

namespace Math {

class Vector3d {
public:
  fp x, y, z;

  Vector3d() : x(0), y(0), z(0) {}
  Vector3d(fp x, fp y, fp z) : x(x), y(y), z(z) {}
};


inline fp dot(const Vector3d &a, const Vector3d &b)
{
  return a.x * b.x + a.y * b.y + a.z * b.z;
}


inline const Vector3d cross(const Vector3d &a, const Vector3d &b)
{
  return Vector3d(a.y * b.z - b.y * a.z,
                  a.z * b.x - b.z * a.x,
                  a.x * b.y - b.x * a.y);
}


inline fp abs(const Vector3d &v)
{
  return static_cast<Math::fp>(std::sqrt(static_cast<double>(dot(v, v))));
}


inline const Vector3d operator+(const Vector3d &a, const Vector3d &b)
{
  return Vector3d(a.x + b.x, a.y + b.y, a.z + b.z);
}


inline const Vector3d operator-(const Vector3d &a, const Vector3d &b)
{
  return Vector3d(a.x - b.x, a.y - b.y, a.z - b.z);
}


inline Vector3d &operator+=(Vector3d &a, const Vector3d &b)
{
  a.x += b.x; a.y += b.y; a.z += b.z;
  return a;
}


inline Vector3d &operator-=(Vector3d &a, const Vector3d &b)
{
  a.x -= b.x; a.y -= b.y; a.z -= b.z;
  return a;
}


inline const Vector3d operator-(const Vector3d &v)
{
  return Vector3d(-v.x, -v.y, -v.z);
}


inline const Vector3d operator*(const Vector3d &v, const fp f)
{
  return Vector3d(v.x * f, v.y * f, v.z * f);
}


inline const Vector3d operator/(const Vector3d &v, const fp f)
{
  return v * (1 / f);
}


inline Vector3d &operator*=(Vector3d &v, const fp f)
{
  v.x *= f; v.y *= f; v.z *= f;
  return v;
}


inline Vector3d &operator/=(Vector3d &v, const fp f)
{
  return v *= (1 / f);
}


inline Vector3d normalize(const Vector3d &v)
{
  return v / abs(v);
}


std::ostream &operator<<(std::ostream &out, const Vector3d &v);
std::istream &operator>>(std::istream &in, Vector3d &v);

}

#endif
