#ifndef OPENGL_CAMERA_H
#define OPENGL_CAMERA_H

#include "math/Vector3d.h"

/**
 * The OpenGL namespace
 */
namespace OpenGL {

/**
 * Represents a Camera with position and orientation.
 * The Camera is defined in a right-handed coordinate system.
 */
class Camera {

public:
	/**
	 * Constructs a new Camera with the given world position.
	 * By default the Camera is looking along the negative Z-axis, with
	 * the pitch and yaw angles set to 0. 
	 * @param pos the world position.
	 */
	Camera(Math::Vector3d pos);

	/**
	 * Moves the Camera forward or backward according to the orientation
	 * and the given distance. 
	 * Positive distance -> forward.
	 * Negative distance -> backward.
	 * @param dist the distance to move the Camera. 
	 */
	void move(float dist);

	/**
	 * Moves the Camera left or right according to the orientation
	 * and the given distance. 
	 * Positive distance -> right.
	 * Negative distance -> left.
	 * @param dist the distance to move the Camera. 
	 */
	void strafe(float dist);

	/**
	 * Moves the Camera up or down according to the orientation
	 * and the given distance. 
	 * Positive distance -> up.
	 * Negative distance -> down.
	 * @param dist the distance to move the Camera. 
	 */
	void up(float dist);

	/**
	 * Increases the pitch angle with the given angle. 
	 * Positive angle -> clockwise around the X-axis.
	 * Negative angle -> counter-clockwise around the X-axis. 
	 * @param angle the amount to increase the pitch angle with.
	 * @param limitRotation limit the rotation, disallowing loops.
	 */
	void pitch(float angle, bool limitRotation = false);

	/**
	 * Increases the yaw angle with the given angle.
	 * Positive angle -> clockwise around the Y-axis.
	 * Negative angle -> counter-clockwise around the Y-axis. 
	 * @param angle the amount to increase the yaw angle with.
	 */
	void yaw(float angle);

	/**
	 * Returns the world position of the Camera.
	 * @return the world position. 
	 */
	Math::Vector3d getPosition() const;

	/**
	 * Returns the normalized view direction vector of the Camera.
	 * @return the normalized view direction vector.
	 */
	Math::Vector3d getViewDirection() const;

	/**
	 * Transforms the Camera according to the position
	 * and the orientation. 
	 */
	void transform() const;

private:
	const float* getMatrix() const;

	float pitchAngle, yawAngle;

	Math::Vector3d pos;
};

};

#endif