#ifndef OPENGL_CGCONTEXT_H
#define OPENGL_CGCONTEXT_H

#include <memory>
#include <string>
#include <vector>

#include <GL/glew.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>


namespace OpenGL {

/**
 * The CgContext class is responsible for initializing a valid 
 * Cg context, and determines which fragment and vertex profiles are supported.
 * The class also contains which texture targets and internal formats to use. 
 */
class CgContext {
public:
	
	/**
	 * Returns the valid Cg context.
	 * @return the valid Cg context.
	 */
	CGcontext getContext() const;

	/**
	 * Returns whether the Cg context was successfully initialized.
	 * @return whether the Cg context was successfully initialized.
	 */
	bool good() const;

	/**
	 * Enables the given Cg profile. 
	 * This profile is then used by Cg programs executed afterwards. 
	 * @param profile the Cg profile to enable.
	 */
	void enableProfile(CGprofile profile) const;

	/**
	 * Disables the given Cg profile. 	
	 * @param profile the Cg profile to disable.
	 */
	void disableProfile(CGprofile profile) const;

	/**
	 * Returns the single CgContext instance. 
	 * An instance is initialized if not already existing.
	 * @return the single CgContext instance.
	 */
	static CgContext &getInstance();

	/**
	 * Returns whether the CgContext has an error.
	 * @return whether the CgContext has an error.
	 */
	bool hasError() const;

	/**
	 * Returns a textual description of the latest error.
	 * Returns an empty string if no error has occured. 
	 * @return a textual description of the latest error.
	 */
	std::string getError() const;

	/**
	 * Returns the Cg profile used for fragment shaders (Cg programs).
	 * @return the fragment shader Cg profile.
	 */
	CGprofile getFragmentShader() const;

	/**
	 * Returns the Cg profile used for vertex shaders (Cg programs).
	 * @return the vertex shader Cg profile.
	 */
	CGprofile getVertexShader() const;

	/**
	 * Returns the path where the Cg program files are stored.
	 * @return the Cg program path. 
	 */
	std::string getCgProgramPath() const;

	/**
	 * Returns the texture target used for samplerRECT textures.
	 * The texture target defines how the texture image
     * is specified and accessed via a set of texture coordinates.
	 * @return the texture rectangle target.
	 */
	GLenum getTexRectTarget() const;

	/**
	 * Returns the internal format used for samplerRECT textures.
	 * The internal format specifies the number of channels 
	 * and the type and precision of each channel.
	 * @return the texture rectangle internal format.
	 */
	GLint getTexRectInternalFormat() const;

	/**
	 * Returns the texture target used for sampler2D textures.
	 * The texture target defines how the texture image
     * is specified and accessed via a set of texture coordinates.
	 * @return the texture 2D target.
	 */
	GLenum getTex2dTarget() const;

	/**
	 * Returns the internal format used for sampler2D textures.
	 * The internal format specifies the number of channels 
	 * and the type and precision of each channel.
	 * @return the texture 2D internal format.
	 */
	GLint getTex2dInternalFormat() const;

private:
	CgContext();
	~CgContext();

	CgContext(const CgContext &context);
	CgContext &operator=(const CgContext &context);

	mutable std::string lastError;

	CGcontext context;
};

}

#endif
