#include <algorithm>
#include <cassert>

#include "opengl/CgOperation.h"
#include "opengl/IoTextureSet.h"

namespace OpenGL {

CgOperation::CgOperation(unsigned int computationWidth, unsigned int computationHeight,
                         CgProgram *interior, CgProgram *boundary,
                         bool normalizedTexCoords)
	: interior(interior),
	  boundary(boundary),
	  ioTextureSet(0),
	  computationWidth(computationWidth),
	  computationHeight(computationHeight),
	  mustSwap(false),
	  outputTextureSet(false),
	  normalizedTexCoords(normalizedTexCoords)
{
	if (!interior)
		throw NullInteriorProgramException();

	ioTextureSet = interior->getIoTextureSet();

	if (boundary && ioTextureSet != boundary->getIoTextureSet())
		throw IoTextureSetMismatchException();
}

CgOperation::~CgOperation() 
{

}


void CgOperation::compute(bool rtt)
{
	setViewport();

	if (rtt && ioTextureSet)
		ioTextureSet->setActiveTexture(outputTexture, mustSwap);

	activateInteriorProgram();
	executeInteriorProgram();
	deactivateInteriorProgram();

	if (hasBoundaryProgram()) {
		activateBoundaryProgram();
		executeBoundaryProgram();
		deactivateBoundaryProgram();
	}

	if (mustSwap && ioTextureSet)
		ioTextureSet->swap();

	resetViewport();
}


void CgOperation::setOutputTexture(unsigned int texture)
{
	setOutputTexture(texture, computationWidth, computationHeight);
}


void CgOperation::setOutputTexture(unsigned int texture, unsigned int width, unsigned int height)
{
	outputTexture = texture;
	outputTextureSet = true;

	outputWidth = width;
	outputHeight = height;

	checkSwapping();
}


void CgOperation::checkSwapping()
{
	mustSwap = interior->isInputTexture(outputTexture) ||
	           boundary && boundary->isInputTexture(outputTexture);
}


void CgOperation::setViewport()
{
	glGetIntegerv(GL_VIEWPORT, oldViewport);
	glViewport(0, 0, outputWidth, outputHeight);
}


void CgOperation::resetViewport()
{
	glViewport(oldViewport[0], oldViewport[1], oldViewport[2], oldViewport[3]);
}


void CgOperation::activateInteriorProgram()
{
	interior->activate();
}


void CgOperation::deactivateInteriorProgram()
{
	interior->deactivate();
}


void CgOperation::activateBoundaryProgram()
{
	assert(hasBoundaryProgram());
	boundary->activate();
}


void CgOperation::deactivateBoundaryProgram()
{
	assert(hasBoundaryProgram());
	boundary->deactivate();
}


bool CgOperation::hasBoundaryProgram() const
{
	return boundary != 0;
}


unsigned int CgOperation::getComputationWidth() const
{
	return computationWidth;
}


unsigned int CgOperation::getComputationHeight() const
{
	return computationHeight;
}


bool CgOperation::hasNormalizedTexCoords() const
{
	return normalizedTexCoords;
}

}