#ifndef OPENGL_CGOPERATION_H
#define OPENGL_CGOPERATION_H

#include "opengl/CgProgram.h"
#include "opengl/IoTextureSet.h"

#include <exception>
#include <vector>

namespace OpenGL {

struct IoTextureSetMismatchException : public std::exception
{
public:
	const char *what() const {
		return "CgOperation: IoTextureSet mismatch.";
	}
};

struct NullInteriorProgramException  : public std::exception
{
public:
	const char *what() const {
		return "CgOperation: Interior program was null.";
	}
};

/**
 * CgOperation is an abstraction of Cg fragment programs performing operations on textures. It
 * uses instances of CgProgram for interior and optionally boundary operations. The user chooses
 * which output texture to use, and the CgOperation instance determines whether swapping is
 * required (if the output texture is also an input texture for either of the interior or boundary
 * CgProgram instances). To perform the operation represented by an instance of CgOperation
 * the member function compute() is offered, but it can only be called through subclasses since
 * CgOperation is an abstract class. CgOperation is also responsible for setting the correct
 * viewports needed when operating on the textures.
 */
class CgOperation {
public:
	/**
	 * Base constructor.
	 * @param computationWidth the width of the computation domain.
	 * @param computationHeight the height of the computation domain.
	 * @param interior the CgProgram instance used for interior computations.
	 * @param boundary the CgProgram instance used for boundary computations.
	 * @param normalizedTexCoords whether normalized texture coordinates is used.	         
	 */
	CgOperation(unsigned int computationWidth, unsigned int computationHeight,
	            CgProgram *interior, CgProgram *boundary = 0,
	            bool normalizedTexCoords = false);
	
	virtual ~CgOperation() = 0;

	/**
	 * Performs one computation by executing the interior and
	 * boundary CgProgram instances. Before the computation is performed
	 * the viewport is set according to the computation width and height.
	 * @param rtt whether the results from the CgProgram instances are written to
	 *            the output texture or the frame buffer.
	 */
	void compute(bool rtt = true);

	/**
	 * Sets the output texture used as render target when compute() is called.
	 * The texture identifier must be valid in the 
	 * IoTextureSet instance used by the interior and boundary CgProgram instances.
	 * @param texture the texture identifier of the output texture.
	 */
	void setOutputTexture(unsigned int texture);

	/**
	 * Sets the output texture used as render target when compute() is called.
	 * This function should be used if the output texture dimensions differ from 
	 * the computation width and height.
	 * The texture identifier must be valid in the 
	 * IoTextureSet instance used by the interior and boundary CgProgram instances.
	 * @param texture the texture identifier of the output texture.
	 * @param outputWidth the new computation width.
	 * @param outputHeight the new computation height.
	 */
	void setOutputTexture(unsigned int texture,
	                      unsigned int outputWidth,
	                      unsigned int outputHeight);

protected:
	/**
	 * Returns whether the CgOperation instance has a boundary CgProgram;
	 * @return whether the CgOperation instance has a boundary CgProgram;
	 */
	bool hasBoundaryProgram() const;

	/**
	 * Returns the computation width.
	 * @return the computation width.
	 */
	unsigned int getComputationWidth() const;

	/**
	 * Returns the computation height.
	 * @return the computation height.
	 */
	unsigned int getComputationHeight() const;

	/**
	 * Returns whether the CgOperation instance uses normalized texture coordinates.
	 * @return whether the CgOperation instance uses normalized texture coordinates.
	 */
	bool hasNormalizedTexCoords() const;

private:
	void activateInteriorProgram();
	void deactivateInteriorProgram();

	void activateBoundaryProgram();
	void deactivateBoundaryProgram();

	void setViewport();
	void resetViewport();

	void checkSwapping();


	/**
	 * Implementation of interior CgProgram execution, must be supplied by concrete subclasses.	 
	 * Execution is initiated by drawing geometries.
	 */
	virtual void executeInteriorProgram() = 0;

	/**
	 * Implementation of boundary CgProgram execution, must be supplied by concrete subclasses.	 
	 * Execution is initiated by drawing geometries.
	 */
	virtual void executeBoundaryProgram() = 0;

	CgProgram *interior, *boundary;

	IoTextureSet *ioTextureSet;

	unsigned int outputTexture;

	unsigned int computationWidth, computationHeight;
	unsigned int outputWidth, outputHeight;

	bool mustSwap;
	bool outputTextureSet;
	bool normalizedTexCoords;

	int oldViewport[4];
};

}

#endif