#include "CgProgram.h"

#include <iostream>

#include <GL/glew.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#include "CgContext.h"

namespace OpenGL {

CgProgram::ParameterFailedException::ParameterFailedException(const std::string &name) :
	errorStr("CgProgram: Unknown parameter: " + name)
{
}

const char *CgProgram::ParameterFailedException::what() const
{
	return errorStr.c_str();
}

CgProgram::FailedLoadingProgramException::FailedLoadingProgramException(const std::string &cgError,
                                                                        const std::string &program,
																		const std::string &entry) :
	errorStr("CgProgram: Failed loading program " + program + '/' + entry + ": " + cgError)
{
}

const char *CgProgram::FailedLoadingProgramException::what() const
{
	return errorStr.c_str();
}

CgProgram::CgProgram(const std::string& programFile, CGprofile profile,
                     const std::string& entryPoint, IoTextureSet* ioTextureSet) :
	profile(profile),
	ioTextureSet(ioTextureSet)
{

	program = cgCreateProgramFromFile(CgContext::getInstance().getContext(),
		CG_SOURCE,
		programFile.c_str(),
		profile,
		entryPoint.c_str(),
		0);

	if (program)
		cgGLLoadProgram(program);
	else {
		CGerror cgError = cgGetError();

		throw FailedLoadingProgramException(cgGetErrorString(cgError), programFile, entryPoint);
	}
}


CgProgram::~CgProgram()
{
	cgDestroyProgram(program);
}


unsigned int CgProgram::getParameterId(const std::string& name)
{
	if(indexMap.find(name) == indexMap.end()) {

		CGparameter p = cgGetNamedParameter(program, name.c_str());

		if(p)
			return addParameter(p, name);
		else
			throw ParameterFailedException(name);
	}
	else
		return indexMap[name];
}


void CgProgram::setParameter(unsigned int parameterId, float value) const
{
	cgGLSetParameter1f(parameters[parameterId], value);
}


void CgProgram::setParameter(unsigned int parameterId, float values[4]) const
{
	cgGLSetParameter4f(parameters[parameterId], 
	                   values[0],
	                   values[1],
	                   values[2],
	                   values[3]);
}

void CgProgram::setMatrixParameter(unsigned int parameterId, const float values[16]) const
{
	cgGLSetMatrixParameterfr(parameters[parameterId], values);
}

void CgProgram::setMatrixParameter(const std::string &name, const float values[16])
{
	setMatrixParameter(getParameterId(name), values);
}

void CgProgram::setTextureParameter(const std::string &name, unsigned int value)
{
	setTextureParameter(getParameterId(name), value);
}


void CgProgram::setTextureParameter(unsigned int parameterId, unsigned int value)
{
	textureMap[parameters[parameterId]] = value;
}

void CgProgram::setFreeTextureParameter(const std::string &name, GLuint texture)
{
	setFreeTextureParameter(getParameterId(name), texture);
}

void CgProgram::setFreeTextureParameter(unsigned int parameterId, GLuint texture)
{
	freeTextureMap[parameters[parameterId]] = texture;
}

void CgProgram::loadModelViewProjection(unsigned int parameterId) const
{
	cgGLSetStateMatrixParameter(parameters[parameterId],
                                CG_GL_MODELVIEW_PROJECTION_MATRIX,
	                            CG_GL_MATRIX_IDENTITY);
}


void CgProgram::activate()
{
	cgGLBindProgram(program);

	cgGLEnableProfile(profile);

	for(TextureMap::const_iterator iter = textureMap.begin(), end = textureMap.end();
	    iter != end; ++iter)
	{
		GLuint texture = ioTextureSet->lookup(iter->second);
		cgGLSetTextureParameter(iter->first, texture);
		cgGLEnableTextureParameter(iter->first);
	}

	for(FreeTextureMap::const_iterator iter = freeTextureMap.begin(), end = freeTextureMap.end();
	    iter != end; ++iter)
	{		
		cgGLSetTextureParameter(iter->first, iter->second);
		cgGLEnableTextureParameter(iter->first);
	}
}


void CgProgram::deactivate()
{
	for(FreeTextureMap::const_iterator iter = freeTextureMap.begin(), end = freeTextureMap.end();
	    iter != end; ++iter)
	{		
		cgGLDisableTextureParameter(iter->first);
	}

	for(TextureMap::const_iterator iter = textureMap.begin(), end = textureMap.end();
	    iter != end; ++iter)
	{
		cgGLDisableTextureParameter(iter->first);
	}

	cgGLDisableProfile(profile);
}


IoTextureSet *CgProgram::getIoTextureSet() const
{
	return ioTextureSet;
}

bool CgProgram::isInputTexture(unsigned int texture) const
{
	TextureMap::const_iterator iter, end = textureMap.end();

	for (iter = textureMap.begin(); iter != end; ++iter)
		if (iter->second == texture)
			break;

	return iter != end;
}


unsigned int CgProgram::addParameter(CGparameter parameter, const std::string& name)
{
	parameters.push_back(parameter);
	unsigned int result = static_cast<unsigned int>(parameters.size() - 1);
	indexMap[name] = result;
	return result;
}

}