#ifndef OPENGL_CGPROGRAM_H
#define OPENGL_CGPROGRAM_H

#include <GL/glew.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#include "opengl/IoTextureSet.h"

#include <exception>
#include <string>
#include <memory>
#include <vector>
#include <map>

namespace OpenGL {

/**
 * CgProgram is a simple wrapper around a Cg fragment or vertex program. It loads and initializes
 * Cg programs, and makes it easy to activate and deactivate them by using the member functions
 * activate() and deactivate() respectively. In addition, it contains support for setting various
 * uniform parameters like floats, vectors, and textures. If the Cg program needs to read from
 * textures, a reference to an IoTextureSet instance must be supplied when initializing the
 * CgProgram instance.
 */
class CgProgram {
public:
	struct ParameterFailedException {
	public:
		ParameterFailedException(const std::string &name);
		const char *what() const;
	private:
		std::string errorStr;
	};

	struct FailedLoadingProgramException : public std::exception {
	public:
		FailedLoadingProgramException(const std::string &cgError,
		                              const std::string &program,
                                      const std::string &entry);
		const char *what() const;
	private:
		std::string errorStr;
	};
	
	/**
	 * Constructs a new CgProgram instance that acts as a wrapper
	 * for the Cg program given by the entry point in the Cg program file. 	 
	 * @param programFile the name of the Cg program file (including the path).
	 * @param profile the Cg profile used by the Cg program.
	 * @param entryPoint the entry point (function) that represents the Cg program.
	 * @param ioTextureSet the texture set containing textures used by the Cg program.
	 */
	CgProgram(const std::string& programFile, CGprofile profile,
	          const std::string& entryPoint,
	          IoTextureSet* ioTextureSet);
	
	~CgProgram();

	/**
	 * Returns a parameter identifier associated with a Cg parameter with the given name. 
	 * The Cg parameter is bound to the Cg program.
	 * @param name the name of the Cg parameter.
	 * @return the parameter identifier.
	 */
	unsigned int getParameterId(const std::string& name);

	/**
	 * Sets a value on the Cg parameter associated with the given parameter identifier.
	 * @param parameterId the parameter identifier.
	 * @param value the new value to be set on the Cg parameter.
	 */
	void setParameter(unsigned int parameterId, float value) const;

	/**
	 * Sets a vector on the Cg parameter associated with the given parameter identifier.
	 * @param parameterId the parameter identifier.
	 * @param values the new vector to be set on the Cg parameter.
	 */
	void setParameter(unsigned int parameterId, float values[4]) const;

	/**
	 * Sets a matrix on the Cg parameter associated with the given parameter identifier.
	 * @param parameterId the parameter identifier.
	 * @param values the new matrix to be set on the Cg parameter.
	 */
	void setMatrixParameter(unsigned int parameterId, const float values[16]) const;

	/**
	 * Sets a matrix on the Cg parameter associated with the given parameter name.
	 * @param name the parameter name.
	 * @param values the new matrix to be set on the Cg parameter.
	 */
	void setMatrixParameter(const std::string &name, const float values[16]);

	/**
	 * Sets a value on the Cg parameter associated with the given parameter name.
	 * @param name the parameter name.
	 * @param value the new value to be set on the Cg parameter.
	 */
	template <typename T>
	void setParameter(const std::string &name, T value);

	/**
	 * Sets a texture on the Cg parameter associated with the given parameter name.
	 * @param name the parameter name.
	 * @param value the texture identifier of the texture to be set on the Cg parameter.
	 */
	void setTextureParameter(const std::string &name, unsigned int value);

	/**
	 * Sets a texture on the Cg parameter associated with the given parameter identifier.
	 * @param parameterId the parameter identifier.
	 * @param value the texture identifier of the texture to be set on the Cg parameter.
	 */
	void setTextureParameter(unsigned int parameterId, unsigned int value);

	/**
	 * Sets a free texture directly on the Cg parameter associated with the given parameter name.
	 * The free texture does not have to be a part of the IoTextureSet used by the CgProgram instance.
	 * @param name the parameter name.
	 * @param texture the texture to be set on the Cg parameter.
	 */
	void setFreeTextureParameter(const std::string &name, GLuint texture);

	/**
	 * Sets a free texture directly on the Cg parameter associated with the given parameter identifier.
	 * The free texture does not have to be a part of the IoTextureSet used by the CgProgram instance.
	 * @param parameterId the parameter identifier.
	 * @param texture the texture to be set on the Cg parameter.
	 */
	void setFreeTextureParameter(unsigned int parameterId, GLuint texture);

	/**
	 * Loads the OpenGL modelview projection matrix into the Cg parameter associated 
	 * with the given parameter identifier.
	 * @param parameterId the parameter identifier.
	 */
	void loadModelViewProjection(unsigned int parameterId) const;

	/**
	 * Returns a pointer to the IoTextureSet instance used by the CgProgram instance.
	 * @return a pointer to the IoTextureSet instance.
	 */
	IoTextureSet *getIoTextureSet() const;

	/**
	 * Returns whether the texture with the given texture identifier is set
	 * to one of the Cg parameters of this CgProgram instance. 
	 * @param texture the texture identifier.
	 * @return whether the texture is a input texture. 
	 */
	bool isInputTexture(unsigned int texture) const;

	/**
	 * Activates the Cg program.
	 */
	void activate();

	/**
	 * Deactivates the Cg program.
	 */
	void deactivate();

private:
	typedef std::map<CGparameter, unsigned int> TextureMap;
	TextureMap textureMap;

	typedef std::map<CGparameter, GLuint> FreeTextureMap;
	FreeTextureMap freeTextureMap;
	
	CGprogram program;
	CGprofile profile;
	IoTextureSet *ioTextureSet;

	std::map<std::string, unsigned int> indexMap;
	std::vector<CGparameter> parameters;

	unsigned int addParameter(CGparameter parameter, const std::string& name);
};

template <typename T>
void CgProgram::setParameter(const std::string &name, T value)
{
	setParameter(getParameterId(name), value);
}

};

#endif