#include "Flat3dTextureOperation.h"

#include "CgProgram.h"

namespace OpenGL {
	
Flat3dTextureOperation::Flat3dTextureOperation(unsigned int tileWidth, unsigned int tileHeight, 
                                               unsigned int layoutWidth, unsigned int layoutHeight, 
                                               unsigned int numTiles,
                                               CgProgram *interior, CgProgram *boundary,
                                               OperationType operationType,
											   bool normalizedTexCoords) :
	CgOperation(tileWidth * layoutWidth, tileHeight * layoutHeight, interior, boundary, normalizedTexCoords),
	layoutWidth(layoutWidth), layoutHeight(layoutHeight),
	tileWidth(tileWidth), tileHeight(tileHeight), numTiles(numTiles), operationType(operationType)
{
	initExecutionParameters();
}

Flat3dTextureOperation::Flat3dTextureOperation(unsigned int tileWidth, unsigned int tileHeight, 
                                               unsigned int layoutWidth, unsigned int layoutHeight, 
                                               unsigned int numTiles,
                                               CgProgram *interior, OperationType operationType,
											   bool normalizedTexCoords) :
	CgOperation(tileWidth * layoutWidth, tileHeight * layoutHeight, interior, 0, normalizedTexCoords),
	layoutWidth(layoutWidth), layoutHeight(layoutHeight),
	tileWidth(tileWidth), tileHeight(tileHeight), numTiles(numTiles), operationType(operationType)
{
	initExecutionParameters();
}

void Flat3dTextureOperation::initExecutionParameters()
{
	float simWidth = static_cast<float>(getComputationWidth());
	float simHeight = static_cast<float>(getComputationHeight());

	tileWidthOrtho = 1.0f / static_cast<float>(layoutWidth);
	tileHeightOrtho = 1.0f / static_cast<float>(layoutHeight);	

	float texDelta = hasBoundaryProgram() || operationType != INTERIOR_FULL ? 1.0f : 0.0f;
	float vertexDelta = hasBoundaryProgram() || operationType != INTERIOR_FULL ? 1.0f: 0.0f;	

	minX  = vertexDelta / simWidth;
	maxX  = tileWidthOrtho - minX;
	minY  = vertexDelta / simHeight;
	maxY  = tileHeightOrtho - minY;

	if (hasNormalizedTexCoords()) {

		minS = texDelta / simWidth;
		maxS = tileWidthOrtho - minX;
		minT = texDelta / simHeight;
		maxT = tileHeightOrtho - minY;		

	} else {

		minS = texDelta;
		maxS = tileWidth - texDelta;
		minT = texDelta;
		maxT = tileHeight - texDelta;

	}
}


void Flat3dTextureOperation::executeInteriorProgram()
{
	unsigned int currentTile = 0;

	bool hasNormalizedTexCoords = this->hasNormalizedTexCoords();

	for (unsigned int y = 0; y < layoutHeight && currentTile < numTiles; ++y) {

		float offsetY = tileHeightOrtho * static_cast<float>(y);
		float offsetT = hasNormalizedTexCoords ? offsetY : 
				static_cast<float>(tileHeight * y);

		for (unsigned int x = 0; x < layoutWidth; ++x) {

			float offsetX = tileWidthOrtho * static_cast<float>(x);
			float offsetS = hasNormalizedTexCoords ? offsetX : 
				static_cast<float>(tileWidth * x);

			bool isBoundarySlice = (currentTile == 0 || currentTile == numTiles - 1);

			if (!isBoundarySlice || operationType != INTERIOR_3D) {
				// draw interior quad				
				drawQuad(offsetX, offsetY, offsetS, offsetT);						
			}

			if (++currentTile >= numTiles)
				break;
		}
	}
}

void Flat3dTextureOperation::executeBoundaryProgram()
{
	unsigned int currentTile = 0;

	bool hasNormalizedTexCoords = this->hasNormalizedTexCoords();

	for (unsigned int y = 0; y < layoutHeight && currentTile < numTiles; ++y) {

		float offsetY = tileHeightOrtho * static_cast<float>(y);
		float offsetT = hasNormalizedTexCoords ? offsetY : 
				static_cast<float>(tileHeight * y);

		for (unsigned int x = 0; x < layoutWidth; ++x) {

			float offsetX = tileWidthOrtho * static_cast<float>(x);			
			float offsetS = hasNormalizedTexCoords ? offsetX : 
				static_cast<float>(tileWidth * x);

			bool isBoundarySlice = (currentTile == 0 || currentTile == numTiles - 1);
			
			if (isBoundarySlice && operationType == INTERIOR_3D) {
				// draw boundary quad				
				glMultiTexCoord3f(GL_TEXTURE1_ARB, 0, 0, currentTile ? -1 : 1); // offset amount
				drawQuad(offsetX, offsetY, offsetS, offsetT);				
			}				
			
			// draw frame				
			drawFrame(offsetX, offsetY, offsetS, offsetT);				

			if (++currentTile >= numTiles)
				break;
		}
	}
}

void Flat3dTextureOperation::drawQuad(float offsetX, float offsetY, float offsetS, float offsetT)
{
	float minX = this->minX + offsetX;
	float maxX = this->maxX + offsetX;
	float minY = this->minY + offsetY;
	float maxY = this->maxY + offsetY;

	float minS = this->minS + offsetS;
	float maxS = this->maxS + offsetS;
	float minT = this->minT + offsetT;
	float maxT = this->maxT + offsetT;

	glBegin(GL_QUADS);

	glTexCoord2f(minS, minT);
	glVertex2f(minX, minY);

	glTexCoord2f(maxS, minT);
	glVertex2f(maxX, minY);

	glTexCoord2f(maxS, maxT);
	glVertex2f(maxX, maxY);

	glTexCoord2f(minS, maxT);
	glVertex2f(minX, maxY);

	glEnd();
}


void Flat3dTextureOperation::drawFrame(float offsetX, float offsetY, float offsetS, float offsetT)
{
	float halfX = minX / 2.0f;
	float halfY = minY / 2.0f;

	float halfS = minS / 2.0f;
	float halfT = minT / 2.0f;

	float minS = this->minS - halfS + offsetS;
	float maxS = this->maxS + halfS + offsetS;

	float minT = this->minT - halfT + offsetT;
	float maxT = this->maxT + halfT + offsetT;

	float minX = this->minX - halfX + offsetX;
	float maxX = this->maxX + halfX + offsetX;

	float minY = this->minY - halfY + offsetY;
	float maxY = this->maxY + halfY + offsetY;

	glBegin(GL_LINES);

	// left boundary
	glMultiTexCoord3f(GL_TEXTURE1_ARB, this->minS, 0, 0); // offset amount
	glTexCoord2f(minS, minT);
	glVertex2f(minX, minY);
	glTexCoord2f(minS, maxT);
	glVertex2f(minX, maxY);

	// top boundary
	glMultiTexCoord3f(GL_TEXTURE1_ARB, 0, -this->minT, 0); // offset amount
	glTexCoord2f(minS, maxT);
	glVertex2f(minX, maxY);
	glTexCoord2f(maxS, maxT);
	glVertex2f(maxX, maxY);

	// right boundary
	glMultiTexCoord3f(GL_TEXTURE1_ARB, -this->minS, 0, 0); // offset amount
	glTexCoord2f(maxS, maxT);
	glVertex2f(maxX, maxY);
	glTexCoord2f(maxS, minT);
	glVertex2f(maxX, minY);

	// bottom boundary
	glMultiTexCoord3f(GL_TEXTURE1_ARB, 0, this->minT, 0); // offset amount
	glTexCoord2f(maxS, minT);
	glVertex2f(maxX, minY);
	glTexCoord2f(minS, minT);
	glVertex2f(minX, minY);

	glMultiTexCoord2f(GL_TEXTURE1_ARB, 0, 0); // clear texture coordinates

    glEnd();
}


}
