#ifndef OPENGL_FLAT3DTEXTUREOPERATION_H
#define OPENGL_FLAT3DTEXTUREOPERATION_H

#include "CgOperation.h"

namespace OpenGL {

class CgProgram;

/**
 * Flat3dTextureOperation is a specialization of CgOperation for performing operations on
 * flat 3D textures (Flat3dTextureSet). For interior programs it draws several quads, one for each slice, in order to
 * compute the interior CgProgram. For boundary fragments, it draws frames surrounding each
 * quad to compute the boundary CgProgram. In addition, for 3D volume domains the front and
 * back quads are also computed using the boundary CgProgram.
 */
class Flat3dTextureOperation : public CgOperation {
public:
	/**
	 * Operation types
	 */
	enum OperationType {
		INTERIOR_3D,  /// does not perform interior operation on boundary tiles or boundary frames.
		INTERIOR_2D,  /// performs interior operation on all tiles but not on boundary frames.
		INTERIOR_FULL /// performs interior operation on all tiles, including boundary frames.
	};

	/**
	 * Creates a Flat3dTextureOperation instance with both interior and boundary CgProgram instances.
	 * @param tileWidth the width of one tile in the Flat3dTextureSet.
	 * @param tileHeight the height of one tile in the Flat3dTextureSet.
	 * @param layoutWidth the number of columns in the tile layout of the Flat3dTextureSet.
	 * @param layoutHeight the number of rows in the tile layout of the Flat3dTextureSet.
	 * @param numTiles the total number of tiles in the Flat3dTextureSet.
	 * @param interior the CgProgram instance used for interior computations.
	 * @param boundary the CgProgram instance used for boundary computations.
	 * @param operationType the operation type of the Flat3dTextureOperation instance.
	 * @param normalizedTexCoords whether normalized texture coordinates is used.
	 */
	Flat3dTextureOperation(unsigned int tileWidth, unsigned int tileHeight, 
                           unsigned int layoutWidth, unsigned int layoutHeight, unsigned int numTiles,
                           CgProgram *interior, CgProgram *boundary, OperationType operationType,
                           bool normalizedTexCoords = false);

	/**
	 * Creates a Flat3dTextureOperation instance with only an interior CgProgram instance.
	 * @param tileWidth the width of one tile in the Flat3dTextureSet.
	 * @param tileHeight the height of one tile in the Flat3dTextureSet.
	 * @param layoutWidth the number of columns in the tile layout of the Flat3dTextureSet.
	 * @param layoutHeight the number of rows in the tile layout of the Flat3dTextureSet.
	 * @param numTiles the total number of tiles in the Flat3dTextureSet.
	 * @param interior the CgProgram instance used for interior computations.	 
	 * @param operationType the operation type of the Flat3dTextureOperation instance.
	 * @param normalizedTexCoords whether normalized texture coordinates is used.
	 */
	Flat3dTextureOperation(unsigned int tileWidth, unsigned int tileHeight, 
                           unsigned int layoutWidth, unsigned int layoutHeight, unsigned int numTiles,
                           CgProgram *interior, OperationType operationType,
                           bool normalizedTexCoords = false);

private:
	void executeInteriorProgram();
	void executeBoundaryProgram();
	void initExecutionParameters();
	void drawQuad(float offsetX, float offsetY, float offsetS, float offsetT);
	void drawFrame(float offsetX, float offsetY, float offsetS, float offsetT);

	unsigned int numTiles;
	unsigned int layoutWidth, layoutHeight;
	float tileWidthOrtho, tileHeightOrtho;
	unsigned int tileWidth, tileHeight;
	OperationType operationType;

	// execution parameters
	float minX, minY, maxX, maxY;
	float minS, minT, maxS, maxT;

	CgProgram *interior;
	CgProgram *boundary;	
};

}

#endif