#include "Flat3dTextureSet.h"

#include <algorithm>
#include <cassert>

using std::max;
using std::min;

namespace OpenGL {

std::pair<unsigned int, unsigned int>
calculateLayoutDimensions(unsigned int computationWidth,
                          unsigned int computationHeight,
                          unsigned int computationDepth)
{
	std::pair<unsigned int, unsigned int> layout;

	double bestRatio = 0.0;

	for (unsigned int h = 1; h <= computationDepth; ++h) {
		for (unsigned int w = 1; w <= computationDepth; ++w) {
			unsigned int totalWidth = w * computationWidth;
			unsigned int totalHeight = h * computationHeight;

			unsigned int numTiles = w * h;

			if (numTiles < computationDepth)
				continue;

			if (numTiles - computationDepth >= w)
				continue;

			double ratio = static_cast<double>(min(totalWidth, totalHeight)) /
			               static_cast<double>(max(totalWidth, totalHeight));

			if (ratio > bestRatio) {
				layout = std::make_pair(totalWidth, totalHeight);
				bestRatio = ratio;
			}
		}
	}

	return layout;
}


Flat3dTextureSet::Flat3dTextureSet(unsigned int computationWidth,
                                   unsigned int computationHeight,
								   unsigned int computationDepth,
								   GLenum target,
                                   GLint internalFormat) :
	computationWidth(computationWidth),
	computationHeight(computationHeight),
	computationDepth(computationDepth),
	IoTextureSet(calculateLayoutDimensions(computationWidth,
	                                       computationHeight,
	                                       computationDepth),
	             target,
	             internalFormat)
{
	layoutHeight = getPrimaryHeight() / computationHeight;
	layoutWidth = getPrimaryWidth() / computationWidth;

	initializeDomainLookup();
	initializeNeighborLookup();
	initializeOffsetLookup();
	initializeSliceLookup();
}

Flat3dTextureSet::~Flat3dTextureSet() {}

void Flat3dTextureSet::fillTexture(unsigned int textureId,
                                   unsigned int numChannels,
                                   const std::vector<float> &values)
{
	std::vector<float> 
		pixels(4 * getPrimaryWidth() * getPrimaryHeight());

	unsigned int index = 0;

	for (unsigned int z = 0; z < computationDepth; ++z) {
		for (unsigned int y = 0; y < computationHeight; ++y) {
			for (unsigned int x = 0; x < computationWidth; ++x) {
				unsigned int globalX = 
					(z % layoutWidth) * computationWidth + x;
				unsigned int globalY = 
					(z / layoutWidth) * computationHeight + y;

				unsigned int globalIndex = 
					4 * (globalY * getPrimaryWidth() + globalX);

				for (unsigned int i = 0; i < numChannels; ++i) {
					if (index < values.size())
						pixels[globalIndex + i] = values[index++];
				}
			}
		}
	}

	IoTextureSet::fillTexture(textureId, GL_RGBA, GL_FLOAT, &*pixels.begin());
}


void Flat3dTextureSet::fillTexture(unsigned int textureId,
								   const Util::VoxelGrid<Util::Color> &values)
{
	assert(computationWidth == values.getWidth() &&
	       computationHeight == values.getHeight() &&
	       computationDepth == values.getDepth());

	std::vector<float> 
		pixels(4 * getPrimaryWidth() * getPrimaryHeight());

	unsigned int index = 0;

	for (unsigned int z = 0; z < computationDepth; ++z) {
		for (unsigned int y = 0; y < computationHeight; ++y) {
			for (unsigned int x = 0; x < computationWidth; ++x) {
				unsigned int globalX = 
					(z % layoutWidth) * computationWidth + x;
				unsigned int globalY = 
					(z / layoutWidth) * computationHeight + y;

				unsigned int globalIndex = 
					4 * (globalY * getPrimaryWidth() + globalX);
				
				Util::Color color = values(x, y, z);

				pixels[globalIndex] = color.getRed();
				pixels[globalIndex + 1] = color.getGreen();
				pixels[globalIndex + 2] = color.getBlue();
				pixels[globalIndex + 3] = color.getAlpha();
			}
		}
	}

	IoTextureSet::fillTexture(textureId, GL_RGBA, GL_FLOAT, &*pixels.begin());
}

Util::VoxelGrid<Util::Color> Flat3dTextureSet::getVoxelGrid(unsigned int textureId)
{
	Util::VoxelGrid<Util::Color> result(computationWidth, computationHeight, computationDepth);

	GLuint texture = IoTextureSet::lookup(textureId);

	unsigned int width = getWidth(textureId);
	unsigned int height = getHeight(textureId);

	GLenum target = getTarget(textureId);

	std::vector<float> v(4 * width * height);

	glBindTexture(target, texture);
	glGetTexImage(target, 0, GL_RGBA, GL_FLOAT, &*v.begin());

	for (unsigned int i = 0; i < v.size(); i+=4) {

		unsigned int globalX = (i / 4) % width;
		unsigned int globalY = (i / 4) / width;

		unsigned int localX = globalX % computationWidth;
		unsigned int localY = globalY % computationHeight;
		unsigned int localZ = globalX / computationWidth + layoutWidth * (globalY / computationHeight);

		if (localZ < computationDepth)
			result(localX, localY, localZ) = Util::Color(v[i], v[i+1], v[i+2], v[i+3]);
	}

	return result;
}


unsigned int Flat3dTextureSet::getDomainLookup() const
{
	return domainLookup;
}


unsigned int Flat3dTextureSet::getNeighborLookup() const
{
	return neighborLookup;
}


unsigned int Flat3dTextureSet::getOffsetLookup() const
{
	return offsetLookup;
}

unsigned int Flat3dTextureSet::getSliceLookup() const
{
	return sliceLookup;
}

unsigned int Flat3dTextureSet::getLayoutWidth() const
{
	return layoutWidth;
}

unsigned int Flat3dTextureSet::getLayoutHeight() const
{
	return layoutHeight;
}

unsigned int Flat3dTextureSet::getComputationWidth() const
{
	return computationWidth;
}

unsigned int Flat3dTextureSet::getComputationHeight() const
{
	return computationHeight;
}

unsigned int Flat3dTextureSet::getComputationDepth() const
{
	return computationDepth;
}

std::pair<unsigned int, unsigned int> Flat3dTextureSet::getGlobalCoordinates(int x, int y, int z) const
{
	unsigned int globalX = 
		(z % layoutWidth) * computationWidth + x;
	unsigned int globalY = 
		(z / layoutWidth) * computationHeight + y;	

	return std::pair<unsigned int, unsigned int>(globalX, globalY);
}

void Flat3dTextureSet::initializeDomainLookup()
{
	domainLookup = addPrimaryTexture("Domain lookup texture");

	std::vector<float> pixels(3 * getPrimaryWidth() * getPrimaryHeight());

	for (unsigned int y = 0; y < getPrimaryHeight(); ++y) {
		for (unsigned int x = 0; x < getPrimaryWidth(); ++x) {
			unsigned int tile = (y / computationHeight) * layoutWidth +
			                    (x / computationWidth);

			if (tile < computationDepth) {
				unsigned int index = 3 * (y * getPrimaryWidth() + x);

				pixels[index] = 
					static_cast<float>(x % computationWidth) + 0.5f;
				pixels[index + 1] = 
					static_cast<float>(y % computationHeight) + 0.5f;
				pixels[index + 2] = 
					static_cast<float>(tile) + 0.5f;
			}
		}
	}

	IoTextureSet::fillTexture(domainLookup, GL_RGB, GL_FLOAT, &*pixels.begin());
}


void Flat3dTextureSet::initializeNeighborLookup()
{
	neighborLookup = addPrimaryTexture("Neighbor lookup texture");

	std::vector<float> pixels(4 * getPrimaryWidth() * getPrimaryHeight());

	for (unsigned int tile = 0; tile < computationDepth; ++tile) {
		for (unsigned int y = 0; y < computationHeight; ++y) {
			for (unsigned int x = 0; x < computationWidth; ++x) {
				unsigned int offsetX = x + computationWidth * (tile % layoutWidth);
				unsigned int offsetY = y + computationHeight * (tile / layoutWidth);

				unsigned index = 4 * (offsetY * getPrimaryWidth() + offsetX);

				if (tile > 0) {
					pixels[index] =
						static_cast<float>(x + computationWidth * ((tile - 1) % layoutWidth)) + 0.5f;

					pixels[index + 1] = 
						static_cast<float>(y + computationHeight * ((tile - 1) / layoutWidth)) + 0.5f;
				}

				if (tile < computationDepth - 1) {
					pixels[index + 2] =
						static_cast<float>(x + computationWidth * ((tile + 1) % layoutWidth)) + 0.5f;

					pixels[index + 3] =
						static_cast<float>(y + computationHeight * ((tile + 1) / layoutWidth)) + 0.5f;
				}
			}
		}
	}

	IoTextureSet::fillTexture(neighborLookup, GL_RGBA, GL_FLOAT, &*pixels.begin());
}


void Flat3dTextureSet::initializeOffsetLookup()
{
	offsetLookup =
		addSecondaryTexture(computationDepth, 1,
		                    getPrimaryTarget(), getPrimaryInternalFormat(),
		                    "Offset lookup texture");

	std::vector<float> pixels(3 * computationDepth);

	for (unsigned int i = 0; i < computationDepth; ++i) {
		pixels[3 * i] = 
			static_cast<float>(computationWidth * (i % layoutWidth));
		pixels[3 * i + 1] = 
			static_cast<float>(computationHeight * (i / layoutWidth));
	}

	IoTextureSet::fillTexture(offsetLookup, GL_RGB, GL_FLOAT, &*pixels.begin());
}


void Flat3dTextureSet::initializeSliceLookup()
{
	sliceLookup =
		addSecondaryTexture(2 * computationDepth + 1, 1, getPrimaryTarget(), getPrimaryInternalFormat(),
		                    "Slice lookup texture");

	glBindTexture(getPrimaryTarget(), lookup(sliceLookup));
	glTexParameteri(getPrimaryTarget(), GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(getPrimaryTarget(), GL_TEXTURE_WRAP_T, GL_REPEAT);

	std::vector<float> pixels(3 * (2 * computationDepth + 1));

	for (unsigned int i = 0; i <= 2 * computationDepth; ++i) {
		pixels[3 * i] = 
			static_cast<float>(computationWidth * ((i % computationDepth) % layoutWidth)) + computationWidth / 2.0f;
		pixels[3 * i + 1] = 
			static_cast<float>(computationHeight * ((i % computationDepth) / layoutWidth));
		pixels[3 * i + 2] =
			-2.0f * static_cast<float>((i % (2 * computationDepth)) / computationDepth) + 1.0f;
	}

	IoTextureSet::fillTexture(sliceLookup, GL_RGB, GL_FLOAT, &*pixels.begin());
}

}