#ifndef OPENGL_FLAT3DTEXTURESET_H
#define OPENGL_FLAT3DTEXTURESET_H

#include "IoTextureSet.h"

#include <utility>

#include "util/Color.h"
#include "util/VoxelGrid.h"

namespace OpenGL {

/**
 * Calculates optimal layout dimensions for a flat 3d texture.
 * Chooses layouts which are as square-like as possible.
 * @param computationWidth the width of the computation domain.
 * @param computationHeight the height of the computation domain.
 * @param computationDepth the depth of the computation domain.
 * @return the flat 3d texture layout dimensions.
 */
std::pair<unsigned int, unsigned int> 
		calculateLayoutDimensions(unsigned int computationWidth,
		                          unsigned int computationHeight,
		                          unsigned int computationDepth);

/**
 * Flat3dTextureSet is a specialization of IoTextureSet for flat 3D textures. 
 * It maintains flat 3d texture representation of a 3d computation domain,
 * and determines an efficient layout for texture tiles and creates domain, slice, 
 * and offset lookup textures needed by Cg programs operating on flat 3D textures.
 * The depth slices of the 3d computation domain are stored as tiles in
 * the flat 3d texture.
 * In addition it provides functionality for writing voxel grid values to flat 3D 
 * textures without needing to know the internal representation.
 */
class Flat3dTextureSet : public IoTextureSet {
public:
	/**
	 * Constructor.
	 * @param computationWidth the width of the computation domain.
	 * @param computationHeight the height of the computation domain.
	 * @param computationDepth the depth of the computation domain.
	 * @param target the OpenGL texture target.
	 * @param internalFormat the internal format of the texture.
	 */
	Flat3dTextureSet(unsigned int computationWidth,
	                 unsigned int computationHeight,
	                 unsigned int computationDepth,
					 GLenum target, 
	                 GLint internalFormat);

	~Flat3dTextureSet();

	/**
	 * Fills a flat 3d texture with values from a float vector representing the
	 * computation domain. The vector layout should be on the following form:
	 * int offset = ((z * width + y) * height + x) * numChannels + channel
	 * @param textureId the identifier of the texture to be filled.
	 * @param numChannels the number of channels per pixels in the float vector.
	 * (Must be between 1 and 4).
	 * @param values the pixel values the texture should be filled with.
	 */
	void fillTexture(unsigned int textureId,
	                 unsigned int numChannels,
	                 const std::vector<float> &values);

	/**
	 * Fills a flat 3d texture with Color values from a VoxelGrid representing the
	 * computation domain. 
	 * @param textureId the identifier of the texture to be filled.
	 * @param values the Color values the texture should be filled with.
	 */
	void fillTexture(unsigned int textureId, 
	                 const Util::VoxelGrid<Util::Color> &values);

	using IoTextureSet::fillTexture;

	/**
	 * Returns a VoxelGrid representation of a given flat 3d texture. 
	 * @param textureId the identifier of the texture.
	 * @return the VoxelGrid representation.
	 */
	Util::VoxelGrid<Util::Color> getVoxelGrid(unsigned int textureId);

	/**
	 * The domain lookup texture has the same dimension as the flat 3d
	 * textures, and contains for each 2d coordinate the global 3d coordinates
	 * of the corresponding voxel in the computation domain.
	 * @return the identifier of the domain lookup texture.
	 */
	unsigned int getDomainLookup() const;

	/**
	 * The neighbor lookup texture has the same dimension as the flat 3d
	 * textures, and contains for each 2d coordinate the 2d coordinates of the
	 * two neighboring voxels in the depth direction (xy and zw).
	 * @return the identifier of the neighbor lookup texture.
	 */
	unsigned int getNeighborLookup() const;

	/**
	 * The offset lookup texture has dimension depth x 1, and contains for each
	 * depth the offsets of the lower left corner of the corresponding tiles in the flat 3d textures.
	 * @return the identifier of the offset lookup texture.
	 */
	unsigned int getOffsetLookup() const;

	/**
	 * The slice lookup texture has dimension 2*depth x 1, and contains
	 * offsets and directions for the various slices of the flat 3d textures.
	 * @return the identifier of the slice lookup texture.
	 */
	unsigned int getSliceLookup() const;

	/** 
	 * Returns the number of layout tiles in the x-direction of the flat 3d textures.
	 * @return the layout width.
	 */
	unsigned int getLayoutWidth() const;

	/** 
	 * Returns the number of layout tiles in the y-direction of the flat 3d textures.
	 * @return the layout height.
	 */	
	unsigned int getLayoutHeight() const;

	/** 
	 * Returns the width of the computation domain.
	 * @return the computation width.
	 */	
	unsigned int getComputationWidth() const;

	/** 
	 * Returns the height of the computation domain.
	 * @return the computation height.
	 */	
	unsigned int getComputationHeight() const;

	/** 
	 * Returns the depth of the computation domain.
	 * @return the computation depth.
	 */	
	unsigned int getComputationDepth() const;

	/**
	 * Returns the local texture coordinates given the global x, y and z coordinate.
	 * @param x the global x coordinate.
	 * @param y the global y coordinate.
	 * @param z the global z coordinate.
	 * @return the local texture coordinates (x and y).
	 */
	std::pair<unsigned int, unsigned int> getGlobalCoordinates(int x, int y, int z) const; 

private:
	void initializeDomainLookup();
	void initializeNeighborLookup();
	void initializeOffsetLookup();
	void initializeSliceLookup();

	unsigned int computationWidth;
	unsigned int computationHeight;
	unsigned int computationDepth;

	unsigned int layoutWidth;
	unsigned int layoutHeight;

	unsigned int domainLookup;
	unsigned int neighborLookup;
	unsigned int offsetLookup;
	unsigned int sliceLookup;
};

}

#endif