#include <cmath>
#include <iomanip>
#include <numeric>
#include <sstream>
#include <string>

#include "FpsCounter.h"

#include "opengl/OpenGL.h"
#include "opengl/Rendering.h"

namespace OpenGL {

FpsCounter::FpsCounter(unsigned int numTimeSteps) : numTimeSteps(numTimeSteps), currentIndex(0)
{
}


void FpsCounter::update(float timestep)
{
	if (timesteps.size() < numTimeSteps)
		timesteps.push_back(timestep);
	else {
		timesteps[currentIndex] = timestep;
		currentIndex = (currentIndex + 1) % numTimeSteps;
	}
}


void FpsCounter::render(unsigned int x, unsigned int y) const
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, 1, 0, 1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	float viewport[4];
	glGetFloatv(GL_VIEWPORT, viewport);

	int stddev = static_cast<int>(std::sqrt(calculateVariance()) * 10.0f);

	int fps = static_cast<int>(calculateFps() * 10.0f);

	std::stringstream ss;
	ss << "FPS: " << fps / 10 << '.' << fps % 10 << " STDDEV: " << stddev / 10 << '.' << stddev % 10;

	glColor3f(1, 1, 1);
	glRasterPos2f(static_cast<float>(x) / viewport[2],
	              static_cast<float>(y) / viewport[3]);
	displayString(GLUT_BITMAP_TIMES_ROMAN_24, ss.str());

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
}


float FpsCounter::calculateFps() const
{
	return static_cast<float>(timesteps.size()) / 
		std::accumulate(timesteps.begin(), timesteps.end(), 0.0f);
}

float FpsCounter::calculateVariance() const
{
	float variance = 0.0f;

	float mean = calculateFps();

	for (unsigned int i = 0; i < static_cast<unsigned int>(timesteps.size()); ++i)
		variance += (1.0f / timesteps[i] - mean) * (1.0f / timesteps[i] - mean);

	variance /= static_cast<float>(timesteps.size());

	return variance;
}

}
