#include "IoTextureSet.h"

#include <algorithm>
#include <cassert>
#include <iostream>
#include <utility>

#include <GL/glew.h>
#include <GL/gl.h>

#include "gpgpu/framebufferObject.h"

namespace OpenGL {

bool IoTextureSet::isFboBound = false;

GLuint IoTextureSet::currentAttachment;

IoTextureSet::IoTextureSet(unsigned int width, unsigned int height,
                           GLenum target, GLint internalFormat)
	: width(width),
	  height(height),
	  target(target),
	  internalFormat(internalFormat),	  	  
	  fbo(getFbo())
{
	initialize();
}

IoTextureSet::IoTextureSet(std::pair<unsigned int, unsigned int> dimensions, 
                           GLenum target, GLint internalFormat)
	: width(dimensions.first),
	  height(dimensions.second),
	  target(target),
	  internalFormat(internalFormat),	  	  
	  fbo(getFbo())
{
	initialize();
}

IoTextureSet::~IoTextureSet()
{
}

unsigned int IoTextureSet::addPrimaryTexture(const std::string &description)
{
	unsigned int textureId = addTexture(width, height, target, internalFormat, description, true);
	primaryTextures.push_back(textureId);

	return textureId;
}

unsigned int IoTextureSet::addSecondaryTexture(unsigned int width, unsigned int height,
                                               GLenum target, GLint internalFormat,
                                               const std::string &description)
{
	unsigned int textureId = addTexture(width, height, target, internalFormat, description, false);
	secondaryTextures.push_back(textureId);

	return textureId;
}

unsigned int IoTextureSet::importSecondaryTexture(IoTextureSet *textureSet, unsigned int textureId,
                                    const std::string &description)
{
	unsigned int newTextureId = static_cast<unsigned int>(textures.size());	
	
	TextureData data(textureSet->getWidth(textureId), textureSet->getHeight(textureId), 
	                 textureSet->getTarget(textureId), textureSet->getInternalFormat(textureId),
	                 textureSet->lookup(textureId), description, false, textureSet, textureId);
	textures.push_back(data);

	secondaryTextures.push_back(newTextureId);

	return newTextureId;
}

void IoTextureSet::resizeSecondaryTexture(unsigned int textureId, unsigned int width, 
                                          unsigned int height)
{
	assert(textures[textureId].actualTextureSet == this);

	textures[textureId].width = width;
	textures[textureId].height = height;
}

void IoTextureSet::clearTexture(unsigned int textureId)
{
	setActiveTexture(textureId);

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	glClear(GL_COLOR_BUFFER_BIT);
}

std::vector<Util::Color> IoTextureSet::getTexture(unsigned int textureId)
{
	GLuint texture = IoTextureSet::lookup(textureId);

	unsigned int width = getWidth(textureId);
	unsigned int height = getHeight(textureId);

	std::vector<Util::Color> result(width * height);

	GLenum target = getTarget(textureId);

	std::vector<float> v(4 * width * height);

	glBindTexture(target, texture);
	glGetTexImage(target, 0, GL_RGBA, GL_FLOAT, &*v.begin());

	for (unsigned int i = 0; i < v.size(); i+=4)
		result[i / 4] = Util::Color(v[i], v[i+1], v[i+2], v[i+3]);

	return result;
}

void IoTextureSet::fillTexture(unsigned int textureId, GLenum format, GLenum type, const GLvoid *pixels)
{	
	const TextureData &data = textures[textureId];	

	glBindTexture(data.target, lookup(textureId));
	glTexImage2D(data.target, 0, data.internalFormat, data.width, data.height, 
	             0, format, type, pixels);
}


void IoTextureSet::swap()
{
	// Only primary textures can be swapped.
	assert(textures[tempTexture].primary);
	assert(textures[swapTexture].primary);
	std::swap(textures[tempTexture].texture, textures[swapTexture].texture);
}


void IoTextureSet::setActiveTexture(unsigned int textureId, bool swap)
{
	unsigned int outputTexture = swap ? tempTexture : textureId;

	if (swap)
		swapTexture = textureId;

	unsigned int output = lookup(outputTexture);

	if (currentAttachment != output) {
		currentAttachment = output;

		fbo->AttachTexture(GL_COLOR_ATTACHMENT0_EXT, target, currentAttachment);
	}
}


GLuint IoTextureSet::lookup(unsigned int textureId) const
{
	unsigned int actualId = textures[textureId].actualTextureId;
	IoTextureSet *actualTextureSet = textures[textureId].actualTextureSet;

	return actualTextureSet->textures[actualId].texture;
}


void IoTextureSet::bindFbo()
{
	if (!isFboBound) {
		isFboBound = true;
		getFbo()->Bind();
	}
}


void IoTextureSet::disableFbo()
{
	if (isFboBound) {
		isFboBound = false;
		getFbo()->Disable();
	}
}

unsigned int IoTextureSet::getPrimaryWidth() const
{
	return width;
}
	
unsigned int IoTextureSet::getPrimaryHeight() const
{
	return height;
}

GLenum IoTextureSet::getPrimaryTarget() const
{
	return target;
}

GLint IoTextureSet::getPrimaryInternalFormat() const
{
	return internalFormat;
}

unsigned int IoTextureSet::getWidth(unsigned int textureId) const
{
	return textures[textureId].width;
}

unsigned int IoTextureSet::getHeight(unsigned int textureId) const
{
	return textures[textureId].height;
}

GLenum IoTextureSet::getTarget(unsigned int textureId) const
{
	return textures[textureId].target;
}

GLint IoTextureSet::getInternalFormat(unsigned int textureId) const
{
	return textures[textureId].internalFormat;
}

std::string IoTextureSet::getDescription(unsigned int textureId) const
{
	return textures[textureId].description;
}

std::vector<unsigned int> IoTextureSet::getPrimaryTextures() const
{
	return primaryTextures;
}

std::vector<unsigned int> IoTextureSet::getSecondaryTextures() const
{
	return secondaryTextures;
}

unsigned int IoTextureSet::getPrimaryTextureCount() const
{
	return static_cast<unsigned int>(primaryTextures.size());
}

unsigned int IoTextureSet::getSecondaryTextureCount() const
{
	return static_cast<unsigned int>(secondaryTextures.size());
}


//// PRIVATE FUNCTIONS ////

FramebufferObject *IoTextureSet::getFbo()
{
	static FramebufferObject *fbo = new FramebufferObject;

	return fbo;	
}

void IoTextureSet::initialize()
{
	tempTexture = addPrimaryTexture("Temporary texture");
	swapTexture = tempTexture;

	std::vector<float> v(4 * getPrimaryWidth() * getPrimaryHeight());

	fillTexture(tempTexture, GL_RGBA, GL_FLOAT, &*v.begin());

	currentAttachment = lookup(tempTexture);
}

void IoTextureSet::initializeTexture(unsigned int textureId)
{
	glBindTexture(getTarget(textureId), lookup(textureId));

	glTexParameteri(getTarget(textureId), GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(getTarget(textureId), GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(getTarget(textureId), GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(getTarget(textureId), GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	glTexImage2D(getTarget(textureId), 0, getInternalFormat(textureId), 
	             getWidth(textureId), getHeight(textureId), 0, GL_RGBA, GL_FLOAT, 0);
}

unsigned int IoTextureSet::addTexture(unsigned int width, unsigned int height,
                                      GLenum target, GLint internalFormat,
                                      const std::string &description, bool primary)
{
	unsigned int textureId = static_cast<unsigned int>(textures.size());
	
	GLuint texture;
	glGenTextures(1, &texture);

	TextureData data(width, height, target, internalFormat, texture, description, primary, this, textureId);
	textures.push_back(data);

	initializeTexture(textureId);

	return textureId;	
}


}
