#ifndef OPENGL_IOTEXTURESET_H
#define OPENGL_IOTEXTURESET_H

#include "GL/glew.h"

#include "opengl/OpenGL.h"

#include "util/Color.h"

#include <memory>
#include <vector>

class FramebufferObject;

namespace OpenGL {

/**
 * IoTextureSet represents a set of textures which can be used as input or output by fragment
 * operations. Each texture gets an associated texture identifier which is used to refer to it. Textures
 * are used as output by enabling a framebuffer object and attaching the texture to it. Enabling
 * and disabling the framebuffer object is done by the two static member functions bindFbo()
 * and disableFbo() respectively, whereas textures are attached using the member function
 * setActiveTexture(). Additionally, IoTextureSet offers member functions for accessing the
 * texture data of the contained textures. The textures contained in an IoTextureSet instance
 * are divided into secondary and primary textures. Primary textures all have the same dimensions
 * and format, whereas the secondary texture can be of different dimensions and formats.
 * Swapping (or ping-ponging) is supported by adding an additional temporary texture of the
 * same dimensions and format as the primary textures. To swap the currently active texture
 * with the temporary texture the member function swap() is used. Secondary textures are not
 * supported for swapping, but can be imported from other IoTextureSet instances in order to
 * share a texture between several instances. 
 */
class IoTextureSet {
public:
	/**
	 * Default constructor.
	 * @param width the primary texture width.
	 * @param height the primary texture height.
	 * @param target the OpenGL texture target. 
	 * @param internalFormat the texture internal format (number of channels etc).
	 */
	IoTextureSet(unsigned int width, unsigned int height, GLenum target, GLint internalFormat);

	/**
	 * Constructor using dimensions. 
	 * @param dimensions the dimensions of the primary texture, where
	 * width = dimensions.first and height = dimensions.second.
	 * @param target the OpenGL texture target. 
	 * @param internalFormat the texture internal format (number of channels etc).
	 */
	IoTextureSet(std::pair<unsigned int, unsigned int> dimensions, 
	             GLenum target, GLint internalFormat);

	~IoTextureSet();

	/**
	 * Adds a primary texture.	 
	 * @param description a textual description of the contents of the texture.
	 * @return a texture identifier used for looking up the texture.
	 */
	unsigned int addPrimaryTexture(const std::string &description);

	/**
	 * Adds a secondary texture.
	 * @param width the texture's width.
	 * @param height the texture's height.
	 * @param target the OpenGL texture target.
	 * @param internalFormat the texture's internal format (number of channels etc).
	 * @param description a textual description of the contents of the texture.
	 * @return a texture identifier used for looking up the texture.
	 */
	unsigned int addSecondaryTexture(unsigned int width, unsigned int height,
	                                 GLenum target, GLint internalFormat,
	                                 const std::string &description);

	/**
	 * Imports a secondary texture from a primary or secondary texture in another texture set.
	 * @param textureSet the texture set containing the texture.
	 * @param textureId the identifier of the texture in the other texture set.
	 * @param description a textual description of the contents of the texture.
	 * @return a texture identifier used for looking up the texture.
	 */
	unsigned int importSecondaryTexture(IoTextureSet *textureSet, unsigned int textureId,
	                                    const std::string &description);

	/**
	 * Resizes a secondary texture owned by this IoTextureSet.
	 * @param textureId the texture identifier of the texture to be resized.
	 * @param width the new texture width.
	 * @param height the new texture height.
	 */
	void resizeSecondaryTexture(unsigned int textureId, unsigned int width, 
	                            unsigned int height);


	/**
	 * Clears a texture by setting all channel values to 0.
	 * @param textureId the identifier of the texture to be cleared.
	 */
	void clearTexture(unsigned int textureId);

	/**
	 * Gets the contents of a texture.
	 * @param textureId the identifier of the texture.
	 * @return a vector containing the texture's contents.
	 */
	std::vector<Util::Color> getTexture(unsigned int textureId);

	/**
	 * Fills a texture with pixel values.
	 * @param textureId the identifier of the texture to be filled.
	 * @param format the format of pixel values (number of channels).
	 * @param type the type of the pixel values.
	 * @param pixels pointer to the pixel values. 
 	 */
	void fillTexture(unsigned int textureId, GLenum format, GLenum type,
	                 const GLvoid *pixels);
	
	/**
	 * Swaps temporary texture with swap texture (active texture). 
	 */
	void swap();

	/**
	 * Sets a texture as the active texture on the Frame Buffer Object. 
	 * The active texture is then the current render target. 
	 * @param textureId the identifier of the texture to be set active.
	 * @param swap if swap is needed. 
	 */
	void setActiveTexture(unsigned int textureId, bool swap = false);

	/**
	 * Performs a lookup to retrieve the real GLuint texture.
	 * @param textureId the identifier of the texture to be retrieved.
	 * @return the real GLuint texture. 
	 */
	GLuint lookup(unsigned int textureId) const;

	/**
	 * Binds the Frame Buffer Object. 
	 * Results from the fragment shaders are written to the active texture 
	 * (render target) instead of the frame buffer.
	 */
	static void bindFbo();

	/**
	 * Disables the Frame Buffer Object.
	 */
	static void disableFbo();

	/**
	 * Returns the width of the primary textures.
	 * @return the width of the primary textures.
	 */
	unsigned int getPrimaryWidth() const;

	/**
	 * Returns the height of the primary textures.
	 * @return the height of the primary textures.
	 */
	unsigned int getPrimaryHeight() const;

	/**
	 * Returns the target of the primary textures.
	 * @return the target of the primary textures.
	 */
	GLenum getPrimaryTarget() const;

	/**
	 * Returns the internal format of the primary textures.
	 * @return the internal format of the primary textures.
	 */
	GLint getPrimaryInternalFormat() const;

	/**
	 * Returns the width of a given texture.
	 * @param textureId the identifier of the texture.
	 * @return the width of the texture.
	 */
	unsigned int getWidth(unsigned int textureId) const;

	/**
	 * Returns the height of a given texture.
	 * @param textureId the identifier of the texture.
	 * @return the height of the texture.
	 */
	unsigned int getHeight(unsigned int textureId) const;

	/**
	 * Returns the target of a given texture.
	 * @param textureId the identifier of the texture.
	 * @return the target of the texture.
	 */
	GLenum getTarget(unsigned int textureId) const;

	/**
	 * Returns the internal format of a given texture.
	 * @param textureId the identifier of the texture.
	 * @return the internal format of the texture.
	 */
	GLint getInternalFormat(unsigned int textureId) const;

	/**
	 * Returns the description of a given texture.
	 * @param textureId the identifier of the texture.
	 * @return the description of the texture.
	 */
	std::string getDescription(unsigned int textureId) const;

	/**
	 * Returns the primary texture identifiers.
	 * @return the primary texture identifiers.
	 */
	std::vector<unsigned int> getPrimaryTextures() const;

	/**
	 * Returns the secondary texture identifiers.
	 * @return the secondary texture identifiers.
	 */
	std::vector<unsigned int> getSecondaryTextures() const;

	/**
	 * Returns the number of primary textures.
	 * @return the number of primary textures.
	 */
	unsigned int getPrimaryTextureCount() const;

	/**
	 * Returns the number of secondary textures.
	 * @return the number of secondary textures.
	 */
	unsigned int getSecondaryTextureCount() const;

private:
	struct TextureData {	

		TextureData(unsigned int width, unsigned int height, GLenum target, GLint internalFormat,
		            GLuint texture, std::string description, bool primary, IoTextureSet *actualTextureSet,
		            unsigned int actualTextureId) :
			width(width), height(height), target(target), internalFormat(internalFormat),
			texture(texture), description(description), primary(primary), actualTextureSet(actualTextureSet),
			actualTextureId(actualTextureId) {}

		unsigned int width, height;
		GLenum target;
		GLint internalFormat;
		GLuint texture;
		std::string description;
		bool primary;
		IoTextureSet *actualTextureSet;
		unsigned int actualTextureId;
	};

	static FramebufferObject *getFbo();

	static bool isFboBound;

	void initialize();
	void initializeTexture(unsigned int textureId);	
	unsigned int addTexture(unsigned int width, unsigned int height,
                    GLenum target, GLint internalFormat,
                    const std::string &description, bool primary);

	unsigned int tempTexture, swapTexture;

	unsigned int width, height;

	GLenum target;
	GLint internalFormat;

	static GLuint currentAttachment;
		
	std::vector<TextureData> textures;
	std::vector<unsigned int> primaryTextures;
	std::vector<unsigned int> secondaryTextures;

	FramebufferObject *fbo;
};

}

#endif