#include "Texture2dOperation.h"

namespace OpenGL {

Texture2dOperation::Texture2dOperation(unsigned int computationWidth, unsigned int computationHeight,
                                       CgProgram *interior, CgProgram *boundary,
									   bool normalizedTexCoords) :
	CgOperation(computationWidth, computationHeight,
	            interior, boundary, normalizedTexCoords),
	interiorOnly(true)
{
	initExecutionParameters();
}

Texture2dOperation::Texture2dOperation(unsigned int computationWidth, unsigned int computationHeight,
                                       CgProgram *interior, bool interiorOnly,
									   bool normalizedTexCoords) :
	CgOperation(computationWidth, computationHeight,
	            interior, 0, normalizedTexCoords), 
	interiorOnly(interiorOnly)
{
	initExecutionParameters();
}

void Texture2dOperation::initExecutionParameters()
{
	float simWidth = static_cast<float>(getComputationWidth());
	float simHeight = static_cast<float>(getComputationHeight());

	float texDelta = hasBoundaryProgram() || interiorOnly ? 1.0f : 0.0f;
	float vertexDelta = hasBoundaryProgram() || interiorOnly ? 1.0f: 0.0f;	

	minX  = vertexDelta / simWidth;
	maxX  = 1.0f - minX;
	minY  = vertexDelta / simHeight;
	maxY  = 1.0f - minY;

	if (hasNormalizedTexCoords()) {

		minS = texDelta / simWidth;
		maxS = 1.0f - minX;
		minT = texDelta / simHeight;
		maxT = 1.0f - minY;		

	} else {

		minS = texDelta;
		maxS = simWidth - texDelta;
		minT = texDelta;
		maxT = simHeight - texDelta;

	}
}

void Texture2dOperation::executeInteriorProgram()
{
	glBegin(GL_QUADS);

	glTexCoord2f(minS, minT);
	glVertex2f(minX, minY);

	glTexCoord2f(maxS, minT);
	glVertex2f(maxX, minY);

	glTexCoord2f(maxS, maxT);
	glVertex2f(maxX, maxY);

	glTexCoord2f(minS, maxT);
	glVertex2f(minX, maxY);

	glEnd();
}

void Texture2dOperation::executeBoundaryProgram()
{
	float halfX = minX / 2.0f;
	float halfY = minY / 2.0f;

	float halfS = minS / 2.0f;
	float halfT = minT / 2.0f;

	float minS = this->minS - halfS;
	float maxS = this->maxS + halfS;

	float minT = this->minT - halfT;
	float maxT = this->maxT + halfT;

	float minX = this->minX - halfX;
	float maxX = this->maxX + halfX;

	float minY = this->minY - halfY;
	float maxY = this->maxY + halfY;

	glBegin(GL_LINES);

	// left boundary
	glMultiTexCoord2f(GL_TEXTURE1_ARB, this->minS, 0); // offset amount
	glTexCoord2f(minS, minT);
	glVertex2f(minX, minY);
	glTexCoord2f(minS, maxT);
	glVertex2f(minX, maxY);

	// top boundary
	glMultiTexCoord2f(GL_TEXTURE1_ARB, 0, -this->minT); // offset amount
	glTexCoord2f(minS, maxT);
	glVertex2f(minX, maxY);
	glTexCoord2f(maxS, maxT);
	glVertex2f(maxX, maxY);

	// right boundary
	glMultiTexCoord2f(GL_TEXTURE1_ARB, -this->minS, 0); // offset amount
	glTexCoord2f(maxS, maxT);
	glVertex2f(maxX, maxY);
	glTexCoord2f(maxS, minT);
	glVertex2f(maxX, minY);

	// bottom boundary
	glMultiTexCoord2f(GL_TEXTURE1_ARB, 0, this->minT); // offset amount
	glTexCoord2f(maxS, minT);
	glVertex2f(maxX, minY);
	glTexCoord2f(minS, minT);
	glVertex2f(minX, minY);

	glMultiTexCoord2f(GL_TEXTURE1_ARB, 0, 0); // clear texture coordinates

    glEnd();
}

}