#include <algorithm>
#include <cassert>

#include "TextureContainer.h"

#include "util/Util.h"

#include "IoTextureSet.h"
#include "CgContext.h"
#include "CgProgram.h"
#include "OrthonormalProjection.h"
#include "Rendering.h"
#include "Texture2dOperation.h"

using namespace Util;

namespace OpenGL {

TextureContainer::TextureContainer(IoTextureSet *textureSet) :
	textureSet(textureSet),
	titleHeight(24),
	layoutWidth(0),
	layoutHeight(0),
	grid(false)
{
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	viewportX = viewport[0];
	viewportY = viewport[1];
	viewportWidth = viewport[2];
	viewportHeight = viewport[3];

	initialize();
}

TextureContainer::TextureContainer(IoTextureSet *textureSet,
								   unsigned int x, unsigned int y,
	                               unsigned int width, unsigned int height) :
	textureSet(textureSet),
	viewportX(x),
	viewportY(y),
	viewportWidth(width),
	viewportHeight(height),
	titleHeight(24),
	layoutWidth(0),
	layoutHeight(0),
	grid(true)
{
	initialize();
}

TextureContainer::~TextureContainer()
{
}

void TextureContainer::initialize()
{
	cgDisplayTex2d = autoPtr(
		new CgProgram("src/cg/texture.cg", OpenGL::CgContext::getInstance().getFragmentShader(), 
	                  "display2D", textureSet));

	cgDisplayTexRect = autoPtr(
		new CgProgram("src/cg/texture.cg", OpenGL::CgContext::getInstance().getFragmentShader(), 
	                  "displayRECT", textureSet));

	cgDisplayTex2dSmooth =
		autoPtr(new CgProgram("src/cg/texture.cg", OpenGL::CgContext::getInstance().getFragmentShader(), 
	                          "display2Dsmooth", textureSet));

	cgDisplayTexRectSmooth = 
		autoPtr(new CgProgram("src/cg/texture.cg", OpenGL::CgContext::getInstance().getFragmentShader(), 
	                          "displayRECTsmooth", textureSet));

	activateSingleChannel(RED);
	setRange(0.0f, 1.0f);

	setSeparatorColor(1.0f, 0.0f, 0.0f);
	setGridColor(1.0f, 1.0f, 1.0f);
	setTextColor(1.0f, 1.0f, 1.0f);
}

void TextureContainer::insertTexture(unsigned int texture)
{
	textures.insert(texture);
}


bool TextureContainer::isGridActive() const
{
	return grid;
}


void TextureContainer::setGridActive(bool gridActive)
{
	grid = gridActive;
}

void TextureContainer::dumpTexture(unsigned int textureNumber, Channel channel, std::ostream &output)
{
	TextureSet::const_iterator iter = textures.begin();
	std::advance(iter, textureNumber);	

	GLuint texture = textureSet->lookup(*iter);

	unsigned int width = getWidth(textureNumber), height = getHeight(textureNumber);

	GLenum target = getTarget(textureNumber);

	std::vector<float> v(4 * width * height);

	glBindTexture(target, texture);
	glGetTexImage(target, 0, GL_RGBA, GL_FLOAT, &*v.begin());

	for (unsigned int y = height; y > 0; --y) {
		for (unsigned int x = 0; x < width; ++x) {			

			unsigned int index = 4 * ((y - 1) * width + x) + channel;

			output << v[index] << ' '; 			
		}
		output << std::endl;
	}
}

unsigned int TextureContainer::getTextureId(unsigned int textureNumber)
{
	TextureSet::const_iterator iter = textures.begin();
	std::advance(iter, textureNumber);

	return textureSet->lookup(*iter);
}

void TextureContainer::setTextColor(float red, float green, float blue)
{
	textColor[0] = red; textColor[1] = green; textColor[2] = blue;
}


void TextureContainer::setGridColor(float red, float green, float blue)
{
	gridColor[0] = red; gridColor[1] = green; gridColor[2] = blue;
}


void TextureContainer::setSeparatorColor(float red, float green, float blue)
{
	separatorColor[0] = red; separatorColor[1] = green; separatorColor[2] = blue;
}


void TextureContainer::setDescription(const std::string &description)
{
	this->description = description;
}


void TextureContainer::render() const
{
	GLint viewport[4];

	glGetIntegerv(GL_VIEWPORT, viewport);

	glViewport(viewportX, viewportY, viewportWidth, viewportHeight - titleHeight);

	OrthoNormalProjection projection;

	if (!activeTextures.empty()) {
		std::set<unsigned int>::const_iterator iter = activeTextures.begin();
		
		unsigned int texWidth = getActiveWidth() * layoutWidth;
		unsigned int texHeight = getActiveHeight() * layoutHeight; 

		float viewportAspect = static_cast<float>(viewportWidth) /
							static_cast<float>(viewportHeight - titleHeight);

		float actualAspect = static_cast<float>(texWidth) /
							static_cast<float>(texHeight);

		float width, height, offsetX, offsetY;
		
		const float viewScale = 0.95f;

		if (viewportAspect > actualAspect) {

			height = viewScale / static_cast<float>(layoutHeight);
			width = height * static_cast<float>(getActiveWidth()) /
							(static_cast<float>(getActiveHeight()) * viewportAspect);
			offsetX = (viewScale - width * layoutWidth) / 2.0f;
			offsetY = 0.0f;		

		} else {
			width = viewScale / static_cast<float>(layoutWidth);
			height = width * (static_cast<float>(getActiveHeight()) * viewportAspect) /
							static_cast<float>(getActiveWidth());
			offsetX = 0.0f;
			offsetY = (viewScale - height * layoutHeight) / 2.0f;
		}

		float border = (1.0f - viewScale) / 2.0f;

		for (unsigned int y = layoutHeight; y > 0; --y) {
			for (unsigned int x = 0; x < layoutWidth; ++x) {
				if (iter != activeTextures.end()) {
					if (layoutHeight == 1 && layoutWidth == 1 && texHeight == 1) {
						drawTexture(*(iter++), x * width + offsetX + border,
									border, width, viewScale);
					} else
						drawTexture(*(iter++), x * width + offsetX + border,
									(y - 1) * height + offsetY + border, width, height);
				}
			}
		}

		float horizontalDelta = height;
		float verticalDelta = width;

		glColor3fv(separatorColor);

		if (layoutHeight == 1 && layoutWidth == 1 && texHeight == 1) {
			for (unsigned int i = 0; i < 2; ++i) {
				glBegin(GL_LINES);
				glVertex2f(offsetX + border, i * viewScale + border);
				glVertex2f(offsetX + width * layoutWidth + border, i * viewScale + border);

				glVertex2f(offsetX + i * verticalDelta + border, border);
				glVertex2f(offsetX + i * verticalDelta + border, viewScale + border);
				glEnd();
			}
		} else {
			for (unsigned int i = 0; i <= layoutHeight; ++i) {
				glBegin(GL_LINES);
				glVertex2f(offsetX + border, offsetY + i * horizontalDelta + border);
				glVertex2f(offsetX + width * layoutWidth + border, offsetY + i * horizontalDelta + border);
				glEnd();
			}

			for (unsigned int i = 0; i <= layoutWidth; ++i) {
				glBegin(GL_LINES);
				glVertex2f(offsetX + i * verticalDelta + border, offsetY + border);
				glVertex2f(offsetX + i * verticalDelta + border, offsetY + height * layoutHeight + border);
				glEnd();
			}
		}

		glColor3f(1, 1, 1);

	}

	glViewport(viewportX, viewportY + (viewportHeight - titleHeight),
	           viewportWidth, titleHeight);

	float fontDeltaX = 2.0f / static_cast<float>(viewportWidth);
	float fontDeltaY = 2.0f / static_cast<float>(titleHeight);

	glColor4f(1, 1, 1, 0);

	glRasterPos2f(fontDeltaX, fontDeltaY);
	OpenGL::displayString(GLUT_BITMAP_HELVETICA_18, description);

	glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
}


unsigned int TextureContainer::getTextureCount() const
{
	return static_cast<unsigned int>(textures.size());
}


bool TextureContainer::isActive(unsigned int textureNumber) const
{
	TextureSet::const_iterator iter = textures.begin();

	std::advance(iter, textureNumber);

	return activeTextures.find(*iter) != activeTextures.end();
}


void TextureContainer::activateTexture(unsigned int textureNumber)
{
	TextureSet::const_iterator iter = textures.begin();

	std::advance(iter, textureNumber);

	activeTextures.insert(*iter);

	handleTextureActivation(textureNumber);

	calculateDimensions();
}


void TextureContainer::deactivateTexture(unsigned int textureNumber)
{
	TextureSet::const_iterator iter = textures.begin();

	std::advance(iter, textureNumber);

	TextureSet::iterator activeIter = activeTextures.find(*iter);

	if (activeIter != activeTextures.end())
		activeTextures.erase(activeIter);

	calculateDimensions();
}


void TextureContainer::drawTexture(unsigned int texture, float x, float y, 
                                   float width, float height) const 
{
	CgProgram *cgDisplayTex = 
		getActiveTarget() == GL_TEXTURE_2D ? 
		(getSmoothDisplay() ? cgDisplayTex2dSmooth.get() : cgDisplayTex2d.get()) : 
		(getSmoothDisplay() ? cgDisplayTexRectSmooth.get() : cgDisplayTexRect.get());

	cgDisplayTex2dSmooth->setParameter("dimX", static_cast<float>(getActiveWidth()));
	cgDisplayTex2dSmooth->setParameter("dimY", static_cast<float>(getActiveHeight()));

	cgDisplayTex->setTextureParameter("texture", texture);
	cgDisplayTex->setParameter("maxValue", maxValue);
	cgDisplayTex->setParameter("minValue", minValue);
	cgDisplayTex->setMatrixParameter("mixingMatrix", mixMatrix);

	cgDisplayTex->activate();

	float maxS = 
		textureSet->getPrimaryTarget() == GL_TEXTURE_2D
		? 1.0f 
		: static_cast<float>(getActiveWidth());

	float maxT = 
		textureSet->getPrimaryTarget() == GL_TEXTURE_2D 
		? 1.0f 
		: static_cast<float>(getActiveHeight());

	glBegin(GL_QUADS);

	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(x, y);
	
	glTexCoord2f(maxS, 0.0f);
	glVertex2f(x + width, y);

	glTexCoord2f(maxS, maxT);
	glVertex2f(x + width, y + height);

	glTexCoord2f(0.0f, maxT);
	glVertex2f(x, y + height);

	glEnd();

	cgDisplayTex->deactivate();

	if (grid) {
		
		float horizontalDelta = height / static_cast<float>(getActiveHeight());
		float verticalDelta = width / static_cast<float>(getActiveWidth());

		glColor3fv(gridColor);

		for (unsigned int i = 0; i <= getActiveHeight(); ++i) {
			glBegin(GL_LINES);
			glVertex2f(x, y + i * horizontalDelta);
			glVertex2f(x + width, y + i * horizontalDelta);
			glEnd();
		}

		for (unsigned int i = 0; i <= getActiveWidth(); ++i) {
			glBegin(GL_LINES);
			glVertex2f(x + i * verticalDelta, y);
			glVertex2f(x + i * verticalDelta, y + height);
			glEnd();
		}
	}

	TextureSet::const_iterator iter = textures.find(texture);

	assert(iter != textures.end());

	unsigned int fontHeight = 10;

	glColor3fv(textColor);

	unsigned int index = 
		std::distance(textures.begin(), textures.find(texture));

	std::string description = Util::toString(index) + ": " + textureSet->getDescription(*iter);

	glRasterPos2f(x + 2.0f / viewportWidth, y + fontHeight / (2.0f * viewportHeight));
	OpenGL::displayString(GLUT_BITMAP_HELVETICA_10, description, width * viewportWidth);
}


void TextureContainer::calculateDimensions()
{
	if (activeTextures.empty())
		return;

	unsigned int texWidth = getActiveWidth();
	unsigned int texHeight = getActiveHeight();

	double viewportAspect = static_cast<double>(viewportWidth) /
	                       static_cast<double>(viewportHeight - titleHeight);

	double bestArea = 0.0;

	unsigned int numTextures = static_cast<unsigned int>(activeTextures.size());

	layoutWidth = 0;
	layoutHeight = 0;

	for (unsigned int w = 1; w <= numTextures; ++w) {
		for (unsigned int h = 1; h <= numTextures; ++h) {

			if (w * h < numTextures)
				continue;

			unsigned int actualWidth = w * texWidth;
			unsigned int actualHeight = h * texHeight;

			double aspect = static_cast<double>(actualWidth) / static_cast<double>(actualHeight);

			double area = min(viewportAspect, aspect) / max(viewportAspect, aspect);

			area *= static_cast<double>(numTextures) / static_cast<double>(w * h);

			if (area > bestArea) {
				layoutWidth = w;
				layoutHeight = h;
				bestArea = area;
			}
		}
	}
}

void TextureContainer::activateSingleChannel(Channel channel)
{
	eraseMixMatrix();

	for (unsigned int i = 0; i < 12; i+=4) {
		mixMatrix[i + channel] = 1.0f;
	}
}

void TextureContainer::activateDualChannels(Channel first, Channel second)
{
	eraseMixMatrix();

	mixMatrix[first] = 1.0f;
	mixMatrix[4 + second] = 1.0f;
}

void TextureContainer::activateTripleChannels(Channel first, Channel second, Channel third)
{
	eraseMixMatrix();
	
	mixMatrix[first] = 1.0f;
	mixMatrix[4 + second] = 1.0f;
	mixMatrix[8 + third] = 1.0f;
}

void TextureContainer::setRange(float minValue, float maxValue)
{
	this->minValue = minValue;
	this->maxValue = maxValue;
}	

void TextureContainer::eraseMixMatrix()
{
	for (unsigned int i = 0; i < 16; ++i) {
		mixMatrix[i] = 0.0f;
	}
}

void TextureContainer::setSmoothDisplay(bool smoothDisplay)
{
	TextureContainer::smoothDisplay = smoothDisplay;
}

bool TextureContainer::getSmoothDisplay()
{
	return smoothDisplay;
}

bool TextureContainer::smoothDisplay = false;

}
