#ifndef OPENGL_TEXTURECONTAINER_H
#define OPENGL_TEXTURECONTAINER_H

#include <iosfwd>
#include <set>
#include <string>
#include <utility>
#include <vector>

#include "OpenGL.h"

namespace OpenGL {

class IoTextureSet;
class CgProgram;
class Texture2dOperation;

/**
 * Contains a set of textures from an IoTextureSet and displays them on the
 * screen. The textures in the TextureContainer are numbered from 0
 * to getTextureCount() - 1 and these texture numbers are used as
 * identifiers for the actual textures. 
 */
class TextureContainer {
public:
	/**
	 * Texture channels.
	 */
	enum Channel {
		RED,   /// the red channel
		GREEN, /// the green channel
		BLUE,  /// the blue channel
		ALPHA  /// the alpha channel
	};

	/**
	 * Initializes the TextureContainer based on the given IoTextureSet and 
	 * viewport dimensions.
	 * @param textureSet the given IoTextureSet.
	 * @param viewportX the left position of the viewport.
	 * @param viewportY the top position of the viewport.
	 * @param viewportWidth the width of the viewport.
	 * @param viewportHeight the height of the viewport.
	 */
	TextureContainer(IoTextureSet *textureSet,
	                 unsigned int viewportX, unsigned int viewportY,
	                 unsigned int viewportWidth, unsigned int viewportHeight);

	/**
	 * Initializes the TextureContainer based on the given IoTextureSet,
	 * and the current OpenGL viewport coordinates.
	 * @param textureSet the given IoTextureSet.
	 */
	TextureContainer(IoTextureSet *textureSet);

	virtual ~TextureContainer() = 0;

	/**
	 * Renders the currently active textures to the viewport, based on the
	 * selected channels, range and colors.
	 */
	void render() const;

	/**
	 * Gets the number of textures in the TextureContainer.
	 * The textures in the TextureContainer are numbered from 0
	 * to number of textures - 1. 
	 * @return the number of textures.
	 */
	unsigned int getTextureCount() const;

	/**
	 * Checks if the given texture is active.
	 * @param textureNumber the number of the given texture.
	 * @return if the texture is active or not.
	 */
	bool isActive(unsigned int textureNumber) const;

	/**
	 * Activates the given texture.
	 * @param textureNumber the number of the given texture.
	 */
	void activateTexture(unsigned int textureNumber);

	/**
	 * Deactivates the given texture.
	 * @param textureNumber the number of the given texture.
	 */
	void deactivateTexture(unsigned int textureNumber);

	/**
	 * Dumps the selected channel of the given texture to the output stream.
	 * @param textureNumber the number of the given texture.
	 * @param channel the channel to dump.
	 * @param output the output stream to dump the texture to.
	 */
	void dumpTexture(unsigned int textureNumber, Channel channel, std::ostream &output);

	/**
	 * Gets the OpenGL texture identifier (GLuint) of the given texture.
	 * @param textureNumber the number of the given texture.
	 * @return the OpenGL texture's identifier.
	 */
	unsigned int getTextureId(unsigned int textureNumber);

	/**
	 * Activates a single channel of the textures.
	 * This channel will be displayed in grayscale.
	 * @param channel the given channel (to be displayed in grayscale).
	 */
	void activateSingleChannel(Channel channel);

	/**
	 * Activates two channels of the textures.
	 * The first channel becomes the red channel of the output, whereas the
	 * second channel becomes the green channel of the output.
	 * @param first the first channel (to be displayed in red).
	 * @param second the second channel (to be displayed in green).
	 */
	void activateDualChannels(Channel first, Channel second);

	/**
	 * Activates three channels of the textures.
	 * The first channel becomes the red channel of the output,
	 * the second channel becomes the green channel of the output, and
	 * the third channel becomes the blue channel of the output.
	 * @param first the first channel (to be displayed in red).
	 * @param second the second channel (to be displayed in green).
	 * @param third the third channel (to be displayed in blue).
	 */
	void activateTripleChannels(Channel first, Channel second, Channel third);

	/**
	 * Sets the range of values to be displayed.
	 * Values below minValue will become 0, whereas values above maxValue become 1.
	 * @param minValue the minimum value of the range to be displayed.
	 * @param maxValue the maximum value of the range to be displayed.
	 */
	void setRange(float minValue, float maxValue);

	/**
	 * Checks if the texture grid is currently active. 
	 * The texture grid is drawn on top of each active texture, 
	 * separating the texture texels.
	 * @return if the grid is active.
	 */
	bool isGridActive() const;

	/**
	 * Sets the active status of the texture grid.
	 * @param gridActive the new grid active status.
	 */
	void setGridActive(bool gridActive);

	/**
	 * Sets the text color used when displaying texture descriptions.
	 * @param red the red component.
	 * @param green the green component.
	 * @param blue the blue component.
	 */
	void setTextColor(float red, float green, float blue);

	/**
	 * Sets the texture grid color used when drawing the texture grid on top
	 * of each active texture, separating the texture texels.
	 * @param red the red component.
	 * @param green the green component.
	 * @param blue the blue component.
	 */
	void setGridColor(float red, float green, float blue);

	/**
	 * Sets the separator color used when drawing a frame around each active texture.	 
	 * @param red the red component.
	 * @param green the green component.
	 * @param blue the blue component.
	 */
	void setSeparatorColor(float red, float green, float blue);

	/**
	 * Set the texture container's description, displayed when render is called.
	 * @param description the texture container's description.
	 */
	void setDescription(const std::string &description);

	/** 
	 * Checks whether interpolation is used when displaying the active textures.
	 * @return whether interpolation is used. 
	 */
	static bool getSmoothDisplay();

	/**
	 * Sets whether interpolation should be used when displaying the active textures.
	 * @param smoothDisplay whether interpolation should be used.
	 */
	static void setSmoothDisplay(bool smoothDisplay);

protected:
	/**
	 * Used by subclasses to insert textures.
	 * @param textureId the texture's identifier, used to lookup in the IoTextureSet.
	 */
	void insertTexture(unsigned int textureId);

private:
	/**
	 * Gets the texture target of the active texture(s).
	 * @return the texture target of the active texture(s).
	 */
	virtual GLenum getActiveTarget() const = 0;

	/**
	 * Gets the width of the active texture(s).
	 * @return the width of the active texture(s).
	 */
	virtual unsigned int getActiveWidth() const = 0;

	/**
	 * Gets the height of the active texture(s).
	 * @return the height of the active texture(s).
	 */
	virtual unsigned int getActiveHeight() const = 0;

	/**
	 * Gets the texture target of the given texture.
	 * @param textureNumber the number of the given texture.
	 * @return the texture target of the given texture.
	 */	
	virtual GLenum getTarget(unsigned int textureNumber) const = 0;

	/**
	 * Gets the width of the given texture.
	 * @param textureNumber the number of the given texture.
	 * @return the width of the given texture.
	 */
	virtual unsigned int getWidth(unsigned int textureNumber) const = 0;

	/**
	 * Gets the height of the given texture.
	 * @param textureNumber the number of the given texture.
	 * @return the height of the active texture(s).
	 */
	virtual unsigned int getHeight(unsigned int textureNumber) const = 0;

	/**
	 * Called whenever the given texture is activated.
	 * @param textureNumber the number of the given texture.
	 */
	virtual void handleTextureActivation(unsigned int textureNumber) = 0;

	void initialize();
	void calculateDimensions();

	void drawTexture(unsigned int texture, float x, float y, 
	                 float width, float height) const;

	void eraseMixMatrix();

	typedef std::set<unsigned int> TextureSet;

	TextureSet textures, activeTextures;
	IoTextureSet *textureSet;
	std::auto_ptr<CgProgram> cgDisplayTex2d, cgDisplayTexRect;
	std::auto_ptr<CgProgram> cgDisplayTex2dSmooth, cgDisplayTexRectSmooth;

	unsigned int layoutWidth, layoutHeight;
	unsigned int viewportX, viewportY, viewportWidth, viewportHeight;
	unsigned int titleHeight;

	bool grid;

	std::string description;

	float textColor[3];
	float separatorColor[3];
	float gridColor[3];

	float mixMatrix[16];
	float minValue;
	float maxValue;

	static bool smoothDisplay;
};

}

#endif