#include "TextureContainerActions.h"

#include <sstream>

#include "util/Util.h"

using namespace Util;

namespace OpenGL {

ChannelSelector::ChannelSelector(TextureContainer *container) : TextureContainerAction(container)
{
}

bool parseChannel(char ch, TextureContainer::Channel &channel)
{
	switch (ch) {
	case 'r':
		channel = TextureContainer::RED;
		return true;
	case 'g':
		channel = TextureContainer::GREEN;
		return true;
	case 'b':
		channel = TextureContainer::BLUE;
		return true;
	case 'a':
		channel = TextureContainer::ALPHA;
		return true;
	default:
		return false;
	}
}

std::string ChannelSelector::run(const std::vector<std::string> &parameters)
{
	if (parameters.empty() || parameters[0].empty())
		return "Channel must be specified.";

	if (parameters[0] == "range") {
		if (parameters.size() < 3)
			return "Min and max values must be specified.";
		else {
			float minValue, maxValue;

			if (!lexical_cast(parameters[1], minValue) ||
				!lexical_cast(parameters[2], maxValue))
			{
				return "Parse error.";
			} else {
				getContainer()->setRange(minValue, maxValue);

				return "Range set to [" + toString(minValue) + ", " +
					toString(maxValue) + "].";
			}
		}
	} else {
		if (parameters[0].size() == 1) {
			TextureContainer::Channel channel;

			if (parseChannel(parameters[0][0], channel))
				getContainer()->activateSingleChannel(channel);
			else
				return "Invalid channel.";
		} else if (parameters[0].size() == 2) {
			TextureContainer::Channel first, second;

			if (parseChannel(parameters[0][0], first) &&
				parseChannel(parameters[0][1], second))
			{
				getContainer()->activateDualChannels(first, second);
			}
			else
				return "Invalid channel.";
		} else if (parameters[0].size() == 3) {
			TextureContainer::Channel first, second, third;

			if (parseChannel(parameters[0][0], first) &&
				parseChannel(parameters[0][1], second) &&
				parseChannel(parameters[0][2], third))
			{
				getContainer()->activateTripleChannels(first, second, third);
			}
			else
				return "Invalid channel.";
		} else {
			return "Must specify one, two or three channels. (Example: rgb)";
		}
	}

	return "Channels updated.";
}


TextureSelector::TextureSelector(TextureContainer *container) : TextureContainerAction(container)
{
}

std::string TextureSelector::run(const std::vector<std::string> &parameters)
{
	if (parameters.empty())
		return "Texture must be specified.";


	bool toggle = false, activate = false;

	if (parameters[0] == "toggle")
		toggle = true;
	else if (parameters[0] == "activate")
		activate = true;
	else if (parameters[0] == "deactivate")
		activate = false;
	else
		return "Valid parameters: activate <texture>, deactivate <texture>, toggle <texture>.";

	if (parameters.size() < 2) 
		return "Texture must be specified.";
	else {			
		unsigned int texture;

		if (lexical_cast(parameters[1], texture)) {

			if (texture >= getContainer()->getTextureCount())
				return "Illegal texture, max is " + toString(getContainer()->getTextureCount() - 1) + '.';

			if (toggle && getContainer()->isActive(texture) || !toggle && !activate) {
				getContainer()->deactivateTexture(texture);
				return "Texture " + toString(texture) + " deactivated.";
			} else {
				getContainer()->activateTexture(texture);
				return "Texture " + toString(texture) + " activated.";
			}
		} else
			return "Parse error.";
	}

}



GridToggler::GridToggler(TextureContainer *container) : TextureContainerAction(container)
{
}


std::string GridToggler::run(const std::vector<std::string> &/*parameters*/)
{
	getContainer()->setGridActive(!getContainer()->isGridActive());

	return getContainer()->isGridActive() ? "Grid activated" : "Grid deactivated";
}


TextureDumper::TextureDumper(TextureContainer *container) : TextureContainerAction(container)
{

}

std::string TextureDumper::run(const std::vector<std::string> &parameters)
{
	if (parameters.size() < 2) 
		return "Must specify texture number and channel.";

	unsigned int texture;
	TextureContainer::Channel channel;

	if (lexical_cast(parameters[0], texture) && parseChannel(parameters[1][0], channel)) {
		
		if (texture >= getContainer()->getTextureCount())
			return "Illegal texture, max is " + toString(getContainer()->getTextureCount() - 1) + '.';

		std::ostringstream oss;
		getContainer()->dumpTexture(texture, channel, oss);
		return "Texture " + toString(texture) + ", channel " + 
			parameters[1].substr(0, 1) + ":\n" + oss.str();
	} else 
		return "Parse error.";
}

TextureId::TextureId(TextureContainer *container) : TextureContainerAction(container)
{
}

std::string TextureId::run(const std::vector<std::string> &parameters)
{
	if (parameters.size() < 1) 
		return "Must specify texture number.";

	unsigned int texture;

	if (lexical_cast(parameters[0], texture)) {
		if (texture >= getContainer()->getTextureCount())
			return "Illegal texture, max is " + toString(getContainer()->getTextureCount() - 1) + '.';

		return "Texture " + toString(texture) + " has OpenGL identifier " +
			toString(getContainer()->getTextureId(texture));
	} else
		return "Parse error.";

}

}