#ifndef OPENGL_TEXTURECONTAINERACTIONS_H
#define OPENGL_TEXTURECONTAINERACTIONS_H

#include "console/Action.h"

#include "opengl/TextureContainer.h"

namespace OpenGL {

/**
 * Parses a character into a Channel enumerator. Valid character values are:
 * 'r' -> RED, 'g' -> GREEN, 'b' -> BLUE, and 'a' -> ALPHA.
 * @param ch the character to parse.
 * @param channel the Channel enumerator to store the result.
 * @return whether the parsing was successful.
 */
bool parseChannel(char ch, TextureContainer::Channel &channel);

/**
 * Represents an action to be performed on a TextureContainer. 
 */
class TextureContainerAction : public Console::Action {
public:
	/**
	 * Base constructor.
	 * @param container the TextureContainer to affect.
	 */
	TextureContainerAction(TextureContainer *container) : container(container) {}
	virtual ~TextureContainerAction() {}

	/**
	 * Sets a new TextureContainer to affect.
	 */
	void setContainer(TextureContainer *container) { this->container = container; }

protected:
	/**
	 * Gets the TextureContainer to affect.
	 */
	TextureContainer *getContainer() const { return container; }

private:
	TextureContainer *container;
};

/**
 * Represents channel selector actions for a specified TextureContainer instance.
 *
 * The run() function can be called with the following arguments: <br>
 * "range <minValue> <maxValue>" - calls setRange(). <br>
 * "<channel>" - calls activateSingleChannel(). <br>
 * "<channel><channel>" - calls activateDualChannels(). <br>
 * "<channel><channel><channel>" - calls activateTripleChannels().
 *
 * <channel> must be a character with value 'r', 'g', 'b', or 'a'. 
 */
class ChannelSelector : public TextureContainerAction {
public:
	/**
	 * Creates a ChannelSelector for the given TextureContainer.
	 * @param container the TextureContainer to affect.
	 */
	ChannelSelector(TextureContainer *container);

	std::string run(const std::vector<std::string> &parameters);
};


/**
 * Represents texture selector actions for a specified TextureContainer instance.
 * 
 * The run() function can be called with the following arguments: <br>
 * "activate <texture>" - calls activateTexture(). <br>
 * "deactivate <texture>" - calls deactivateTexture(). <br>
 * "toggle <texture>" - calls activateTexture() if deactivated and vice versa.
 *
 * <texture> must be a valid texture number of the TextureContainer instance. 
 */
class TextureSelector : public TextureContainerAction {
public:
	/**
	 * Creates a TextureSelector for the given TextureContainer.
	 * @param container the TextureContainer to affect.
	 */
	TextureSelector(TextureContainer *container);

	std::string run(const std::vector<std::string> &parameters);
};

/**
 * Represents a grid toggler action for a specified TextureContainer instance.
 * 
 * The run() function can be called with the following argument: <br>
 * "" - calls setGridActive(!isGridActive()).
 */
class GridToggler : public TextureContainerAction {
public:
	/**
	 * Creates a GridToggler for the given TextureContainer.
	 * @param container the TextureContainer to affect.
	 */
	GridToggler(TextureContainer *container);

	std::string run(const std::vector<std::string> &parameters);
};

/**
 * Represents a texture dumper action for a specified TextureContainer instance.
 * 
 * The run() function can be called with the following argument: <br>
 * "<texture> <channel>" - calls dumpTexture() and returns the texture dump as a string.
 *
 * <texture> must be a valid texture number of the TextureContainer instance. <br>
 * <channel> must be a character with value 'r', 'g', 'b', or 'a'. 
 */
class TextureDumper : public TextureContainerAction {
public:
	/**
	 * Creates a TextureDumper for the given TextureContainer.
	 * @param container the TextureContainer to affect.
	 */
	TextureDumper(TextureContainer *container);

	std::string run(const std::vector<std::string> &parameters);
};

/**
 * Represents a texture id action for a specified TextureContainer instance.
 * 
 * The run() function can be called with the following argument: <br>
 * "<texture>" - calls getTextureId() and returns the OpenGL identifier as a string.
 * 
 * <texture> must be a valid texture number of the TextureContainer instance.
 */
class TextureId : public TextureContainerAction {
public:
	/**
	 * Creates a TextureDumper for the given TextureContainer.
	 * @param container the TextureContainer to affect.
	 */
	TextureId(TextureContainer *container);

	std::string run(const std::vector<std::string> &parameters);
};

}

#endif