#include "XmlDocument.h"

#include <fstream>
#include <iostream>
#include <stack>

#include "XmlNode.h"

#include "util/Util.h"

namespace SimpleXml {

XmlDocument::XmlDocument(const std::string &xmlFile) :
	xmlFile(xmlFile)
{
	loadFile();
}

void XmlDocument::loadFile()
{
	std::ifstream inputFile(xmlFile.c_str());

	if (!inputFile)
		throw SimpleXmlLoadException(xmlFile);

	std::string document = readDocument(inputFile);	

	parseDocument(document);
}

XmlNode* XmlDocument::getRootNode()
{
	return rootNode;
}

void XmlDocument::parseDocument(const std::string &document)
{
	int currentIndex = 0;
	int startOfTag = 0;
	int endOfTag = 0;
	std::string currentTag;
	std::string rootTag;
	bool isStartTag;
	bool hasFoundRootEnd = false;

	std::stack<XmlNode*> nodeStack;

	if (!findNextTag(document, currentIndex, &rootTag, &startOfTag, &endOfTag, &isStartTag))
		throw SimpleXmlParseException("Root start tag element not found.");

	rootNode = new XmlNode(rootTag);
	nodeStack.push(rootNode);
	currentIndex = endOfTag + 1;	

	while (!hasFoundRootEnd) {

		XmlNode *lastNode = nodeStack.top();

		if (!lastNode)
			throw SimpleXmlParseException("Null pointer to last xml node.");

		if (!findNextTag(document, currentIndex, &currentTag, &startOfTag, &endOfTag, &isStartTag))
			throw SimpleXmlParseException("Missing end tag for <" + lastNode->getName() + ">.");

		if (isStartTag) {
			XmlNode *currentNode = lastNode->appendChild(currentTag);			
			nodeStack.push(currentNode);			
		} else {
			if (currentTag == lastNode->getName()) {
				if (!lastNode->hasChildNodes()) {
					std::string nodeValue = document.substr(currentIndex, startOfTag - currentIndex);
					lastNode->setValue(nodeValue);					
				}
				
				nodeStack.pop();
				if (nodeStack.empty())
					hasFoundRootEnd = true;
			} else {
				throw SimpleXmlParseException("Current end tag element </" + currentTag + 
					"> doesn't match last xml node <" + lastNode->getName() + ">.");
			}
		}

		currentIndex = endOfTag + 1;
	}	
}

bool XmlDocument::findNextTag(const std::string &document, int startIndex,
                              std::string *tagName, int *startOfTag, int *endOfTag, bool *isStartTag)
{
	*tagName = "";	
	*isStartTag = false;

	*startOfTag = static_cast<int>(document.find_first_of('<', startIndex));
	if (*startOfTag == std::string::npos)
		return false;

	*endOfTag = static_cast<int>(document.find_first_of('>', *startOfTag));
	if (*endOfTag == std::string::npos)
		return false;

	int startOfTagName = *startOfTag + 1;
	if (document[startOfTagName] == '/') {
		*isStartTag = false;
		++startOfTagName;
	}
	else 
		*isStartTag = true;
	
	std::string result; 
	result += document[startOfTagName];

	char invalidNameChar = 0;

	for (int i = startOfTagName + 1; i < *endOfTag; ++i) {

		if (!isLegalNameChar(document[i]))
			invalidNameChar = document[i];

		result += document[i];
	}

	std::string tagElement = "<";
	if (!*isStartTag)
		tagElement += '/';
	tagElement += result + ">";

	if (!isLegalFirstLetter(document[startOfTagName])) {
		throw SimpleXmlParseException("Invalid first letter in tag element " + tagElement + ".");
	}

	if (invalidNameChar != 0) {
		throw SimpleXmlParseException("Invalid character " + Util::toString(invalidNameChar) + 
			" in tag element " + tagElement + ".");
	}

	
	*tagName = result;	

	return true;
}

bool XmlDocument::isLegalFirstLetter(char c)
{
	if (c >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z' || c == '_' || c == ':')
		return true;
	else 
		return false;
}

bool XmlDocument::isLegalNameChar(char c) 
{
	if (c >= 'A' && c <= 'Z' || c >= 'a' && c <= 'z' || 
		c >= '0' && c <= '9' ||
		c == '_' || c == ':' || c == '.' || c == '-')
		return true;
	else 
		return false;
}


std::string XmlDocument::readDocument(std::istream &is)
{
	std::string line;
	std::string result;

	while (std::getline(is, line))
		result += line;
	
	return result;
}


}

