#include "Perlin.h"

#include <cmath>
#include <ctime>
#include <iostream>

#include "Random.h"

#include "math/Definitions.h"

Perlin2D::Perlin2D(unsigned int numOctaves, float persistence, int seed) :
  mNumOctaves(numOctaves),
  mPersistence(persistence),
  mSeed(seed)
{
}



float Perlin2D::sample(float x, float y) const
{
  float result = 0;

  for (unsigned int i = 0; i < mNumOctaves; ++i) {
    float frequency = static_cast<float>(1 << i);
    float amplitude = std::pow(mPersistence, static_cast<float>(i));

    result += interpolateNoise(x * frequency, y * frequency, i) * amplitude;
  }
  
  return result;
}
 


float Perlin2D::random(int x, int y, unsigned int octave) const
{
  int n = mSeed + x + y * 1337 + octave * 1337 * 1337;

  n = (n << 13) ^ n;
 
  return 1.0f - ((n * (n * n * 15731 + 789221) + 1376312589) & 0x7fffffff) /
    1073741824.0f;
}



float Perlin2D::smoothedRandom(int x, int y, unsigned int octave) const
{
  float corners = 
    random(x - 1, y - 1, octave) + random(x + 1, y + 1, octave) + 
    random(x + 1, y - 1, octave) + random(x - 1, y + 1, octave);
  float sides =
    random(x, y - 1, octave) + random(x - 1, y, octave) +
    random(x, y + 1, octave) + random(x + 1, y, octave);
  float centre =
    random(x, y, octave);
  
  return corners / 16.0f + sides / 8.0f + centre / 4.0f;
}



float Perlin2D::interpolate(float a, float b, float fraction) const
{
	float f = (1.0f - std::cos(Math::pi * fraction)) * 0.5f;

  return a * (1.0f - f) + b * f;
}



float Perlin2D::interpolateNoise(float x, float y, unsigned int octave) const
{
  int integerX = static_cast<int>(x);
  int integerY = static_cast<int>(y);
  
  float fractionalX = x - static_cast<float>(integerX);
  float fractionalY = y - static_cast<float>(integerY);

  float v1 = smoothedRandom(integerX,     integerY,     octave);
  float v2 = smoothedRandom(integerX + 1, integerY,     octave);
  float v3 = smoothedRandom(integerX,     integerY + 1, octave);
  float v4 = smoothedRandom(integerX + 1, integerY + 1, octave);

  float interpolateTop = interpolate(v1, v2, fractionalX);
  float interpolateBottom = interpolate(v3, v4, fractionalX);

  return interpolate(interpolateTop, interpolateBottom, fractionalY);
}



Perlin1D::Perlin1D(unsigned int numOctaves, float persistence, int seed) :
  mNumOctaves(numOctaves),
  mPersistence(persistence),
  mSeed(seed)
{
}



float Perlin1D::sample(float x) const
{
  float result = 0;

  float max = 0.0f;

  for (unsigned int i = 0; i < mNumOctaves; ++i) {
    float frequency = static_cast<float>(1 << i);
    float amplitude = std::pow(mPersistence, static_cast<float>(i));

    result += interpolateNoise(x * frequency, i) * amplitude;

	max += amplitude;
  }
  
  return result / max;
}



float Perlin1D::random(int x, unsigned int octave) const
{
	int n = mSeed + x + octave * 1337;

	n = (n << 13) ^ n;

	return 1.0f - ((n * (n * n * 15731 + 789221) + 1376312589) & 0x7fffffff) / static_cast<float>(0x7fffffff); //1073741824.0f;
}



float Perlin1D::smoothedRandom(int x, unsigned int octave) const
{  
	float corners = 
		random(x - 1, octave) + random(x + 1, octave);
	float centre =
		random(x, octave);
  
  return corners / 4.0f + centre / 2.0f;
}



float Perlin1D::interpolate(float a, float b, float fraction) const
{
	float f = (1.0f - std::cos(Math::pi * fraction)) * 0.5f;

	return a * (1.0f - f) + b * f;
}



float Perlin1D::interpolateNoise(float x, unsigned int octave) const
{
  int integerX = static_cast<int>(x);
  
  float fractionalX = x - static_cast<float>(integerX);

  float v1 = smoothedRandom(integerX,     octave);
  float v2 = smoothedRandom(integerX + 1, octave);

  return interpolate(v1, v2, fractionalX);
}
