#include "Random.h"

#include <cassert>
#include <cmath>
#include <cstdlib>
#include <limits>

const int Random::mModulus = static_cast<unsigned int>(1 << 31) - 1;
const int Random::mMultiplier = 16807;



Random::Random(int seed) : mCurrent(seed), mSeed(seed)
{
  assert(std::numeric_limits<int>::max() >= mModulus);
}



int Random::next()
{
  // Lehmer generator (multiplicative linear congruential generator)
  const int quotient = mModulus / mMultiplier;
  const int remainder = mModulus % mMultiplier;

  int hi = mCurrent / quotient;
  int lo = mCurrent % quotient;

  int test = mMultiplier * lo - remainder * hi;

  if (test > 0)
    mCurrent = test;
  else
    mCurrent = test + mModulus;

  return mCurrent;
}



void Random::reseed(int seed)
{
  mCurrent = mSeed = seed;
}



Random &Random::global()
{
  static Random instance;
  return instance;
}



int Random::getMaxValue()
{
  return mModulus - 1;
}



double nextDouble(Random &rnd)
{
  return rnd.next() / static_cast<double>(Random::getMaxValue() + 1);
}



float nextFloat(Random &rnd)
{
  return rnd.next() / static_cast<float>(Random::getMaxValue() + 1);
}



bool nextBool(Random &rnd)
{
  return rnd.next() == 0;
}



double nextGaussian(Random &rnd)
{
  // Box-Muller transformation
  static bool useLast = false;
  static double y;

  if (useLast) {
    useLast = false;
    return y;
  } else
    useLast = true;
  
  double x1, x2, w;

  do {
    x1 = 2.0 * nextDouble(rnd) - 1.0;
    x2 = 2.0 * nextDouble(rnd) - 1.0;
    w = x1 * x1 + x2 * x2;
  } while(w >= 1.0);

  w = std::sqrt(-2.0 * std::log(w) / w);

  y = x1 * w;

  return x2 * w;
}



double nextExponential(Random &rnd, double gamma)
{
  return -gamma * std::log(1.0 - nextDouble(rnd));
}
