#ifndef UTIL_H
#define UTIL_H

#include <cstdlib>
#include <iostream>
#include <memory>
#include <set>
#include <sstream>
#include <string>

namespace Util {

/**
 * Shortcut for creating an std::auto_ptr without having to specify template parameters.
 * @param t the pointer to embed in an auto_ptr.
 * @return an auto_ptr containing the supplied pointer.
 */
template <typename T>
std::auto_ptr<T> autoPtr(T *t)
{
	return std::auto_ptr<T>(t);
}

/**
 * Lexical cast between types.
 * Uses std::stringstream to perform the conversion.
 * @param source the source variable.
 * @param target the target variable.
 * @return if the conversion was successful.
 */
template <typename Target, typename Source>
bool lexical_cast(const Source &source, Target &target)
{
	std::stringstream ss;

	return (ss << source && ss >> target);
}

/**
 * Overload of lexical cast for string targets.
 * Uses toString to make sure that the source is not split by whitespace.
 * @param source the source variable.
 * @param target the target string.
 * @return true, as this conversion is always successful.
 */
template <typename Source>
bool lexical_cast(const Source &source, std::string &target)
{
	target = toString(source);

	return true;
}

/**
 * Overload of lexical cast for bools.
 * Allows bools to be represented as 0/1 or true/false.
 * @param source the source string.
 * @param target the bool target.
 * @return if conversion was successful (source was 0/1 or true/false).
 */
inline bool lexical_cast(const std::string &source, bool &target)
{
	
	if (source == "true" || source == "1") {
		target = true;
		return true;
	} else if (source == "false" || source == "0") {
		target = false;
		return true;
	} else 
		return false;		
}

/**
 * Conversion to std::string.
 * @param source the variable to convert.
 * @return a string representing the variable.
 */
template <typename Source>
std::string toString(const Source &source)
{
	std::stringstream ss;

	ss << source;

	return ss.str();
}

/**
 * Reads a std::set of values from an std::istream.
 * @param in the input stream.
 * @return the set containing the values read from the istream.
 */
template <typename T>
std::set<T> readSet(std::istream &in)
{
	std::set<T> result;

	T temp;

	while (in >> temp)
		result.insert(temp);

	return result;
}

/**
 * Returns random float values in the range [0.0f, 1.0f>.
 * @return the random float value.
 */
inline float randf()
{
	return std::rand() / (1.0f + RAND_MAX);
}

/**
 * Deletes the supplied pointer.
 * @param t the pointer to the object to delete.
 */
template <typename T>
void destroy(T *t)
{
	delete t;
}

}

#endif