#ifndef UTIL_VOXELGRID_H
#define UTIL_VOXELGRID_H

#include <vector>

namespace Util {

/**
 * Class template for storing a voxel grid of values.
 */
template <typename T>
class VoxelGrid {
public:
	/**
	 * Creates a voxel grid of the specified dimensions.
	 * @param width the voxel grid width.
	 * @param height the voxel grid height.
	 * @param depth the voxel grid depth.
	 */
	VoxelGrid(unsigned int width, unsigned int height, unsigned int depth);

	/**
	 * Gets the element at the specified location.
	 * @param x the x coordinate.
	 * @param y the y coordinate.
	 * @param z the z coordinate.
	 * @return The element at the specified location.
	 */
	T &operator()(unsigned int x, unsigned int y, unsigned int z);

	/**
	 * Gets the element at the specified location.
	 * @param x the x coordinate.
	 * @param y the y coordinate.
	 * @param z the z coordinate.
	 * @return The element at the specified location.
	 */
	const T &operator()(unsigned int x, unsigned int y, unsigned int z) const;

	/**
	 * Gets the width of the voxel grid.
	 * @return The voxel grid width.
	 */
	unsigned int getWidth() const;

	/**
	 * Gets the height of the voxel grid.
	 * @return The voxel grid height.
	 */
	unsigned int getHeight() const;

	/**
	 * Gets the depth of the voxel grid.
	 * @return The voxel grid depth.
	 */
	unsigned int getDepth() const;

private:
	unsigned int getIndex(unsigned int x,
	                      unsigned int y,
	                      unsigned int z) const;

	unsigned int width, height, depth;

	std::vector<T> values;
};

template <typename T>
VoxelGrid<T>::VoxelGrid(unsigned int width,
                        unsigned int height,
						unsigned int depth) :
	width(width),
	height(height),
	depth(depth),
	values(width * height * depth)
{
}

template <typename T>
T &VoxelGrid<T>::operator()(unsigned int x,
                            unsigned int y,
                            unsigned int z)
{
	return values.at(getIndex(x, y, z));
}

template <typename T>
const T &VoxelGrid<T>::operator ()(unsigned int x,
                                   unsigned int y,
                                   unsigned int z) const
{
	return values.at(getIndex(x, y, z));
}

template <typename T>
unsigned int VoxelGrid<T>::getWidth() const
{
	return width;
}

template <typename T>
unsigned int VoxelGrid<T>::getHeight() const
{
	return height;
}

template <typename T>
unsigned int VoxelGrid<T>::getDepth() const
{
	return depth;
}

template <typename T>
unsigned int VoxelGrid<T>::getIndex(unsigned int x, 
                                    unsigned int y,
                                    unsigned int z) const
{
	return x + width * (y + height * z);
}

}

#endif